#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>
#include <unistd.h>
#include "WinObj.h"

static void draw_button_label(ARButtonWindow* b, Bool invert)
{
    Pixel fg;

    if(b->label == NULL)
	return;

    if(invert)
	fg = b->bg;
    else
	fg = b->fg;

    DrawString(b->disp, b->win, b->font, fg,
	       0, 0, b->width, b->height, b->label, Center);
}

static void draw_pressed_button(ARButtonWindow* b)
{
    XSetForeground(b->disp, b->gc, b->mark_color);
    XFillRectangle(b->disp, b->win, b->gc,
		   0, 0, b->width, b->height);
    draw_button_label(b, IsMonoDisplay(b));
    b->border_skeleton->style = WINOBJ_BORDER_CONCAVE;
    RedrawBorderSkeletion((WinObj *)b->border_skeleton);
}

static void arbutton_event_handler(ARButtonWindow* obj, XEvent* e)
{
    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count > 0)
	    return;
	draw_button_label(obj, False);
	break;
      case ButtonPress:
	obj->count = 0;
	draw_pressed_button(obj);
	while(XCheckTypedEvent(obj->disp, ButtonRelease, e) == False)
	{
	    CallCallbackProc((WinObj *)obj, e);
	    if(obj->disp == NULL)
		return;
	    obj->count++;
	    if(obj->delay > 0)
	    {
		XSync(obj->disp, False);
		usleep((unsigned long)(obj->delay * 1000000));
	    }
	}
	XClearWindow(obj->disp, obj->win);
	draw_button_label(obj, False);
	obj->border_skeleton->style = WINOBJ_BORDER_CONVEX;
	RedrawBorderSkeletion((WinObj *)obj->border_skeleton);
	if(obj->stop_arbutton_handler != NULL)
	    obj->stop_arbutton_handler(obj);
	break;
    }
}

static void arbutton_destruction(WinObj* obj)
{
    winobj_free(obj->arb.label);
}

WinObj* create_ARButton(WinObj* parent,
			int x,
			int y,
			int width,
			int height,
			int bd_width,
			CharFont* font,
			const char* label)
{
    ARButtonWindow* arb;

    if(width <= 0 || height <= 0)
    {
	const int gap = 4;

	width  = TextWidth(font, label) + 2 * gap;
	height = TextHeight(font)       + 2 * gap;
    }

    arb = (ARButtonWindow *)create_SimpleWindow(parent, x, y, width, height,
						bd_width,
						WINOBJ_BORDER_CONVEX);
    if(arb == NULL)
	return NULL;

    arb->type = ARButton_T;
    arb->font = font;
    if(label != NULL)
    {
	if((arb->label = winobj_strdup(label)) == NULL)
	{
	    DestroyWinObj((WinObj *)arb);
	    return NULL;
	}
    }
    else
	arb->label = NULL;
    arb->count = 0;
    arb->delay = 0.0;
    arb->stop_arbutton_handler = NULL;
    arb->winobj_destruction = arbutton_destruction;
    SetEventMask((WinObj *)arb, ExposureMask | ButtonPressMask
		               | ButtonReleaseMask);
    SetEventHandler((WinObj *)arb, arbutton_event_handler);

    return (WinObj *)arb;
}
