#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "WinObj.h"

static void border_none(BorderSkeleton* border);
static void border_normal_expose(BorderSkeleton* border);
static void border_convex_expose(BorderSkeleton* border);
static void border_concave_expose(BorderSkeleton* border);

static void (* default_border_expose_handlers[WINOBJ_NUMBER_OF_BORDER])
(BorderSkeleton *) =
{
    border_none,
    border_normal_expose,
    border_convex_expose,
    border_concave_expose,
};

#if 0
static void (** border_expose_handlers)(BorderSkeleton *) = NULL;
static int number_of_border_handlers = WINOBJ_NUMBER_OF_BORDER;
static void init_border_once(void)
{
    if(border_expose_handlers == NULL)
    {
	int i;

	border_expose_handlers = (void (**)(BorderSkeleton *))
	    winobj_malloc(sizeof(void (*)(BorderSkeleton *)));
	for(i = 0; i < WINOBJ_NUMBER_OF_BORDER; i++)
	    border_expose_handlers[i] = default_border_expose_handlers[i];
    }
}
#endif

static void border_none(BorderSkeleton* border)
{
    /* Do nothing */
}

static void border_normal_expose(BorderSkeleton* border)
{
    int i;

    XSetForeground(border->disp, border->gc, border->fg);
    for(i = 0; i < border->bd_width; i++)
	XDrawRectangle(border->disp, border->win, border->gc,
		       i, i, border->width - 1 - i, border->height - 1 - i);
}

static void border_convex_expose(BorderSkeleton* border)
{
    DrawBorder(border->disp, border->win, border->gc,
	       0, 0, border->width, border->height, border->bd_width,
	       border->lbg, border->hbg, True);
}

static void border_concave_expose(BorderSkeleton* border)
{
    DrawBorder(border->disp, border->win, border->gc,
	       0, 0, border->width, border->height, border->bd_width,
	       border->lbg, border->hbg, False);
}

static void border_event_handler(BorderSkeleton* border, XEvent* e)
{
    if(e->type == Expose && e->xexpose.count == 0 && border->bd_width > 0)
	default_border_expose_handlers[border->style](border);
}

WinObj* create_BorderSkeleton(WinObj* parent,
			      int x, int y, int width, int height,
			      int bd_width, int border_style)
{
    BorderSkeleton* border = (BorderSkeleton *)
	create_AnyWindow(parent, x, y,
			 width + 2 * bd_width, height + 2 * bd_width);

    if(border == NULL)
	return NULL;

    border->type = BorderSkeleton_T;
    border->bd_width = bd_width;
    border->style = border_style;
    SetEventMask((WinObj *)border, ExposureMask);
    SetEventHandler((WinObj *)border, border_event_handler);

    return (WinObj *)border;
}

void RedrawBorderSkeletion(WinObj* obj)
{
    BorderSkeleton* border = (BorderSkeleton *)obj;
    XClearWindow(obj->any.disp, obj->any.win);
    default_border_expose_handlers[border->style](border);
}
