#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>
#include "WinObj.h"

static void draw_button_label(ButtonWindow* b, Bool invert)
{
    Pixel fg;

    if(b->label == NULL)
	return;

    if(invert)
	fg = b->bg;
    else
	fg = b->fg;

    DrawString(b->disp, b->win, b->font, fg,
	       0, 0, b->width, b->height, b->label, Center);
    XSync(b->disp, False);
}

static void draw_pressed_button(ButtonWindow* b)
{
    XSetForeground(b->disp, b->gc, b->mark_color);
    XFillRectangle(b->disp, b->win, b->gc,
                   0, 0, b->width, b->height);
    draw_button_label(b, IsMonoDisplay(b));
    b->border_skeleton->style = WINOBJ_BORDER_CONCAVE;
    RedrawBorderSkeletion((WinObj *)b->border_skeleton);
    XSync(b->disp, False);
}

static void button_event_handler(ButtonWindow* obj, XEvent* e)
{
    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count > 0)
	    return;
	draw_button_label(obj, False);
	break;
      case ButtonPress:
	draw_pressed_button(obj);
	obj->enter_flag = True;
	obj->button_flag = True;

	break;
      case ButtonRelease:
	obj->border_skeleton->style = WINOBJ_BORDER_CONVEX;
	RedrawBorderSkeletion((WinObj *)obj->border_skeleton);
	if(obj->button_flag && obj->enter_flag)
	{
	    obj->button_flag = obj->enter_flag = False;
	    if(IsMonoDisplay(obj))
	    {
		XClearWindow(obj->disp, obj->win);
		draw_button_label(obj, False);
	    }
	    XFlush(obj->disp);
	    CallCallbackProc((WinObj *)obj, e);
	    if(obj->disp == NULL)
		return;
	}
	XClearWindow(obj->disp, obj->win);
	draw_button_label(obj, False);
	obj->button_flag = obj->enter_flag = False;
	break;
      case EnterNotify:
	if(obj->button_flag == True)
	{
	    obj->enter_flag = True;
	    draw_pressed_button(obj);
	}
	break;
      case LeaveNotify:
	if(obj->button_flag == True)
	{
	    obj->enter_flag = False;
	    obj->border_skeleton->style = WINOBJ_BORDER_CONVEX;
	    RedrawBorderSkeletion((WinObj *)obj->border_skeleton);
	    XClearWindow(obj->disp, obj->win);
	    draw_button_label(obj, False);
	}
	break;
    }
}

static void button_destruction(WinObj* obj)
{
    winobj_free(obj->button.label);
}

WinObj* create_Button(WinObj* parent,
		      int x,
		      int y,
		      int width,
		      int height,
		      int bd_width,
		      CharFont* font,
		      const char* label)
{
    ButtonWindow* b;

    if(width <= 0 || height <= 0)
    {
	const int gap = 4;

	width  = TextWidth(font, label) + 2 * gap;
	height = TextHeight(font)       + 2 * gap;
    }

    b = (ButtonWindow *)create_SimpleWindow(parent, x, y, width, height,
					    bd_width, WINOBJ_BORDER_CONVEX);
    if(b == NULL)
	return NULL;

    b->type = Button_T;
    b->font = font;
    if(label != NULL)
    {
	if((b->label = winobj_strdup(label)) == NULL)
	{
	    DestroyWinObj((WinObj *)b);
	    return NULL;
	}
    }
    else
	b->label = NULL;

    b->winobj_destruction = button_destruction;

    SetEventMask((WinObj *)b, ExposureMask | ButtonPressMask
		 | ButtonReleaseMask
		 | EnterWindowMask | LeaveWindowMask);

    SetEventHandler((WinObj *)b, button_event_handler);

    return (WinObj *)b;
}
