#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>
#include "WinObj.h"

#define SIDE_GAP 8

static void redraw_label(ChoiceButton* c, int select_number, Bool invert)
{
    Pixel fg;
    if(invert)
	fg = c->bg;
    else
	fg = c->fg;

    DrawString(c->disp, c->choice_win[select_number], c->font, fg,
	       0, 0, c->width_list[select_number], c->height,
	       c->choice_strings[select_number], Center);
}

void change_ChoiceButton(WinObj* obj, int select_number)
{
    ChoiceButton* c = (ChoiceButton *)obj;

    if(select_number == c->selected)
	return;

    XSetForeground(c->disp, c->gc, c->bg);
    XFillRectangle(c->disp, c->choice_win[c->selected], c->gc,
		   0, 0, c->width_list[c->selected], c->height);
    XSetForeground(c->disp, c->gc, c->mark_color);
    XFillRectangle(c->disp, c->choice_win[select_number], c->gc,
		   0, 0, c->width_list[select_number], c->height);

    redraw_label(c, c->selected, False);
    redraw_label(c, select_number, IsMonoDisplay(c));

    DrawBorder(c->disp, c->choice_win[c->selected], c->gc,
	       0, 0, c->width_list[c->selected], c->height,
	       c->bd_width, c->lbg, c->hbg,
	       True);
    DrawBorder(c->disp, c->choice_win[select_number], c->gc,
	       0, 0, c->width_list[select_number], c->height,
	       c->bd_width, c->lbg, c->hbg,
	       False);

    c->selected = select_number;
    XFlush(c->disp);
}

static void choice_event_handler(ChoiceButton* c, XEvent* e)
{
    int i;

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count > 0)
	    return;

	if(e->xany.window == c->win)
	    DrawString(c->disp, c->win, c->font, c->fg,
		       0, 0, 0, c->height, c->label_string, West);
	else
	{
	    for(i = 0; i < c->n; i++)
		if(c->choice_win[i] == e->xany.window)
		{
		    if(i == c->selected)
			XSetForeground(c->disp, c->gc, c->mark_color);
		    else
			XSetForeground(c->disp, c->gc, c->bg);
		    XFillRectangle(c->disp, c->choice_win[i], c->gc,
				   0, 0,
				   c->width_list[c->selected], c->height);
		    DrawBorder(c->disp, e->xany.window, c->gc,
			       0, 0, c->width_list[i], c->height,
			       c->bd_width, c->lbg, c->hbg,
			       i != c->selected);
		    redraw_label(c, i, IsMonoDisplay(c) && i == c->selected);
		}
	}
	break;
      case ButtonPress:
	for(i = 0; i < c->n; i++)
	    if(c->choice_win[i] == e->xany.window)
	    {
		change_ChoiceButton((WinObj *)c, i);
		if(c->call_back_proc != NULL)
		    c->call_back_proc((WinObj *)c, e);
		break;
	    }
    }
}

static void choice_button_destruction(WinObj* obj)
{
    ChoiceButton* c = (ChoiceButton *)obj;
    int i;

    for(i = 0; i < c->n; i++)
	winobj_free(c->choice_strings[i]);
    winobj_free(c->choice_strings);
    winobj_free(c->label_string);
    winobj_free(c->choice_win);
    winobj_free(c->width_list);
}

WinObj* create_ChoiceButton(WinObj* parent,
			    int x,
			    int y,
			    int bd_width,
			    int button_height,
			    CharFont* font,
			    char* label_string,
			    char** choice_strings,
			    int n,
			    int default_choice_string)
{
    ChoiceButton* c;
    int width;
    Display* disp;
    Window win;
    int start_x;
    int i;
    int wx;
    int* width_list;

    if((width_list = (int *)winobj_malloc(n * sizeof(int))) == NULL)
	return NULL;

    for(i = width = 0; i < n; i++)
    {
	int w;

	w = SIDE_GAP * 2 + TextWidth(font, choice_strings[i]) + 2 * bd_width;
	width_list[i] = w;
	width += w;
    }
    if(label_string != NULL)
	start_x = TextWidth(font, label_string);
    else
	start_x = 0;

    if(button_height == 0)
	button_height = TextHeight(font) + 2 * bd_width;
    else
	button_height += 2 * bd_width;

    c = (ChoiceButton *)create_AnyWindow(parent, x, y,
					 width + start_x,
					 button_height);
    if(c == NULL)
    {
	winobj_free(width_list);
	return NULL;
    }

    disp = c->disp;
    win  = c->win;

    c->choice_strings = (char **)winobj_malloc(n * sizeof(char *));
    c->choice_win = (Window *)winobj_malloc(n * sizeof(Window *));
    c->font = font;
    c->n = n;
    if(label_string != NULL)
	c->label_string = winobj_strdup(label_string);
    else
	c->label_string = winobj_null_string;

    if(c->choice_strings == NULL ||
       c->choice_win == NULL ||
       c->label_string == NULL)
    {
	winobj_free(c->choice_strings);
	winobj_free(c->choice_win);
	winobj_free(c->label_string);
	DestroyWinObj((WinObj *)c);
	return NULL;
    }

    c->selected = default_choice_string;
    c->width_list = width_list;
    c->bd_width = bd_width;

    wx = start_x;
    for(i = 0; i < n; i++)
    {
	if((c->choice_strings[i] = winobj_strdup(choice_strings[i])) == NULL)
	{
	    while(i--)
		winobj_free(c->choice_strings[i]);
	    winobj_free(c->choice_strings);
	    winobj_free(c->choice_win);
	    winobj_free(c->label_string);
	    DestroyWinObj((WinObj *)c);
	    return NULL;
	}

	c->choice_win[i] = XCreateSimpleWindow(disp, win,
					       wx, 0,
					       width_list[i],
					       button_height,
					       0, 0, c->bg);

	XSelectInput(disp, c->choice_win[i], ButtonPressMask | ExposureMask);
	install_event_handler(disp, c->choice_win[i], (WinObj *)c);
	XMapWindow(disp, c->choice_win[i]);
	wx += width_list[i];
    }

    c->winobj_destruction = choice_button_destruction;
    SetEventMask((WinObj *)c, ExposureMask);
    SetEventHandler((WinObj *)c, choice_event_handler);

    return (WinObj *)c;
}
