#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <unistd.h>
#include "WinObj.h"

#define GAP 8

static void draw_menu_title(DrawerPopupMenu* menu, int ih)
{
    int bd = menu->bd_width;
    DrawRuledHLine(menu->disp, menu->win, menu->gc, menu->lbg, menu->hbg,
		   bd + GAP/2, ih + GAP/2, menu->width - 2 * bd - GAP);

    DrawString(menu->disp, menu->win, menu->font, menu->fg,
	       0, bd, menu->width, ih + GAP/2 - bd, menu->title, Center);
}


static void draw_all_menu_items(DrawerPopupMenu* menu,
				Bool initialize_background)
{
    Window win = menu->win;
    Display* disp = menu->disp;

    if(initialize_background)
    {
	int bd = menu->bd_width;
	int i;
	int item_height = menu->item_height;
	int n = menu->nitems;

	menu->win = menu->popup_background;
	DrawBorder(disp, menu->popup_background, menu->gc,
		   0, 0, menu->width, menu->height, bd,
		   menu->lbg, menu->hbg, True);
	if(menu->title != NULL)
	    draw_menu_title(menu, item_height);
	for(i = 0; i < n; i++)
	    menu->draw_item_proc((WinObj *)menu, i, False);
	menu->win = win;
    }
    else
	XCopyArea(disp, menu->popup_background, win, menu->gc,
		  0, 0, menu->width, menu->height, 0, 0);
}


static void drawer_move_popup_window(DrawerPopupMenu* menu, int x, int y)
{
    Display* disp = menu->disp;
    int scr = DefaultScreen(disp);
    int w = menu->width;
    int h = menu->height;

    if(x < 0)
	x = 0;
    if(y < 0)
	y = 0;

    else if(x + w >= DisplayWidth(disp, scr))
	x = DisplayWidth(disp, scr) - w - 1;

    if(y + h >= DisplayHeight(disp, scr))
	y = DisplayHeight(disp, scr) - h - 1;
    MoveWinObj((WinObj *)menu, x, y);
}

static void animation_drawer_map(DrawerPopupMenu* menu)
{
    int i, j;
    Bool call_hook;

    if(menu->nitems == 0)
	return;

    if(menu->selecting_hook != NULL)
	call_hook = !menu->selecting_hook((PopupMenuBase *)menu);
    else
	call_hook = False;

    for(i = menu->item_height + menu->bd_width, j = 0;
	i < menu->height;
	i += menu->item_height, j++)
    {
	XResizeWindow(menu->disp, menu->win, menu->width, i);
	XCopyArea(menu->disp, menu->popup_background, menu->win, menu->gc,
		  0, menu->height - i - 1,
		  menu->width, i,
		  0, 0);
	DrawBorder(menu->disp, menu->win, menu->gc, 0, 0, menu->width, i,
		   menu->bd_width, menu->lbg, menu->hbg, True);
	XSync(menu->disp, False);
	if(call_hook)
	{
	    if(j & 1)
		call_hook = !menu->selecting_hook((PopupMenuBase *)menu);
	}
	else
	    usleep(100000 / menu->nitems);
    }
}

static void animation_drawer_unmap(DrawerPopupMenu* menu)
{
    int i, j;
    Bool call_hook;

    if(menu->nitems == 0)
	return;

    if(menu->selecting_hook != NULL)
	call_hook = !menu->selecting_hook((PopupMenuBase *)menu);
    else
	call_hook = False;
    for(i = menu->height - menu->item_height, j = 0;
	i > 0;
	i -= menu->item_height, j++)
    {
	XResizeWindow(menu->disp, menu->win, menu->width, i);
	XCopyArea(menu->disp, menu->popup_background, menu->win, menu->gc,
		  0, menu->height - i - 1,
		  menu->width, i,
		  0, 0);
	DrawBorder(menu->disp, menu->win, menu->gc, 0, 0, menu->width, i,
		   menu->bd_width, menu->lbg, menu->hbg, True);
	XSync(menu->disp, False);
	if(call_hook && (j & 1))
	    call_hook = !menu->selecting_hook((PopupMenuBase *)menu);
    }
}

static void drawer_popup_map_proc(WinObj* win_obj, int x, int y)
{
    DrawerPopupMenu* menu = (DrawerPopupMenu *)win_obj;
    Display* disp = menu->disp;
    Window win = menu->win;

    XResizeWindow(disp, win, menu->width, 1);
    drawer_move_popup_window(menu, x, y);
    XRaiseWindow(disp, win);
    XClearWindow(disp, win);
    XMapWindow(disp, win);

    if(menu->update_flag)
    {
	if(menu->popup_background != None)
	    XFreePixmap(disp, menu->popup_background);
	menu->popup_background =
	    CreatePixmap(disp, win, menu->width, menu->height, menu->bg);
	draw_all_menu_items(menu, True);
    }
    else
    {
	if(menu->popup_background == None)
	{
	    menu->popup_background =
		CreatePixmap(disp, win, menu->width, menu->height, menu->bg);
	    draw_all_menu_items(menu, True);
	}
    }
    animation_drawer_map(menu);
    ResizeWinObj((WinObj *)menu, menu->width, menu->height);
    draw_all_menu_items(menu, False);
}

static void drawer_popup_unmap_proc(WinObj* obj)
{
    DrawerPopupMenu* menu = (DrawerPopupMenu *)obj;
    if(menu->hierarchical_target)
	unmap_popup_menu_internal(menu->hierarchical_target);
    animation_drawer_unmap(menu);
    XUnmapWindow(menu->disp, menu->win);
    menu->popup_flag = False;
    menu->hierarchical_target = NULL;
    menu->selected_item_index = -1;
}

WinObj* create_DrawerPopupMenu(Display* disp,
			   const char* title,
			   int bd_width,
			   Pixel fg,
			   Pixel bg,
			   CharFont* font)
{
    DrawerPopupMenu* menu = (DrawerPopupMenu *)
	create_PopupMenu(disp, title, bd_width, fg, bg, font);

    if(menu == NULL)
	return NULL;

    menu->type = DrawerPopupMenu_T;
    menu->popup_map_proc = drawer_popup_map_proc;
    menu->popup_unmap_proc = drawer_popup_unmap_proc;

    return (WinObj *)menu;
}
