#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#if !defined(sun) && (defined(SVR4) || defined(SYSTYPE_SVR4) || defined(__svr4__))
#include <sys/dir.h>
#define dirent direct
#else
#include <dirent.h>
#endif

#include <sys/param.h>
#include <unistd.h>
#include <fcntl.h>
#include "WinObj.h"
#include "mblock.h"

#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif

//#ifdef __linux__
#if 0
#include <linux/unistd.h> /* for __NR_fchdir */
extern	int syscall __P ((int __number, ...));
#define fchdir(f) syscall(__NR_fchdir, f)
#endif

static char match_buffer[MAXPATHLEN];
static MBlockList name_buffer;

struct dir_cache_t
{
    char* dirname;
    int n;
    char** fnames;
    char* name_buff;
    time_t dir_ctime;
    long ino;
    struct dir_cache_t* next;
};

#define IsEndOfLine(text) ((text)->cur == (text)->textlen)

#define DEFAULT_COMPLETION_BUFFER_SIZE (1024*8)
#define DEFAULT_NAME_LIST_SIZE         128

static struct dir_cache_t* dir_cache = NULL;

static void load_short_cut(PopupMenuBase* menu, char* name, int id)
{
    SetLineText((LineText *)menu->ext_data, name, strlen(name));
}

Bool AddDirLineTextShortCut(WinObj* text, const char* fname)
{
    MCPopupMenu *p = ((DirLineText *)text)->short_cut_list;
    int i;
    char* expn_name;

    if(p == NULL || fname[0] == '\0') 
	return False;
    if(!strcmp(fname, "-"))
	return False;

    expn_name = ExpandHomeDirectory(fname);
    for(i = 0; i < p->nitems; i++)
	if(!strcmp(expn_name, p->name_list[i]))
	    return False;
    return add_MCPopupMenu((WinObj *)p, expn_name, 0, load_short_cut);
}

void SetDirLineTextFileSelector(WinObj* text,
				Bool (* dir_file_selector)
						(const char* pathname))
{
    ((DirLineText *)text)->dir_file_selector = dir_file_selector;
}

void ClearDirLineTextCache(void)
{
    while(dir_cache)
    {
	struct dir_cache_t* p;

	p = dir_cache;
	dir_cache = dir_cache->next;
	if(p->dirname != NULL)
	{
	    winobj_free(p->name_buff);
	    winobj_free(p->fnames);
	    winobj_free(p->dirname);
	}
	winobj_free(p);
    }
}

static Bool directory_info(const char* dirname, time_t* ct, long* ino)
{
    struct stat stb;
    char* p;

    if((p = ExpandHomeDirectory(dirname)) == NULL)
	return False;

    if(stat(p, &stb) < 0)
    {
	CallWinObjSysErrorHandler(p);
	return False;
    }
    if(!S_ISDIR(stb.st_mode))
	return False;
    *ct = stb.st_ctime;
    *ino = (long)stb.st_ino;
    return True;
}

static struct dir_cache_t* search_dir_cache(const char* dirname,
					    time_t ct, long ino)
{
    struct dir_cache_t* p;

    for(p = dir_cache; p; p = p->next)
    {
	if(p->dirname == NULL)
	    continue;
	if(ino == p->ino || !strcmp(p->dirname, dirname))
	{
	    if(p->dir_ctime == ct)
		return p; /* found */

	    winobj_free(p->name_buff);
	    winobj_free(p->fnames);
	    winobj_free(p->dirname);
	    p->dirname = NULL;
	    return NULL;
	}
    }
    return NULL;
}

static Bool check_is_directory(const char* pathname)
{
    struct stat stb;

    if(stat(pathname, &stb) < 0)
    {
	CallWinObjSysErrorHandler(pathname);
	return False;
    }

    return S_ISDIR(stb.st_mode);
}

static int compare_proc(const char** s1, const char** s2)
{
    return strcmp(*s1, *s2);
}

static char** ScanDir(const char* dirname, int* n)
{
    struct dir_cache_t* p;
    struct name_list
    {
	char* name;
	int namelen;
	struct name_list* next;
    }* name_list;
    int dirlen, nfiles, name_buff_len, i;
    DIR *dirp;
    struct dirent *d;
    char* new_dirname;
    int cwd_fd;			/* current working directory */
    time_t dir_ctime;
    long ino;

    if(dirname == NULL)
    {
	new_dirname = new_segment(&name_buffer, 2);
	new_dirname[0] = '/';
	new_dirname[1] = '\0';
	dirlen = 2;
    }

    if((dirname = (const char*)ExpandHomeDirectory(dirname)) == NULL)
	return NULL;

    dirlen = strlen(dirname);
    if(dirlen == 0)
    {
	new_dirname = new_segment(&name_buffer, 2);
	new_dirname[0] = '.';
	new_dirname[1] = '\0';
	dirlen = 2;
    }
    else
    {
	new_dirname = new_segment(&name_buffer, dirlen + 1);
	memcpy(new_dirname, dirname, dirlen);
	if(dirlen > 1 && dirname[dirlen - 1] == '/')
	    new_dirname[--dirlen] = '\0';
	else
	    new_dirname[dirlen] = '\0';
    }

    if(directory_info(new_dirname, &dir_ctime, &ino) == False)
    {
	reuse_mblock(&name_buffer);
	return NULL;
    }

    p = search_dir_cache(new_dirname, dir_ctime, ino);
    if(p != NULL)
    {
	*n = p->n;
	reuse_mblock(&name_buffer);
	return p->fnames;
    }

    if((cwd_fd = open(".", O_RDONLY)) < 0)
    {
	CallWinObjSysErrorHandler(new_dirname);
	reuse_mblock(&name_buffer);
	return NULL;
    }

    if(chdir(new_dirname) < 0)
    {
	CallWinObjSysErrorHandler(new_dirname);
	reuse_mblock(&name_buffer);
	fchdir(cwd_fd);
	close(cwd_fd);
	return NULL;
    }

    if((dirp = opendir(".")) == NULL)
    {
	CallWinObjSysErrorHandler(new_dirname);
	reuse_mblock(&name_buffer);
	return NULL;
    }

    nfiles = 0;
    name_buff_len = 0;
    name_list = NULL;
    while((d = readdir(dirp)) != NULL)
    {
	struct name_list* node;
	int dlen;

	dlen = strlen(d->d_name);
	node = new_segment(&name_buffer, sizeof(struct name_list));
	if(check_is_directory(d->d_name))
	{
	    node->name = new_segment(&name_buffer, dlen + 1);
	    memcpy(node->name, d->d_name, dlen);
	    node->name[dlen] = '/';
	    node->namelen = dlen + 1;
	    name_buff_len += dlen + 2;
	}
	else
	{
	    node->name = new_segment(&name_buffer, dlen);
	    memcpy(node->name, d->d_name, dlen);
	    node->namelen = dlen;
	    name_buff_len += dlen + 1;
	}
	node->next = name_list;
	name_list = node;
	nfiles++;
    }

    if(fchdir(cwd_fd) < 0)
    {
	perror("fchdir");
	exit(1);		/* FIXME */
    }
    close(cwd_fd);

    if(name_list == NULL)
    {
	reuse_mblock(&name_buffer);
	return NULL;
    }

    if((p = (struct dir_cache_t *)winobj_malloc(sizeof(struct dir_cache_t)))
       == NULL)
    {
	reuse_mblock(&name_buffer);
	return NULL;
    }

    p->n = nfiles;
    p->dir_ctime = dir_ctime;
    p->ino = ino;
    if((p->dirname = (char *)winobj_strdup(new_dirname)) == NULL)
    {
	winobj_free(p);
	reuse_mblock(&name_buffer);
	return NULL;
    }

    if((p->fnames = (char **)winobj_malloc(nfiles * sizeof(char *))) == NULL)
    {
	winobj_free(p->dirname);
	winobj_free(p);
	reuse_mblock(&name_buffer);
	return NULL;
    }

    if((p->name_buff = (char *)winobj_malloc(name_buff_len)) == NULL)
    {
	winobj_free(p->dirname);
	winobj_free(p->fnames);
	winobj_free(p);
	reuse_mblock(&name_buffer);
	return NULL;
    }

    i = nfiles = 0;
    while(name_list)
    {
	p->fnames[nfiles] = p->name_buff + i;
	memcpy(p->fnames[nfiles], name_list->name, name_list->namelen);
	p->fnames[nfiles][name_list->namelen] = '\0';
	nfiles++;
	i += name_list->namelen + 1;
	name_list = name_list->next;
    }

    reuse_mblock(&name_buffer);
    qsort((char *)p->fnames, nfiles, sizeof(char *),
	  (int (*)(const void *, const void *))compare_proc);
    p->next = dir_cache;
    dir_cache = p;
    *n = nfiles;

    return p->fnames;
}

static Bool is_prefix(const char* src, const char* tar)
{
    if(*src == '\0')
	return True;
    while(*src != '\0')
	if(*src++ != *tar++)
	    return False;
    return True;
}

static void prefix_matching(char* match_buffer, const char* name)
{
    while(*match_buffer)
    {
	if(*match_buffer++ != *name++)
	{
	    match_buffer[-1] = '\0';
	    return;
	}
    }
}

static void add_file_name(PopupMenuBase* menu, char* name, int id)
{
    WinObj* text = menu->ext_data;
    char path[BUFSIZ];

    GetLineText((LineText *)text, path, BUFSIZ);
    path[text->dirtext.end + 1] = '\0';
    if(text->dirtext.end + 1 + strlen(name) >= MAXPATHLEN)
    {
	CallWinObjErrorHandler(WOE_FNAME_TOO_LONG, NULL);
	return;
    }
    strcat(path, name);
    SetLineText((LineText *)text, path, strlen(path));
    while(LineTextForwardChar((LineText *)text))
	;
    ((DirLineText *)text)->completion_failure = 0;
}

static void DirLineText_Popup(DirLineText* text, char** name_list, int n,
			      const char* prefix)
{
    int i, orig_end;
    int root_pointer_x = 0, root_pointer_y = 0;
    Bool keyflag = False;
    Display* disp = text->disp;

    if(prefix == NULL)
    {
	text->dir->nitems = n;
	text->dir->name_list = name_list;
    }
    else
    {
	int j, e;

	orig_end = text->end;
	text->dir->name_list =
	    (char **)new_segment(&name_buffer, n * sizeof(char *));

	for(i = j = 0; i < n; i++)
	    if(is_prefix(prefix, name_list[i]))
		text->dir->name_list[j++] = name_list[i];
	text->dir->nitems = n = j;

	for(e = text->cur; e >= 0 && text->text[e] != '/'; e--)
	    ;
	text->end = e;
	keyflag = True;
	QueryPointer(disp, DefaultRootWindow(disp),
                     &root_pointer_x, &root_pointer_y);
    }

    text->dir->idlist = (int *)new_segment(&name_buffer, n * sizeof(int));
    text->dir->selected_proc_list =
	(void (**)(struct _PopupMenuBase *, char *, int))
	    new_segment(&name_buffer,
			n * sizeof(void
				   (*)(struct _PopupMenuBase *, char *, int)));
    for(i = 0; i < n; i++)
	text->dir->selected_proc_list[i] = add_file_name;

    LineTextFocusOut((LineText *)text);
    if(n > 100)
	MCPopupMenu_popup_proc((WinObj *)((DirLineText *)text)->dir, -1);
    else
	MCPopupMenu_popup_proc((WinObj *)((DirLineText *)text)->dir, 25);
    LineTextFocusIn((LineText *)text);
    reuse_mblock(&name_buffer);
    text->dir->nitems = 0;
    text->dir->name_list = NULL;
    text->dir->idlist = NULL;
    text->dir->selected_proc_list = NULL;

    if(keyflag == True)
    {
	XWarpPointer(disp, None, DefaultRootWindow(disp), 0, 0, 0, 0,
		     root_pointer_x, root_pointer_y);
	XSync(disp, False);
	text->end = orig_end;
    }
}

static void completion_filename_matching(LineText* text,
					 char* src,
					 char** name_list,
					 int n)
{
    int i, src_len;
    int nmatch;

    src_len = strlen(src);

/*
    if(n == 2 && src_len == 0 && !strcmp(name_list[0], "../"))
	strcpy(match_buffer, name_list[1]);
    else if(n == 2 && src_len == 0 && !strcmp(name_list[1], "../"))
	strcpy(match_buffer, name_list[0]);
    else
*/

    {
	nmatch = 0;
	for(i = 0; i < n; i++)
	{
	    if(src_len == 0 && name_list[i][0] == '.')
		continue;
	    if(!is_prefix(src, name_list[i]))
		continue;

	    if(nmatch == 0)
		strcpy(match_buffer, name_list[i]);
	    else
		prefix_matching(match_buffer + src_len,
				name_list[i] + src_len);
	    nmatch++;
	}

	if(nmatch == 0 || src_len == strlen(match_buffer))
	{
	    if(nmatch == 0)
	    {
		XBell(text->disp, 0);
		XSync(text->disp, False);
	    }
	    else if(nmatch >= 2)
	    {
		((DirLineText *)text)->completion_failure++;
		if(((DirLineText *)text)->completion_failure == 2)
		{
		    DirLineText_Popup((DirLineText *)text, name_list, n,
				      match_buffer);
		    ((DirLineText *)text)->completion_failure = 0;
		}
	    }

	    return;
	}
    }

    while(LineTextForwardChar(text))
	;
    for(i = 0; i < src_len; i++)
    {
	LineTextBackwardChar(text);
	LineTextDeleteChar(text);
    }

    for(i = 0; match_buffer[i]; i++)
	LineTextInsertChar(text, match_buffer[i]);
}

static void completion_filename(LineText* text)
{
    char dirname[BUFSIZ];
    char matched[256];
    char** fnames;
    int i, j, n, dirlen;

    GetLineText(text, dirname, BUFSIZ);
    dirlen = strlen(dirname);

    for(i = dirlen - 1; i > 0 && dirname[i] != '/'; i--)
	;
    matched[0] = '\0';

    if(dirname[i] == '/')
    {
	for(j = i + 1; j < dirlen; j++)
	    matched[j - i - 1] = dirname[j];
	matched[j - i - 1] = '\0';
	dirname[i + 1] = '\0';
	if((fnames = ScanDir(dirname, &n)) == NULL)
	    return;
    }
    else
    {
	strcpy(matched, dirname);
	if((fnames = ScanDir(".", &n)) == NULL)
	    return;
    }

    completion_filename_matching(text, matched, fnames, n);
}

static void obtain_filename_region(const char* path, int pos,
			    int* start_ret, int* end_ret)
{
    int start, end;

    if(path[0] == '\0')
    {
	start = 0;
	end = -1;
    }
    else if((pos <= 0 && path[0] == '/') || !strcmp(path, "/"))
    {	/* root */
	start = 0;
	end = 0;
    }
    else
    {
	int len = strlen(path);

	if(pos >= len)
	    pos = len - 1;
	else if(pos < 0)
	    pos = 0;

	if(path[pos] == '/' && path[pos] == '/')
	    pos--;
	end = start = pos;

	while(path[end] != '\0' && path[end] != '/')
	    end++;
	while(start > 0 && path[start] != '/')
	    start--;

	if(path[end] != '/')
	{
	    if(path[start] != '/') /* only specified filename*/
	    {
		start = 0;
		end = -1;
	    }
	    else if(start == 0) /* /foo (root)*/
	    {
		start = 0;
		end = 0;
	    }
	    else		/* ???/foo */
	    {
		end = start;
		if(start > 0)
		{
		    start--;
		    while(start > 0 && path[start] != '/')
			start--;
		}
		if(path[start] == '/')
		    start++;
	    }
	}
	else			/*  ???/ */
	{
	    if(path[start] == '/') /* ???/foo/ */
		start++;
	}
    }
    *start_ret = start;
    *end_ret = end;
}

static void clear_dir_mark(DirLineText* text, char* path)
{
    int start, end, d = 0, s = 0;

    start = text->start - text->start_char;
    end   = text->end - text->start_char;

    if(text->end < 0)
	return;

    if(end > text->show_chars)
	s = end - text->show_chars;

    if(start < 0)
    {
	d = start;
	start = 0;
    }

    if(end < start)
	return;

    XClearArea(text->disp, text->win, 
	       text->start_origin + text->char_width * start,
	       0,
	       (end - start + 1) * text->char_width,
	       text->height,
	       False);

    DrawStringN(text->disp, text->win, text->font, text->fg,
		text->start_origin + start * text->char_width, 0,
		0, text->height,
		path + text->start - d, end - start + 1 - s, West);

    if(text->start <= text->cur && text->cur <= text->end)
	text->draw_cursor((LineText *)text);
}

static void draw_dir_mark(DirLineText* text, char* path)
{
    int start, end, d = 0, s = 0;

    start = text->start - text->start_char;
    end   = text->end - text->start_char;

    if(end > text->show_chars)
	s = end - text->show_chars;

    if(start < 0)
    {
	d = start;
	start = 0;
    }
    if(end < start)
	return;

    if(end >= 0)
    {
	DrawBorder(text->disp, text->win, text->gc,
		   text->start_origin + text->char_width * start,
		   text->bd_width + 1,
		   (end - start + 1) * text->char_width - 1,
		   text->height - text->bd_width * 2 - 2,
		   text->bd_width, text->lbg, text->hbg, True);

	DrawStringN(text->disp, text->win, text->font, text->fg,
		    text->start_origin + start * text->char_width, 0,
		    0, text->height,
		    path + text->start - d, end - start + 1 - s, West);
	if(text->start <= text->cur && text->cur <= text->end)
	    text->draw_cursor((LineText *)text);
    }
}

static void redraw_proc(DirLineText* text, int pointer_x)
{
    char path[BUFSIZ];
    int pos;
    int start, end;

    pos = (int)floor((pointer_x - text->start_origin)
		     / (double)text->char_width) + text->start_char;
    GetLineText((LineText *)text, path, BUFSIZ);
    obtain_filename_region(path, pos, &start, &end);

    if(text->start != start || text->end != end)
    {
	clear_dir_mark(text, path);
	text->start = start;
	text->end = end;
    }
    draw_dir_mark(text, path);

    if(IsEndOfLine(text))
    {
	int n;

	n = text->show_chars - text->cur + text->start_char;
	if(n > 0)
	    XClearArea(text->disp, text->win, 
		       text->start_origin +
		       (text->cur - text->start_char + 1) * text->char_width,
		       0,
		       n * text->char_width, text->height, False);
	text->draw_cursor((LineText *)text);
    }
}

static void key_storke_proc(DirLineText* text, XEvent* e)
{
    char path[BUFSIZ];
    int pos;
    int start, end;

    pos = (int)floor((e->xkey.x - text->start_origin)
		     / (double)text->char_width) + text->start_char;

    GetLineText((LineText *)text, path, BUFSIZ);

    obtain_filename_region(path, pos, &start, &end);
    clear_dir_mark(text, path);
    text->start = start;
    text->end = end;
    text->key_press_handler((LineText *)text, e);
    draw_dir_mark(text, path);

    if(IsEndOfLine(text))
    {
	XClearArea(text->disp, text->win, 
		   text->start_origin + text->char_width
		   * (text->cur + 1 - text->start_char),
		   0, 0, 0, False);
	text->draw_cursor((LineText *)text);
    }
}

static void move_dir_line_text_cursor(DirLineText* text, int x)
{
    char path[BUFSIZ];
    int pos;
    int start, end;

    pos = (int)floor((x - text->start_origin)
		     / (double)text->char_width) + text->start_char;
    obtain_filename_region(path, pos, &start, &end);
    clear_dir_mark(text, path);
    text->start = start;
    text->end = end;

    while(pos < text->cur && LineTextBackwardChar((LineText *)text))
	;
    while(pos > text->cur && LineTextForwardChar((LineText *)text))
	;
}

static Bool default_text_file_selector(const char* pathname)
{
    return True;
}

static void text_button_press_proc(DirLineText* text, XEvent* e)
{
    int n;
    char buff[BUFSIZ];
    int orig_end;
    int root_pointer_x = 0, root_pointer_y = 0;
    Bool keyflag = False;
    Display* disp = text->disp;
    char** fnames;

    orig_end = text->end;
    GetLineText((LineText *)text, buff, BUFSIZ);

    if(e->type == KeyPress)
    {
	int e;

	for(e = text->cur; e >= 0 && buff[e] != '/'; e--)
	    ;
	text->end = e;
	keyflag = True;
	QueryPointer(disp, DefaultRootWindow(disp),
                     &root_pointer_x, &root_pointer_y);
    }
    if(text->end < 0)
    {
	buff[0] = '.';
	buff[1] = '/';
	buff[2] = '\0';
    }
    else
    {
	buff[text->end + 1] = '\0';
    }

    if((fnames = ScanDir(buff, &n)) == NULL)
    {
	XBell(disp, 0);
	XSync(disp, False);
	return;
    }

    DirLineText_Popup(text, fnames, n, NULL);

    if(keyflag == True)
    {
	XWarpPointer(disp, None, DefaultRootWindow(disp), 0, 0, 0, 0,
		     root_pointer_x, root_pointer_y);
	XSync(disp, False);
	text->end = orig_end;
    }
}

static void dir_line_text_handler(LineText* text, XEvent* e)
{
    static Bool drag_flag = False;
    int x, y;

    switch(e->type)
    {
    case Expose:
      if(e->xexpose.count == 0)
	{
	  char path[BUFSIZ];

	  GetLineText((LineText *)text, path, BUFSIZ);
	  LineTextRedraw(text);
	  draw_dir_mark((DirLineText *)text, path);
	}
      break;
      case MappingNotify:
	XRefreshKeyboardMapping(&e->xmapping);
	break;
      case EnterNotify:
	LineTextFocusIn(text);
	redraw_proc((DirLineText *)text, e->xcrossing.x);
	break;
      case LeaveNotify:
	LineTextFocusOut(text);
	break;
      case KeyPress:
	{
	    KeySym sym;

	    sym = XKeycodeToKeysym(text->disp, e->xkey.keycode, 0);
	    if(sym == XK_Tab
#ifdef XK_KP_Tab
	       || sym == XK_KP_Tab
#endif
	       )
	    {
		if(IsEndOfLine(text))
		    completion_filename(text);
		else
		{
		    XBell(text->disp, 0);
		    XSync(text->disp, False);
		}
	    }
	    else
	    {
		key_storke_proc((DirLineText *)text, e);
		((DirLineText *)text)->completion_failure = 0;
	    }
	    redraw_proc((DirLineText *)text, e->xkey.x);
	}
	break;
      case MotionNotify:
	if(drag_flag == True)
	    move_dir_line_text_cursor((DirLineText *)text, e->xmotion.x);
	redraw_proc((DirLineText *)text, e->xmotion.x);
	break;
      case ButtonPress:
	switch(e->xbutton.button)
	{
	  case Button1:
	    move_dir_line_text_cursor((DirLineText *)text, e->xmotion.x);
	    redraw_proc((DirLineText *)text, e->xmotion.x);
	    drag_flag = True;
	    break;
	  case Button2:
	    if(((DirLineText *)text)->short_cut_list == NULL)
		break;
	    ((DirLineText *)text)->short_cut_list->ext_data = (void *)text;
	    MCPopupMenu_popup_proc((WinObj *)((DirLineText *)text)->
				   short_cut_list, 25);
	    break;
	  case Button3:
	    ((DirLineText *)text)->completion_failure = 0;

	    LineTextFocusOut(text);
	    text_button_press_proc((DirLineText *)text, e);
	    QueryPointer(text->disp, text->win, &x, &y);
	    redraw_proc((DirLineText *)text, x);
	    LineTextFocusIn(text);

	    break;
	}
	break;
      case ButtonRelease:
	drag_flag = False;
	break;
      case SelectionClear:
      case SelectionRequest:
      case SelectionNotify:
	text->selection_event_handler(text, e);
	break;
    }
}

void SetPopupSelectingHook(WinObj* obj,
			   Bool (* selecting_hook)(PopupMenuBase* menu))
{
    DirLineText* text = (DirLineText *)obj;

    if(text->short_cut_list != NULL)
	text->short_cut_list->selecting_hook = selecting_hook;
    text->dir->selecting_hook = selecting_hook;
}

void SetDirLineTextShortCut(DirLineText* dirtext, MCPopupMenu* menu)
{
    dirtext->short_cut_list = menu;
}

static void dirline_text_destruction(WinObj* obj)
{
    DirLineText* text = (DirLineText *)obj;
    winobj_free(text->prompt);
    winobj_free(text->text);
    winobj_free(text->seldata->buffer);
    winobj_free(text->seldata);
    DestroyWinObj((WinObj *)text->dir);
}

WinObj* create_DirLineText(WinObj* parent,
			   int x,
			   int y,
			   int width,
			   int height,
			   int bd_width,
			   CharFont* font,
			   const char* prompt)
{
    DirLineText* text;

    text = (DirLineText *)create_LineText(parent, x, y, width, height,
					  bd_width, font, prompt);

    if(text == NULL)
	return NULL;

    text->type = DirLineText_T;
    text->bd_width = bd_width;
    text->start = 0;
    text->end = -1;
    text->dir_file_selector = default_text_file_selector;
    text->short_cut_list = NULL;

    text->dir
	= (MCPopupMenu *)create_MCPopupMenu(text->disp, NULL, bd_width,
					    text->fg, text->bg, font);

    if(text->dir == NULL)
    {
	DestroyWinObj((WinObj *)text->dir);
	DestroyWinObj((WinObj *)text);
	return NULL;
    }

    text->dir->ext_data = (void *)text;
    text->completion_failure = 0;

    SetEventMask((WinObj *)text, KeyPressMask | ButtonPressMask |
		 ButtonReleaseMask | EnterWindowMask | LeaveWindowMask |
		 PointerMotionMask | ExposureMask);
		 
    SetEventHandler((WinObj *)text, dir_line_text_handler);

    text->winobj_destruction = dirline_text_destruction;

    return (WinObj *)text;
}
