#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "WinObj.h"

static void redraw_gp_bar(GraphPanel* gp)
{
    int x, y;

    XSetForeground(gp->disp, gp->gc, gp->bg);
    for(x = 0; x < gp->width; x++)
    {
	y = gp->gpbuff[x];
	XDrawLine(gp->disp, gp->win, gp->gc,
		  x, 0, x, gp->height - y - 1);
    }

    XSetForeground(gp->disp, gp->gc, gp->fg);
    for(x = 0; x < gp->width; x++)
    {
	y = gp->gpbuff[x];
	if(y > 0)
	{
	    XDrawLine(gp->disp, gp->win, gp->gc,
		      x, gp->height - y - 1,
		      x, gp->height - 1);
	}
    }
}

static void set_gp_bar(GraphPanel* gp, int x, int y)
{
    int x0, x1;

    if(x < 0 || x >= gp->width)
	return;

    y = (y + gp->yresolution) / gp->yresolution * gp->yresolution;
    if(y < 0)
	y = 0;
    else if(y >= gp->height)
	y = gp->height - 1;

    x0 = x - x % gp->xresolution;
    x1 = x0 + gp->xresolution;
    if(x1 > gp->width)
	x1 = gp->width;

    XSetForeground(gp->disp, gp->gc, gp->bg);
    for(x = x0; x < x1; x++)
    {
	if(gp->gpbuff[x] > y)
	    XDrawLine(gp->disp, gp->win, gp->gc,
		      x, gp->height - gp->gpbuff[x] - 1,
		      x, gp->height - y - 2);
    }

    if(y > 0)
    {
	XSetForeground(gp->disp, gp->gc, gp->fg);
	for(x = x0; x < x1; x++)
	{
	    if(gp->gpbuff[x] < y)
		XDrawLine(gp->disp, gp->win, gp->gc,
			  x, gp->height - y - 1,
			  x, gp->height - gp->gpbuff[x] - 1);
	}
    }
    for(x = x0; x < x1; x++)
	gp->gpbuff[x] = y;
}

static void set_gp_bar_line(GraphPanel* gp, int x1, int y1, int x2, int y2)
{
    int x;
    double a;

    if(x1 == x2)
    {
	set_gp_bar(gp, x1, y1);
	return;
    }

    if(x2 < x1)
    {
	int tmp;

	tmp = x2;
	x2 = x1;
	x1 = tmp;

	tmp = y2;
	y2 = y1;
	y1 = tmp;
    }

    a = (y2 - y1) / (double)(x2 - x1);

    for(x = x1; x < x2; x += gp->xresolution)
	set_gp_bar(gp, x, y1 + (int)(a * (x - x1)));
    set_gp_bar(gp, x2, y2);
}

void gp_event_handler(WinObj* obj, XEvent* e)
{
    GraphPanel* gp = (GraphPanel *)obj;

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count == 0)
	    redraw_gp_bar(gp);
	if(gp->call_back_proc != NULL)
	    gp->call_back_proc((WinObj *)gp, e);
	break;
      case ButtonPress:
	gp->prev_x = gp->start_x = e->xbutton.x;
	gp->prev_y = gp->start_y = gp->height - e->xbutton.y;
	gp->prev_flag = True;
	set_gp_bar(gp, gp->prev_x, gp->prev_y);
	if(gp->call_back_proc != NULL)
	    gp->call_back_proc((WinObj *)gp, e);
	break;
      case MotionNotify:
	if(gp->prev_flag == True)
	{
	    if(e->xmotion.state & Button1Mask)
	    {
		set_gp_bar_line(gp,
				e->xbutton.x,
				gp->height - e->xbutton.y,
				gp->prev_x,
				gp->prev_y);
	    }
	    else if (e->xmotion.state & Button2Mask)
	    {
		CompressMotionEvent(e->xany.display, e->xany.window, e);
		set_gp_bar_line(gp,
				e->xbutton.x,
				gp->height - e->xbutton.y,
				gp->prev_x,
				gp->prev_y);
		set_gp_bar_line(gp,
				e->xbutton.x,
				gp->height - e->xbutton.y,
				gp->start_x,
				gp->start_y);
	    }
	}
	else
	{
	    set_gp_bar(gp,
		       e->xbutton.x,
		       gp->height - e->xbutton.y);
	}

	if(gp->call_back_proc != NULL)
	    gp->call_back_proc((WinObj *)gp, e);

	gp->prev_x = e->xbutton.x;
	gp->prev_y = gp->height - e->xbutton.y;
	gp->prev_flag = True;
	break;
      case ButtonRelease:
	if(gp->call_back_proc != NULL)
	    gp->call_back_proc((WinObj *)gp, e);
	gp->prev_flag = False;
	break;
    }
}

static void graphpanel_destruction(WinObj* obj)
{
    GraphPanel* gp = (GraphPanel *)obj;

    winobj_free(gp->gpbuff);
}

WinObj* create_GraphPanel(WinObj* parent,
			  int x,
			  int y,
			  int width,
			  int height,
			  int bd_width,
			  int xresolution,
			  int yresolution)
{
    GraphPanel* gp;

    gp = (GraphPanel *)create_SimpleWindow(parent, x, y, width, height,
					   bd_width, WINOBJ_BORDER_CONCAVE);

    if(gp == NULL)
	return NULL;

    gp->type = GraphPanel_T;
    gp->xresolution = xresolution;
    gp->yresolution = yresolution;
    if((gp->gpbuff = (int *)winobj_malloc(gp->width * sizeof(int))) == NULL)
    {
	DestroyWinObj((WinObj *)gp);
	return NULL;
    }
    memset(gp->gpbuff, 0, gp->width * sizeof(int));

    gp->winobj_destruction = graphpanel_destruction;
    SetEventMask((WinObj *)gp, ButtonPressMask | ButtonReleaseMask
		 | ButtonMotionMask | ExposureMask);
    SetEventHandler((WinObj *)gp, gp_event_handler);

    return (WinObj *)gp;
}

const int* GetGraphPanelValueList(WinObj* gp, int* nelem_return)
{
    if(nelem_return != NULL)
	*nelem_return = ((GraphPanel *)gp)->width;
    return (const int *)((GraphPanel *)gp)->gpbuff;
}
