#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <math.h>
#include "WinObj.h"

static void redraw_gear(HAdjustGear* ha)
{
    Display* disp = ha->disp;
    Window win = ha->win;
    Pixmap pm = ha->pm;
    GC gc = ha->gc;
    int w, h;
    double r;
    double dr;

    w = ha->width;
    h = ha->height;
    r = 2.0 * ha->r / w;

    if(r >= 0)
	r = fmod(r, 2 * M_PI);
    else
	r = 2 * M_PI - fmod(-r, 2 * M_PI);
    dr = 2.0 / ha->measure;
    r = fmod(r, dr);

    XSetForeground(disp, gc, ha->bg);
    XFillRectangle(disp, pm, gc, 0, 0, w, h);
    while(r < M_PI)
    {
	DrawRuledVLine(disp, pm, gc, ha->lbg, ha->hbg,
		       (int)((1.0 + cos(M_PI - r)) * w * 0.5),
		       0,
		       h);
	r += dr;
    }

    XCopyArea(disp, pm, win, gc,
	      0, 0, w, h, 0, 0);
}

static void hadjust_event_handler(HAdjustGear* ha, XEvent* e)
{
    int diff;

    switch(e->type)
    {
      case ButtonPress:
	ha->cur = e->xbutton.x;
	ha->border_skeleton->style = WINOBJ_BORDER_CONVEX;
	RedrawBorderSkeletion((WinObj *)ha->border_skeleton);
	break;
      case ButtonRelease:
	ha->border_skeleton->style = WINOBJ_BORDER_CONCAVE;
	RedrawBorderSkeletion((WinObj *)ha->border_skeleton);
	break;
      case MotionNotify:
	CompressMotionEvent(ha->disp, ha->win, e);
	diff = e->xmotion.x - ha->cur;
	ha->cur = e->xbutton.x;
	ha->r += diff;
	ha->diff = diff;
	redraw_gear(ha);
	if(ha->call_back_proc != NULL)
	    ha->call_back_proc((WinObj *)ha, e);
	break;
      case Expose:
	if(e->xexpose.count > 0)
	    return;
	redraw_gear(ha);
	break;
    }
}

static void hadjustgear_destruction(WinObj* obj)
{
    XFreePixmap(obj->ha.disp, obj->ha.pm);
}

WinObj* create_HAdjustGear(WinObj* parent,
			   int x,
			   int y,
			   int width,
			   int height,
			   int bd_width,
			   int measure)
{
    HAdjustGear* ha = (HAdjustGear *)
	create_SimpleWindow(parent, x, y, width, height, bd_width,
			    WINOBJ_BORDER_CONCAVE);
    Display* disp;
    Window win;

    if(ha == NULL)
	return NULL;

    disp = ha->disp;
    win = ha->win;
    ha->type = HAdjustGear_T;
    ha->winobj_destruction = hadjustgear_destruction;
    ha->measure = measure;
    ha->cur = 0;
    ha->pm = CreatePixmap(disp, win, width, height, ha->bg);
    ha->diff = ha->r = 0;

    SetEventMask((WinObj *)ha, ButtonPressMask | ButtonMotionMask |
		 ButtonReleaseMask | ExposureMask);
    SetEventHandler((WinObj *)ha, hadjust_event_handler);

    return (WinObj *)ha;
}
