#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "WinObj.h"

static void redraw_slider(HLimitedSlider* slider)
{
    int slider_len, bd;
    bd = slider->bd_width;

    XSetForeground(slider->disp, slider->gc, slider->bg);

    XFillRectangle(slider->disp, slider->tmp_pm, slider->gc,
		   0, 0, slider->limit + 2 * bd, slider->height);

    XSetForeground(slider->disp, slider->gc, slider->mark_color);
    XFillRectangle(slider->disp, slider->tmp_pm, slider->gc,
		   slider->limit + 2 * bd, 0,
		   slider->width - slider->limit - 2 * bd, slider->height);

    slider_len = slider->pos;
    if(slider_len > 0)
	XFillRectangle(slider->disp, slider->tmp_pm, slider->gc,
		       slider->bd_width, bd,
		       slider_len,
		       slider->height - 2 * bd);

    DrawBorder(slider->disp, slider->tmp_pm, slider->gc,
	       0, 0,
	       slider_len + 2 * bd,
	       slider->height,
	       bd,
	       slider->lbg, slider->hbg, True);

    XCopyArea(slider->disp, slider->tmp_pm, slider->win, slider->gc,
	      0, 0,
	      slider->width,
	      slider->height,
	      0, 0);
}

void SetHLimitedSliderPos(WinObj *slider, int pos)
{
    if(pos < 0)
	pos = 0;
    else if(pos > slider->limited_slider.limit)
	pos = slider->limited_slider.limit;

    slider->limited_slider.pos = pos;
    redraw_slider((HLimitedSlider *)slider);
}

static void hlimited_slider_event_handler(HLimitedSlider* slider, XEvent* e)
{
    static int drag_flag = False;

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count > 0)
	    return;
	redraw_slider(slider);
	break;
      case ButtonRelease:
	if(!drag_flag)
	    break;
      case ButtonPress:
	drag_flag = (e->type == ButtonPress);
	if(!slider->move_able)
	    break;
	SetHLimitedSliderPos((WinObj *)slider,
			    e->xbutton.x - slider->bd_width);
	if(slider->call_back_proc != NULL)
	    slider->call_back_proc((WinObj *)slider, e);
	break;
      case MotionNotify:
	if(!drag_flag || !slider->move_able)
	    break;
	if(slider->motion_compress)
	    CompressMotionEvent(slider->disp, slider->win, e);
	SetHLimitedSliderPos((WinObj *)slider,
			    e->xmotion.x - slider->bd_width);
	if(slider->call_back_proc != NULL)
	    slider->call_back_proc((WinObj *)slider, e);
	break;
      case KeyPress:
	if(slider->key_press_handler != NULL)
	    slider->key_press_handler((WinObj *)slider, e);
	break;
    }
}

static void hlimslider_destruction(WinObj* obj)
{
    HLimitedSlider* slider = (HLimitedSlider *)obj;

    XFreePixmap(slider->disp, slider->tmp_pm);
}

WinObj* create_HLimitedSlider(WinObj* parent,
		      int x,
		      int y,
		      int width, /* len */
		      int height, /* thick */
		      int bd_width,
		      int default_pos,
		      int limit)

{
    HLimitedSlider* slider
	= (HLimitedSlider *)create_SimpleWindow(parent,
					       x, y,
					       width, height,
					       bd_width,
					       WINOBJ_BORDER_CONCAVE);
    if(slider == NULL)
	return NULL;

    slider->type = HLimitedSlider_T;
    slider->winobj_destruction = hlimslider_destruction;
    slider->maxlen = width - 2 * bd_width;
    slider->motion_compress = True;
    slider->tmp_pm = XCreatePixmap(slider->disp, slider->win,
				   width,
				   height,
				   DefaultDepth(slider->disp,
						DefaultScreen(slider->disp)));
    slider->move_able = True;
    slider->bd_width = bd_width;
    slider->limit = limit;

    if(default_pos < 0)
	slider->pos = 0;
    else if(default_pos > slider->maxlen)
	slider->pos = slider->maxlen;
    else
	slider->pos = default_pos;

    SetEventMask((WinObj *)slider, ButtonPressMask | ButtonReleaseMask |
		 ButtonMotionMask | KeyPressMask | ExposureMask);
    SetEventHandler((WinObj *)slider, hlimited_slider_event_handler);

    SetHLimitedSliderLimit((WinObj *)slider, limit);

    return (WinObj *)slider;
}

void SetHLimitedSliderLimit(WinObj* obj, int limit)
{
    HLimitedSlider* slider = (HLimitedSlider *)obj;

    if(limit < 0)
	limit = 0;
    else if(limit > slider->maxlen)
	limit = slider->maxlen;
    slider->limit = limit;
    if(slider->pos > limit)
	slider->pos = limit;
    redraw_slider(slider);
}
