#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "WinObj.h"

static void redraw_slider(HSimpleSlider* slider)
{
    int slider_len;

    XSetForeground(slider->disp, slider->gc, slider->bg);
    XFillRectangle(slider->disp, slider->tmp_pm, slider->gc,
		   0, 0, slider->width, slider->height);
    XSetForeground(slider->disp, slider->gc, slider->mark_color);
    slider_len = slider->pos;
    if(slider_len > 0)
	XFillRectangle(slider->disp, slider->tmp_pm, slider->gc,
		       slider->bd_width, slider->bd_width,
		       slider_len,
		       slider->height - 2 * slider->bd_width);

    DrawBorder(slider->disp, slider->tmp_pm, slider->gc,
	       0, 0,
	       slider_len + 2 * slider->bd_width,
	       slider->height,
	       slider->bd_width,
	       slider->lbg, slider->hbg, True);

    XCopyArea(slider->disp, slider->tmp_pm, slider->win, slider->gc,
	      0, 0,
	      slider->width,
	      slider->height,
	      0, 0);
}

void SetHSimpleSliderPos(WinObj *slider, int pos)
{
    if(pos < 0)
	pos = 0;
    else if(pos > slider->sslider.maxlen)
	pos = slider->sslider.maxlen;

    slider->sslider.pos = pos;
    redraw_slider((HSimpleSlider *)slider);
}

static void hsimple_slider_event_handler(HSimpleSlider* slider, XEvent* e)
{
    static int drag_flag = False;

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count > 0)
	    return;
	redraw_slider(slider);
	break;
      case ButtonRelease:
	if(!drag_flag)
	    break;
      case ButtonPress:
	drag_flag = (e->type == ButtonPress);
	if(!slider->move_able)
	    break;
	SetHSimpleSliderPos((WinObj *)slider,
			    e->xbutton.x - slider->bd_width);
	if(slider->call_back_proc != NULL)
	    slider->call_back_proc((WinObj *)slider, e);
	break;
      case MotionNotify:
	if(!drag_flag || !slider->move_able)
	    break;
	if(slider->motion_compress)
	    CompressMotionEvent(slider->disp, slider->win, e);
	SetHSimpleSliderPos((WinObj *)slider,
			    e->xmotion.x - slider->bd_width);
	if(slider->call_back_proc != NULL)
	    slider->call_back_proc((WinObj *)slider, e);
	break;
      case KeyPress:
	if(slider->key_press_handler != NULL)
	    slider->key_press_handler((WinObj *)slider, e);
	break;
    }
}

static void hsslider_destruction(WinObj* obj)
{
    HSimpleSlider* slider = (HSimpleSlider *)obj;

    XFreePixmap(slider->disp, slider->tmp_pm);
}

WinObj* create_HSimpleSlider(WinObj* parent,
		      int x,
		      int y,
		      int width, /* len */
		      int height, /* thick */
		      int bd_width,
		      int default_pos)
{
    HSimpleSlider* slider
	= (HSimpleSlider *)create_SimpleWindow(parent,
					       x, y,
					       width, height,
					       bd_width,
					       WINOBJ_BORDER_CONCAVE);
    if(slider == NULL)
	return NULL;

    slider->type = HSimpleSlider_T;
    slider->winobj_destruction = hsslider_destruction;

    slider->maxlen = width - 2 * bd_width;
    slider->motion_compress = True;
    slider->tmp_pm = XCreatePixmap(slider->disp, slider->win,
				   width,
				   height,
				   DefaultDepth(slider->disp,
						DefaultScreen(slider->disp)));
    slider->move_able = True;
    slider->bd_width = bd_width;

    if(default_pos < 0)
	slider->pos = 0;
    else if(default_pos > slider->maxlen)
	slider->pos = slider->maxlen;
    else
	slider->pos = default_pos;

    SetEventMask((WinObj *)slider, ButtonPressMask | ButtonReleaseMask |
		 ButtonMotionMask | KeyPressMask | ExposureMask);
    SetEventHandler((WinObj *)slider, hsimple_slider_event_handler);

    return (WinObj *)slider;
}
