#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "WinObj.h"

static void hscroll_draw_thumb(HScrollbar* hs)
{
    XCopyArea(hs->disp, hs->pm, hs->win, hs->gc,
	      0, 0,
	      hs->len + 2 * hs->bd_width, hs->thick,
	      hs->top, 0);
}

static void set_pos(HScrollbar* hs, double pos)
{
    int top;

    if(pos < 0.0)
    {
	pos = 0.0;
	top = 0;
    }
    else if(pos > 1.0)
    {
	pos = 1.0;
	top = hs->maxlen - hs->len;
    }
    else
	top = (int)((hs->maxlen - hs->len) * pos);

    hs->pos = pos;
    hs->top = top;
    XClearWindow(hs->disp, hs->win);
    hscroll_draw_thumb(hs);
}

void SetHScrollbarThumbPos(HScrollbar* hs, double pos)
{
    hs->set_pos(hs, pos);
}

double GetHScrollbarThumbPos(HScrollbar* hs)
{
    return hs->pos;
}

static void init_draw_thumb(HScrollbar* hs)
{
    Display* disp = hs->disp;
    int bd = hs->bd_width;
    int bd2 = 2 * bd;

    XSetForeground(disp, hs->gc, hs->bg);
    XFillRectangle(disp, hs->pm1, hs->gc, 0, 0, hs->width, hs->thick);
    XSetForeground(disp, hs->gc, hs->mark_color);
    XFillRectangle(disp, hs->pm2, hs->gc, 0, 0, hs->width, hs->thick);

    DrawBorder(disp, hs->pm1, hs->gc, 0, 0, hs->len + bd2, hs->thick, bd,
	       hs->lbg, hs->hbg, True);
    DrawBorder(disp, hs->pm2, hs->gc, 0, 0, hs->len + bd2, hs->thick, bd,
	       hs->lbg, hs->hbg, False);

    if(hs->len > 4)
    {
	DrawRuledVLine(disp, hs->pm1, hs->gc,
		       hs->lbg, hs->hbg, hs->len / 2 + bd - 1, bd - 1,
		       hs->thick - bd2 + 1);
	if(IsMonoDisplay(hs))
	    DrawRuledVLine(disp, hs->pm2, hs->gc,
			   hs->bg, hs->bg, hs->len / 2 + bd - 1, bd,
			   hs->thick - bd2 - 1);
	else
	    DrawRuledVLine(disp, hs->pm2, hs->gc,
			   hs->lbg, hs->hbg, hs->len / 2 + bd - 1, bd,
			   hs->thick - bd2 - 1);
    }
}

void SetHScrollbarShown(HScrollbar* hs, double shown)
{
    if(shown < 0.0)
	shown = 0.0;
    else if(shown > 1.0)
	shown = 1.0;
    hs->len = (int)(hs->maxlen * shown);
    hs->shown = shown;
    hs->top = (int)((hs->maxlen - hs->len) * hs->pos);
    init_draw_thumb(hs);
    XClearWindow(hs->disp, hs->win);
    hscroll_draw_thumb(hs);
}

double GetHScrollbarShown(HScrollbar* hs)
{
    return hs->shown;
}

static void move_thumb(HScrollbar* hs, int p)
{
    int ptr, max;
    double pos;

    max = hs->width - hs->len;
    ptr = p - hs->len / 2;

    pos = (double)ptr / max;

    SetHScrollbarThumbPos(hs, pos);
}

static void hscroll_event_handler(HScrollbar* hs, XEvent* e)
{
    static Bool in;

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count > 0)
	    return;
	hscroll_draw_thumb(hs);
	break;
      case ButtonPress:
	hs->pm = hs->pm2;
	hs->top = -1;
	move_thumb(hs, e->xbutton.x);
	CallCallbackProc((WinObj *)hs, e);
	in = True;
	break;
      case MotionNotify:
	if(!in)
	    break;
	hs->nmoves++;
	CompressMotionEvent(hs->disp, hs->win, e);
	move_thumb(hs, e->xmotion.x);
	CallCallbackProc((WinObj *)hs, e);
	break;
      case ButtonRelease:
	hs->pm = hs->pm1;
	XClearWindow(hs->disp, hs->win);
	hscroll_draw_thumb(hs);
	hs->nmoves = 0;
	in = False;
	break;
    }
}

static void init_pixmap(HScrollbar* hs)
{
    hs->pm = hs->pm1 = CreatePixmap(hs->disp, hs->win,
				    hs->width, hs->thick, hs->bg);
    hs->pm2 = CreatePixmap(hs->disp, hs->win,
			   hs->width, hs->thick, hs->mark_color);
    init_draw_thumb(hs);
}

static void hscrollbar_resize(WinObj* obj, int width, int height)
{
    HScrollbar* hs = (HScrollbar *)obj;
    Bool pm1_flag;

    if(hs->width >= width && hs->height >= height)
	return;
    hs->width = width;
    hs->height = hs->thick = height;
    hs->len = (width * hs->shown);

    pm1_flag = (hs->pm == hs->pm1);
    XFreePixmap(hs->disp, hs->pm1);
    XFreePixmap(hs->disp, hs->pm2);
    init_pixmap(hs);
    if(pm1_flag)
	hs->pm = hs->pm1;
    else
	hs->pm = hs->pm2;

    XResizeWindow(hs->disp, hs->win, width, height);
    XClearWindow(hs->disp, hs->win);
    hscroll_draw_thumb(hs);
}

static void hscrollbar_destruction(WinObj* obj)
{
    XFreePixmap(obj->hscr.disp, obj->hscr.pm1);
    XFreePixmap(obj->hscr.disp, obj->hscr.pm2);
}

WinObj* create_HScrollbar(WinObj* parent,
			  int x,
			  int y,
			  int width,
			  int height,
			  int bd_width,
			  double shown)
{
    HScrollbar* hs = (HScrollbar *)
	create_SimpleWindow(parent, x, y, width, height, bd_width,
			    WINOBJ_BORDER_CONCAVE);

    if(hs == NULL)
	return NULL;

    hs->type = HScrollbar_T;
    hs->thick = hs->height;
    hs->maxlen = hs->width - 2 * bd_width;
    hs->len = (int)(hs->maxlen * shown);
    hs->pos = 0.0;
    hs->top = 0;
    hs->nmoves = 0;
    hs->set_pos = set_pos;
    hs->shown = shown;
    hs->bd_width = bd_width;

    init_pixmap(hs);
    hs->pm = hs->pm1;
    hs->winobj_resize_proc = hscrollbar_resize;
    hs->winobj_destruction = hscrollbar_destruction;
    SetEventMask((WinObj *)hs, ExposureMask | ButtonMotionMask |
		               ButtonPressMask | ButtonReleaseMask);
    SetEventHandler((WinObj *)hs, hscroll_event_handler);

    return (WinObj *)hs;
}
