#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "WinObj.h"

static void redraw_icon_button(IconButton* icon)
{
    if(icon->redraw_proc != NULL)
    {
	icon->redraw_proc(icon, icon->pressd);
	return;
    }

    if(icon->pressd)
    {
	XSetForeground(icon->disp, icon->gc, icon->mark_color);
	XFillRectangle(icon->disp, icon->win, icon->gc,
		       0, 0, icon->width, icon->height);
	if(icon->pressd_pixmap != None)
	    XCopyArea(icon->disp, icon->pressd_pixmap, icon->win, icon->gc,
		      0, 0, icon->pixmap_width, icon->pixmap_height,
		      (icon->width - icon->pixmap_width) / 2,
		      (icon->height - icon->pixmap_height) / 2);
		      
    }
    else
    {
	XClearWindow(icon->disp, icon->win);
	if(icon->released_pixmap != None)
	    XCopyArea(icon->disp, icon->released_pixmap, icon->win, icon->gc,
		      0, 0, icon->pixmap_width, icon->pixmap_height,
		      (icon->width - icon->pixmap_width) / 2,
		      (icon->height - icon->pixmap_height) / 2);
    }
    XSync(icon->disp, False);
}

void icon_event_handler(IconButton* icon, XEvent* e)
{
    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count > 0)
	    break;
	redraw_icon_button(icon);
	break;
      case ButtonPress:
	icon->pressd = True;
	icon->border_skeleton->style = WINOBJ_BORDER_CONCAVE;
	RedrawBorderSkeletion((WinObj *)icon->border_skeleton);
	redraw_icon_button(icon);
	CallCallbackProc((WinObj *)icon, e);
	break;
      case ButtonRelease:
	if(icon->reset_when_released)
	{
	    icon->pressd = False;
	    icon->border_skeleton->style = WINOBJ_BORDER_CONVEX;
	    RedrawBorderSkeletion((WinObj *)icon->border_skeleton);
	    redraw_icon_button((IconButton *)icon);
	}
	CallCallbackProc((WinObj *)icon, e);
	break;
    }
}

void SetPressdIconButton(WinObj* icon, Bool pressd)
{
    if(icon->icon.pressd == pressd)
	return;
    icon->icon.pressd = pressd;
    if(pressd)
	icon->icon.border_skeleton->style = WINOBJ_BORDER_CONCAVE;
    else
	icon->icon.border_skeleton->style = WINOBJ_BORDER_CONVEX;
    RedrawBorderSkeletion((WinObj *)icon->icon.border_skeleton);
    redraw_icon_button((IconButton *)icon);
}

WinObj* create_IconButton(WinObj* parent,
			  int x,
			  int y,
			  int width,
			  int height,
			  int bd_width,
			  Pixmap pixmap1,
			  Pixmap pixmap2,
			  int pixmap_width,
			  int pixmap_height)
{
    IconButton* icon = (IconButton *)
	create_SimpleWindow(parent, x, y, width, height, bd_width,
			    WINOBJ_BORDER_CONVEX);

    icon->released_pixmap = pixmap1;
    icon->pressd_pixmap   = pixmap2;
    icon->pixmap_width    = pixmap_width;
    icon->pixmap_height   = pixmap_height;
    icon->pressd          = False;
    icon->reset_when_released = True;
    icon->icon_button_id = 0;
    icon->redraw_proc = NULL;

    SetEventHandler((WinObj *)icon, icon_event_handler);
    SetEventMask((WinObj *)icon, ExposureMask |
		 ButtonPressMask | ButtonReleaseMask);

    return (WinObj *)icon;
}
