#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>
#include "WinObj.h"

static void redraw_label(LabeldWindow* lw)
{
    if(lw->label[0] != '\0')
	DrawString(lw->disp, lw->win, lw->font, lw->fg,
		   0, 0, lw->width, lw->height,
		   lw->label, lw->where);
}

static void labeld_window_event_handler(LabeldWindow* lw, XEvent* e)
{
    if(e->type == Expose && e->xexpose.count == 0)
	redraw_label(lw);
}

static void labeld_window_destruction(WinObj* obj)
{
    winobj_free(obj->label.label);
}

WinObj* create_LabeldWindow(WinObj* parent,
			    int x,
			    int y,
			    int width,
			    int height,
			    int bd_width,
			    CharFont* font,
			    const char* label,
			    int border_type,
			    int where)
{
    LabeldWindow* lw;

    if(width <= 0 || height <= 0)
    {
	const int gap = 4;

	width  = TextWidth(font, label) + 2 * gap;
	height = TextHeight(font)       + 2 * gap;
    }

    if(bd_width == 0)
	lw = (LabeldWindow *)create_AnyWindow(parent, x, y, width, height);
    else
	lw = (LabeldWindow *)create_SimpleWindow(parent, x, y, width, height,
						 bd_width, border_type);

    if(lw == NULL)
	return NULL;

    if(label == NULL)
	lw->label = winobj_null_string;
    else if((lw->label = winobj_strdup(label)) == NULL)
	lw->label = winobj_null_string;

    lw->type = LabeldWindow_T;
    lw->where = where;
    lw->font = font;
    lw->winobj_destruction = labeld_window_destruction;
    SetEventMask((WinObj *)lw, ExposureMask);
    SetEventHandler((WinObj *)lw, labeld_window_event_handler);

    return (WinObj *)lw;
}

void SetLabeldWindowLabel(WinObj* obj,
			  const char* label,
			  int border_type)
{
    LabeldWindow* lw = (LabeldWindow *)obj;
    if(label == NULL)
	label = winobj_null_string;
    if(strcmp(label, lw->label) != 0)
    {
	winobj_free(lw->label);
	if(label != winobj_null_string &&
	   (lw->label = winobj_strdup(label)) == NULL)
	    lw->label = winobj_null_string;
    }

    if(lw->border_skeleton != NULL &&
       lw->border_skeleton->style != border_type)
    {
	lw->border_skeleton->style = border_type;
	RedrawBorderSkeletion((WinObj *)lw->border_skeleton);
    }

    XClearWindow(lw->disp, lw->win);
    redraw_label(lw);
}
