#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <X11/Xatom.h>
#include <string.h>
#include <ctype.h>
#include "WinObj.h"
#include "jutils.h"

#ifdef sun
#ifndef memmove
#define memmove(a, b, n) bcopy(b, a, n)
#endif /* memmove */
#endif /* sun */

#define IsCharKey(sym) (XK_space <= (sym) && (sym) <= XK_asciitilde)
#define IsEndOfLine(text) ((text)->cur == (text)->textlen)
#define TextLen(text)     ((text)->textlen)
#define TextChar(t, i) ((t)->text[i])

static void text_scroll_right(LineText* text, int cnt)
{
    XCopyArea(text->disp, text->win,  text->win, text->gc,
	      text->start_origin + text->char_width * cnt,
	      0,
	      (text->show_chars - cnt) * text->char_width,
	      text->height,
	      text->start_origin,
	      0);
    XClearArea(text->disp, text->win,
	       text->start_origin +
	           (text->show_chars - cnt) * text->char_width, 0,
	       cnt * text->char_width, text->height, False);
    text->start_char += cnt;
}

static void text_scroll_left(LineText* text, int cnt)
{
    XCopyArea(text->disp, text->win,  text->win, text->gc,
	      text->start_origin,
	      0,
	      (text->show_chars - cnt) * text->char_width,
	      text->height,
	      text->start_origin + cnt * text->char_width,
	      0);
    XClearArea(text->disp, text->win,
	       text->start_origin, 0,
	       cnt * text->char_width, text->height, False);
    text->start_char -= cnt;
}

static void draw_text_char(LineText* text, Pixel fg,
			   int pos, char ch)
{
    if(ch != '\0')
	DrawStringN(text->disp, text->win, text->font, fg,
		    text->start_origin + pos * text->char_width,
		    0,
		    0, text->height,
		    &ch, 1, West);
    else
    {
	XSetForeground(text->disp, text->gc, text->bg);
	XFillRectangle(text->disp, text->win, text->gc,
		       text->start_origin + pos * text->char_width, 0,
		       text->char_width, text->height);
    }
}

static void draw_text_wchar(LineText* text, Pixel fg,
			    int pos, char ch1, char ch2)
{
  char ch[2];

  ch[0] = ch1;
  ch[1] = ch2;

    if(ch != '\0')
      {
	DrawStringN(text->disp, text->win, text->font, fg,
		    text->start_origin + pos * text->char_width,
		    0,
		    0, text->height,
		    ch, 2, West);
      }
    else
    {
	XSetForeground(text->disp, text->gc, text->bg);
	XFillRectangle(text->disp, text->win, text->gc,
		       text->start_origin + pos * text->char_width, 0,
		       text->char_width, text->height);
    }
}

static void draw_cursor(LineText* text)
{
    int len, w;
    char ch;

    if(text->focus == True)	/* pointer in the text window */
    {
	if(IsEndOfLine(text))
	{
	    len = TextLen(text) - text->start_char;
	    XSetForeground(text->disp, text->gc, text->fg);
	    XFillRectangle(text->disp, text->win, text->gc,
			   text->start_origin + len * text->char_width,
			   (text->height - text->char_height) / 2,
			   text->char_width,
			   text->char_height);
	}
	else
	{
	    ch = TextChar(text, text->cur);
	    w = ((ch & 0x80) ? 2 : 1);

	    XSetForeground(text->disp, text->gc, text->fg);
	    XFillRectangle(text->disp, text->win, text->gc,
			   text->start_origin +
			   (text->cur - text->start_char) * text->char_width,
			   (text->height - text->char_height) / 2,
			   text->char_width * w,
			   text->char_height);
	    if(w == 1)
	      draw_text_char(text, text->bg, text->cur - text->start_char, ch);
	    else
	      draw_text_wchar(text, text->bg, text->cur - text->start_char,
			      ch, TextChar(text, text->cur + 1));
	}
    }
    else
    {
	if(IsEndOfLine(text))
	{
	    len = TextLen(text) - text->start_char;
	    XSetForeground(text->disp, text->gc, text->bg);
	    XFillRectangle(text->disp, text->win, text->gc,
			   text->start_origin + len * text->char_width,
			   (text->height - text->char_height) / 2,
			   text->char_width,
			   text->char_height);

	    XSetForeground(text->disp, text->gc, text->fg);
	    XDrawRectangle(text->disp, text->win, text->gc,
			   text->start_origin + len * text->char_width,
			   (text->height - text->char_height) / 2,
			   text->char_width - 1,
			   text->char_height - 1);
	}
	else
	{
	    ch = TextChar(text, text->cur);
	    w = ((ch & 0x80) ? 2 : 1);

	    XSetForeground(text->disp, text->gc, text->bg);
	    XFillRectangle(text->disp, text->win, text->gc,
			   text->start_origin +
			   (text->cur - text->start_char) * text->char_width,
			   (text->height - text->char_height) / 2,
			   text->char_width * w,
			   text->char_height);
	    XSetForeground(text->disp, text->gc, text->fg);
	    XDrawRectangle(text->disp, text->win, text->gc,
			   text->start_origin +
			   (text->cur - text->start_char)* text->char_width,
			   (text->height - text->char_height) / 2,
			   text->char_width * w - 1,
			   text->char_height - 1);
	    if(w == 1)
	      draw_text_char(text, text->fg, text->cur - text->start_char, ch);
	    else
	      draw_text_wchar(text, text->fg, text->cur - text->start_char,
			      ch, TextChar(text, text->cur + 1));
	}
    }
}

static void erase_cursor(LineText* text)
{
    int len;

    if(IsEndOfLine(text))
    {
	len = TextLen(text) - text->start_char;
	XSetForeground(text->disp, text->gc, text->bg);
	XFillRectangle(text->disp, text->win, text->gc,
		       text->start_origin + len * text->char_width,
		       (text->height - text->char_height) / 2,
		       text->char_width,
		       text->char_height);
	XSetForeground(text->disp, text->gc, text->fg);
    }
    else
    {
	char ch;
	int w;

	ch = TextChar(text, text->cur);
	w = ((ch & 0x80) ? 2 : 1);

	XSetForeground(text->disp, text->gc, text->bg);
	XFillRectangle(text->disp, text->win, text->gc,
		       text->start_origin +
		       (text->cur - text->start_char) * text->char_width,
		       (text->height - text->char_height) / 2,
		       text->char_width * w,
		       text->char_height);
	if(w == 1)
	  draw_text_char(text, text->fg, text->cur - text->start_char, ch);
	else
	  draw_text_wchar(text, text->fg, text->cur - text->start_char,
			  ch, TextChar(text, text->cur + 1));
    }
}

Bool LineTextForwardChar(LineText* text)
{
  int e;

    if(IsEndOfLine(text))
	return False;
    erase_cursor(text);

  if(TextChar(text, text->cur) & 0x80)
      text->cur += 2;
  else
      text->cur++;

  if(TextChar(text, text->cur) & 0x80)
    e = text->cur - text->start_char + 2;
  else
    e = text->cur - text->start_char + 1;

  if(e > text->show_chars)
    text_scroll_right(text, e - text->show_chars);

    draw_cursor(text);
    return True;
}

Bool LineTextBackwardChar(LineText* text)
{
  int cnt;

    if(text->cur == 0)
	return False;
  if(text->cur > 1 && (TextChar(text, text->cur - 2) & 0x80))
    cnt = 2;
  else
    cnt = 1;

    erase_cursor(text);
    text->cur -= cnt;

    if(text->cur < text->start_char)
	text_scroll_left(text, text->start_char - text->cur);

    draw_cursor(text);
    return True;
}

Bool LineTextDeleteChar(LineText* text)
{
    int cur, len, pos, after, cnt;

    if(IsEndOfLine(text))
	return False;

    cur = text->cur;
    len = TextLen(text);
    pos = cur - text->start_char;
    after = len - cur;

  if(TextChar(text, text->cur) & 0x80)
    cnt = 2;
  else
    {
      cnt = 1;
      if(pos + 1 == text->show_chars &&
	 TextChar(text, text->cur + 1) & 0x80)
	{
	  text_scroll_right(text, 1);
	  pos--;
	}
    }

    erase_cursor(text);

    memmove(text->text + cur, text->text + cur + 1, after - 1);

    XCopyArea(text->disp, text->win, text->win, text->gc,
	      text->start_origin + (pos + 1) * text->char_width,
	      0,
	      (text->show_chars - pos - 1) * text->char_width,
	      text->height,
	      text->start_origin + pos * text->char_width,
	      0);

    text->text[--len] = '\0';
    text->textlen = len;

    if(text->start_char + text->show_chars > text->textlen)
	draw_text_char(text, text->fg, text->show_chars - 1, '\0');
    else
    {
	int last;

	draw_text_char(text, text->fg, text->show_chars - 1, '\0');
	last = text->start_char + text->show_chars - 1;
	if(!(TextChar(text, last) & 0x80))
	  if(!(TextChar(text, last - 1) & 0x80))
	    draw_text_char(text, text->fg, text->show_chars - 1,
			   TextChar(text, last));
	  else
	    draw_text_wchar(text, text->fg, text->show_chars - 2,
			    TextChar(text, last - 1), TextChar(text, last));
    }
    draw_cursor(text);

    if(cnt == 1)
      return True;
    else
      return LineTextDeleteChar(text);
}

Bool LineTextInsertChar(LineText* text, char ch)
{
    int pos, cur;

    if(text->textlen == text->maxlen)
	return False;

    cur = text->cur;
    pos = cur - text->start_char;

    erase_cursor(text);

    if(IsEndOfLine(text))
    {
	text->text[text->cur] = ch;
	draw_text_char(text, text->fg, pos, ch);
    }
    else
    {
      if(TextChar(text, cur) & 0x80 && pos + 2 == text->show_chars)
	{
	  text_scroll_right(text, 1);
	  pos--;
	}

	XCopyArea(text->disp, text->win, text->win, text->gc,
		  text->start_origin + pos * text->char_width,
		  0,
		  (text->show_chars - pos - 1) * text->char_width,
		  text->height,
		  text->start_origin + (pos + 1) * text->char_width,
		  0);
	draw_text_char(text, text->fg, pos, '\0');
	draw_text_char(text, text->fg, pos, ch);
	memmove(text->text + cur + 1, text->text + cur,
		text->textlen - cur);
	text->text[cur] = ch;

    }
    if(pos == text->show_chars - 1)
	text_scroll_right(text, 1);
    text->textlen++;
    text->cur++;
    draw_cursor(text);
    return True;
}

Bool LineTextInsertWChar(LineText* text, char ch1, char ch2)
{
    int pos, cur, cnt;

    if(text->textlen + 1 >= text->maxlen)
	return False;

    cur = text->cur;
    pos = cur - text->start_char;

    erase_cursor(text);

    if(pos == text->show_chars - 1)
      {
	text_scroll_right(text, 2);
	cnt = 2;
      }
    else if(pos == text->show_chars - 2)
      {
	text_scroll_right(text, 1);
	cnt = 1;
      }
    else
      cnt = 0;

    if(IsEndOfLine(text))
    {
	text->text[text->cur]     = ch1;
	text->text[text->cur + 1] = (ch2 & 0x7f);
	draw_text_wchar(text, text->fg, pos - cnt, ch1, ch2);
    }
    else
    {
      if(cnt == 0)
	{
	  if(TextChar(text, cur) & 0x80 && pos + 3 == text->show_chars)
	    {
	      text_scroll_right(text, 1);
	      pos--;
	    }

	  XCopyArea(text->disp, text->win, text->win, text->gc,
		    text->start_origin + pos * text->char_width,
		    0,
		    (text->show_chars - pos - 2) * text->char_width,
		    text->height,
		    text->start_origin + (pos + 2) * text->char_width,
		    0);
	}
      draw_text_char(text, text->fg, pos - cnt, '\0');
      draw_text_char(text, text->fg, pos - cnt + 1, '\0');
      draw_text_wchar(text, text->fg, pos - cnt, ch1, ch2);
      memmove(text->text + cur + 2, text->text + cur,
	      text->textlen - cur);
      text->text[text->cur]     = ch1;
      text->text[text->cur + 1] = (ch2 & 0x7f);
    }
    text->textlen += 2;
    text->cur += 2;
    draw_cursor(text);
    return True;
}

void LineTextFocusIn(LineText* text)
{
    XSetInputFocus(text->disp, text->win, RevertToParent, None);
    text->focus = True;
    draw_cursor(text);
}

void LineTextFocusOut(LineText* text)
{
    XSetInputFocus(text->disp, PointerRoot, RevertToPointerRoot, None);
    text->focus = False;
    draw_cursor(text);
}

static void key_press_handler(LineText* text, XEvent* e)
{
    KeySym sym;
    char s[10];
    int n;

    if(e->xkey.state & ControlMask)
    {
	Bool state = True;

	sym = XKeycodeToKeysym(text->disp, e->xkey.keycode, 0);

	switch(sym)
	{
	  case XK_f:
	    state = LineTextForwardChar(text);
	    break;
	  case XK_b:
	    state = LineTextBackwardChar(text);
	    break;
	  case XK_d:
	    state = LineTextDeleteChar(text);
	    break;
	  case XK_a:
	    while(LineTextBackwardChar(text))
		;
	    break;
	  case XK_e:
	    while(LineTextForwardChar(text))
		;
	    break;
	  case XK_k:
	    state = LineTextKillLine(text);
	    break;

	  case XK_w:
	    state = LineTextKillRegion(text);
	    break;

	  case XK_y:
	    LineTextYank(text);
	    break;

#ifdef XK_KP_Space
	  case XK_KP_Space:
#endif
#ifdef XK_space
	  case XK_space:
#endif
	    text->mark_pos = text->cur;
#ifdef DEBUG
	    printf("Mark set to %d\n", text->mark_pos);
#endif
	    break;
	}

	if(state == False)
	{
	    XBell(text->disp, 0);
	    XSync(text->disp, False);
	}

	return;
    }

    s[0] = '\0';
    n = XLookupString(&e->xkey, s, 10, &sym, NULL);

    switch(sym)
    {
      case XK_Shift_L:
      case XK_Shift_R:
      case XK_Caps_Lock:
      case XK_Control_L:
      case XK_Control_R:
	return;
      case XK_Right:
#ifdef XK_KP_Right
      case XK_KP_Right:
#endif
	if(LineTextForwardChar(text) == False)
	{
	    XBell(text->disp, 0);
	    XSync(text->disp, False);
	}
	return;
      case XK_Left:
#ifdef XK_KP_Left
      case XK_KP_Left:
#endif
	if(LineTextBackwardChar(text) == False)
	{
	    XBell(text->disp, 0);
	    XSync(text->disp, False);
	}
	return;

      case XK_BackSpace:
      case XK_Delete:
#ifdef XK_KP_Delete
      case XK_KP_Delete:
#endif
	if(LineTextBackwardChar(text) == False)
	{
	    XBell(text->disp, 0);
	    XSync(text->disp, False);
	    return;
	}
	if(LineTextDeleteChar(text) == False)
	{
	    XBell(text->disp, 0);
	    XSync(text->disp, False);
	}
	return;
      case XK_Return:
      case XK_KP_Enter:
	CallCallbackProc((WinObj *)text, e);
	return;
      case XK_space:
      case XK_Tab:
	if(LineTextInsertChar(text, ' ') == False)
	{
	    XBell(text->disp, 0);
	    XSync(text->disp, False);
	}
	return;
    }

    if(n == 0 || !isprint(s[0]))
    {
	XBell(text->disp, 0);
	XSync(text->disp, False);
	return;
    }

    if(LineTextInsertChar(text, s[0]) == False)
    {
	XBell(text->disp, 0);
	XSync(text->disp, False);
	return;
    }
}

static void selection_clear_proc(LineText* text, XEvent* e)
{
    if(e->xselectionclear.window == text->win &&
       e->xselection.selection == XA_PRIMARY)
	text->seldata->selected = False;
}

static void selection_request(LineText* text, XEvent* e)
{
    XSelectionEvent se;

#ifdef USE_LOCALE
    Atom atom_list[3];
    int na = 3;
    char* cl[1];
    XTextProperty ct;
#else
    Atom atom_list[2];
    int na = 2;
#endif

    Display* disp = text->disp;

    atom_list[0] = text->seldata->targets;
    atom_list[1] = XA_STRING;

#ifdef USE_LOCALE
    atom_list[2] = text->seldata->compound_text;
#endif

    se.type = SelectionNotify;
    se.requestor = e->xselectionrequest.requestor;
    se.selection = e->xselectionrequest.selection;
    se.target    = e->xselectionrequest.target;
    se.time      = e->xselectionrequest.time;
    se.property  = e->xselectionrequest.property;

    if(e->xselectionrequest.selection == XA_PRIMARY)
    {
	if(e->xselectionrequest.target == text->seldata->targets)
	{
	    XChangeProperty(disp, se.requestor, se.property,
			    XA_ATOM, 32, PropModeReplace,
			    (const unsigned char *)atom_list, na);
	}
	else if(e->xselectionrequest.target == XA_STRING)
	{
	    XChangeProperty(disp, se.requestor, se.property, se.target,
			    8, PropModeReplace,
			    (const unsigned char *)text->seldata->buffer,
			    strlen(text->seldata->buffer));
	}

#ifdef USE_LOCALE
	else if(e->xselectionrequest.target == text->seldata->compound_text)
	{
	    XmbTextListToTextProperty(disp, cl, 1,
				      XCompoundTextStyle, &ct);
	    XChangeProperty(disp, se.requestor, se.property, se.target,
			    8, PropModeReplace, ct.value,
			    (int)ct.nitems);
	}
#endif

	else
	    se.property = None;
    }
    else
	se.property = None;

    XSendEvent(disp, se.requestor, False, None, (XEvent* )&se);
}
    
static void selection_notify(LineText* text, XEvent* e)
{
    Display* disp = text->disp;
    Window win = text->win;
    Atom* al;
    Atom type;
    int i, fmt;
    unsigned long nitems, left;

    if(e->xselection.target == text->seldata->targets)
    {
      al = NULL;
	if(XGetWindowProperty(disp, win, text->seldata->copy_text,
			      0, 1024, False, XA_ATOM, &type, &fmt, &nitems,
			      &left,
			      (unsigned char **)&al) == Success && nitems > 0)
	{
	  Atom a;

#ifdef USE_LOCALE
	  a = None;
	  for(i = 0; i < nitems; i++)
	    if(al[i] == text->seldata->compound_text)
	      {
		a = al[i];
		break;
	      }
	  if(a == None)
	    a = XA_STRING;
#else
	  a = XA_STRING;
#endif
	  XConvertSelection(disp, XA_PRIMARY, a,
			    text->seldata->copy_text, win,
			    e->xselection.time);
	  if(al != NULL)
	    XFree(al);
	}
    }
    else if(e->xselection.target == XA_STRING
#ifdef USE_LOCALE
	    || e->xselection.target == text->seldata->compound_text
#endif
	    )
    {
	char* buff = NULL;

	if(XGetWindowProperty(disp, win, text->seldata->copy_text,
			      0, 1024, False,
			      e->xselection.target,
			      &type, &fmt, &nitems,
			      &left,
			      (unsigned char **)&buff) == Success
	   && nitems > 0)
	{
	  nitems = jistoeuc((unsigned char *)buff,
			    (unsigned char *)buff, nitems);
	    for(i = 0; i < nitems; i++)
	      {
		if(buff[i] & 0x80)
		  {
		    LineTextInsertWChar(text, buff[i], buff[i + 1]);
		    i++;
		  }
		else
		  LineTextInsertChar(text, buff[i]);
	      }
	    if(buff != NULL)
	      XFree(buff);
	}
    }
}

static void selection_event_handler(LineText* text, XEvent* e)
{
    switch(e->type)
    {
      case SelectionClear:
	selection_clear_proc(text, e);
	break;
      case SelectionRequest:
	selection_request(text, e);
	break;
      case SelectionNotify:
	selection_notify(text, e);
	break;
    }
}

void LineTextRedraw(LineText* text)
{
  int i, last;
  DrawString(text->disp, text->win, text->font, text->fg,
	       0, 0, 0, text->height, text->prompt, West);
  i = text->start_char;
  if(i + text->show_chars < text->textlen)
    last = i + text->show_chars;
  else
    last = text->textlen;
  for(; i < last; i++)
    draw_text_char(text, text->fg, i - text->start_char, TextChar(text, i));
  draw_cursor(text);
}

static void text_event_handler(LineText* text, XEvent* e)
{
    switch(e->type)
    {
    case Expose:
      if(e->xexpose.count == 0)
	LineTextRedraw(text);
      break;
      case MappingNotify:
	XRefreshKeyboardMapping(&e->xmapping);
	break;
      case EnterNotify:
	LineTextFocusIn(text);
	break;
      case LeaveNotify:
	LineTextFocusOut(text);
	break;
      case KeyPress:
	text->key_press_handler(text, e);
	break;
      case SelectionClear:
      case SelectionRequest:
      case SelectionNotify:
	text->selection_event_handler(text, e);
	break;
    }
}

static void line_text_destruction(WinObj* obj)
{
    LineText* text = (LineText *)obj;
    winobj_free(text->prompt);
    winobj_free(text->text);
    winobj_free(text->seldata->buffer);
    winobj_free(text->seldata);
}

WinObj* create_LineText(WinObj* parent,
			int x,
			int y,
			int width,
			int height,
			int bd_width,
			CharFont* font,
			const char* prompt)
{
    Display* disp;
    Window win;
    LineText* text = (LineText *)
	create_SimpleWindow(parent, x, y, width, height, bd_width,
			    WINOBJ_BORDER_CONCAVE);
    int prompt_width;

    if(text == NULL)
	return NULL;

    disp = text->disp;
    win = text->win;
    text->type = LineText_T;
    text->font = font;
    if(prompt == NULL)
      prompt = text->prompt = winobj_null_string;
    else if((text->prompt = winobj_strdup(prompt)) == NULL)
	text->prompt = winobj_null_string;
    text->char_width = MAX(text->font->font8->max_bounds.width,
			   text->font->font16->max_bounds.width/2);

    text->char_height = TextHeight(font);
    prompt_width = TextWidth(font, prompt);

    text->focus = False;
    text->start_origin = prompt_width;
    text->show_chars = (text->width - text->start_origin) / text->char_width;
    text->maxlen = MAX(text->show_chars, BUFSIZ);
    text->cur = 0;
    text->textlen = 0;
    text->start_char = 0;

    text->text = (char *)winobj_malloc(text->maxlen + 1);
    if(text->text == NULL)
    {
	DestroyWinObj((WinObj *)text);
	return NULL;
    }
    memset(text->text, 0, text->maxlen + 1);

    text->key_press_handler = key_press_handler;
    text->draw_cursor = draw_cursor;
    text->selection_event_handler = selection_event_handler;
    text->mark_pos = -1;
    text->seldata = (SelectionData *)winobj_malloc(sizeof(SelectionData));
    text->seldata->buffer = (char *)winobj_malloc(text->maxlen + 1);
    if(text->seldata == NULL ||
       text->seldata->buffer == NULL)
    {
	winobj_free(text->text);
	winobj_free(text->seldata);
	winobj_free(text->seldata->buffer);
	DestroyWinObj((WinObj *)text);
	return NULL;
    }
    text->seldata->selected = False;

    text->seldata->copy_text     = XInternAtom(disp, "__COPY_TEXT", False);
    text->seldata->targets       = XInternAtom(disp, "TARGETS", False);
#ifdef USE_LOCALE
    text->seldata->compound_text = XInternAtom(disp, "COMPOUND_TEXT", False);
#else
    text->seldata->compound_text = None;
#endif

    SetEventMask((WinObj *)text, KeyPressMask |
		 EnterWindowMask | LeaveWindowMask | ExposureMask);

    SetEventHandler((WinObj *)text, text_event_handler);
    DrawString(disp, win, font, text->fg, 0, 0, 0, text->height,
	       prompt, West);
    LineTextFocusOut(text);
    XDefineCursor(disp, text->win, XCreateFontCursor(disp, XC_xterm));
    text->winobj_destruction = line_text_destruction;

    return (WinObj *)text;
}

Bool LineTextKillRegion(LineText* text)
{
    if(text->mark_pos == -1)
	return False;
    if(text->mark_pos == text->cur)
	text->seldata->buffer[0] = '\0';
    else
    {
	int i, j;

	if(text->mark_pos < text->cur)
	{
	    int len;

	    len = text->cur - text->mark_pos;
	    for(j = 0, i = text->mark_pos; j < len; i++, j++)
		text->seldata->buffer[j] = TextChar(text, i);
	    text->seldata->buffer[j] = '\0';
	    for(j = 0; j < len; j++)
	    {
		LineTextBackwardChar(text);
		if(TextChar(text, text->cur) & 0x80)
		  j++;
		LineTextDeleteChar(text);
	    }
	}
	else			/* text->mark_pos >= text->cur */
	{
	    int len;

	    len = text->mark_pos - text->cur;
	    for(j = 0, i = text->mark_pos - len; j < len; i++, j++)
		text->seldata->buffer[j] = TextChar(text, i);
	    text->seldata->buffer[j] = '\0';
	    for(j = 0; j < len; j++)
	      {
		if(TextChar(text, text->cur) & 0x80)
		  j++;
		LineTextDeleteChar(text);
	      }
	}
    }

    XSetSelectionOwner(text->disp, XA_PRIMARY, text->win, None);
    if(XGetSelectionOwner(text->disp, XA_PRIMARY) == text->win)
	text->seldata->selected = True;

    return True;
}


Bool LineTextKillLine(LineText* text)
{
    int i, j, len;

    if(IsEndOfLine(text))
	return False;

    len = TextLen(text) - text->cur;

    for(j = 0, i = text->cur; j < len; i++, j++)
	text->seldata->buffer[j] = TextChar(text, i);
    text->seldata->buffer[j] = '\0';
    for(j = 0; j < len; j++)
	LineTextDeleteChar(text);

    XSetSelectionOwner(text->disp, XA_PRIMARY, text->win, None);
    if(XGetSelectionOwner(text->disp, XA_PRIMARY) == text->win)
	text->seldata->selected = True;

    return True;
}

void LineTextYank(LineText* text)
{
    text->mark_pos = text->cur;
#ifdef DEBUG
    printf("Mark set to %d\n", text->mark_pos);
#endif

    if(text->seldata->selected == True)
    {
	char* str;
	for(str = text->seldata->buffer; *str; str++)
	  if(*str & 0x80)
	    {
	      LineTextInsertWChar(text, str[0], str[1]);
	      str++;
	    }
	  else
	    LineTextInsertChar(text, *str);
    }
    else
    {
	XConvertSelection(text->disp, XA_PRIMARY, text->seldata->targets,
			  text->seldata->copy_text, text->win, None);
    }
}

void SetLineText(LineText* text, char* buff, int len)
{
    int orig_cur = text->cur;

    memset(text->text, 0, text->maxlen);
    if(buff == NULL)
    {
	buff = "";
	len = 0;
    }
    if(len > text->maxlen)
	len = text->maxlen;
    text->cur = len;
    text->textlen = len;
    if(len <= text->show_chars)
	text->start_char = 0;
    else
	text->start_char = len - text->show_chars;
    memcpy(text->text, buff, len);
    memset(text->text + len, 0, text->maxlen - len);
    XSetForeground(text->disp, text->gc, text->bg);
    XFillRectangle(text->disp, text->win, text->gc,
		   text->start_origin, 0,
		   text->width - text->start_origin, text->height);
    buff += text->start_char;
    len -= text->start_char;
    DrawStringN(text->disp, text->win, text->font, text->fg,
		text->start_origin, 0,
		0, text->height, buff, len, West);
    if(orig_cur < text->cur)
	while(orig_cur < text->cur && LineTextBackwardChar(text))
	    ;
    else
	draw_cursor(text);
}

int GetLineText(LineText* text, char* buff, int maxlen)
{
    int len;

    len = TextLen(text);
    if(maxlen <= len)
	len = maxlen - 1;
    memcpy(buff, text->text, len);
    buff[len] = '\0';
    return len;
}
