#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>
#include "WinObj.h"

#define DISPATCH_TABLE_SIZE 1499

#define HASH_DISP(disp) ((unsigned long)((unsigned long)(disp) / sizeof(disp)))
#define HASH(disp, win) ((HASH_DISP(disp) + (win<<4)) % DISPATCH_TABLE_SIZE)
#define REHASH(addr) (((addr) + 1) % DISPATCH_TABLE_SIZE)
static Bool table_over_full = False;

struct dispatch_table_t
{
    Display* disp;
    Window win;
    AnyWindow* obj;
};

static struct dispatch_table_t dispatch_table[DISPATCH_TABLE_SIZE];

double get_dispatch_table_use_ratio()
{
    int i, n;

    n = 0;
    for(i = 0; i < DISPATCH_TABLE_SIZE; i++)
	if(dispatch_table[i].disp != NULL)
	    n++;
    return n / (double)DISPATCH_TABLE_SIZE;
}

Bool install_event_handler(Display* disp,
			   Window win,
			   WinObj* obj)
{
    unsigned long addr = HASH(disp, win);
    int i;

    for(i = 0; i < DISPATCH_TABLE_SIZE; i++)
    {
	if(dispatch_table[addr].disp == NULL)
	{
	    dispatch_table[addr].disp = disp;
	    dispatch_table[addr].win = win;
	    dispatch_table[addr].obj = (AnyWindow *)obj;
	    table_over_full = False;
	    return True;
	}
	addr = REHASH(addr);
    }

    if(table_over_full == False)
    {
	table_over_full = True;
	CallWinObjErrorHandler(WOE_TOO_MANY_HANDLER, NULL);
    }
    return False;
}

void uninstall_event_handler(Display* disp, Window win)
{
    unsigned long addr = HASH(disp, win);
    int i;

    for(i = 0; i < DISPATCH_TABLE_SIZE; i++)
    {
	if(dispatch_table[addr].disp == disp &&
	   dispatch_table[addr].win == win)
	{
	    dispatch_table[addr].disp = NULL;
	    return;
	}
	addr = REHASH(addr);
    }
}

void MainLoop(void)
{
    XEvent e;
    Display* disp;

    while(1)
    {
	disp = CheckMultiDisplayEvent(all_display_list,
				      number_of_display,
				      NULL);
	if(disp == NULL)
	    continue;

	XNextEvent(disp, &e);
	DispatchEvent(&e);
    }
}

void DispatchEvent(XEvent *e)
{
    Display* disp = e->xany.display;
    Window win = e->xany.window;
    unsigned long addr = HASH(disp, win);
    int i;

#ifdef DEBUGALL
    int rehash_cnt = 0;
#endif

    for(i = 0; i < DISPATCH_TABLE_SIZE; i++)
    {
	if(dispatch_table[addr].disp == NULL &&
	   dispatch_table[addr].obj == NULL)
	    break;

	if(dispatch_table[addr].disp == disp &&
	   dispatch_table[addr].win == win)
	{
	    AnyWindow** event_alias;
	    XEvent orig;

#ifdef DEBUGALL
		if(e->type != Expose)
		    printf("DispatchEvent: %p,%ld=>%ld,%d %s (%d), %ld\n",
			   disp, win, addr, rehash_cnt,
			   EventTypeString(e->type),
			   e->type,
			   e->xany.serial);
		else
		    printf("DispatchEvent: %p,%ld=>%ld,%d %s (%d), %ld/%d\n",
			   disp, win, addr, rehash_cnt,
			   EventTypeString(e->type),
			   e->type,
			   e->xany.serial,
			   e->xexpose.count);
#endif


	    memcpy(&orig, e, sizeof(XEvent));
	    event_alias = (AnyWindow **)dispatch_table[addr].obj->event_alias;

	    /* FIXME
	     * There is a bug in event alias.
	     * If the object is deleted, event alias is not workd !
	     */
	    if(event_alias != NULL)
	    {
		int i;

		for(i = 0; event_alias[i] != NULL; i++)
		{
		    if(event_alias[i]->event_handler != NULL)
		    {
			e->xany.display = event_alias[i]->disp;
			e->xany.window = event_alias[i]->win;
			event_alias[i]->event_handler(
			    (union _WinObj *)event_alias[i],
			    e);
		    }
		}
		if(dispatch_table[addr].disp == NULL)
		    return; /* deleted in above event handler */
	    }
	    memcpy(e, &orig, sizeof(XEvent));
	    if(dispatch_table[addr].obj->event_handler != NULL)
	    {
		dispatch_table[addr].obj->event_handler( 
			(union _WinObj *)(dispatch_table[addr].obj),
			e);
	    }
	    break;
	}
	addr = REHASH(addr);
#ifdef DEBUGALL
	rehash_cnt++;
#endif
    }
}
