#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "WinObj.h"

#define POINTER_IN_BAR_MENU(bar, x, i) ((bar)->name_x_list[i] <= (x) && (x) < (bar)->name_x_list[i] + (bar)->name_width_list[i])

static void redraw_menu_bar(MenuBar* bar)
{
    int i;

    for(i = 0; i < bar->nmenu; i++)
	DrawString(bar->disp, bar->win, bar->font, bar->fg,
		   bar->name_x_list[i], 0,
		   bar->name_width_list[i], bar->height,
		   bar->name_list[i], Center);
}

static void popup_mark_set(MenuBar* bar, int x, int w, char* s)
{
    Display* disp = bar->disp;
    Window win = bar->win;
    GC gc = bar->gc;
    Pixel fg;

    XSetForeground(disp, gc, bar->mark_color);
    XFillRectangle(disp, win, gc, x, 0, w, bar->height);
    DrawBorder(disp, win, gc, x, 0, w, bar->height,
	       bar->bd_width, bar->lbg, bar->hbg, True);

    if(IsMonoDisplay(bar))
	fg = bar->bg;
    else
	fg = bar->fg;

    DrawString(disp, win, bar->font, fg,
	       x, 0, w, bar->height, s, Center);
}

static void popup_mark_reset(MenuBar* bar, int x, int w, char* s)
{
    Display* disp = bar->disp;
    Window win = bar->win;
    GC gc = bar->gc;

    XSetForeground(disp, gc, bar->bg);
    XFillRectangle(disp, win, gc, x, 0, w, bar->height);
    DrawString(disp, win, bar->font, bar->fg,
	       x, 0, w, bar->height, s, Center);
}

static void bar_event_handler(MenuBar* bar, XEvent* e)
{
    int i;

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count == 0)
	    redraw_menu_bar(bar);
	break;
      case ButtonPress:
	if(bar->popup_menu_index != -1)
	    return;
	for(i = 0; i < bar->nmenu; i++)
	{
	    if(POINTER_IN_BAR_MENU(bar, e->xbutton.x, i))
	    {
		int rx, ry, x;

		if(bar->popup_list[i] == NULL)
		    continue;

		x = bar->name_x_list[i];
		QueryWindowPositionFromRootWindow(bar->disp, bar->win,
						  &rx, &ry);
		popup_mark_set(bar, x, bar->name_width_list[i],
			       bar->name_list[i]);
		bar->popup_menu_index = i;
#ifdef DEBUG
		puts("MenuBar: PopupMenu_popup_proc()");
#endif
		PopupMenu_popup_proc(bar->popup_list[i],
				     x + rx, ry + bar->height);
#ifdef DEBUG
		puts("MenuBar: PopupMenu_popup_proc() - done");
#endif

		if(bar->disp == NULL)
		    return;	/* Destroyd */
		i = bar->popup_menu_index;
		x = bar->name_x_list[i];
		if(i != -1)
		    popup_mark_reset(bar, x, bar->name_width_list[i],
				     bar->name_list[i]);
		bar->popup_menu_index = -1;
		return;
	    }
	}
    }
}


static Bool switch_menu_proc(MenuBar* bar, XEvent* e, int* status_ret)
{
    int i;

    if(e->type != MotionNotify)
	return True;

    *status_ret = False;
    for(i = 0; i < bar->nmenu; i++)
    {
	if(POINTER_IN_BAR_MENU(bar, e->xmotion.x, i))
	{
	    int rx, ry, j, x;

	    if(bar->popup_list[i] == NULL)
		continue;

	    if(i == bar->popup_menu_index)
		return True;

	    x = bar->name_x_list[i];

	    XUngrabPointer(bar->disp, 0);

	    j = bar->popup_menu_index;
	    unmap_popup_menu_internal((PopupMenu *)bar->popup_list[j]);
	    popup_mark_reset(bar, bar->name_x_list[j], bar->name_width_list[j],
			     bar->name_list[j]);

	    QueryWindowPositionFromRootWindow(bar->disp, bar->win,
					      &rx, &ry);
	    bar->popup_menu_index = i;
	    popup_mark_set(bar, x, bar->name_width_list[i],
			   bar->name_list[i]);
	    *status_ret = PopupMenu_popup_proc(bar->popup_list[i],
					       x + rx, ry + bar->height);
	    if(bar->disp == NULL)
		return False;	/* Destroyd */
	    i = bar->popup_menu_index;
	    if(i != -1)
		popup_mark_reset(bar, x, bar->name_width_list[i],
				 bar->name_list[i]);
	    bar->popup_menu_index = -1;

	    return False;
	}
    }
    return True;
}

static void menubar_destruction(WinObj* obj)
{
    MenuBar* bar = (MenuBar *)obj;
    int i;

    for(i = 0; i < bar->nmenu; i++)
	winobj_free(bar->name_list[i]);
    winobj_free(bar->name_list);
    winobj_free(bar->name_width_list);
    winobj_free(bar->name_x_list);
    winobj_free(bar->popup_list);
}

WinObj* create_MenuBar(WinObj* parent,
		       int x,
		       int y,
		       int width,
		       CharFont* font,
		       int bd_width,
		       MenuBarItem* item,
		       int n,
		       char* help_string,
		       WinObj* help_popup)
{
    char** name_list;
    int* name_width_list;
    int* name_x_list;
    WinObj** popup_list;
    int i, wx, k;
    MenuBar* bar;
    int height;

    if(help_string == NULL)
	k = n;
    else
	k = n + 1;

    name_list = (char **)winobj_malloc(k * sizeof(char *));
    name_width_list = (int *)winobj_malloc(k * sizeof(int));
    name_x_list = (int *)winobj_malloc(k * sizeof(int));
    popup_list = (WinObj **)winobj_malloc(k * sizeof(WinObj *));

    if(name_list == NULL ||
       name_width_list == NULL ||
       name_x_list == NULL ||
       popup_list == NULL)
    {
	winobj_free(name_list);
	winobj_free(name_width_list);
	winobj_free(name_x_list);
	winobj_free(popup_list);
	return NULL;
    }

    height = TextHeight(font);
    if(width == 0)
	width = parent->any.width - bd_width * 2;
    wx = 0;
    for(i = 0; i < n; i++)
    {
	if((name_list[i] = winobj_strdup(item[i].name)) == NULL)
	{
	    while(i--)
		winobj_free(name_list[i]);
	    winobj_free(name_list);
	    winobj_free(name_width_list);
	    winobj_free(name_x_list);
	    winobj_free(popup_list);
	    return NULL;
	}
		
	name_width_list[i] = TextWidth(font, name_list[i]) + height;
	name_x_list[i] = wx;
	popup_list[i] = item[i].popup;
	if(!WinObjPopupAble(popup_list[i]))
	{
	    fprintf(stderr,
"WinObj Fatal Error: Specified PopupMenu is Invalid in create_MenuBar().\n");
	    exit(1);
	}
	wx += name_width_list[i];
    }

    if(n < k)
    {
	if((name_list[i] = winobj_strdup(help_string)) == NULL)
	{
	    while(i--)
		winobj_free(name_list[i]);
	    winobj_free(name_list);
	    winobj_free(name_width_list);
	    winobj_free(name_x_list);
	    winobj_free(popup_list);
	    return NULL;
	}
	name_width_list[i] = TextWidth(font, name_list[i]) + height;
	name_x_list[i] = width - name_width_list[i];
	popup_list[i] = help_popup;
	if(help_popup != NULL && !WinObjPopupAble(popup_list[i]))
	{
	    fprintf(stderr,
"WinObj Fatal Error: Specified PopupMenu is Invalid in create_MenuBar().\n");
	    exit(1);
	}
    }

    bar = (MenuBar *)create_SimpleWindow(parent, x, y,
					 width, height + 16,
					 bd_width, WINOBJ_BORDER_CONVEX);

    if(bar == NULL)
    {
	for(i = 0; i < k; i++)
	    winobj_free(name_list[i]);
	winobj_free(name_list);
	winobj_free(name_width_list);
	winobj_free(name_x_list);
	winobj_free(popup_list);
	DestroyWinObj((WinObj *)bar);
	return NULL;
    }

    for(i = 0; i < n; i++)
	popup_list[i]->popup_menu.bar_ref = (MenuBar *)bar;
    if(n < k && popup_list[n] != NULL)
	popup_list[n]->popup_menu.bar_ref = (MenuBar *)bar;
    bar->type = MenuBar_T;

    bar->nmenu = k;
    bar->name_list = name_list;
    bar->name_width_list = name_width_list;
    bar->name_x_list = name_x_list;
    bar->popup_list = popup_list;
    bar->font = font;
    bar->popup_menu_index = -1;
    bar->bd_width = bd_width;
    bar->switch_menu_proc = switch_menu_proc;
    bar->winobj_destruction = menubar_destruction;

    SetEventMask((WinObj *)bar, ExposureMask | ButtonPressMask |
		 PointerMotionMask);
    SetEventHandler((WinObj *)bar, bar_event_handler);

    return (WinObj *)bar;
}
