#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "WinObj.h"
#include "colcnv.h"

static void hscroll_draw_thumb(NamedHScrollbar* hs)
{
    XCopyArea(hs->disp, hs->pm,
	      hs->win, hs->gc,
	      0, 0,
	      hs->len + 2 * hs->bd_width, hs->thick,
	      hs->top, 0);
}

static void set_pos(NamedHScrollbar* hs, double pos)
{
    int top;

    if(pos < 0.0)
    {
	pos = 0.0;
	top = 0;
    }
    else if(pos > 1.0)
    {
	pos = 1.0;
	top = hs->top = hs->maxlen - hs->len;
    }
    else
	top = (int)((hs->maxlen - hs->len) * pos);

    hs->pos = pos;
    hs->top = top;
    XClearWindow(hs->disp, hs->win);
    hscroll_draw_thumb(hs);
}

static void move_thumb(NamedHScrollbar* hs, int p)
{
    int ptr, max;
    double pos;

    max = hs->width - hs->len;
    ptr = p - hs->len / 2;

    pos = (double)ptr / max;

    SetHScrollbarThumbPos((HScrollbar *)hs, pos);
}

static void hscroll_event_handler(NamedHScrollbar* hs, XEvent* e)
{
    static Bool in;

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count > 0)
	    return;
	hscroll_draw_thumb(hs);
	break;
      case ButtonPress:
	hs->pm = hs->pm2;
	hs->top = -1;
	move_thumb(hs, e->xbutton.x);
	CallCallbackProc((WinObj *)hs, e);
	in = True;
	break;
      case MotionNotify:
	if(!in)
	    break;
	hs->nmoves++;
	CompressMotionEvent(hs->disp, hs->win, e);
	move_thumb(hs, e->xmotion.x);
	CallCallbackProc((WinObj *)hs, e);
	break;
      case ButtonRelease:
	hs->pm = hs->pm1;
	XClearWindow(hs->disp, hs->win);
	hscroll_draw_thumb(hs);
	hs->nmoves = 0;
	in = False;
	break;
    }
}

static void init_draw_thumb(NamedHScrollbar* hs)
{
    Display* disp = hs->disp;
    int bd = hs->bd_width;
    int bd2 = 2 * bd;

    XSetForeground(disp, hs->gc, hs->bg);
    XFillRectangle(disp, hs->pm1, hs->gc, 0, 0, hs->width, hs->thick);
    XSetForeground(disp, hs->gc, hs->mark_color);
    XFillRectangle(disp, hs->pm2, hs->gc, 0, 0, hs->width, hs->thick);

    DrawBorder(disp, hs->pm1, hs->gc, 0, 0, hs->len + bd2, hs->thick, bd,
	       hs->lbg, hs->hbg, True);
    DrawBorder(disp, hs->pm2, hs->gc, 0, 0, hs->len + bd2, hs->thick, bd,
	       hs->lbg, hs->hbg, False);

    if(hs->len > 4)
    {
	DrawRuledVLine(disp, hs->pm1, hs->gc,
		       hs->lbg, hs->hbg, hs->len / 2 + bd - 1, bd - 1,
		       2);
	DrawRuledVLine(disp, hs->pm1, hs->gc,
		       hs->lbg, hs->hbg, hs->len / 2 + bd - 1,
		       hs->thick - bd - 2,
		       2);

	if(IsMonoDisplay(hs))
	{
	    DrawRuledVLine(disp, hs->pm2, hs->gc,
			   hs->bg, hs->bg, hs->len / 2 + bd - 1, bd, 1);
	    DrawRuledVLine(disp, hs->pm2, hs->gc,
			   hs->bg, hs->bg, hs->len / 2 + bd - 1,
			   hs->thick - bd - 2, 1);
	}
	else
	{
	    DrawRuledVLine(disp, hs->pm2, hs->gc,
			   hs->lbg, hs->hbg, hs->len / 2 + bd - 1, bd, 1);
	    DrawRuledVLine(disp, hs->pm2, hs->gc,
			   hs->lbg, hs->hbg, hs->len / 2 + bd - 1,
			   hs->thick - bd - 2, 1);
	}
    }
}

static void init_pixmap(NamedHScrollbar* hs)
{
    hs->pm1 = CreatePixmap(hs->disp, hs->win,
			   hs->width, hs->thick, hs->bg);
    hs->pm2 = CreatePixmap(hs->disp, hs->win,
			   hs->width, hs->thick, hs->mark_color);
    init_draw_thumb(hs);
}

static void nhscrollbar_resize(WinObj* obj, int width, int height)
{
    NamedHScrollbar* hs = (NamedHScrollbar *)obj;
    Bool pm1_flag;

    if(hs->width >= width && hs->height >= height)
	return;
    hs->width = width;
    hs->height = hs->thick = height;
    hs->len = (width * hs->shown);

    pm1_flag = (hs->pm == hs->pm1);
    XFreePixmap(hs->disp, hs->pm1);
    XFreePixmap(hs->disp, hs->pm2);
    init_pixmap(hs);
    if(pm1_flag)
	hs->pm = hs->pm1;
    else
	hs->pm = hs->pm2;

    XResizeWindow(hs->disp, hs->win, width, height);
    XClearWindow(hs->disp, hs->win);
    hscroll_draw_thumb(hs);
}

static void nhscrollbar_destruction(WinObj* obj)
{
    XFreePixmap(obj->named_hscr.disp, obj->named_hscr.pm1);
    XFreePixmap(obj->named_hscr.disp, obj->named_hscr.pm2);
}

WinObj* create_NamedHScrollbar(WinObj* parent,
			  int x,
			  int y,
			  int width,
			  int height,
			  int bd_width,
			  double shown)
{
    NamedHScrollbar* hs = (NamedHScrollbar *)
	create_SimpleWindow(parent, x, y, width, height, bd_width,
			    WINOBJ_BORDER_CONCAVE);
    if(hs == NULL)
	return NULL;

    hs->type = NamedHScrollbar_T;
    hs->thick = hs->height;
    hs->maxlen = hs->width - 2 * bd_width;
    hs->len   = (int)(hs->maxlen * shown);
    hs->pos   = 0.0;
    hs->top   = 0;
    hs->bar_type = '\0';
    hs->nmoves = 0;
    hs->set_pos = set_pos;
    hs->shown = shown;
    hs->bd_width = bd_width;

    init_pixmap(hs);
    hs->pm = hs->pm1;
    hs->winobj_resize_proc = nhscrollbar_resize;
    hs->winobj_destruction = nhscrollbar_destruction;

    SetEventMask((WinObj *)hs, ExposureMask | ButtonMotionMask |
		               ButtonPressMask | ButtonReleaseMask);
    SetEventHandler((WinObj *)hs, hscroll_event_handler);

    return (WinObj *)hs;
}

static double calcLvalue(Display* disp, Pixel x)
{
    double r, g, b;
    rgb_t rgb;
    double h, l, s;

    PixelToRGB(disp, x, &r, &g, &b);
    rgb.r = r;
    rgb.g = g;
    rgb.b = b;
    rgb_to_hls(&rgb, &h, &l, &s);
    return l;
}    

void SetNamedHScrollbarBartype(NamedHScrollbar* hs, char ch,
			       CharFont* font)
{
    Pixel name_bg = 0;

    if(IsMonoDisplay(hs))
    {
	DrawStringN(hs->disp, hs->pm1, font, hs->fg, 0, 0, hs->len, hs->thick,
		    &ch, 1, Center);
	DrawStringN(hs->disp, hs->pm2, font, hs->bg, 0, 0, hs->len, hs->thick,
		    &ch, 1, Center);
    }
    else
    {
	if(calcLvalue(hs->disp, hs->bg) < 0.4)
	    name_bg = hs->hbg;
	else
	    name_bg = hs->lbg;
	DrawStringN(hs->disp, hs->pm1, font, name_bg, 0, 0, hs->len, hs->thick,
		    &ch, 1, Center);
	DrawStringN(hs->disp, hs->pm2, font, name_bg, 0, 0, hs->len, hs->thick,
		    &ch, 1, Center);
    }
}
