#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <string.h>
#include "WinObj.h"

#define NO_BUTTON_WIDTH  50
#define NO_BUTTON_HEIGHT 30
#define YES_BUTTON_WIDTH  50
#define YES_BUTTON_HEIGHT 30
#define VGAP 30
#define HGAP 30
#define MIN_YESNO_DIALOG_WIDTH (3 * HGAP + YES_BUTTON_WIDTH + NO_BUTTON_WIDTH)

static void redraw_dialog(QueryYesNo* qyn)
{
    DrawBorder(qyn->disp, qyn->win, qyn->gc,
	       0, 0, qyn->width, qyn->height,
	       qyn->bd_width, qyn->lbg, qyn->hbg, True);
    DrawBorder(qyn->disp, qyn->win, qyn->gc,
	       qyn->bd_width * 2, 
	       qyn->bd_width * 2,
	       qyn->width  - qyn->bd_width * 4,
	       qyn->height - qyn->bd_width * 4,
	       qyn->bd_width, qyn->lbg, qyn->hbg, False);
    DrawString(qyn->disp, qyn->win, qyn->font, qyn->fg,
	       0,
	       0,
	       qyn->width,
	       SkeletonWinObj(qyn->yes_button)->any.y, qyn->message, Center);
}

static void qyn_event_proc(WinObj* obj, XEvent* e);
static void restack_window(WinObj* obj)
{
    XEvent e;
    Display* disp = obj->any.disp;

    /* remove expose event */
    XSync(disp, False);
    while(XCheckTypedEvent(disp, Expose, &e) == True)
    {
	qyn_event_proc(obj, &e);
	XSync(disp, False);
    }

    /* raise window */
    XRaiseWindow(disp, obj->any.win);
    XSync(disp, False);
    if(XCheckTypedEvent(disp, Expose, &e) == True)
	qyn_event_proc(obj, &e);
    else
	XLowerWindow(disp, obj->any.win);
    XFlush(disp);
}

static void qyn_event_proc(WinObj* obj, XEvent* e)
{
    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count == 0 &&
	   obj->msg_dialog.message != NULL)
	    redraw_dialog((QueryYesNo *)obj);
	break;
      case ButtonPress:
	restack_window(obj);
	break;
    }
}

static void yes_button_proc(WinObj* obj, XEvent* e)
{
    QueryYesNo* qyn = (QueryYesNo *)obj->any.ext_data;
    XUnmapWindow(qyn->disp, qyn->win);
    qyn->mapped = False;
    if(qyn->message != NULL)
    {
	winobj_free(qyn->message);
	qyn->message = NULL;
    }
    XFlush(qyn->disp);
    if(qyn->yes_selected != NULL)
	qyn->yes_selected(qyn);
    XFlush(qyn->disp);
}

static void no_button_proc(WinObj* obj, XEvent* e)
{
    QueryYesNo* qyn = (QueryYesNo *)obj->any.ext_data;
    XUnmapWindow(qyn->disp, qyn->win);
    qyn->mapped = False;
    if(qyn->message != NULL)
    {
	winobj_free(qyn->message);
	qyn->message = NULL;
    }
    XFlush(qyn->disp);
    if(qyn->no_selected != NULL)
	qyn->no_selected(qyn);
    XFlush(qyn->disp);
}

static void qyndestruction(WinObj* obj)
{
    QueryYesNo* qyn = (QueryYesNo *)obj;
    winobj_free(qyn->message);
}

WinObj* create_QueryYesNo(Display* disp,
			  int bd_width,
			  Pixel fg,
			  Pixel bg,
			  CharFont* font)
{
    QueryYesNo* qyn = (QueryYesNo *)
	toplevel_WinObj(disp, NULL, NULL, 1, 1, fg, bg);

    if(qyn == NULL)
	return NULL;

    qyn->yes_button = (ButtonWindow *)create_Button((WinObj *)qyn,
						    0,
						    0,
						    YES_BUTTON_WIDTH,
						    YES_BUTTON_HEIGHT,
						    bd_width,
						    font,
						    " Yes ");
    if(qyn->yes_button == NULL)
    {
	DestroyWinObj((WinObj *)qyn);
	return NULL;
    }

    qyn->no_button = (ButtonWindow *)create_Button((WinObj *)qyn,
						   0,
						   0,
						   YES_BUTTON_WIDTH,
						   NO_BUTTON_HEIGHT,
						   bd_width,
						   font,
						   " No ");
    if(qyn->no_button == NULL)
    {
	DestroyWinObj((WinObj *)qyn->yes_button);
	DestroyWinObj((WinObj *)qyn);
	return NULL;
    }

    qyn->yes_button->ext_data = qyn;
    qyn->no_button->ext_data = qyn;
    SetCallBackProc((WinObj *)qyn->yes_button, yes_button_proc);
    SetCallBackProc((WinObj *)qyn->no_button, no_button_proc);

    qyn->type = QueryYesNo_T;
    qyn->font = font;
    qyn->bd_width = bd_width;
    qyn->when_realize_mapping = False;
    qyn->message = NULL;
    qyn->mapped = False;
    qyn->yes_selected = NULL;
    qyn->no_selected  = NULL;

    {
	XSetWindowAttributes at;
	at.override_redirect = True;
	at.save_under = True;
	XChangeWindowAttributes(qyn->disp, qyn->win,
				CWOverrideRedirect | CWSaveUnder, &at);
    }

    qyn->winobj_destruction = qyndestruction;
    SetEventMask((WinObj *)qyn, ExposureMask | ButtonPressMask);
    SetEventHandler((WinObj *)qyn, qyn_event_proc);

    return (WinObj *)qyn;
}

void QueryYesNoProc(WinObj* obj, const char* msg)
{
    int len;
    int str_width, str_height;
    QueryYesNo* qyn = (QueryYesNo *)obj;
    int win_width, win_height;
    int win_x, win_y;
    int px, py;
    int scr;

    if(msg == NULL)
	return;
    len = strlen(msg);
    if(len == 0)
	return;

    if(obj->any.realized_flag == False)
	WinObjRealize(obj);

    scr = DefaultScreen(qyn->disp);

    str_width  = TextWidth(qyn->font, msg);
    str_height = TextHeight(qyn->font);

    win_width  = MAX(MIN_YESNO_DIALOG_WIDTH, str_width + 2 * HGAP);
    win_height = 3 * VGAP + str_height + YES_BUTTON_HEIGHT;

    MoveWinObj((WinObj*)qyn->yes_button,
	       win_width / 2 - HGAP / 2 - YES_BUTTON_WIDTH,
	       2 * VGAP + str_height);
    MoveWinObj((WinObj*)qyn->no_button,
	       win_width / 2 + HGAP / 2,
	       2 * VGAP + str_height);

    QueryPointer(qyn->disp, DefaultRootWindow(qyn->disp), &px, &py);
    win_x = px - win_width / 2;
    win_y = py - (2 * VGAP + str_height + YES_BUTTON_HEIGHT/2);

    if(win_x < 0)
	win_x = 0;
    else if(win_x + win_width > DisplayWidth(qyn->disp, scr))
	win_x = DisplayWidth(qyn->disp, scr) - win_width;

    if(win_y < 0)
	win_y = 0;
    else if(win_y + win_height > DisplayHeight(qyn->disp, scr))
	win_y = DisplayHeight(qyn->disp, scr) - win_height;

    ResizeWinObj(obj, win_width, win_height);
    MoveWinObj(obj, win_x, win_y);
    XRaiseWindow(qyn->disp, qyn->win);
    XMapWindow(qyn->disp, qyn->win);
    qyn->mapped = True;
    XWarpPointer(qyn->disp, None, qyn->yes_button->win, None, None, None, None,
		 YES_BUTTON_WIDTH / 2, YES_BUTTON_HEIGHT / 2);
    if((qyn->message = winobj_strdup(msg)) == NULL)
    {
	fprintf(stderr, "Can't allocate memory.\n");
	qyn->message = winobj_null_string;
    }

    XFlush(qyn->disp);
}
