#include <stdio.h>
#include <math.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <unistd.h>
#include "WinObj.h"

#define ITEM_TRIG_MARK_SIZE 8
#define ANIMAT_FRAME_USEC 10000
#define ANIMAT_FRAMES     8

static void redraw_select_menu(SelectMenu* sel, int idx)
{
    DrawString(sel->disp, sel->win, sel->font, sel->fg,
	       0, 0, sel->width, sel->height,
	       sel->select_popup->popup_menu.name_list[idx],
	       Center);

    DrawTriangle(sel->disp, sel->win, sel->gc,
		 sel->width - 4 - ITEM_TRIG_MARK_SIZE,
		 sel->height / 2,
		 1, (double)ITEM_TRIG_MARK_SIZE, -M_PI / 2,
		 sel->bg, sel->lbg, sel->hbg, True);
}

static void set_select_item(SelectMenu* sel, int idx)
{
    PopupMenu* popup = (PopupMenu *)sel->select_popup;

    if(popup->idlist[idx] == sel->selected_id)
	return;

    popup->marker[
	PopupMenuIdToIndex((WinObj *)popup, sel->selected_id, True)] = False;
    popup->marker[idx] = True;

    sel->selected_id = popup->idlist[idx];
    XClearWindow(sel->disp, sel->win);
    redraw_select_menu(sel, idx);
}

void SetSelectMenuItem(WinObj* obj, int menuid)
{
    SelectMenu* sel = (SelectMenu *)obj;
    set_select_item(sel, PopupMenuIdToIndex(sel->select_popup, menuid, True));
}

static void select_menu_event_handler(SelectMenu* sel, XEvent* e)
{
    int i, rx, ry;

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count == 0)
	{
	    redraw_select_menu(sel,
			       PopupMenuIdToIndex(sel->select_popup,
						  sel->selected_id, True));
	}
	break;
      case ButtonPress:
	QueryWindowPositionFromRootWindow(sel->disp,
					  SkeletonWinObj(sel)->any.win,
					  &rx, &ry);
	ry += SkeletonWinObj(sel)->any.height;
	XSync(sel->disp, False);
	for(i = 0; i <= ANIMAT_FRAMES; i++)
	{
	    XClearArea(sel->disp, sel->win, 
		       sel->width - 5 - 2 * ITEM_TRIG_MARK_SIZE,
		       0,
		       5 + 2 * ITEM_TRIG_MARK_SIZE,
		       sel->height,
		       False);
	    DrawTriangle(sel->disp, sel->win, sel->gc,
			 sel->width - 4 - ITEM_TRIG_MARK_SIZE,
			 sel->height / 2,
			 1, (double)ITEM_TRIG_MARK_SIZE,
			 -M_PI / 2.0 / ANIMAT_FRAMES * (double)i,
			 sel->bg, sel->lbg, sel->hbg, True);
	    XSync(sel->disp, False);
	    usleep(ANIMAT_FRAME_USEC);
	}

	if(PopupMenu_popup_proc(sel->select_popup, rx, ry))
	    set_select_item(sel,
			    sel->select_popup->popup_menu.selected_item_index);
	break;
    }
}

WinObj* create_SelectMenu(WinObj* parent,
			  int x,
			  int y,
			  CharFont* font,
			  int bd_width,
			  WinObj* popup,
			  int selected_id)
{
    int i, n;
    PopupMenu* menu = (PopupMenu *)popup;
    SelectMenu* sel = (SelectMenu *)
	create_SimpleWindow(parent, x, y,
			    menu->width,
			    8 + TextHeight(font),
			    bd_width, WINOBJ_BORDER_CONVEX);
    if(sel == NULL)
	return NULL;

    SetEventMask((WinObj *)sel, ExposureMask | ButtonPressMask);
    SetEventHandler((WinObj *)sel, select_menu_event_handler);

    sel->selected_id = selected_id;
    n = PopupMenuIdToIndex(popup, selected_id, True);
    for(i = 0; i < menu->nitems; i++)
	popup->popup_menu.marker[i] = False;
    popup->popup_menu.marker[n] = True;
    sel->font = font;
    sel->select_popup = popup;

    return (WinObj *)sel;
}
