#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <math.h>
#include "WinObj.h"

static void redraw_gear(HAdjustGear* va)
{
    Display* disp = va->disp;
    Window win = va->win;
    Pixmap pm = va->pm;
    GC gc = va->gc;
    int w, h;
    double r;
    double dr;

    w = va->width;
    h = va->height;
    r = 2.0 * va->r / h;

    if(r >= 0)
	r = fmod(r, 2 * M_PI);
    else
	r = 2 * M_PI - fmod(-r, 2 * M_PI);
    dr = 2.0 / va->measure;
    r = fmod(r, dr);

    XSetForeground(disp, gc, va->bg);
    XFillRectangle(disp, pm, gc, 0, 0, w, h);
    while(r < M_PI)
    {
	DrawRuledHLine(disp, pm, gc, va->lbg, va->hbg,
		       0,
		       (int)((1.0 + cos(M_PI - r)) * h * 0.5),
		       w);
	r += dr;
    }

    XCopyArea(disp, pm, win, gc,
	      0, 0, w, h, 0, 0);
}

static void vadjust_event_handler(VAdjustGear* va, XEvent* e)
{
    int diff;

    switch(e->type)
    {
      case ButtonPress:
	va->cur = e->xbutton.y;
	va->border_skeleton->style = WINOBJ_BORDER_CONVEX;
	RedrawBorderSkeletion((WinObj *)va->border_skeleton);
	break;
      case ButtonRelease:
	va->border_skeleton->style = WINOBJ_BORDER_CONCAVE;
	RedrawBorderSkeletion((WinObj *)va->border_skeleton);
	break;
      case MotionNotify:
	CompressMotionEvent(va->disp, va->win, e);
	diff = e->xmotion.y - va->cur;
	va->cur = e->xbutton.y;
	va->r += diff;
	va->diff = diff;
	redraw_gear(va);
	if(va->call_back_proc != NULL)
	    va->call_back_proc((WinObj *)va, e);
	break;
      case Expose:
	if(e->xexpose.count > 0)
	    return;
	redraw_gear(va);
	break;
    }
}	

static void vadjustgear_destruction(WinObj* obj)
{
    XFreePixmap(obj->va.disp, obj->va.pm);
}

WinObj* create_VAdjustGear(WinObj* parent,
			   int x,
			   int y,
			   int width,
			   int height,
			   int bd_width,
			   int measure)
{
    VAdjustGear* va = (VAdjustGear *)
	create_SimpleWindow(parent, x, y, width, height, bd_width,
			    WINOBJ_BORDER_CONCAVE);
    Display* disp;
    Window win;

    if(va == NULL)
	return NULL;

    disp = va->disp;
    win = va->win;
    va->type = VAdjustGear_T;
    va->winobj_destruction = vadjustgear_destruction;
    va->measure = measure;
    va->cur = 0;
    va->pm = CreatePixmap(disp, win, width, height, va->bg);
    va->diff = va->r = 0;

    SetEventMask((WinObj *)va, ButtonPressMask | ButtonMotionMask |
		 ButtonReleaseMask | ExposureMask);
    SetEventHandler((WinObj *)va, vadjust_event_handler);

    return (WinObj *)va;
}
