#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "WinObj.h"
#include "WinObjVersion.h"
#ifdef USE_LOCALE
#include <X11/Xlocale.h>
#endif

char* WinObjVersion = WINOBJ_VERSION;

static WinObj* WinObj_free_list = NULL;

/* #define WINOBJ_MALLOC_DEBUG */

#define NUMBER_OF_MAXDISP 10
Display** all_display_list = NULL;
int number_of_display = 0;
int sizeof_WinObj = sizeof(WinObj);
int display_visual_type = DISPLAY_VISUAL_DEFAULT;
char* winobj_null_string = "";
Bool locale_initialized = False;

#ifdef WINOBJ_MALLOC_DEBUG
struct memory_pointer_list
{
    void* ptr;
    Bool free_flag;
    struct memory_pointer_list* next;
};
static struct memory_pointer_list* memptr = NULL;
static struct memory_pointer_list* memptrall = NULL;
static void add_ptr(void* p)
{
    struct memory_pointer_list* v;
    struct memory_pointer_list* v2;

    v = (struct memory_pointer_list *)
	malloc(sizeof(struct memory_pointer_list));
    v2 = (struct memory_pointer_list *)
	malloc(sizeof(struct memory_pointer_list));
    if(v == NULL || v2 == NULL)
    {
	fprintf(stderr, "Can't trace the memory debug.\n");
	exit(1);
    }

    v->ptr = p;
    v->next = memptr;
    memptr = v;

    v2->ptr = p;
    v2->next = memptrall;
    memptrall = v2;
}
static void remove_ptr(void* p)
{
    struct memory_pointer_list* cur;
    struct memory_pointer_list* prev;
    int n;

    prev = NULL;
    for(cur = memptr; cur; cur = cur->next)
    {
	if(p == cur->ptr)
	    break;
	prev = cur;
    }

    if(cur == NULL)
    {
	fprintf(stderr, "%p  malloc Ƥʤ뤤 2  free Ƥޤ\n", p);

	for(cur = memptrall; cur; cur = cur->next)
	    if(cur->ptr == p)
	    {
		fprintf(stderr, "%p  1 ʾ malloc Ƥޤ", p);
		exit(1);
	    }
	exit(1);
    }

    if(prev == NULL)
    {
	memptr = memptr->next;
	free(cur);
    }
    else
    {
	prev->next = cur->next;
	free(cur);
    }

    for(n = 0, cur = memptr; cur; cur = cur->next, n++)
	;
    printf("%d ĤΥǤ\n", n);
}
#endif

void* winobj_malloc(int n)
{
    void* ptr;

    if((ptr = (void *)malloc(n)) == NULL)
    {
	CallWinObjSysErrorHandler(NULL);
	return NULL;
    }
#ifdef WINOBJ_MALLOC_DEBUG
    add_ptr(ptr);
#endif

    return ptr;
}

void* winobj_realloc(void* ptr, int n)
{
    void* new_ptr;

    if((new_ptr = realloc(ptr, n)) == NULL)
    {
	CallWinObjSysErrorHandler(NULL);
	return NULL;
    }
#ifdef WINOBJ_MALLOC_DEBUG
    if(new_ptr != ptr)
    {
	remove_ptr(ptr);
	add_ptr(new_ptr);
    }
#endif
    return new_ptr;
}

void winobj_free(void* ptr)
{
    if(ptr != NULL && ptr != winobj_null_string)
    {
#ifdef WINOBJ_MALLOC_DEBUG
	remove_ptr(ptr);
#endif
	free(ptr);
    }
}

char* winobj_strdup(const char* string)
{
    char* ptr;

    if(string == NULL)
	return winobj_null_string;

    if((ptr = strdup(string)) == NULL)
    {
	CallWinObjSysErrorHandler(NULL);
	return NULL;
    }
#ifdef WINOBJ_MALLOC_DEBUG
    add_ptr(ptr);
#endif
    return ptr;
}

WinObj* new_WinObj(void)
{
    WinObj* obj;

    if(WinObj_free_list == NULL)
	obj = (WinObj *)winobj_malloc(sizeof_WinObj);
    else
    {
	obj = WinObj_free_list;
	WinObj_free_list = WinObj_free_list->any.parent;
    }

    obj->any.parent = NULL;
    return obj;
}

static void reuse_WinObj(WinObj* obj)
{
    obj->any.disp = NULL;
    obj->any.parent = WinObj_free_list;
    WinObj_free_list = obj;
}

void SetCallBackProc(WinObj* obj, call_back_t proc)
{
    obj->any.call_back_proc = proc;
}

void VisitEachWinObj(WinObj* root, void (* visitor)(WinObj* obj))
{
    if(root != NULL)
    {
	WinObjList* children;
	visitor(root);
	for(children = root->any.children; children; children = children->next)
	    VisitEachWinObj(children->obj, visitor);
    }
}

static void WinObjRealize1(WinObj* obj)
{
    AnyWindow* any = (AnyWindow *)obj;

    if(any->realized_flag == True)
    {
	if(any->when_realize_mapping)
	    XMapWindow(any->disp, any->win);
	return;
    }

    XSelectInput(any->disp, any->win, (long)any->event_mask);
    if(any->event_handler != NULL)
	install_event_handler(any->disp, any->win, (WinObj *)any);
    if(any->when_realize_mapping)
	XMapWindow(any->disp, any->win);
    any->realized_flag = True;
}

void WinObjRealize(WinObj* obj)
{
    VisitEachWinObj(obj, WinObjRealize1);
}

static Pixel calc_mark_color_pixel(Display* disp, Pixel fg, Pixel bg)
{
    double r1, g1, b1;
    double r2, g2, b2;
    double r3, g3, b3;

    PixelToRGB(disp, fg, &r1, &g1, &b1);
    PixelToRGB(disp, bg, &r2, &g2, &b2);
    r3 = r2 * 0.8 + r1 * 0.2;
    g3 = g2 * 0.8 + g1 * 0.2;
    b3 = b2 * 0.8 + b1 * 0.2;
    return AllocRGBColor(disp, r3, g3, b3);
}

WinObj* toplevel_WinObj(Display* disp,
			const char* wname,
			const char* iname,
			int width,
			int height,
			Pixel fg,
			Pixel bg)
{
    AnyWindow* obj = (AnyWindow *)new_AnyWindow();
    int visual_class;
    XGCValues gcv;

#ifdef USE_LOCALE
    if(locale_initialized == False)
    {
	locale_initialized = True;
	if(setlocale(LC_ALL, "") == NULL)
	    fprintf(stderr, "Warning: Can't set the locale.\n");
    }
#endif

    if(obj == NULL)
	return NULL;

    visual_class = DefaultVisual(disp, DefaultScreen(disp))->class;

    if(visual_class == StaticGray || visual_class == GrayScale ||
       display_visual_type == DISPLAY_VISUAL_MONO)
	obj->mono_color = True;
    else
	obj->mono_color = False;

    if(all_display_list == NULL)
    {
	all_display_list = (Display **)malloc(NUMBER_OF_MAXDISP * 
					      sizeof(Display *));

	if(all_display_list == NULL) /* Fatal error */
	{
	    fprintf(stderr, "WinObj Fatal Error: Can't allocate memory for the display list.");
	    exit(1);
	}
	number_of_display = 1;
	all_display_list[0] = disp;

#ifdef DEBUG
	printf("sizeof_WinObj = %d\n", sizeof_WinObj);
#endif
    }
    else
    {
	int i;
	Bool newdisp = True;

	for(i = 0; i < number_of_display; i++)
	    if(all_display_list[i] == disp)
	    {
		newdisp = False;
		break;
	    }

	if(newdisp == True)
	{
	    if(number_of_display == NUMBER_OF_MAXDISP)
	    {
		CallWinObjErrorHandler(WOE_TOO_MANY_OPEN_DISPLAY, NULL);
		DestroyWinObj((WinObj *)obj);
		return NULL;
	    }

	    all_display_list[number_of_display++] = disp;
	}
    }

    obj->disp = disp;
    obj->win = CreateWindow(disp, None, 0, 0, width, height, bg);

    if(wname != NULL)
	XStoreName(disp, obj->win, wname);
    if(iname != NULL)
	XSetIconName(disp, obj->win, iname);

    gcv.graphics_exposures = False;
    obj->gc = XCreateGC(disp, obj->win, GCGraphicsExposures, &gcv);
    if(obj->gc == NULL)
    {
	CallWinObjSysErrorHandler(NULL);
	DestroyWinObj((WinObj *)obj);
	return NULL;
    }

    obj->width = width;
    obj->height = height;
    obj->fg = fg;
    obj->bg = bg;

    if(IsMonoDisplay(obj))
    {
	obj->mark_color = fg;
	obj->lbg = fg;
	obj->hbg = fg;
    }
    else
    {
	obj->mark_color = calc_mark_color_pixel(disp, fg, bg);
	AllocLowHighColor(disp, obj->bg, &obj->lbg, &obj->hbg);
    }

    return (WinObj *)obj;
}

void SetEventMask(WinObj* obj, unsigned long mask)
{
    obj->any.event_mask = mask;
    if(obj->any.realized_flag == True)
	XSelectInput(obj->any.disp, obj->any.win, (long)mask);
}

void SetEventHandler(WinObj* obj, void (* event_handler)())
{
    obj->any.event_handler = event_handler;
}

void CallCallbackProc(WinObj* obj, XEvent* e)
{
    AnyWindow* any = (AnyWindow *)obj;

    if(any->call_back_proc != NULL)
	any->call_back_proc(obj, e);
}

void ResizeWinObj(WinObj* obj, int width, int height)
{
    if(obj->any.winobj_resize_proc != NULL)
    {
	obj->any.winobj_resize_proc(obj, width, height);
	return;
    }

    if(obj->any.winobj_resize_proc != NULL)
	obj->any.winobj_resize_proc(obj, width, height);
    else
	XResizeWindow(obj->any.disp, obj->any.win, width, height);
    obj->any.width  = width;
    obj->any.height = height;
    if(obj->any.border_skeleton)
    {
	BorderSkeleton* bs = (BorderSkeleton *)obj->any.border_skeleton;
	bs->width  = width  + 2 * bs->bd_width;
	bs->height = height + 2 * bs->bd_width;
	XResizeWindow(bs->disp, bs->win, bs->width, bs->height);
	RedrawBorderSkeletion((WinObj *)bs);
    }
    if(obj->any.type == LineText_T ||
       obj->any.type == DirLineText_T)
    {
	DrawString(obj->text.disp, obj->text.win, obj->text.font,
		   obj->text.fg, 0, 0, 0, obj->text.height,
		   obj->text.prompt, West);
    }
}

void MoveWinObj(WinObj* obj, int x, int y)
{
    AnyWindow* w = (AnyWindow *)SkeletonWinObj(obj);
    w->x = x;
    w->y = y;
    XMoveWindow(w->disp, w->win, x, y);
}

static void default_destruction_proc_first(WinObj* obj)
{
    AnyWindow* any = (AnyWindow *)obj;

    if(obj->any.winobj_destruction_hook != NULL)
    {
	obj->any.winobj_destruction_hook(obj);
    }

    /* Ƥȼʬ GC ۤʤ GC 롥*/
    if(any->gc != NULL &&
       (any->parent == NULL ||
	any->parent->any.gc != any->gc))
	XFreeGC(any->disp, any->gc);

    if(any->winobj_destruction != NULL)
	any->winobj_destruction(obj);
}

static void default_destruction_proc_second(WinObj* obj)
{
    uninstall_event_handler(obj->any.disp, obj->any.win);
    obj->any.disp = NULL;
}

static void free_win_obj(WinObj* obj)
{
    WinObjList* children;

    if(obj == NULL)
	return;

    children = obj->any.children;
    while(children)
    {
	WinObjList* tmp;

	tmp = children;
	free_win_obj(children->obj);
	children = children->next;
	winobj_free(tmp);
    }

    reuse_WinObj(obj);
}

void DestroyWinObj(WinObj* obj)
{
    if(obj == NULL)
	return;

    obj = SkeletonWinObj(obj);

    /* 桼 destruction hook θƤӽФ
     * GC γ
     * WinObj Ѥ destruction θƤӽФ
     */
    VisitEachWinObj(obj, default_destruction_proc_first);

    /* ҥɥ˲ */
    if(obj->any.win != None)
	XDestroyWindow(obj->any.disp, obj->any.win);

    /* WinObj ٥ȥϥɥ顼鼫ʬ */
    VisitEachWinObj(obj, default_destruction_proc_second);

    /* Ƥ鼫ʬڤΥ */
    if(obj->any.parent != NULL)
    {
	WinObjList* children;
	for(children = obj->any.parent->any.children; children;
	    children = children->next)
	    if(children->obj == obj)
	    {
		children->obj = NULL;
		break;
	    }
    }

    /* WinObj ˳Ƥ줿 */
    free_win_obj(obj);
}

Bool WinObjPopupAble(WinObj* obj)
{
    return obj->any.type == PopupMenu_T ||
	obj->any.type == DrawerPopupMenu_T;
}

#ifdef NoUSleep
#include <sys/time.h>
/*
 * sleep microseconds
 */
static int usleep(unsigned long usec)
{
    struct timeval tv;
    tv.tv_sec = usec / 1000000;
    tv.tv_usec = (usec % 1000000) * 1000000;

    return select(0, NULL, NULL, NULL, &tv);
}
#endif
