#include <stdio.h>
#include <stdlib.h>
#include "jutils.h"
#define ESC 0x1b

/* ascii-mode           ESC        $          [@B]
 *  ----+---->{0}----+----->{1}--+----->{2}--+-----> kanji-mode
 *     /|\           |           |           |
 *      |            |           |           | [(]      [B]
 *      +------------+[^ESC]     |           +----->{3}-----> kanji-mode
 *      |                        |           |
 *      +------------------------+[^$]	     |
 *      |				     |
 *	+------------------------------------+[^@B)]
 *
 * kanji-mode           ESC          (         [JB]
 *  ----+---->{0}----+----->{1}--+----->{2}--+-----> ascii-mode
 *     /|\           |           |           |
 *      |            |           |           |
 *      +------------+[^ESC]     |           |
 *      |                        |           |
 *      +------------------------+[^(]	     |
 *      |			             |
 *	+------------------------------------+[^JB]
 */

/* eptr can point to jptr */
size_t jistoeuc(unsigned char* eptr,
		unsigned char* jptr, size_t jlength)
{
    unsigned char* eptr_begin = eptr;
    size_t j;

/* Be careful: eptr - eptr_begin <= j */
    j = 0;
  ascii:
    while(j < jlength && jptr[j] != ESC)
    {
	int c = jptr[j++];
	if(c >= ' ' || c == '\n' || c == '\t') /* skip control sequences */
	    *eptr++ = c;
	else if(c == '\r' && jptr[j] != '\n') /* for *junet*mac */
	    *eptr++ = '\n';
    }
    if(j >= jlength)
	goto done;

    /* 1 */
    j++;
    if(jptr[j] != '$')		/* ?? \033[^$]  */
    {
	if(j == jlength)
	    goto done;
	goto ascii;
    }

    /* 2 */
    j++;
    if(jptr[j] == '(' || jptr[j] == ')')
      j++;
    if(jptr[j] != '@' && jptr[j] != 'B') /* ?? \033$[^@B] */
    {
	if(j == jlength)
	    goto done;
	goto ascii;
    }

  kanji:
    j++;
    /* -1 for truncated jis character */
    while(j < jlength - 1 && jptr[j] != ESC)
    {
	/* assume that correct jis code */
	*eptr++ = (jptr[j++] | 0x80);
	*eptr++ = (jptr[j++] | 0x80);
    }

    if(j == jlength - 1)
	*eptr++ = jptr[j++];
    if(j >= jlength)
	goto done;

    /* 1 */
    j++;
    if(jptr[j] != '(')
    {
	if(j == jlength)
	    goto done;
	goto kanji;
    }

    /* 2 */
    j++;
    if(jptr[j] != 'J' && jptr[j] != 'B')
    {
	if(j == jlength)
	    goto done;
	goto kanji;
    }

    j++;
    goto ascii;

  done:
    *eptr = '\0';		/* terminate a string */
    return (size_t)(eptr - eptr_begin);
}

int boeucwcp(const unsigned char* body, long pos)
{
    while(pos > 0 && (body[pos] & 0x80))
    {
	if((body[pos - 1] & 0x80) == 0)
	    return 1;
	pos -= 2;
    }
    return pos == 0;
}

#ifdef JISTOEUC
void main(void)
{
    char buff[BUFSIZ];
    int jlen;

    while(fgets(buff, BUFSIZ, stdin) != NULL)
    {
	jlen = strlen(buff);
	jistoeuc((unsigned char *)buff, (unsigned char *)buff, jlen);
	fputs(buff, stdout);
    }
}
#endif /* JISTOEUC */
