#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "mblock.h"

#ifndef False
#define False 0
#endif

#ifndef True
#define True 1
#endif

#ifndef Bool
#define Bool int
#endif


static MBlockNode* free_mblock_list = NULL;
#define ADDRALIGN sizeof(void *) /* must be 2^n */

void init_mblock(MBlockList* mblock)
{
    mblock->first = NULL;
    mblock->allocated = 0;
}

static MBlockNode* new_mblock_node(size_t n)
{
    MBlockNode* p;

    if(n > MIN_MBLOCK_SIZE)
    {
	p = (MBlockNode *)malloc(sizeof(MBlockNode));
	p->buffer = (char *)malloc(n);
	p->block_size = n;
    }
    else if(free_mblock_list == NULL)
    {
	p = (MBlockNode *)malloc(MIN_MBLOCK_SIZE);
	p->buffer = (char *)malloc(MIN_MBLOCK_SIZE);
	p->block_size = MIN_MBLOCK_SIZE;
    }
    else
    {
	p = free_mblock_list;
	free_mblock_list = free_mblock_list->next;
    }

    p->offset = 0;
    p->next = NULL;

    return p;
}

static Bool enough_block_memory(MBlockList* mblock, size_t n)
{
    size_t newoffset;

    if(mblock->first == NULL)
	return False;

    newoffset = mblock->first->offset + n;

    if(newoffset < mblock->first->offset) /* exceed representable in size_t */
	return False;

    if(newoffset > mblock->first->block_size)
	return False;

    return True;
}

void* new_segment(MBlockList* mblock, size_t nbytes)
{
    MBlockNode* p;
    void* addr;

    /* alignment must be multiple of ADDRALIGN */
    nbytes = ((nbytes + ADDRALIGN - 1) & ~(ADDRALIGN - 1));

    if(!enough_block_memory(mblock, nbytes))
    {
	p = new_mblock_node(nbytes);
	p->next = mblock->first;
	mblock->first = p;
	mblock->allocated += p->block_size;
    }
    else
	p = mblock->first;

    addr = (void *)(p->buffer + p->offset);
    p->offset += nbytes;
    return addr;
}

static void reuse_mblock1(MBlockNode* p)
{
    if(p->block_size > MIN_MBLOCK_SIZE)
	free(p->buffer);
    else /* p->block_size == MIN_MBLOCK_SIZE */
    {
	p->next = free_mblock_list;
	free_mblock_list = p;
    }
}

void reuse_mblock(MBlockList* mblock)
{
    MBlockNode* p;

    if((p = mblock->first) == NULL)
	return;			/* There is nothing to collect memory */

    while(p)
    {
	MBlockNode* tmp;

	tmp = p;
	p = p->next;
	reuse_mblock1(tmp);
    }
    init_mblock(mblock);
}

char* strdup_mblock(MBlockList* mblock, const char* str)
{
    int len;
    char* p;

    len = strlen(str);
    p = (char *)new_segment(mblock, len + 1); /* for '\0' */
    memcpy(p, str, len + 1);
    return p;
}
