#ifndef ___MBLOCK_H_
#define ___MBLOCK_H_


/* Memory block for decreasing malloc
 *
 * +------+    +------+             +-------+
 * |BLOCK1|--->|BLOCK2|---> ... --->|BLOCK N|---> NULL
 * +------+    +------+             +-------+
 *
 *
 * BLOCK:
 * +-----------------------+
 * | memory 1              |
 * |                       |
 * +-----------------------+
 * | memory 2              |
 * +-----------------------+
 * | memory 3              |
 * |                       |
 * |                       |
 * +-----------------------+
 * | unused ...            |
 * +-----------------------+
 */


#define MIN_MBLOCK_SIZE 8192

typedef struct _MBlockNode
{
    char* buffer;
    size_t block_size;
    size_t offset;
    struct _MBlockNode* next;
} MBlockNode;

typedef struct _MBlockList
{
    MBlockNode* first;
    size_t allocated;
} MBlockList;

void init_mblock(MBlockList* mblock);
void* new_segment(MBlockList* mblock, size_t nbytes);
void reuse_mblock(MBlockList* mblock);
char* strdup_mblock(MBlockList* mblock, const char* str);


#endif /* ___MBLOCK_H_ */
