#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <errno.h>
#include "WinObj.h"

#define ERROR_TEXT_PADDING ": "
#define ERROR_TEXT_PADDING_SIZE 3 /* ": " + '\0'*/
static char text_buff[BUFSIZ + ERROR_TEXT_PADDING_SIZE];

static const char* error_text_table[] =
{
    "Too many open display",		/* WOE_TOO_MANY_OPEN_DISPLAY */
    "Too many defined event handler",	/* WOE_TOO_MANY_HANDLER */
    "Can't grab the pointer",		/* WOE_CANT_GRAB_POINTER, */
    "Can't find home direcotry",	/* WOE_CANT_GET_HOME */
    "Unknown user name",	        /* WOE_UNKNOWN_USER_NAME */
    "Filename too long",		/* WOE_FNAME_TOO_LONG */
};

static const char* error_jtext_table[] =
{
    "¿Υǥץ쥤򥪡ץ󤷤Ǥ",/* WOE_TOO_MANY_OPEN_DISPLAY */
    "٥ȥϥɥ顼Υơ֥뤬դޤ", /* WOE_TOO_MANY_HANDLER */
    "ݥ󥿤Υ֤˼Ԥޤ", /* WOE_CANT_GRAB_POINTER */
    "ۡǥ쥯ȥ꤬ĤޤǤ",	/* WOE_CANT_GET_HOME */
    "ꤵ줿桼̾ϸĤޤǤ", /* WOE_UNKNOWN_USER_NAME */
    "ե̾Ĺޤ",			/* WOE_FNAME_TOO_LONG */
};

static const char* get_error_text(int error_code)
{
  //extern char *sys_errlist[];
    char* lang;

    if(error_code < WOE_MIN_ERROR_CODE)
	return sys_errlist[error_code];
    if((lang = getenv("LANG")) != NULL &&
       (!strcmp(lang, "japanese") || !strcmp(lang, "ja")))
	return error_jtext_table[error_code - WOE_MIN_ERROR_CODE - 1];
    return error_text_table[error_code - WOE_MIN_ERROR_CODE - 1];
}

static Bool default_woeh(int error_code, const char* error_text)
{
    fprintf(stderr, "%s (WinObj error code = %d)\n", error_text, error_code);
    return False;
}
int (* WinObjErrorHandler)(int error_code, const char* error_text)
    = default_woeh;

Bool (* WinObjSetErrorHandler(int (* handler)(int error_code,
					      const char* error_text)))
    (int error_code, const char* error_text)
{
    int (* orig_woeh)(int error_code, const char* error_text)
	= WinObjErrorHandler;
    WinObjErrorHandler = handler;
    return orig_woeh;
}

void CallWinObjErrorHandler(int error_code, const char* msg)
{
    if(msg == NULL)
	strcpy(text_buff, get_error_text(error_code));
    else
    {
	int msglen = strlen(msg);
	const char* errmsg = get_error_text(error_code);
	int errlen = strlen(errmsg);

	if(msglen + 2 >= BUFSIZ)
	{
	    strncpy(text_buff, msg, BUFSIZ);
	    text_buff[BUFSIZ] = '\0';
	}
	else
	{
	    sprintf(text_buff, "%s" ERROR_TEXT_PADDING, msg);
	    if(msglen + errlen + 2 < BUFSIZ)
		strcat(text_buff, errmsg);
	    else
	    {
		errlen = msglen + errlen - BUFSIZ;
		strncat(text_buff, errmsg, errlen);
		text_buff[BUFSIZ] = '\0';
	    }
	}
    }

    if(WinObjErrorHandler(error_code, text_buff))
	exit(1);
}

void CallWinObjSysErrorHandler(const char* msg)
{
    CallWinObjErrorHandler(errno, msg);
}
