#ifndef ___XMISC_H_
#define ___XMISC_H_

#include <sys/types.h>
#include <sys/time.h>		/* for struct timeval */
#include "winobj_defs.h"

#define TextHeight(font) MAX((font)->font8->max_bounds.ascent + (font)->font8->max_bounds.descent, (font)->font16->max_bounds.ascent + (font)->font16->max_bounds.descent)

#define MAX_PROP_CHUNK_SIZE 32768

enum {
    Center,
    Noth,
    NothEast,
    East,
    SouthEast,
    South,
    SouthWest,
    West,
    NothWest,
    BaseLine
};

typedef struct _CharFont
{
    GC gc8;
    GC gc16;
    XFontStruct* font8;
    XFontStruct* font16;
} CharFont;

void AllocLowHighColor(
	Display* disp,
	Pixel src_color,
	Pixel* low_color_ret,
	Pixel* high_color_ret);

Pixel AllocNamedColor(
	Display* disp,
	const char* color_name);

Pixel AllocRGBColor(
	Display* disp,
	double red,		/* [0.0, 1.0] */
	double green,		/* [0.0, 1.0] */
	double blue);		/* [0.0, 1.0] */

void AppendPropertyValue(
	Display* disp,
	Atom type,
	void* value,
	int nbytes);

Display* CheckMultiDisplayEvent(
	Display** disp_list,
	int ndisp,
	struct timeval* tout);

void CompressEvent(
	Display* disp,
	Window win,
	int event_type,
	XEvent* e);

void CompressMotionEvent(
	Display* disp,
	Window win,
	XEvent* e);

void CompressKeyPressEvent(
	Display* disp,
	Window win,
	XEvent* e);

CharFont* CreateCharFont(
	Display* disp,
	const char* fontname8,
	const char* fontname16);

Pixmap CreateIconImagePixmap(
	Display* disp,
	GC gc,
	Pixel* icon_color_map,
	const char* icon_image,
	int* width_return,
	int* height_return);

Window CreateWindow(
	Display* disp,
	Window parent,		/* None is RootWindow */
	int x,
	int y,
	int width,
	int height,
	Pixel bg);		/* background color */

Pixmap CreatePixmap(
	Display* disp,
	Window parent,		/* None is RootWindow */
	int width,
	int height,
	Pixel bg);		/* background color */

void DrawBorder(
	Display* disp,
	Drawable drawable,
	GC gc,
	int x,
	int y,
	int width,
	int height,
        int bd_width,
	Pixel low_color,
	Pixel high_color,
	Bool status);		/* True if it looks like convex,
				   otherwise False (concave). */

void DrawChecker(
	Display* disp,
	Drawable drawable,
	GC gc,
	int x,
	int y,
	int width,
	int height);

void DrawRuledHLine(
	Display* disp,
	Drawable drawable,
	GC gc,
	Pixel low_color,
	Pixel high_color,
	int x,
	int y,
	int len);

void DrawRuledVLine(
	Display* disp,
	Drawable drawable,
	GC gc,
	Pixel low_color,
	Pixel high_color,
	int x,
	int y,
	int len);

void DrawString(
	Display* disp,
	Drawable win,
	CharFont* font,
	Pixel fg,
	int x, int y,
	int width, int height,
	const char* string,
	int alignment);

void DrawStringN(
	Display* disp,
	Drawable win,
	CharFont* font,
	Pixel fg,
	int x, int y,
	int width, int height,
	const char* string,
	int n,
	int alignment);

void DrawTriangle(
	Display* disp,
	Window win,
	GC gc,
	int x,
	int y,
	int bd,
	double r,
	double th,
	Pixel bg,
	Pixel lbg,
	Pixel hbg,
	Bool isconvex);

const char* EventTypeString(
	int type);

unsigned long GetEventMask(
	Display* disp,
	Window win);

XImage* GetImageData(
	Display* disp,
	Drawable drawable,
	int x,
	int y,
	int width,
	int height);

unsigned long GetPropertyDataSize(
	Display* disp,
	Atom type);

char* GetPropertyValue(
	Display* disp,
	Atom type,
	unsigned long* nbytes_return);

void GetWindowSize(
	Display* disp,
	Drawable drawable,
	int* width_ret,
	int* height_ret);

Bool IsKeyPressed(
	Display* disp,
	KeySym sym);

XFontStruct* LoadFont(
	Display* disp,
	const char* fontname);

void MapWindow(
	Display* disp,
	Window win,
	Bool   iswait);		/* is waiting until window is maped */

Display* OpenDisplay(
	const char* hostname);

void QueryPointer(
	Display* disp,
	Window win,
	int* x_ret,
	int* y_ret);

void QueryRootPointer(
	Display* disp,
	int* x_ret,
	int* y_ret);

void QueryWindowPositionFromRootWindow(
	Display* disp,
	Window win,
	int* x_ret,
	int* y_ret);

Window ParentWindow(
	Display* disp,
	Window win);

void PixelToRGB(
	Display* disp,
	Pixel pixel,
	double* r_ret,		/* [0.0, 1.0] */
	double* g_ret,		/* [0.0, 1.0] */
	double* b_ret);		/* [0.0, 1.0] */

void SetFixWindowSize(
	Display* disp,
	Window win,
	int width,
	int height);

void SetPropertyValue(
	Display* disp,
	Atom type,
	void* value,
	unsigned long nbytes);

void SetWindowBorder(
	Display* disp,
	Window win,
	GC gc,
	int border_width,
	Pixel bg,
	Pixel low_color,
	Pixel high_color,
	Bool status);		/* True if it looks like convex,
				   otherwise False (concave). */

int TextWidth(
	CharFont* font,
	const char* text);

int TextWidthN(
	CharFont* font,
	const char* text,
	int n);

#endif /* ___XMISC_H_ */
