/* abc - audio bucket controller */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "soundscope.h"
#include "bucket.h"
#include "ssframe.h"
#include "audriv.h"
#include "aubuff.h"
#include "soundex.h"
#include "player.h"
#include "timer.h"
#include "usffd.h"

enum device_sample_rate_priority_t
{
    HIGHEST_SAMPLE_RATE,
    LOWEST_SAMPLE_RATE
};

#define FILTER_PLAYER(filter_ptr) ((filter_ptr)->player)

static int device_sample_rate_priority = HIGHEST_SAMPLE_RATE;
static long audio_buffer_max_size_expectation = 0;
static long current_audio_buffer_size = 0;
static long audio_write_size = 0;
static double audio_start_time = 0.0;

AudioBucketFilter* audio_wave_adder;
static AudioBucketFilter* audio_bucket_buffer;
static AudioBucketFilter* audio_bucket_converter;
static AudioBucket* audio_bucket_remains = NULL;

static void (* abc_audio_converter)(void*, void*, int) = NULL;

static AudioBucket* abc_call_bucket_filter(AudioBucketFilter* filter);

/* Хåȥ꡼Ǿ̤Υե륿 */
static AudioBucket* abf_reader_proc(AudioBucketFilter* filter)
{
    AudioPlayer* player = FILTER_PLAYER(filter);
    AudioBucket* bucket;
    long pos_min, pos_max, cur, len;

    cur = filter->u.reader.seek_pointer;

    /* ղǽΰμ */
    player_get_region(player, &pos_min, &pos_max);
    if(player->load_flag)
    {
	if(pos_max > file_loader.read_samples)
	{
	    pos_max = file_loader.read_samples;
	    if(pos_min >= pos_max)
		return NULL;
	}
    }

    if(player->ff_rw_type == FF_RW_UNUSED) /*  */
    {
	if(cur >= pos_max)
	{
	    if(player->looping)
		filter->u.reader.seek_pointer = cur = pos_min;
	    else
		return NULL;
	}

	if(audio_buffer_max_size_expectation == 0 ||
	   current_audio_buffer_size < audio_buffer_max_size_expectation / 2)
	    len = AUDIO_BUCKET_LEN;
	else
	    len = AUDIO_BUCKET_LEN / 2;

	if(cur + len > pos_max)
	    len = pos_max - cur;

	if(player->frame_reference != NULL &&
	   player->frame_reference->equlizing_flag)
	{
#ifdef DEBUG
	    double t;
	    t = get_current_calender_time();
#endif /* DEBUG */
	    if((bucket = new_allocated_bucket()) == NULL)
		return NULL;
	    convolution_proc(filter->u.reader.buffer, player->aubuff->nsamples,
			     cur, bucket->data, len, 
			     player->frame_reference->convolution);
	    bucket->const_data = False;
#ifdef DEBUG
	    t = get_current_calender_time() - t;
	    if(t > 0)
		printf("convolution %ld-%ld (%.2fK samples/sec)\n",
		       cur, cur + len, 0.001 * (double)len / t);
#endif /* DEBUG */

	}
	else
	{
	    if((bucket = new_audio_bucket()) == NULL)
		return NULL;
	    bucket->data = filter->u.reader.buffer + cur;
	    bucket->const_data = True;
	}

	bucket->pos = 0;
	bucket->len = len;
	filter->u.reader.seek_pointer += len;
    }
    else if(player->ff_rw_type == FAST_FORWARD_PLAY) /*  */
    {
	int i, j, step;
	short* to;
	short* from;

	if(cur >= pos_max)
	{
	    if(player->looping)
		filter->u.reader.seek_pointer = cur = pos_min;
	    else
		return NULL;
	}

	if((bucket = new_allocated_bucket()) == NULL)
	    return NULL;
	step = player->ffre_step;
	len = AUDIO_BUCKET_LEN - AUDIO_BUCKET_LEN % step;

	if(cur + len > pos_max)
	{
	    len = pos_max - cur;
	    filter->u.reader.seek_pointer += len;
	    len -= AUDIO_BUCKET_LEN % step;
	}
	else
	    filter->u.reader.seek_pointer += len;

	bucket->pos = 0;
	bucket->len = len / step;
	to = bucket->data;
	from = filter->u.reader.buffer + cur;
	for(i = j = 0; i < len; i += step, j++)
	    to[j] = from[i];
    }
    else			/*  */
    {
	int i, j, step;
	short* to;
	short* from;

	if(cur <= pos_min)
	{
	    if(player->looping)
		filter->u.reader.seek_pointer = cur = pos_max;
	    else
		return NULL;
	}

	if((bucket = new_allocated_bucket()) == NULL)
	    return NULL;

	step = player->ffre_step;
	len = AUDIO_BUCKET_LEN - AUDIO_BUCKET_LEN % step;
	if(cur - len < pos_min)
	{
	    len = cur - pos_min;
	    filter->u.reader.seek_pointer -= len;
	    len -= AUDIO_BUCKET_LEN % step;
	}
	else
	    filter->u.reader.seek_pointer -= len;

	bucket->pos = 0;
	bucket->len = len / step;
	to = bucket->data;
	from = filter->u.reader.buffer + cur;
	for(i = j = 0; i < len; i += step, j++)
	    to[j] = from[-i];
    }

    return bucket;
}


/* ץ졼Ĵե륿 */
static AudioBucket* sample_change_filter_proc(AudioBucketFilter* filter)
{
    long play_r, dev_r;
    AudioBucket* source;

    play_r = FILTER_PLAYER(filter)->sample_rate;
    dev_r = device_sample_rate;
    source = abc_call_bucket_filter(filter->superior);
    source = convert_sample_rate(source, play_r, dev_r);
    return source;
}

/* ͥߥ */
static AudioBucket* abf_channel_mixer_proc(AudioBucketFilter* filter)
{
    AudioBucket* channel_bucket[MAX_BUCKET_NCHANNELS];
    int ch, nch;

    nch = filter->u.channel_mixer.nchannels;
    for(ch = 0; ch < nch; ch++)
	channel_bucket[ch] = abc_call_bucket_filter(
	    filter->u.channel_mixer.superior_filters[ch]);
    return mix_audio_channel(channel_bucket, nch, device_nchannels);
}

/*  */
static AudioBucket* abf_wave_adder_proc(AudioBucketFilter* filter)
{
    int i, n, len;
    int bucket_empty_count;
    AudioBucket* bucket;
    long* adder_buffer = (long *)temp_buff;

    n = filter->u.wave_adder.nchannels;

#ifdef DEBUGALL
    printf("Wave adder (channel: %d)\n", n);
#endif

    if(n == 1)
	return abc_call_bucket_filter(
		filter->u.wave_adder.superior_filters[0]);

    len = AUDIO_BUCKET_LEN;
    bucket_empty_count = 0;
    for(i = 0; i < n; i++)
    {
	if(filter->u.wave_adder.remains[i] == NULL)
	    if((filter->u.wave_adder.remains[i]
		= abc_call_bucket_filter(
		    filter->u.wave_adder.superior_filters[i])) == NULL)
	    {
		bucket_empty_count++;
		continue;
	    }
	if(filter->u.wave_adder.remains[i]->len < len)
	    len = filter->u.wave_adder.remains[i]->len;
    }

    if(bucket_empty_count == n)
	return NULL;

    if((bucket = new_allocated_bucket()) == NULL)
    {
	/* 㳰 */
	for(i = 0; i < n; i++)
	    reuse_audio_bucket_list(filter->u.wave_adder.remains[i]);
	return NULL;
    }

    memset(adder_buffer, 0, len * sizeof(long));
    n = filter->u.wave_adder.nchannels;
    for(i = 0; i < n; i++)
    {
	int j;
	short* data;

	if(filter->u.wave_adder.remains[i] == NULL)
	    continue;
	data = filter->u.wave_adder.remains[i]->data
	    + filter->u.wave_adder.remains[i]->pos;

	for(j = 0; j < len; j++)
	    adder_buffer[j] += data[j];

	filter->u.wave_adder.remains[i]->pos += len;
	filter->u.wave_adder.remains[i]->len -= len;

	if(filter->u.wave_adder.remains[i]->len == 0)
	{
	    AudioBucket* tmp;

	    tmp = filter->u.wave_adder.remains[i];
	    filter->u.wave_adder.remains[i]
		= filter->u.wave_adder.remains[i]->next;
	    reuse_audio_bucket(tmp);
	}
    }
    
    for(i = 0; i < len; i++)
	bucket->data[i] = CLIP_SHORT(adder_buffer[i]);
    bucket->len = len;
    return bucket;
}

/* ١ХåȤ򽸤ơ̤ˤޤȤή(Хåե) */
static AudioBucket* bucket_buffer_filter_proc(AudioBucketFilter* filter)
{
    long size, buffer_size;
    AudioBucket* bucket_head;
    AudioBucket* bucket_tail;

#ifdef DEBUGALL
    puts("->bucket_buffer_filter_proc()");
#endif

    bucket_head = bucket_tail = abc_call_bucket_filter(filter->superior);
    if(audio_buffer_max_size_expectation == 0 ||
       current_audio_buffer_size < audio_buffer_max_size_expectation / 2)
	buffer_size = 2 * AUDIO_BUCKET_LEN;
    else
      {
#ifdef DEBUGALL
	puts("<-bucket_buffer_filter_proc()");
#endif
	return bucket_head;
      }
    if(bucket_head == NULL)
      {
#ifdef DEBUGALL
	puts("<-bucket_buffer_filter_proc()");
#endif
	return NULL;
      }

    size = bucket_head->len;
    if(size < buffer_size)
    {
	int loop_cnt = 100;	/* ̵¥롼פ򤱤뤿Υ */

#ifdef DEBUG
	if(bucket_tail->next == bucket_tail)
	  {
	    fprintf(stderr, "Fatal error: Enter infinite loop!!\n");
	    exit(1);
	  }
#endif
	while(bucket_tail->next)
	{

	    size += bucket_tail->len;
	    bucket_tail = bucket_tail->next;
	}
	while(size < buffer_size && loop_cnt-- > 0)
	{
	    if((bucket_tail->next = abc_call_bucket_filter(filter->superior))
	       == NULL)
		break;
	    while(bucket_tail->next)
	    {
		size += bucket_tail->len;
		bucket_tail = bucket_tail->next;
	    }
	}
    }
#ifdef DEBUGALL
    puts("<-bucket_buffer_filter_proc()");
#endif
    return bucket_head;
}

/* ǥХݡȤƤ벻Ѵե륿 */
static AudioBucket* bucket_converter_filter_proc(AudioBucketFilter* filter)
{
    AudioBucket* p;
    AudioBucket* bucket_head;
    AudioBucket* bucket_tail;

#ifdef DEBUGALL
    puts("->bucket_converter_filter_proc()");
#endif

    p = abc_call_bucket_filter(filter->superior);
#ifdef DEBUGALL
    puts("abc_call_bucket_filter...done");
#endif
    if(internal_encoding == device_data_encoding)
    {
	for(bucket_tail = p; bucket_tail; bucket_tail = bucket_tail->next)
	    bucket_tail->len *=
		usffd_encoding_sample_size[device_data_encoding];
#ifdef DEBUGALL
	puts("<-bucket_converter_filter_proc()");
#endif
	return p;
    }

    BUCKET_LIST_INIT(bucket_head, bucket_tail);

    while(p)
    {
	AudioBucket* bucket;
	AudioBucket* tmp;

	/* ǡϤʤΤǡ夫ΥХåȤƱΥХå
	 * ѰդƤнʬǤ롥
	 */
	if((bucket = new_allocated_bucket()) == NULL)
	{
	    reuse_audio_bucket_list(p);
#ifdef DEBUGALL
	puts("<-bucket_converter_filter_proc()");
#endif
	    return NULL;
	}

#ifdef DEBUGALL
	printf("abc_audio_converter() ... %p\n", abc_audio_converter);
#endif

	abc_audio_converter(p->data, bucket->data, p->len * sizeof(short));

#ifdef DEBUGALL
	printf("abc_audio_converter() - Done\n");
#endif

	bucket->pos = 0;
	bucket->len =
	    p->len * usffd_encoding_sample_size[device_data_encoding];
	/*  bucket->len ϥǡǤʤ
	 * ХȿǤ롥̤ΥեϤ
	 * ХȿȤư
	 */

	/* bucket ɲ */
	ADD_BUCKET(bucket_head, bucket_tail, bucket);

	/* ѺѤΥХåȤ */
	tmp = p;
	p = p->next;
	reuse_audio_bucket(tmp);
    }
#ifdef DEBUGALL
    puts("<-bucket_converter_filter_proc()");
#endif

    return bucket_head;
}

static AudioBucket* abc_call_bucket_filter(AudioBucketFilter* filter)
{
    AudioBucket* bucket;
#ifdef DEBUGALL
    static int depth = 0;
    printf("->abc_call_bucket_filter(%d),%d\n", filter->filter_type, depth++);
#endif

    bucket = filter->filter_proc(filter);
#if 0
    switch(filter->filter_type)
    {
      case ABF_normal:
	bucket = filter->u.normal_filter(filter);
	break;
      case ABF_reader:
	bucket = abf_reader_proc(filter);
	break;
      case ABF_channel_mixer:
	bucket = abf_channel_mixer_proc(filter);
	break;
      case ABF_wave_adder:
	bucket = abf_wave_adder_proc(filter);
	break;
      default:
	fprintf(stderr, "Fatal error in abc_call_bucket_filter().\n");
	exit(1);
    }
#endif


#ifdef DEBUGALL
    printf("<-abc_call_bucket_filter(%d),%d\n", filter->filter_type, --depth);
#endif

    return bucket;
}

static void abc_stop_all_audio(void)
{
    int i;

    abc_stop_audio();
    for(i = 0; i < audio_wave_adder->u.wave_adder.nchannels; i++)
    {
	AudioPlayer* player
	   = FILTER_PLAYER(audio_wave_adder->u.wave_adder.superior_filters[i]);
	abc_remove_player(player);
	SetPressdIconButton(player->play_button, False);
	SetPressdIconButton(player->stop_button, False);
	if(player->frame_reference)
	    repaint_frame_scope(player->frame_reference, player->pos, False);
    }
    audriv_play_stop();
    audio_active_flag = False;
    update_ss_active_flag();
}

static void abc_stop_rewind_all_audio(void)
{
    int i;

    abc_stop_audio();
    for(i = 0; i < audio_wave_adder->u.wave_adder.nchannels; i++)
    {
	AudioPlayer* player
	   = FILTER_PLAYER(audio_wave_adder->u.wave_adder.superior_filters[i]);
	abc_remove_player(player);
	SetPressdIconButton(player->play_button, False);
	SetPressdIconButton(player->stop_button, False);
	player->pos = AUDIO_PLAYER_POS_MIN(player);
	if(player->frame_reference)
	    repaint_frame_scope(player->frame_reference, player->pos, False);
    }
    audriv_play_stop();
    audio_active_flag = False;
    update_ss_active_flag();
}

/* ƥ֤ʥХåեСդ롥
 * ֤ͤϡºݤ audio ˽񤭹ޤ줿Хȿ
 */
PUBLIC int play_buffer(void)
{
    int write_cnt;

    write_cnt = 0;
    if(audio_bucket_remains == NULL)
    {
	audio_bucket_remains
	    = bucket_converter_filter_proc(audio_bucket_converter);
	if(audio_bucket_remains == NULL && audriv_play_active() == False)
	{
	    abc_stop_rewind_all_audio();
	    return 0;
	}
    }

    /* bucket ĤŤ audio ήǹԤ
     * ήήǡήʤʤ(ʤ audriv_write  0
     * ֤)롼פȴĤΥǡϼθƤӽФ˸󤷤롥
     */
    while(audio_bucket_remains)
    {
	AudioBucket* tmp;
	long n;

	if(audio_write_size == 0)
	    audio_start_time = get_current_calender_time();


	if(audio_bucket_remains->len > 0)
	    n = audriv_write(((char *)audio_bucket_remains->data) +
			     audio_bucket_remains->pos,
			     audio_bucket_remains->len);
	else
	    n = 0;

	audio_write_size += n;
	write_cnt += n;
	current_audio_buffer_size =
	    (long)
		(audio_write_size - 
		 device_sample_rate *
		 device_nchannels *
		 usffd_encoding_sample_size[device_data_encoding] *
		 (get_current_calender_time() - audio_start_time));

	if(n == audio_bucket_remains->len)
	{
	    tmp = audio_bucket_remains;
	    audio_bucket_remains = audio_bucket_remains->next;
	    reuse_audio_bucket(tmp);
	}
	else
	{
	    if(n == 0)
	    {
		if(current_audio_buffer_size < 0)
		    current_audio_buffer_size = 0;
		else if(current_audio_buffer_size > audio_buffer_max_size_expectation)
		    audio_buffer_max_size_expectation
			= current_audio_buffer_size;
		break;
	    }
	    audio_bucket_remains->pos += n;
	    audio_bucket_remains->len -= n;
	}
    }
    return write_cnt;
}

/* audio νλΤ audio bucket controller 뤳 */
PUBLIC void initialize_abc(void)
{
    audio_wave_adder = alloc_audio_bucket_filter(ABF_wave_adder);
    audio_bucket_buffer = alloc_audio_bucket_filter(ABF_normal);
    audio_bucket_converter = alloc_audio_bucket_filter(ABF_normal);

    if(audio_wave_adder == NULL ||
       audio_bucket_buffer == NULL ||
       audio_bucket_converter == NULL)
    {
	fprintf(stderr, "Can't initialize audio bucket controller.\n");
	exit(1);
    }

    audio_wave_adder->u.wave_adder.nchannels = 0;
    audio_wave_adder->superior = NULL;
    audio_wave_adder->filter_proc = abf_wave_adder_proc;

    audio_bucket_buffer->filter_proc = bucket_buffer_filter_proc;
    audio_bucket_buffer->superior = audio_wave_adder;

    audio_bucket_converter->filter_proc = bucket_converter_filter_proc;
    audio_bucket_converter->superior = audio_bucket_buffer;
    abc_update_audio_converter();
}

static AudioBucketFilter* make_player_buffer_reader(AudioPlayer* player,
						    long ch)
{
    AudioBucketFilter* reader;
    AudioBuffer* b = player->aubuff;

    if((reader = alloc_audio_bucket_filter(ABF_reader)) == NULL)
	return NULL;
    FILTER_PLAYER(reader) = player;
    reader->u.reader.buffer = b->buffer[ch];
    reader->u.reader.seek_pointer = player->pos;
    reader->filter_proc = abf_reader_proc;
    player->bucket_reader[ch] = reader;

    return reader;
}

PUBLIC AudioBucketFilter* make_normal_filter(AudioBucketFilter* superior,
		AudioBucket* (* filter_proc)(struct _AudioBucketFilter *))
{
    AudioBucketFilter* filter;

    if((filter = alloc_audio_bucket_filter(ABF_normal)) == NULL)
	return NULL;
    filter->filter_proc = filter_proc;
    filter->superior = superior;
    return filter;
}

static AudioBucketFilter* make_player_sample_filter(AudioPlayer* player,
						    long ch)
{
    AudioBucketFilter* sample_filter;
    AudioBucketFilter* superior;

    if((superior = make_player_buffer_reader(player, ch)) == NULL)
	return NULL;

    if((sample_filter =
	make_normal_filter(superior, sample_change_filter_proc)) == NULL)
    {
	reuse_audio_bucket_filter(superior);
	return NULL;
    }

    FILTER_PLAYER(sample_filter) = player;
    return sample_filter;
}

PUBLIC AudioBucketFilter* make_mixer_filter(AudioBucketFilter** superior,
					    int nchannels,
		AudioBucket* (* filter_proc)(struct _AudioBucketFilter *))
{
    int ch;
    AudioBucketFilter* mixer_filter;
    AudioBucketFilter** channel_mixer;

    if((mixer_filter = alloc_audio_bucket_filter(ABF_channel_mixer)) == NULL)
	return NULL;
    mixer_filter->filter_proc = filter_proc;
    mixer_filter->u.channel_mixer.nchannels = nchannels;

    channel_mixer = mixer_filter->u.channel_mixer.superior_filters;
    for(ch = 0; ch < nchannels; ch++)
	channel_mixer[ch] = superior[ch];
    return mixer_filter;
}


static AudioBucketFilter* make_player_mixer_filter(AudioPlayer* player)
{
    AudioBucketFilter* mixer_filter;
    AudioBucketFilter* superior[MAX_BUCKET_NCHANNELS];
    int ch, nchannels;

    nchannels = player->aubuff->nchannels;
    for(ch = 0; ch < nchannels; ch++)
	if((superior[ch] = make_player_sample_filter(player, ch)) == NULL)
	{
	    while(--ch >= 0)
		reuse_audio_bucket_filter_list(superior[ch]);
	    return NULL;
	}
    if((mixer_filter = make_mixer_filter(superior, nchannels,
					 abf_channel_mixer_proc)) == NULL)
    {
	while(--ch >= 0)
	    reuse_audio_bucket_filter_list(superior[ch]);
	return NULL;
    }

    player->play_flag = True;
    FILTER_PLAYER(mixer_filter) = player;
    return mixer_filter;
}

static void abc_stop_audio_recursive(AudioBucketFilter* filter, long samples)
{
    int i;

    switch(filter->filter_type)
    {
      case ABF_normal:
	abc_stop_audio_recursive(filter->superior, samples);
	break;
      case ABF_reader:
	filter->u.reader.seek_pointer =
	    calc_player_samples(FILTER_PLAYER(filter), samples);
	break;
      case ABF_channel_mixer:
	for(i = 0; i < filter->u.channel_mixer.nchannels; i++)
	    abc_stop_audio_recursive(
		filter->u.channel_mixer.superior_filters[i], samples);
	FILTER_PLAYER(filter)->pos = calc_player_samples(FILTER_PLAYER(filter),
							 samples);
	FILTER_PLAYER(filter)->pos_offset = 0;
	break;
      case ABF_wave_adder:
	for(i = 0; i < filter->u.wave_adder.nchannels; i++)
	{
	    reuse_audio_bucket_list(filter->u.wave_adder.remains[i]);
	    filter->u.wave_adder.remains[i] = NULL;
	    abc_stop_audio_recursive(
		filter->u.wave_adder.superior_filters[i], samples);
	}
	break;
    }
}

/* դߤΥХåե򤹤٤ƥեå夹롥*/
PUBLIC void abc_stop_audio(void)
{
    long samples;
    reuse_audio_bucket_list(audio_bucket_remains);
    audio_bucket_remains = NULL;
    samples = audriv_play_stop();
    if(samples < 0)
	samples = 0;
    abc_stop_audio_recursive(audio_bucket_converter, samples);
    audio_write_size = 0;
    current_audio_buffer_size = 0;
}

/* player  audio_wave_adder ɲäƥ־֤ˤ롥*/
PUBLIC Bool abc_add_player(AudioPlayer* player)
{
    AudioBucketFilter* mixer_filter;
    int n;

    if(audio_wave_adder->u.wave_adder.nchannels > MAX_BUCKET_NCHANNELS ||
       player->aubuff->nchannels > MAX_BUCKET_NCHANNELS)
    {
	write_error_message("Too many open channels");
	return False;
    }

    if((mixer_filter = make_player_mixer_filter(player)) == NULL)
	return False;

    n = audio_wave_adder->u.wave_adder.nchannels++;
    audio_wave_adder->u.wave_adder.superior_filters[n] = mixer_filter;
    audio_wave_adder->u.wave_adder.players[n] = player;
    return True;
}

/* player  audio_wave_adder Υ󥢥ƥ־֤ˤ롥*/
PUBLIC void abc_remove_player(AudioPlayer* player)
{
    int i, n, ch;

    n = audio_wave_adder->u.wave_adder.nchannels;
    for(i = 0; i < n; i++)
    {
	if(audio_wave_adder->u.wave_adder.players[i] == player)
	{
	    reuse_audio_bucket_filter_list(audio_wave_adder->u.wave_adder.
					   superior_filters[i]);
	    while(++i < n)
	    {
		audio_wave_adder->u.wave_adder.superior_filters[i - 1] =
		    audio_wave_adder->u.wave_adder.superior_filters[i];
		audio_wave_adder->u.wave_adder.players[i - 1] =
		    audio_wave_adder->u.wave_adder.players[i];
	    }
	    audio_wave_adder->u.wave_adder.nchannels--;
	    for(ch = 0; ch < player->aubuff->nchannels; ch++)
		player->bucket_reader[i] = NULL;
	    player->play_flag = False;
	    player->ff_rw_type = FF_RW_UNUSED;
	    return;		/* found */
	}
    }
}

PUBLIC Bool abc_set_appropriate_device_sample_rate(void)
{
    int ns;
    const long* slist;
    long new_sample_rate;

    slist = audriv_available_sample_rates(&ns);

    if(device_sample_rate_priority == LOWEST_SAMPLE_RATE)
    {
	long smin = 99999;
	int i;

	for(i = 0; i < audio_wave_adder->u.wave_adder.nchannels; i++)
	    if(smin > audio_wave_adder->u.wave_adder.players[i]->sample_rate)
		smin = audio_wave_adder->u.wave_adder.players[i]->sample_rate;
	for(i = ns - 1; i > 0; i--)
	    if(slist[i] <= smin)
		break;
	new_sample_rate = slist[i];
    }
    else			/* HIGHEST_SAMPLE_RATE */
    {
	long smax = 1000;
	int i;

	for(i = 0; i < audio_wave_adder->u.wave_adder.nchannels; i++)
	    if(smax < audio_wave_adder->u.wave_adder.players[i]->sample_rate)
		smax = audio_wave_adder->u.wave_adder.players[i]->sample_rate;
	for(i = 0; i < ns - 1; i++)
	    if(slist[i] >= smax)
		break;
	new_sample_rate = slist[i];
    }

    device_sample_rate = new_sample_rate;
    return audriv_set_play_sample_rate(device_sample_rate);
}

PUBLIC Bool abc_restart_audio(void)
{
    update_ss_active_flag();
    if(audio_active_flag)
	if(abc_set_appropriate_device_sample_rate() == False ||
	   audriv_set_play_encoding(device_data_encoding) == False ||
	   audriv_set_play_channels(device_nchannels) == False ||
	   audriv_play_open() == False)
	{
	    abc_stop_all_audio();
	    return False;
	}
    abc_update_audio_converter();
    return True;
}

PUBLIC void abc_update_audio_converter(void)
{
#ifdef DEBUG
    printf("abc_update_audio_converter(): %s(internal) -> %s(device)\n",
	   usffd_encoding_jname(internal_encoding),
	   usffd_encoding_jname(device_data_encoding));
#endif
    abc_audio_converter = select_converter(internal_encoding,
					   device_data_encoding,
					   NULL, NULL);
    if(abc_audio_converter == NULL)
    {
	fprintf(stderr, "Fatal error in abc_update_audio_converter(): %s to %s is not defined.\n",
		usffd_encoding_name(internal_encoding),
		usffd_encoding_name(device_data_encoding));
	exit(1);
    }

#ifdef DEBUGALL
    printf("abc_audio_converter = %p\n", abc_audio_converter);
#endif
}
