#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include "winobj_defs.h"
#include "ssdefs.h"
#include "misc.h"
#include "bucket.h"

#define ALLOCATED_N_BUCKET   128
#define NOALLOCATED_N_BUCKET 512

#ifdef DEBUGALL
#define BUCKET_REMAIN_DEBUG
#endif

AudioBucket* bucket_list = NULL;
AudioBucket* allocated_bucket_list = NULL;
AudioBucketFilter* audio_bucket_filter_free_list = NULL;
static short* initial_allocated_data;
static long initial_allocated_data_size;

#ifdef BUCKET_REMAIN_DEBUG
static int bucket_list_length = ALLOCATED_N_BUCKET + NOALLOCATED_N_BUCKET;
#endif

PUBLIC void initialize_audio_bucket(void)
{
    short* data;
    int i;

    initial_allocated_data_size = ALLOCATED_N_BUCKET * AUDIO_BUCKET_LEN;
    if((data = (short *)xmalloc(initial_allocated_data_size
			       * sizeof(short))) == NULL)
    {
	perror("malloc");
	exit(1);
    }
    initial_allocated_data = data;

    for(i = 0; i < ALLOCATED_N_BUCKET; i++)
    {
	AudioBucket* b;

	if((b = (AudioBucket *)xmalloc(sizeof(AudioBucket))) == NULL)
	{
	    perror("malloc");
	    exit(1);
	}
	b->data = data + i * AUDIO_BUCKET_LEN;
	b->pos = 0;
	b->len = AUDIO_BUCKET_LEN;
	b->allocated = True;

	reuse_audio_bucket(b);
    }

    for(i = 0; i < NOALLOCATED_N_BUCKET; i++)
    {
	AudioBucket* b;

	if((b = (AudioBucket *)xmalloc(sizeof(AudioBucket))) == NULL)
	{
	    perror("malloc");
	    exit(1);
	}
	b->data = NULL;
	b->pos = 0;
	b->len = 0;
	b->allocated = False;
	b->const_data = True;

	reuse_audio_bucket(b);
    }
}


#ifdef BUCKET_REMAIN_DEBUG
static void check_reuse_audio_bucket(AudioBucket* bucket)
{
    AudioBucket* p;

    for(p = allocated_bucket_list; p; p = p->next)
    {
	if(p == bucket)
	{
	    fprintf(stderr, "%p ϴ˺ѤǤ\n", p);
	    exit(1);
	}
    }

    for(p = bucket_list; p; p = p->next)
    {
	if(p == bucket)
	{
	    fprintf(stderr, "%p ϴ˺ѤǤ\n", p);
	    exit(1);
	}
    }
}
#endif

static Bool is_initial_bucket(AudioBucket* bucket)
{
    short* p;

    p = bucket->data;
    return initial_allocated_data <= p &&
	p < initial_allocated_data + initial_allocated_data_size;
}

PUBLIC void reuse_audio_bucket(AudioBucket* bucket)
{
#ifdef BUCKET_REMAIN_DEBUG
    check_reuse_audio_bucket(bucket);
#endif

    if(bucket == NULL)
	return;
    if(bucket->allocated)
    {
	if(is_initial_bucket(bucket))
	{
	    bucket->next = allocated_bucket_list;
	    allocated_bucket_list = bucket;
	}
	else
	{
	    xfree(bucket->data);
	    xfree(bucket);
	}
    }
    else
    {
	bucket->next = bucket_list;
	bucket_list = bucket;
    }
#ifdef BUCKET_REMAIN_DEBUG
    bucket_list_length--;
    printf("-Bucket %d\n", bucket_list_length);
#endif
}

PUBLIC void reuse_audio_bucket_list(AudioBucket* bucket)
{
    while(bucket)
    {
	AudioBucket* tmp;

	tmp = bucket;
	bucket = bucket->next;
	reuse_audio_bucket(tmp);
    }
}

PUBLIC AudioBucket* new_audio_bucket(void)
{
    AudioBucket* b;

    if(bucket_list == NULL)
    {
	if((b = (AudioBucket *)xmalloc(sizeof(AudioBucket))) == NULL)
	    return NULL;
    }
    else
    {
	b = bucket_list;
	bucket_list = bucket_list->next;
    }

    b->data = NULL;
    b->pos = b->len = 0;
    b->allocated = False;
    b->const_data = True;
    b->next = NULL;

#ifdef BUCKET_REMAIN_DEBUG
    bucket_list_length++;
    printf("+Bucket %d\n", bucket_list_length);
#endif

    return b;
}

PUBLIC AudioBucket* new_allocated_bucket(void)
{
    AudioBucket* b;

    if(allocated_bucket_list == NULL)
    {
	if((b = new_audio_bucket()) == NULL)
	    return NULL;
	if((b->data = (short *)xmalloc(AUDIO_BUCKET_LEN * sizeof(short)))
	   == NULL)
	{
	    xfree(b);
	    return NULL;
	}
	b->allocated = True;
    }
    else
    {
	b = allocated_bucket_list;
	allocated_bucket_list = allocated_bucket_list->next;
    }
    b->pos = 0;
    b->len = 0;
    b->const_data = False;
    b->next = NULL;

#ifdef BUCKET_REMAIN_DEBUG
    bucket_list_length++;
    printf("*Bucket %d\n", bucket_list_length);
#endif

    return b;
}

PUBLIC long audio_bucket_length(AudioBucket* bucket)
{
    long n = 0;

    while(bucket)
    {
	bucket = bucket->next;
	n++;
    }
    return n;
}

PUBLIC AudioBucketFilter* alloc_audio_bucket_filter(int type)
{
    AudioBucketFilter* filter;

#ifdef DEBUGALL
    puts("alloc_audio_bucket_filter()");
#endif

    if(audio_bucket_filter_free_list == NULL)
    {
	filter = (AudioBucketFilter *)xmalloc(sizeof(AudioBucketFilter));
	if(filter == NULL)
	    return NULL;
    }
    else
    {
	filter = audio_bucket_filter_free_list;
	audio_bucket_filter_free_list =
	    audio_bucket_filter_free_list->free_chain;
    }

    /* ƤΥݥ󥿤 NULL ˽ */
    memset(filter, 0, sizeof(AudioBucketFilter));
    filter->filter_type = type;

    return filter;
}

PUBLIC void reuse_audio_bucket_filter(AudioBucketFilter* filter)
{
#ifdef DEBUGALL
    puts("reuse_audio_bucket_filter()");
#endif
    if(filter == NULL)
	return;
    filter->free_chain = audio_bucket_filter_free_list;
    audio_bucket_filter_free_list = filter;
}

PUBLIC void reuse_audio_bucket_filter_list(AudioBucketFilter* filter)
{
    int i;
#ifdef DEBUGALL
    puts("reuse_audio_bucket_filter_list()");
#endif

    while(filter)
    {
	AudioBucketFilter* tmp;

	tmp = filter;
	switch(tmp->filter_type)
	{
	  case ABF_normal:
	    filter = filter->superior;
	    reuse_audio_bucket_filter(tmp);
	    break;
	  case ABF_reader:
	    reuse_audio_bucket_filter(tmp);
	    return;
	  case ABF_channel_mixer:
	    for(i = 0; i < tmp->u.channel_mixer.nchannels; i++)
		reuse_audio_bucket_filter(tmp->u.channel_mixer.
					  superior_filters[i]);
	    return;
	  case ABF_wave_adder:
	    for(i = 0; i < tmp->u.wave_adder.nchannels; i++)
	    {
		reuse_audio_bucket_filter(tmp->u.wave_adder.
					  superior_filters[i]);
		reuse_audio_bucket_list(tmp->u.wave_adder.remains[i]);
	    }
	    return;
	}
    }
}

PUBLIC AudioBucket* buffer_to_bucket_list(short* buffer, long len)
{
    AudioBucket *head, *tail, *bucket;
    long n;

    if(len == 0 || buffer == NULL)
	return NULL;

    BUCKET_LIST_INIT(head, tail);
    while(len > 0)
    {
	bucket = new_audio_bucket();
	bucket->data = buffer;
	n = MIN(len, AUDIO_BUCKET_LEN);
	bucket->len = n;
	len    -= n;
	buffer += n;
	ADD_BUCKET(head, tail, bucket);
    }
    return head;
}

/* linear completion */
static AudioBucket* sample_rate_linear_compl(AudioBucket* source, double ratio)
{
    double dx, a, b, x, d, ab;
    AudioBucket* bucket_head;
    AudioBucket* bucket_tail;
    AudioBucket* bucket = NULL;
    short* buff;
    short* from;
    int spos, tpos, len;

    if(ratio <= 1.0)
    {
	fprintf(stderr,
		"Can't convert sample rate: sample_rate_linear_compl()\n");
	exit(1);
    }

    dx = 1.0 / ratio;		/* dx < 1.0 */
    BUCKET_LIST_INIT(bucket_head, bucket_tail);

    if(source == NULL)
	return NULL;

    /* ץ 1 ΥХåȤ̵뤹 */
    while(source->len <= 1)
    {
	reuse_audio_bucket(source);
	source = source->next;
	if(source == NULL)
	    return new_audio_bucket();
    }

    a = source->data[0];
    b = source->data[1];
    spos = 1;
    bucket = new_allocated_bucket();
    buff = bucket->data;
    tpos = 0;
    from = source->data;
    ADD_BUCKET(bucket_head, bucket_tail, bucket);
    x = 0.0;
    d = 0.0;
    ab = b - a;
    len = source->len;

    while(1)
    {
	int dbgcnt = 0;
	while(x < spos)
	{
	    double y;
	    dbgcnt++;
	    y = a + ab * d;

	    d += dx;
	    x += dx;

	    if(tpos == AUDIO_BUCKET_LEN)
	    {
		bucket->len = tpos;
		bucket = new_allocated_bucket();
		ADD_BUCKET(bucket_head, bucket_tail, bucket);
		buff = bucket->data;
		buff[0] = y;
		tpos = 1;
	    }
	    else
		buff[tpos++] = y;
	}
	d--;
	spos++;

	if(spos >= len)
	{
	    source = source->next;

	    if(source == NULL)
		break;
	    x -= spos;
	    spos = 0;
	    len = source->len;
	    from = source->data;
	}
	a = b;
	b = from[spos];
	ab = b - a;
    }

    bucket->len = tpos;
    return bucket_head;
}

/* ץ졼Ĵ */
PUBLIC AudioBucket* convert_sample_rate(AudioBucket* source,
					long source_sample_rate,
					long target_sample_rage)
{
    double ratio;		/* bucket ΥǡΰФΨ */
    AudioBucket* bucket_head;
    AudioBucket* bucket_tail;
    AudioBucket* bucket = NULL;
    double r;			/* 1.0 / ratio */

    /* Don't mind follows warning:
     * `bucket_remain_len' might be used uninitialized in this function
     */

    if(target_sample_rage == source_sample_rate || source == NULL)
	return source;		/* Ʊ or λ */
    BUCKET_LIST_INIT(bucket_head, bucket_tail);
    ratio = (double)target_sample_rage / (double)source_sample_rate;
    r     = 1.0 / ratio;
    if(target_sample_rage < source_sample_rate)	/* ratio < 1.0 */
    {
	/* ǡϸΤǡñ˴ְ롥*/
	while(source)
	{
	    long n;
	    AudioBucket* tmp;

	    if((n = (long)(source->len * ratio)) > 0)
	    {
		short* from;
		short* to;
		long i;

		if((bucket = new_allocated_bucket()) == NULL)
		{
		    reuse_audio_bucket_list(source);
		    reuse_audio_bucket_list(bucket_head);
		    return NULL;
		}
		bucket->len = n;
		from = source->data + source->pos;
		to = bucket->data;
		for(i = 0; i < n; i++)
		    to[i] = from[(long)(i * r)];
		ADD_BUCKET_LIST(bucket_head, bucket_tail, bucket);
	    }
	    tmp = source;
	    source = source->next;
	    reuse_audio_bucket(tmp);
	}
    }
    else			/* ratio > 1.0 */
    {
      bucket_head = sample_rate_linear_compl(source, ratio);
      
#if 0
	/* source  ratio ܤ䤷 bucket_head ˷ҤǤ椯
	 */
	long source_remain_len, bucket_remain_len = 0;
	Bool need_new_bucket = True;
	long from_pos = 0;

	while(source)
	{
	    source_remain_len = source->len;

	    while(source_remain_len > 0)
	    {
		long slen;	/* Ѵ source Υǡ */
		long blen;	/* ǡ */
		long i;
		short* to;
		short* from;

	      new_bucket_label:
		if(need_new_bucket)
		{
		    if((bucket = new_allocated_bucket()) == NULL)
		    {
			reuse_audio_bucket_list(source);
			reuse_audio_bucket_list(bucket_head);
			return NULL;
		    }
		    bucket_remain_len = AUDIO_BUCKET_LEN;
		    bucket->len = 0;
		    ADD_BUCKET(bucket_head, bucket_tail, bucket);
		}

		slen = source_remain_len;
		blen = source_remain_len * ratio;
		if(blen > bucket_remain_len)
		{
		    need_new_bucket = True;
		    slen = bucket_remain_len * r; /* r < 1.0 */
		    blen = slen * ratio;

		    if(slen == 0)
		    {
			need_new_bucket = True;
			goto new_bucket_label;
		    }
		}
		else
		    need_new_bucket = False;

		to = bucket->data + bucket->len;
		from = source->data + from_pos;
		for(i = 0; i < blen; i++)
		    to[i] = from[(long)(i * r)]; /* r < 1.0 */

		from_pos          += slen;
		source_remain_len -= slen;
		bucket_remain_len -= blen;
		bucket->len       += blen;
	    }
	    source = source->next;
	    from_pos = 0;
	}
#endif
    }
    if(bucket_head == NULL)
	return new_audio_bucket();
    return bucket_head;
}

/* ͥߥ */
PUBLIC AudioBucket* mix_audio_channel(AudioBucket** channel_bucket,
				      int nch_from, int nch_to)
{
    int ch;
    long* adder_buffer = (long *)temp_buff;

    if(channel_bucket[0] == NULL)
	return NULL;

    if(nch_to == 1 && nch_from == 1)
	return channel_bucket[0];

    if(nch_to == 1)
    {
	AudioBucket* bucket_head;
	AudioBucket* bucket_tail;

	/* ƤΥͥ sample ­碌롥*/
	BUCKET_LIST_INIT(bucket_head, bucket_tail);
	while(channel_bucket[0])
	{
	    AudioBucket* bucket;
	    long len, i;

	    if((len = channel_bucket[0]->len) > 0)
	    {
		if((bucket = new_allocated_bucket()) == NULL)
		{
		    /* 㳰*/
		    for(ch = 0; ch < nch_from; ch++)
			reuse_audio_bucket_list(channel_bucket[ch]);
		    return NULL;
		}
		bucket->len = len;

		memset(adder_buffer, 0, len * sizeof(long));
		for(ch = 0; ch < nch_from; ch++)
		{
		    long* x = adder_buffer;
		    short* y = channel_bucket[ch]->data;
		    
		    for(i = 0; i < len; i++)
			x[i] += y[i];
		}
		for(i = 0; i < len; i++)
		    bucket->data[i] = CLIP_SHORT(adder_buffer[i]/nch_from);
		ADD_BUCKET_LIST(bucket_head, bucket_tail, bucket);
	    }
	    for(ch = 0; ch < nch_from; ch++)
	    {
		AudioBucket* tmp;

		tmp = channel_bucket[ch];
		channel_bucket[ch] = channel_bucket[ch]->next;
		reuse_audio_bucket(tmp);
	    }
	}
	if(bucket_head == NULL)
	    return new_audio_bucket();
	return bucket_head;
    }
    else
    {

/***************************************************************************
 *
 *		           -- FIXME --
 *		ϤޤǥХåƤޤ
 *
 ***************************************************************************/
	AudioBucket* bucket_temp[MAX_BUCKET_NCHANNELS];
	short* bucket_temp_data[MAX_BUCKET_NCHANNELS];
	AudioBucket* bucket_head;
	AudioBucket* bucket_tail;
	AudioBucket* bucket;
	int maxch;
	long bucket_remain_len;
	short* sptr;

	/* algorithm
	 *
	 * step 1
	 * bucket_temp[ch] (ch = 0 ... nch_to - 1) Фб
	 * ͥ sample ­碌롥
	 * for(j = ch; j < MAX(nch_from, nch_to; j += nch_to)
	 * 	bucket_temp[ch] += channel_bucket[j % nch_from];
	 *
	 *
	 * step 2
	 * bucket_temp[0]: aaaaaaaa
	 * bucket_temp[1]: bbbbbbbb   => abcabcabcabcabcabcabcabc
	 * bucket_temp[2]: cccccccc
	 */

	/*  */
	maxch = MAX(nch_from, nch_to);
	BUCKET_LIST_INIT(bucket_head, bucket_tail);
	for(ch = 0; ch < nch_to; ch++)
	{
	    if((bucket_temp[ch] = new_allocated_bucket()) == NULL)
	    {
		/* 㳰 */
		int j;

		while(ch--)
		    reuse_audio_bucket(bucket_temp[ch]);
		for(j = 0; j < nch_from; j++)
		    reuse_audio_bucket_list(channel_bucket[j]);
		return NULL;
	    }
	    bucket_temp_data[ch] = bucket_temp[ch]->data;
	}
	bucket_remain_len = 0;
	bucket = NULL;
	sptr = NULL;

	while(channel_bucket[0])
	{
	    long len = channel_bucket[0]->len;
	    long j;
	    long r = nch_from / nch_to;

	    /* step 1
	     * bucket_temp[0 ... nch_to - 1] κ
	     */
	    for(ch = 0; ch < nch_to; ch++)
	    {
		/* device  ch ͥб channel_bucket
		 * ­碌롥
		 *
		 * bucket_head[0] <=== ch0 + ch3 + ...
		 * bucket_head[1] <=== ch1 + ch4 + ...
		 * bucket_head[2] <=== ch2 + ch5 + ...
		 *
		 * 뤤
		 *
		 * bucket_head[0] <=== ch0
		 * bucket_head[1] <=== ch1
		 * bucket_head[2] <=== ch0
		 * bucket_head[3] <=== ch1
		 * ...		  <=== ...
		 *
		 * Τ褦ʤ
		 */

		memset(adder_buffer, 0, len * sizeof(long));
		for(j = ch; j < maxch; j += nch_to)
		{
		    long k;
		    short* p;

		    p = channel_bucket[j % nch_from]->data;
		    for(k = 0; k < len; k++)
			adder_buffer[k] += p[k];
		}

		if(r > 1)
		    for(j = 0; j < len; j++)
			bucket_temp_data[ch][j] = CLIP_SHORT(adder_buffer[j]/r);
		else
		    for(j = 0; j < len; j++)
			bucket_temp_data[ch][j] = CLIP_SHORT(adder_buffer[j]);
	    }

	    /* step 2
	     * bucket_temp[0 ... nch_to - 1]  1 ˺롥
	     */

	    for(j = 0; j < len; j++)
	    {
		if(bucket_remain_len == 0)
		{
		    if(bucket != NULL)
		    {
			ADD_BUCKET(bucket_head, bucket_tail, bucket);
		    }

		    if((bucket = new_allocated_bucket()) == NULL)
		    {
			/* 㳰 */
			int k;

			for(k = 0; k < nch_to; k++)
			    reuse_audio_bucket(bucket_temp[k]);
			for(k = 0; k < nch_from; k++)
			    reuse_audio_bucket_list(channel_bucket[j]);
			reuse_audio_bucket_list(bucket_head);

			return NULL;
		    }

		    bucket->pos = bucket->len = 0;

		    /* nch_to ܤˤʤ褦 */
		    bucket_remain_len = nch_to * 
			(AUDIO_BUCKET_LEN / nch_to);
		    sptr = bucket->data;
		}

		for(ch = 0; ch < nch_to; ch++)
		    sptr[ch] = bucket_temp_data[ch][j];
		sptr += nch_to;
		bucket->len += nch_to;
		bucket_remain_len -= nch_to;
	    }
	    for(ch = 0; ch < nch_from; ch++)
		channel_bucket[ch] = channel_bucket[ch]->next;
	}
	if(bucket != NULL)
	{
	    ADD_BUCKET(bucket_head, bucket_tail, bucket);
	}
	for(ch = 0; ch < nch_to; ch++)
	    reuse_audio_bucket(bucket_temp[ch]);
	if(bucket_head == NULL)
	    return new_audio_bucket();

	return bucket_head;
/***************************************************************************
 *
 *		ޤ
 *
 ***************************************************************************/

    }
}
