#ifndef ___BUCKET_H_
#define ___BUCKET_H_

/* Bucket Delivery Phase

A:	                      B:
+---+---+                     +---+
|ch1|ch2|                     |ch1|		- AudioBuffer
+---+---+                     +---+
  |   |        	                |
  |   +------------+	        |
  |	    	   |            |
+-------+      +-------+    +-------+
|buffer |      |buffer |    |buffer |		- reader
|reader1|      |reader2|    |reader3|
+-------+      +-------+    +-------+
  |         	   |            |
+-----------+ +-----------+ +-----------+
|sample rate| |sample rate| |sample rate|	- sample rate adjustment
|exchanger  | |exchanger  | |exchanger  |
+-----------+ +-----------+ +-----------+
        |          |             |
+-------------------------+ +-------------+
|      channel mixer      | |channel mixer|	- channel adjustment
+-------------------------+ +-------------+
             |                   |
+---------------------------------------+
|             wave adder                |	- compound audio
+---------------------------------------+
                   |
               +------+
               |buffer|				- buffering
               +------+
                   |
               +---------+
               |converter|			- convert to supported audio
               +---------+
                   |
               +-----+
               |audio|				- audio device
               +-----+

buffer reader:
	AudioBuffer  ǡɤߡѥåȤʬ򤵤줿 AudioBucket 
	̤Υե륿Ϥ

sample rate exchanger:
	Ŭڤʥץ졼Ȥˤʤ褦ˡץǡνʣְ
	Ԥʤ

channel mixer:
	ͥι

wave adder:
	νŤ͹碌

buffer:
	ʾǡίޤǡ̤˥ǡϤʤ

converter:
	ǥХ˰¸粽Ѵ

audio:
	ǥХ
*/

typedef struct _AudioBucket
{
    short* data;		/* ǡ */
    int pos;			/* ϰ */
    int len;			/* ǡ(ǸΥեǤϥХȿɽ)*/

    short allocated;		/* True ʤ ư˳ݤ줿ǡǤꡤ
				 * data  free ƤϤʤʤ
				 */
    short const_data;		/* True ʤ data ľѹƤϤʤʤ
				 * ԡƤѹ뤳ȡ
				 */

    struct _AudioBucket* next;
} AudioBucket;

enum audio_bucket_filter_t
{
    ABF_normal,
    ABF_reader,
    ABF_channel_mixer,
    ABF_wave_adder
};

typedef struct _AudioBucketFilter
{
    int filter_type;	/* ABF_normal/ABF_channel/ABF_reader/ABF_mixer */
    struct _AudioPlayer* player;
    void* ext_data;

    union
    {
	struct
	{
	    short* buffer;	/* PCM ǡ(1 ͥʬ) */
	    long seek_pointer;	/* ɤ߹߰ */
	} reader;

	struct
	{
	    struct _AudioBucketFilter* superior_filters[MAX_BUCKET_NCHANNELS];
	    int nchannels;
	} channel_mixer;

	struct
	{
	    struct _AudioBucketFilter* superior_filters[MAX_BUCKET_NCHANNELS];
	    struct _AudioBucket* remains[MAX_BUCKET_NCHANNELS];
	    struct _AudioPlayer* players[MAX_BUCKET_NCHANNELS];
	    int nchannels;
	} wave_adder;
    } u;

    AudioBucket* (* filter_proc)(struct _AudioBucketFilter *); /* ʬ */
    struct _AudioBucketFilter* superior; /* AudioBuffer ˶ᤤ */
    struct _AudioBucketFilter* free_chain;
} AudioBucketFilter;

extern AudioBucket* bucket_list;
extern AudioBucket* allocated_bucket_list;
extern AudioBucketFilter* audio_bucket_filter_free_list;

void initialize_audio_bucket(void);
/* ǽ 1 ٤ƤӽФޤ
 */

void reuse_audio_bucket(AudioBucket* bucket);
void reuse_audio_bucket_list(AudioBucket* bucket);
/* bucket Ѳǽˤޤ
 * bucket->allocated  True ʤ顤bucket  allocated_bucket_list 
 * Ĥʤ졤Ǥʤ bucket_list ˤĤʤޤ
 */

AudioBucket* new_audio_bucket(void);
AudioBucket* new_allocated_bucket(void);
/*  AudioBucket ޤ*/

long audio_bucket_length(AudioBucket* bucket);
/* bucket Ĺ֤ޤ */

#define is_bucket_list_empty(list) ((list)->head == NULL)
/* list  ʤ鿿Ǥ*/

AudioBucketFilter* alloc_audio_bucket_filter(int type);
/* ե륿ޤ
 */

AudioBucketFilter* make_normal_filter(AudioBucketFilter* superior,
		AudioBucket* (* filter_proc)(struct _AudioBucketFilter *));

AudioBucketFilter* make_mixer_filter(AudioBucketFilter** superior_list,
				     int nchannels,
		AudioBucket* (* filter_proc)(struct _AudioBucketFilter *));


/* ե륿Ѳǽˤޤ*/
void reuse_audio_bucket_filter(AudioBucketFilter* filter);
void reuse_audio_bucket_filter_list(AudioBucketFilter* filter);

/* Ѵؿ򹹿ޤ*/
void abc_update_audio_converter(void);

/* ץ졼Ȥѹޤ*/
AudioBucket* convert_sample_rate(AudioBucket* source,
				 long source_sample_rate,
				 long target_sample_rate);

/* ʣΥͥߤ¤٤ޤ*/
AudioBucket* mix_audio_channel(AudioBucket** channel_bucket,
			       int nch_from, int nch_to);

/*  buffer ХåȤΥꥹȤʬ䤷ޤ*/
AudioBucket* buffer_to_bucket_list(short* buffer, long buffer_len);

#define BUCKET_LIST_INIT(head, tail) head = tail = NULL

#define ADD_BUCKET(head, tail, bucket)	\
    if(head == NULL)			\
	head = tail = bucket;		\
    else				\
	tail = tail->next = bucket

#define ADD_BUCKET_LIST(head, tail, bucket)	\
    if(head == NULL)				\
	head = tail = bucket;			\
    else					\
	tail = tail->next = bucket;		\
    while(tail->next) tail = tail->next

#endif /* ___BUCKET_H_ */
