#include <math.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include "fft.h"
#include "convolution.h"

#ifndef CLIP_SHORT
#define CLIP_SHORT(x) ((x) < SHRT_MIN ? SHRT_MIN : (x) > SHRT_MAX ? SHRT_MAX : (short)(x))
#endif
#ifndef MAX
#define MAX(a, b) ((a) < (b) ? (b) : (a))
#endif

void convolution_proc(short* s, long s_len, long start,
		      short* t, long t_len,
		      double* h)
{
    static short* s_tmp = NULL;
    static long s_tmp_len;
    long i, lim;
    short* p;

    if(s_tmp != NULL && s_tmp_len < t_len + CONVOLUTION_SIZE)
    {
	free(s_tmp);
	s_tmp = NULL;
    }
    if(s_tmp == NULL)
    {
	s_tmp_len = MAX(8192, t_len) + CONVOLUTION_SIZE;
	s_tmp = (short *)malloc(s_tmp_len * sizeof(short));
    }

    if(s == NULL)
	return;			/* only initializing */

    start -= CONVOLUTION_SIZE / 2;

    if(start < 0 || start + t_len + CONVOLUTION_SIZE > s_len)
    {
	int j;

	lim = start + t_len + CONVOLUTION_SIZE;

	p = (short *)s_tmp;
	for(i = start, j = 0; i < 0; i++, j++)
	    p[j] = 0;
	for(; i < lim && i < s_len; i++, j++)
	    p[j] = s[i];
	for(; i < lim; i++, j++)
	    p[j] = 0;
    }
    else
	p = s + start;

/* Here is non unroll version
    for(i = 0; i < t_len; i++)
    {
	double v;
	int j, k;
	v = h[0] * p[i] + h[CONVOLUTION_SIZE/2] * p[i + CONVOLUTION_SIZE/2];
	for(j = 1, k = CONVOLUTION_SIZE - 1; j < CONVOLUTION_SIZE/2; j++, k--)
	    v += h[j] * ((int)p[i + j] + (int)p[i + k]);
	*t++ = CLIP_SHORT(v);
    }
*/

    for(i = 0; i < t_len - 1; i += 2)
    {
	double v1, v2;
	int j, k;

	v1 = h[0] * p[i] + h[CONVOLUTION_SIZE/2] * p[i + CONVOLUTION_SIZE/2]
	   + h[1] * ((int)p[i + 1] + (int)p[i + CONVOLUTION_SIZE - 1])
	   + h[2] * ((int)p[i + 2] + (int)p[i + CONVOLUTION_SIZE - 2])
	   + h[3] * ((int)p[i + 3] + (int)p[i + CONVOLUTION_SIZE - 3]);

	v2 = h[0] * p[i+1] + h[CONVOLUTION_SIZE/2] * p[i+1+CONVOLUTION_SIZE/2]
	   + h[1] * ((int)p[i + 2] + (int)p[i + CONVOLUTION_SIZE])
	   + h[2] * ((int)p[i + 3] + (int)p[i + CONVOLUTION_SIZE - 1])
	   + h[3] * ((int)p[i + 4] + (int)p[i + CONVOLUTION_SIZE - 2]);

	for(j = 4, k = CONVOLUTION_SIZE - 4;
	    j < CONVOLUTION_SIZE/2;
	    j += 4, k -= 4)
	{
	    v1 += h[j  ] * ((int)p[i + j  ] + (int)p[i + k  ])
	       +  h[j+1] * ((int)p[i + j+1] + (int)p[i + k-1])
	       +  h[j+2] * ((int)p[i + j+2] + (int)p[i + k-2])
	       +  h[j+3] * ((int)p[i + j+3] + (int)p[i + k-3]);
	    v2 += h[j  ] * ((int)p[i + j+1] + (int)p[i + k+1])
	       +  h[j+1] * ((int)p[i + j+2] + (int)p[i + k  ])
	       +  h[j+2] * ((int)p[i + j+3] + (int)p[i + k-1])
	       +  h[j+3] * ((int)p[i + j+4] + (int)p[i + k-2]);

	}
	*t++ = CLIP_SHORT(v1);
	*t++ = CLIP_SHORT(v2);
    }
    if(i < t_len)
    {
	double v;
	int j, k;
	v = h[0] * p[i] + h[CONVOLUTION_SIZE/2] * p[i + CONVOLUTION_SIZE/2];
	for(j = 1, k = CONVOLUTION_SIZE - 1; j < CONVOLUTION_SIZE/2; j++, k--)
	    v += h[j] * ((int)p[i + j] + (int)p[i + k]);
	*t++ = CLIP_SHORT(v);
    }
}

void mkconvolution(double* h)
{
    static double hx[CONVOLUTION_SIZE];
    static double hy[CONVOLUTION_SIZE];
    int i;
    double t;

    for(i = 0; i <= CONVOLUTION_SIZE/2; i++)
	hx[i] = h[i];
    for(i = 1; i < CONVOLUTION_SIZE/2; i++)
	hx[CONVOLUTION_SIZE - i] = hx[i];
    for(i = 0; i < CONVOLUTION_SIZE; i++)
	hy[i] = 0.0;

    fft(hx, hy, -CONVOLUTION_SIZE);

    for(i = 0; i < CONVOLUTION_SIZE/2; i++)
    {
	double tmp;

	tmp = hx[i];
	hx[i] = hx[i + CONVOLUTION_SIZE/2];
	hx[i + CONVOLUTION_SIZE/2] = tmp;
    }

    /* Window */
    t = -M_PI;
    for(i = 0; i <= CONVOLUTION_SIZE/2; i++)
    {
	hx[i] *= (0.50 + 0.50 * cos(t)) / CONVOLUTION_SIZE;
	t += 2.0 * M_PI / CONVOLUTION_SIZE;
    }

    memcpy(h, hx, (CONVOLUTION_SIZE/2 + 1) * sizeof(double));
}
