#include <stdio.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include "soundscope.h"
#include "ssframe.h"
#include "player.h"
#include "convolution.h"

#define EQULIZER_MIN_RESOLUTION      4
#define EQULIZER_WINDOW_WIDTH      512
#define EQULIZER_YRESOLUTION        10
#define EQULIZER_DECIBEL_RESOLUTION 25
#define EQULIZER_DECIBEL_ZERO       13
#define EQULIZER_WINDOW_HEIGHT (EQULIZER_DECIBEL_RESOLUTION * EQULIZER_YRESOLUTION)
#define EQULIZER_WINDOW_X           58
#define EQULIZER_DECIBEL_MAX        24
#define EQULIZER_DECIBEL_MIN       -24
#define EQULIZER_MEASSURE_WIDTH (EQULIZER_WINDOW_X - GAP)
#define MEASSURE_SIDE_X             31

static void equlizer_event_handler(AnyWindow* obj, XEvent* e)
{
    int i, j, h; 
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);
    WinObj* gp_win = frame->equlizer_gp;
    long sample_rate;

    if(e != NULL &&
       (e->type != Expose || e->xexpose.count > 0))
	return;

    h = TextHeight(font);
    sample_rate = frame->player->sample_rate;

    DrawRuledHLine(obj->disp, obj->win, obj->gc, obj->lbg, obj->hbg,
		   GAP / 2, h + 3 * GAP, obj->width - GAP);

    DrawRuledHLine(obj->disp, obj->win, obj->gc, obj->lbg, obj->hbg,
		   WinObjLeft(gp_win) - 6,
		   WinObjBottom(gp_win) - BD_WIDTH - 1, 3);

    XSetForeground(disp, gc, obj->fg);
    XDrawString(disp, obj->win, gc,
		WinObjLeft(gp_win) - 16,
		WinObjBottom(gp_win) - BD_WIDTH + 5,
		"0", 1);

    /* draw x messure */
    XSetForeground(disp, gc, obj->fg);
    for(i = WinObjLeft(gp_win) + BD_WIDTH;
	i < WinObjRight(gp_win);
	i += gp_win->gp.xresolution)
    {
	DrawRuledVLine(obj->disp, obj->win, obj->gc, obj->lbg, obj->hbg,
		       i, WinObjBottom(gp_win) + 1, 3);
    }
    for(i = WinObjLeft(gp_win) + BD_WIDTH, j = 0;
	i < WinObjRight(gp_win);
	i += 64, j += 64)
    {
	DrawRuledVLine(obj->disp, obj->win, obj->gc, obj->lbg, obj->hbg,
		       i, WinObjBottom(gp_win), 5);
	if(j > 0)
	{
	    sprintf(temp_buff, "%dHz", (int)(0.5 +
		    exp(j * log(CONVOLUTION_SIZE/2.0) / gp_win->gp.width)
		    / CONVOLUTION_SIZE * sample_rate));
	    XDrawString(disp, obj->win, gc,
			i - 13, WinObjBottom(gp_win) + 18,
			temp_buff, strlen(temp_buff));
	}
	else
	{
	    XDrawString(disp, obj->win, gc,
			i - 1,
			gp_win->any.y + gp_win->any.height + 18,
			"0Hz", 3);
	}
    }
}

void redraw_equlizer_window(SSFrame* frame)
{
    WinObj* eq;

    eq = frame->equlizer;
    XClearWindow(disp, eq->any.win);
    equlizer_event_handler((AnyWindow *)eq, NULL);
}

static void close_equlizer_proc(ButtonWindow* btn, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(btn);
    XUnmapWindow(disp, frame->equlizer->any.win);
}

static double convert_graph_panel_data(SSFrame* frame, int y)
{
    double decibel;

    if(y == 0)
	return 0.0;

    decibel = (double)(y - frame->equlizer_measure_zero)
	/ (double)EQULIZER_YRESOLUTION;

    return pow(10.0, decibel * 0.05);
}

static void get_graph_panel_data(SSFrame* frame, double* h)
{
    int i;
    const int* hlist;
    int j0, j1;
    double a;

    hlist = GetGraphPanelValueList(frame->equlizer_gp, NULL);
    a = frame->equlizer_gp->gp.width / log(CONVOLUTION_SIZE/2.0);

    j0 = 0;
    for(i = 1; i < CONVOLUTION_SIZE/2; i++)
    {
	int j, m;
	double v;
	j1 = (int)(a * log((double)i));
	m = 1;
	v = convert_graph_panel_data(frame, hlist[j1]);
	for(j = j0 + 1; j < j1; j++)
	{
	    m++;
	    v += convert_graph_panel_data(frame, hlist[j]);
	}
	j0 = j1;
	h[i] = v / m;
    }

    h[0] = h[1];
    h[CONVOLUTION_SIZE/2] = h[CONVOLUTION_SIZE/2 - 1];
}



static void use_equlizer(SSFrame* frame)
{
    get_graph_panel_data(frame, frame->convolution);
    mkconvolution(frame->convolution);

    frame->equlizing_flag = True;
    if(audio_active_flag)
    {
	abc_stop_audio();
	abc_restart_audio();
    }
    else
	repaint_frame_scope(frame, frame->player->pos, False);
}

static void gp_call_back_handler(WinObj* obj, XEvent* e)
{
    int i;
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count > 0)
	    break;
      case ButtonPress:
      case MotionNotify:
	XSetForeground(disp, dashgc, obj->any.mark_color);
	XSetForeground(disp, gc,     obj->any.mark_color);
	for(i = EQULIZER_YRESOLUTION;
	    i < obj->any.height;
	    i += EQULIZER_YRESOLUTION)
	{
	    XDrawLine(disp, obj->any.win, gc,
		      0, i - 1,
		      obj->any.width - 1, i - 1);
	}
	i = obj->any.height - EQULIZER_YRESOLUTION * EQULIZER_DECIBEL_ZERO;
	XDrawLine(disp, obj->any.win, gc,
		  0, i, obj->any.width - 1, i);

	for(i = obj->gp.xresolution;
	    i < obj->any.width;
	    i += obj->gp.xresolution)
	{
	    XDrawLine(disp, obj->any.win, dashgc,
		      i, 0,
		      i, obj->any.height - 1);
	}
	XSync(obj->any.disp, False);
	break;
      case ButtonRelease:
	XSync(obj->any.disp, False);
	if(frame->equlizer_query->choice.selected == 1)
	    use_equlizer(frame);
	break;
    }
}


static void redraw_equlizer_measure(SSFrame* frame, AnyWindow* obj)
{
    Display* disp = obj->disp;
    Window win = obj->win;
    int width = obj->width, height = obj->height; 
    Pixmap pm;
    const int mx = MEASSURE_SIDE_X;
    const int bd_width = 1;
    int i;
    int decibel, measure;
    static Pixmap equlizer_measure_pm = None;

    if(equlizer_measure_pm == None)
	equlizer_measure_pm
	    = XCreatePixmap(disp, win, width, height,
			    DefaultDepth(disp, DefaultScreen(disp)));
    pm = equlizer_measure_pm;
    XSetForeground(disp, gc, obj->bg);
    XFillRectangle(disp, pm, gc, 0, 0, width, height);
    DrawRuledVLine(disp, pm, gc, obj->lbg, obj->hbg,
		   mx, bd_width, height - 1 - 2 * bd_width);
    for(i = EQULIZER_YRESOLUTION + bd_width; i < height - bd_width;
	i += EQULIZER_YRESOLUTION)
	DrawRuledHLine(disp, pm, gc, obj->lbg, obj->hbg, mx + 1, i, 5);

    decibel = frame->equlizer_measure_start / EQULIZER_YRESOLUTION + 1;
    measure = frame->equlizer_measure_start % EQULIZER_YRESOLUTION + bd_width - EQULIZER_YRESOLUTION;

    for(i = 0; i < EQULIZER_DECIBEL_RESOLUTION + 1;
	i++, decibel--, measure += EQULIZER_YRESOLUTION)
    {
	int len;

	DrawRuledHLine(disp, pm, gc, obj->lbg, obj->hbg,
		       mx - 6, measure, 5);

	if(decibel % 2)
	    continue;

	if(decibel > 0)
	    sprintf(temp_buff, "+%d", decibel);
	else
	    sprintf(temp_buff, "%d", decibel);
	len = strlen(temp_buff);
	XSetForeground(disp, gc, obj->fg);
	XDrawString(disp, pm, gc,
		    mx - 8 - GC_CHAR_WIDTH * len,
		    measure + 6,
		    temp_buff, len);
    }
    DrawBorder(disp, pm, gc, 0, 0, width, height,
	       1, obj->lbg, obj->hbg, False);
    XCopyArea(disp, equlizer_measure_pm, win, gc,
	      0, 0, width, height, 0, 0);
}

static void move_equlizer_measure(SSFrame* frame, AnyWindow* obj, int diff)
{
    frame->equlizer_measure_start += diff;
    frame->equlizer_measure_zero  -= diff;
    redraw_equlizer_measure(frame, obj);
}

static void equlizer_measure_proc(AnyWindow* obj, XEvent* e)
{
    static int    equlizer_measure_y;
    static Bool   equlizer_measure_button_flag = False;
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count == 0)
	    redraw_equlizer_measure(frame, obj);
	break;
      case ButtonPress:
	if(e->xbutton.x < MEASSURE_SIDE_X)
	{
	    equlizer_measure_button_flag = True;
	    equlizer_measure_y = e->xbutton.y;
	}
	break;
      case ButtonRelease:
	equlizer_measure_button_flag = False;
	if(frame->equlizer_gp->any.ext_data != NULL)
	    use_equlizer(frame);
	break;
      case MotionNotify:
	if(equlizer_measure_button_flag == True)
	{
	    CompressMotionEvent(e->xany.display, e->xany.window, e);
	    move_equlizer_measure(frame, obj,
				  e->xmotion.y - equlizer_measure_y);
	    equlizer_measure_y = e->xmotion.y;
	}
	break;
    }
}

static void equlizer_use_switch(ChoiceButton* c, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(c);

    if(c->selected == 0)
    {
	frame->equlizing_flag = False;
	ResizeWinObj(frame->equlizer,
		     WinObjRight(frame->equlizer_gp) + 2 * GAP,
		     WinObjBottom(frame->equlizer_query) + GAP + GAP / 2);
	if(audio_active_flag)
	{
	    abc_stop_audio();
	    abc_restart_audio();
	}
	else
	    repaint_frame_scope(frame, frame->player->pos, False);
    }
    else
    {
	use_equlizer(frame);
	ResizeWinObj(frame->equlizer,
		     WinObjRight(frame->equlizer_gp) + 2 * GAP,
		     WinObjBottom(frame->equlizer_gp) + 4 * GAP + 2 * BD_WIDTH
		     + TextHeight(font));
    }
}

static void redraw_gp_win(SSFrame* frame)
{
    XEvent ev;

    XClearWindow(disp, frame->equlizer_gp->any.win);
    ev.type = Expose;
    ev.xexpose.count = 0;
    frame->equlizer_gp->any.event_handler(frame->equlizer_gp, &ev);

    redraw_equlizer_window(frame);
}

static void change_equlizer_resolution(ChoiceButton* c, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(c);
    int resolution = EQULIZER_MIN_RESOLUTION;
    int i;

    for(i = 0; i < c->selected; i++)
	resolution <<= 1;
    frame->equlizer_gp->gp.xresolution = resolution;

    redraw_gp_win(frame);
}

void create_equlizer(SSFrame* frame)
{
    WinObj* close_btn;
    WinObj* equlizer_resolution;
    static WinObj* equlizer_measure;
    WinObj* equlizer;
    WinObj* equlizer_query;
    WinObj* gp_win;
    int i, h;
    char* query_strings[2] = {"Unuse", "Use"};
    char* resolution_names[] =
    { "128", "64", "32", "16", "8", "4" };
#define NRESOLUTION (sizeof(resolution_names) / sizeof(resolution_names[0]))

    h = TextHeight(font) + GAP;

    equlizer = toplevel_WinObj(disp, "equlizer", "Equlizer", 1, 1,
			       WinObj_fg, WinObj_bg);
    if(ISMONODISP)
    {
	equlizer->any.mono_color = True;
	equlizer->any.mark_color = WinObj_fg;
    }

    equlizer->any.when_realize_mapping = False;
    SetEventMask(equlizer, ExposureMask);
    SetEventHandler(equlizer, equlizer_event_handler);

    close_btn = create_Button(equlizer, GAP, GAP,
			      TextWidth(font, "Ĥ") + GAP, h,
			      BD_WIDTH,
			      font, "Ĥ");
    SetCallBackProc(close_btn, (call_back_t)close_equlizer_proc);

    gp_win = create_GraphPanel(equlizer,
			       EQULIZER_WINDOW_X,
			       h + 3 * GAP,
			       EQULIZER_WINDOW_WIDTH,
			       EQULIZER_WINDOW_HEIGHT,
			       BD_WIDTH, 128, EQULIZER_YRESOLUTION);
    SetCallBackProc(gp_win, (call_back_t)gp_call_back_handler);
    for(i = 0; i < gp_win->gp.width; i++)
	gp_win->gp.gpbuff[i]
	    = EQULIZER_DECIBEL_ZERO * EQULIZER_YRESOLUTION;

    equlizer_query =
	create_ChoiceButton(equlizer, WinObjRight(close_btn) + GAP, GAP,
			    BD_WIDTH, h, font, "Equlizer:",
			    query_strings, 2, 0);
    SetCallBackProc(equlizer_query, (call_back_t)equlizer_use_switch);

    equlizer_resolution =
	create_ChoiceButton(equlizer,
			    WinObjLeft(gp_win), WinObjBottom(gp_win) + 2 * GAP,
			    BD_WIDTH, h, font, "Resolution:",
			    resolution_names, NRESOLUTION, 5);
    SetCallBackProc(equlizer_resolution,
		    (call_back_t)change_equlizer_resolution);

    equlizer_measure
	= create_AnyWindow(equlizer,
			   EQULIZER_WINDOW_X - EQULIZER_MEASSURE_WIDTH,
			   WinObjTop(gp_win),
			   EQULIZER_MEASSURE_WIDTH,
			   gp_win->any.height - EQULIZER_YRESOLUTION);
    SetEventMask(equlizer_measure, ButtonPressMask | ButtonReleaseMask
		 | ButtonMotionMask | ExposureMask);
    SetEventHandler(equlizer_measure, equlizer_measure_proc);

    ResizeWinObj(equlizer,
		 WinObjRight(gp_win) + 2 * GAP,
		 WinObjBottom(equlizer_query) + GAP + GAP / 2);

    frame->equlizing_flag = False;
    frame->equlizer = equlizer;
    frame->equlizer_gp = gp_win;
    frame->equlizer_query = equlizer_query;
    frame->equlizer_measure_start = (EQULIZER_DECIBEL_RESOLUTION -
	EQULIZER_DECIBEL_ZERO) * EQULIZER_YRESOLUTION;
    frame->equlizer_measure_zero
	= EQULIZER_DECIBEL_ZERO * EQULIZER_YRESOLUTION;
    WinObjRealize(equlizer);
}
 
void open_equlizer_win_popup_proc(PopupMenuBase* menu, char* name, int id)
{
    Window win = WINOBJ_EVENT_FRAME(menu)->equlizer->any.win;
    static Bool init_flag = False;

    if(!init_flag)
    {
	XSizeHints sizehints;

	init_flag = True;
	sizehints.flags = USPosition;
	QueryPointer(disp, DefaultRootWindow(disp),
		     &sizehints.x, &sizehints.y);
	XSetNormalHints(disp, win, &sizehints);
	XMoveWindow(disp, win, sizehints.x, sizehints.y);
    }
    XMapWindow(disp, win);
}
