#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "soundscope.h"
#include "aubuff.h"
#include "ssframe.h"
#include "soundex.h"
#include "player.h"
#include "timer.h"

#ifdef USE_FILE_MAP
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>
#endif /* USE_FILE_MAP */

#define FILE_LOAD_WINDOW_WIDTH 500
#define FILE_INDICATOR_WINDOW_WIDTH 600
static Bool indicator_map_flag = False;

struct _file_loader file_loader =
{
    False,			/* load_flag */
    NULL,			/* data */
    1,				/* nchannels */
    1,				/* block_align (= byte*nchannels/sample */
    0,				/* read_samples */
    0,				/* rest_samples */
    NULL,			/* load_buffer */
    NULL,			/* converter */
#ifdef USE_FILE_MAP
    NULL,			/* map_file */
    0,				/* map_file_size */
    0,				/* map_file_offset */
#endif /* USE_FILE_MAP */
    NULL,			/* backup_buffer */
    8000,			/* orig_sample_rate */
};


SSFrame* open_file_target_frame = NULL;
static char indicator_message[BUFSIZ];
static USFFD file_usffd;
static WinObj* sound_conversion_method;
static WinObj* sound_conversion_channel;
static WinObj* sound_conversion_encoding;
static WinObj* sound_conversion_sample_rate;
static WinObj* sound_conversion_base_table;
static int sound_conversion_y1, sound_conversion_y2;
static int sound_conversion_h1, sound_conversion_h2;
static WinObj* sound_conversion_ok_button;
static WinObj* sound_conversion_cancel_button;
static WinObj* file_load_window;
static WinObj* secondary_load_text;
static WinObj* file_indicator_window;
static WinObj* file_indicator_slider;
static WinObj* indicator_cancel_button;


static void file_indicator_event_handler(WinObj* obj, XEvent* e)
{
    AnyWindow* any = (AnyWindow *)obj;
    static Window move_window_target = None;
    static int    move_window_x;
    static int    move_window_y;
    static double button_click_time = 0.0;
    double t;

#ifdef DEBUG
    printf("file_indicator_event_handler:%s\n", EventTypeString(e->type));
#endif

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count != 0)
	    break;
	DrawString(disp, any->win, font, any->fg, 2 * GAP, 2 * GAP,
		   0, GAP + TextHeight(font), indicator_message, West);
	DrawBorder(disp, any->win, any->gc,
		   0, 0, any->width, any->height,
		   BD_WIDTH, any->lbg, any->hbg, True);
	DrawBorder(disp, any->win, any->gc,
		   BD_WIDTH * 2, 
		   BD_WIDTH * 2,
		   any->width  - BD_WIDTH * 4,
		   any->height - BD_WIDTH * 4,
		   BD_WIDTH, any->lbg, any->hbg, False);
	XSync(disp, False);
	break;
      case ButtonPress:
	XRaiseWindow(disp, obj->any.win);
	move_window_target = e->xbutton.window;
	move_window_x = e->xbutton.x;
	move_window_y = e->xbutton.y;
	t = get_current_calender_time();
	if(t - button_click_time > 0.5)
	    button_click_time = t;
	else
	{
	    button_click_time = 0;
	    close_file_indicator();
	}
	break;
      case ButtonRelease:
	move_window_target = None;
	break;
      case MotionNotify:
	button_click_time = 0.0;
	if(e->xbutton.window == move_window_target)
	{
	    int x, y;
	    CompressMotionEvent(e->xany.display, e->xany.window, e);

	    QueryPointer(obj->any.disp, DefaultRootWindow(obj->any.disp),
			 &x, &y);
	    MoveWinObj(obj, x - move_window_x, y - move_window_y);
	    XSync(disp, False);
	}
	break;
    }
}

static void reset_hlimited_slider(SSFrame* frame)
{
    if(frame->player->aubuff == open_file_target_frame->player->aubuff)
	SetHLimitedSliderLimit(frame->play_slider,
		open_file_target_frame->play_slider->limited_slider.maxlen);
}

PUBLIC void open_file_indicator(const char* message,
			void (* cancel_handler)(WinObj* obj, XEvent* e))
{
    if(indicator_map_flag == True)
	return;
    XMoveWindow(disp, file_indicator_window->any.win,
		(display_width - file_indicator_window->any.width) / 2,
		(display_height - file_indicator_window->any.height) / 2);
    strcpy(indicator_message, message);
    if(cancel_handler != NULL)
    {
	XMapWindow(disp, SkeletonWinObj(indicator_cancel_button)->any.win);
	SetCallBackProc(indicator_cancel_button, (call_back_t)cancel_handler);
    }
    else
    {
	XUnmapWindow(disp, SkeletonWinObj(indicator_cancel_button)->any.win);
    }
    SetHSimpleSliderPos(file_indicator_slider, 0);
    XRaiseWindow(disp, file_indicator_window->any.win);
    XMapWindow(disp, file_indicator_window->any.win);
    indicator_map_flag = True;
}

PUBLIC void close_file_indicator(void)
{
    if(indicator_map_flag == False)
	return;
    XUnmapWindow(disp, file_indicator_window->any.win);
    indicator_map_flag = False;
}

static void load_file_done(void)
{
    close_file_indicator();
    if(number_of_ssframes() > 1)
	PopupMenuResetNoselect(open_file_target_frame->sound_win_popup,
			       WIN_POPUP_WINCLOSE);
#ifdef MAP_FILE
    if(file_loader.map_file != NULL)
	munmap(file_loader.map_file, file_loader.map_file_size);
#endif /* MAP_FILE */
    usffd_close(&file_usffd);
    file_loader.load_flag = False;
    if(open_file_target_frame != NULL)
    {
	if(open_file_target_frame->player != NULL)
	{
	    open_file_target_frame->player->load_flag = False;
	    if(!open_file_target_frame->player->play_flag)
		repaint_frame_scope(open_file_target_frame,
				    open_file_target_frame->player->pos,
				    False);
	}
	zoom_slider_proc(open_file_target_frame->zoom_slider, NULL);
    }

    open_file_target_frame = NULL;
    set_audio_file_lock(False);
    update_ss_active_flag();
#ifdef DEBUG
    puts("File loading is done.");
#endif
}

static void load_cancel_button_proc(WinObj* obj, XEvent* e)
{
    /* recover backup audio buffer */

#ifdef DEBUG
    puts("load_cancel_button_proc()");
#endif

    SSFrameMapFuntion(reset_hlimited_slider);
    set_initial_player(open_file_target_frame->player);
    kill_audio_buffer(file_loader.load_buffer);
    open_file_target_frame->player->aubuff = file_loader.backup_buffer;
    player_change_sample_rate(open_file_target_frame->player,
			      file_loader.orig_sample_rate);
    player_change_channel(file_loader.orig_channels);
    player_change_encoding(file_loader.orig_encoding);
    if(file_loader.backup_buffer != NULL &&
       file_loader.backup_buffer->name != NULL)
	SetLabeldWindowLabel(open_file_target_frame->buffer_label,
			     file_loader.backup_buffer->name,
			     WINOBJ_BORDER_CONCAVE);
    else
	SetLabeldWindowLabel(open_file_target_frame->buffer_label,
			     SOUNDSCOPE_DEFAULT_BUFFER_NAME,
			     WINOBJ_BORDER_CONCAVE);
    init_scope_pixmap(open_file_target_frame);
    load_file_done();
}

PUBLIC void initialize_file_indicator_window(void)
{
    WinObj* top;
    int height, indicator_cancel_button_width;
    XSetWindowAttributes at;

    height = GAP + TextHeight(font);
    indicator_cancel_button_width = GAP + TextWidth(font, "󥻥");

    top = file_indicator_window =
	toplevel_WinObj(disp, "Sound File Load", "Sound File Load",
			FILE_INDICATOR_WINDOW_WIDTH,
			2 * height + 4 * GAP + 2 * BD_WIDTH,
			WinObj_fg, WinObj_bg);
    top->any.when_realize_mapping = False;
    SetEventMask(top, ExposureMask | ButtonPressMask | ButtonReleaseMask
		 | ButtonMotionMask);
    SetEventHandler(top, file_indicator_event_handler);
    at.override_redirect = True;
    XChangeWindowAttributes(top->any.disp, top->any.win,
			    CWOverrideRedirect, &at);

    indicator_cancel_button =
	create_Button(top,
		      top->any.width - indicator_cancel_button_width
		      - 2 * BD_WIDTH - 2 * GAP,
		      3 * GAP + TextHeight(font),
		      indicator_cancel_button_width, height, BD_WIDTH,
		      font, "󥻥");

    file_indicator_slider =
	create_HSimpleSlider(top, 2 * GAP, 3 * GAP + TextHeight(font),
			     top->any.width - 5 * GAP - 2 * BD_WIDTH
			     - indicator_cancel_button_width,
			     height,
			     BD_WIDTH, 0);
    file_indicator_slider->sslider.move_able = False;

    WinObjRealize(top);
}

static void sound_conversion_method_proc(ChoiceButton* c, XEvent* e)
{
    if(sound_conversion_method->choice.selected == 0)
    {
	XUnmapWindow(disp,
		     SkeletonWinObj(sound_conversion_base_table)->any.win);
	MoveWinObj(sound_conversion_ok_button,
		   WinObjLeft(sound_conversion_ok_button),
		   sound_conversion_y1);
	MoveWinObj(sound_conversion_cancel_button,
		   WinObjLeft(sound_conversion_cancel_button),
		   sound_conversion_y1);
	ResizeWinObj(file_load_window,
		     file_load_window->any.width,
		     sound_conversion_h1);
    }
    else
    {
	XMapWindow(disp,
		   SkeletonWinObj(sound_conversion_base_table)->any.win);
	MoveWinObj(sound_conversion_ok_button,
		   WinObjLeft(sound_conversion_ok_button),
		   sound_conversion_y2);
	MoveWinObj(sound_conversion_cancel_button,
		   WinObjLeft(sound_conversion_cancel_button),
		   sound_conversion_y2);
	ResizeWinObj(file_load_window,
		     file_load_window->any.width,
		     sound_conversion_h2);
    }
}

#if 0
static void setup_file_load_event_handler(WinObj* obj, XEvent* e)
{}
#endif

#ifdef USE_FILE_MAP
static char* open_map_file(int fd, int prot, long* text_size_return)
{
    struct stat buf;
    char* text;

/* FIXME the error handling */
    if(fstat(fd, &buf) < 0)
	return NULL;

    if(!S_ISREG(buf.st_mode))
	return NULL;

    if(buf.st_size == 0)
	return NULL;

    if((text = (char *)mmap(0, buf.st_size, prot, MAP_SHARED, fd, 0))
       == (caddr_t) - 1)
	return NULL;

    *text_size_return = (long)buf.st_size;
    return text;

}
#endif /* USE_FILE_MAP */

static Bool start_load_file_proc(USFFD* usffd, const char* buffer_name)
{
    AudioBuffer* new_buffer;
    AudioPlayer* player = open_file_target_frame->player;

#ifdef DEBUG
    usffd_print(&file_usffd);
#endif

    /* setup windows */
    XUnmapWindow(disp, file_load_window->any.win);

    new_buffer = make_usffd_audio_buffer(usffd, buffer_name);
    if(new_buffer == NULL)
	return False;

    set_initial_player(player);

    /* setup file_loader */
    file_loader.load_flag = True;
    file_loader.data = new_buffer->buffer;
    file_loader.nchannels = usffd->nchannels;
    file_loader.block_align
	= usffd_encoding_sample_size[usffd->data_encoding]
	    * usffd->nchannels;
    file_loader.read_samples = 0;
    file_loader.rest_samples = new_buffer->nsamples;
    file_loader.load_buffer = new_buffer;
#ifdef DEBUG
    printf("load converter: %s -> %s\n",
	   usffd_encoding_jname(usffd->data_encoding),
	   usffd_encoding_jname(internal_encoding));
    printf("nchannels: %ld\n", file_loader.nchannels);
    printf("sample rate: %ld\n", usffd->sample_rate);
#endif
    file_loader.converter = select_converter(usffd->data_encoding,
					     internal_encoding,
					     NULL, NULL);

#ifdef USE_FILE_MAP
    file_loader.map_file_offset = (long)lseek(usffd->fd, 0, SEEK_CUR);
    file_loader.map_file = open_map_file(usffd->fd,
					 PROT_READ,
					 &file_loader.map_file_size);
    if(file_loader.map_file == NULL)
	return False;
#endif /* USE_FILE_MAP */

    /*  */
    file_loader.backup_buffer = player->aubuff;
    file_loader.orig_sample_rate = player->sample_rate;
    file_loader.orig_channels = device_nchannels;
    file_loader.orig_encoding = device_data_encoding;

    /* ե饰Υå */
    ss_active_flag = True;
    player->load_flag = True;

    if(DATA_FRAGMENTATION_SIZE / file_loader.block_align * 5
       < file_loader.rest_samples)
    {
	if(usffd->data_size >= 1000000)
	    sprintf(temp_buff, "Loading %s (%.1f Mbytes)",
		    new_buffer->name, (double)usffd->data_size / 1000000.0);
	else
	    sprintf(temp_buff, "Loading %s (%.1f kbytes)",
		    new_buffer->name, (double)usffd->data_size / 1000.0);
	open_file_indicator(temp_buff, load_cancel_button_proc);
    }

    /* 桢ե졼बĤʤ褦 */
    PopupMenuNoselect(open_file_target_frame->sound_win_popup,
		      WIN_POPUP_WINCLOSE);

    /* Хåե̾Υå */
    SetLabeldWindowLabel(open_file_target_frame->buffer_label,
			 new_buffer->name, WINOBJ_BORDER_CONCAVE);

    /* ¾ */
    player_change_sample_rate(player, new_buffer->sample_rate);
    player_change_channel(new_buffer->nchannels);
    player_change_encoding(new_buffer->file_encoding);
    player->aubuff = new_buffer;
    AddDirLineTextShortCut(secondary_load_text, usffd->filename);
    init_scope_pixmap(open_file_target_frame);

    return True;
}

static char* file_name_no_direcotry(char* p)
{
    char* s;

    s = p;
    if(!*p)
	return p;
    do
	p++;
    while(*p);
    do
	p--;
    while(p != s && *p != '/');
    if(*p == '/')
	return p + 1;
    return p;
}

static void setup_file_load_ok(WinObj* obj, XEvent* e)
{
    char filename[BUFSIZ];

    GetLineText((LineText *)secondary_load_text, filename, BUFSIZ);
    if(filename[0] == '\0')
    {
	write_error_message("File name is empty");
	return;
    }

    if(usffd_open(ExpandHomeDirectory(filename), &file_usffd) == False)
    {
	if(usffd_errno == USFFD_SYS_ERROR)
	    write_syserror_message(filename);
	else
	{
	    sprintf(temp_buff, "%s: %s",
		    filename, usffd_strerror(usffd_errno));
	    write_error_message(temp_buff);
	}
	return;
    }

#if 0
    /* Change the usffd file name */
    xfree(file_usffd.filename);
    file_usffd.filename = xstrdup(filename);
#endif

    if(sound_conversion_method->choice.selected == 0)
    {
	if(!usffd_set_header(&file_usffd))
	{
	    if(usffd_errno != USFFD_NO_ERROR)
		write_error_message(usffd_strerror(usffd_errno));
	    usffd_encoding_recognize(&file_usffd);
	}
    }
    else
    {
	usffd_make_header(&file_usffd,
			sound_conversion_encoding->select_menu.selected_id,
			sound_conversion_channel->select_menu.selected_id,
			sound_conversion_sample_rate->select_menu.selected_id);
    }

    if(start_load_file_proc(&file_usffd,
		file_name_no_direcotry(file_usffd.filename)) == False)
    {
	load_file_done();
	return;
    }

    SetLineText((LineText *)open_file_target_frame->dir_text,
		filename, strlen(filename));
}

PUBLIC void initialize_setup_file_load_window(void)
{
    WinObj* top;
    char* snd_conv_method_msg[2] = {"ư", "ư"};
    WinObj* cancel_button;
    WinObj* ok_button;
    int cancel_button_width;
    int h, height, width;

    cancel_button_width = TextWidth(font, "󥻥") + GAP;
    h = TextHeight(font);
    height = h + GAP;

    width = FILE_LOAD_WINDOW_WIDTH;
    top = file_load_window =
	toplevel_WinObj(disp, "Sound File Selector", "Sound File Selector",
			width,
			3 * height + 5 * GAP + 6 * BD_WIDTH,
			WinObj_fg, WinObj_bg);
    top->any.when_realize_mapping = False;
#if 0
    SetEventMask(top, ExposureMask);
    SetEventHandler(top, setup_file_load_event_handler);
#endif

    secondary_load_text =
	create_DirLineText(top,
			   GAP, GAP,
			   top->any.width - 2 * GAP, height,
			   BD_WIDTH, font, " ե:");
    SetCallBackProc(secondary_load_text, (call_back_t)setup_file_load_ok);
    secondary_load_text->dirtext.short_cut_list =
	(MCPopupMenu *)text_short_cut_list_popup;
    SetPopupSelectingHook(secondary_load_text, popup_waiting_hook);

    sound_conversion_method =
	create_ChoiceButton(top,
			    GAP, WinObjBottom(secondary_load_text) + GAP,
			    BD_WIDTH, height, font,
			    "Ѵ ", snd_conv_method_msg, 2, 0);
    SetCallBackProc(sound_conversion_method,
		    (call_back_t)sound_conversion_method_proc);

    sound_conversion_base_table =
	create_SimpleWindow(top,
			    GAP, WinObjBottom(sound_conversion_method) + GAP,
			    1, 1, 1, WINOBJ_BORDER_CONCAVE);

    {
	WinObj* popup;
	WinObj* label;
	int i;
	popup = create_PopupMenu(disp, NULL, BD_WIDTH,
				 WinObj_fg, WinObj_bg, font);
	for(i = 1; i <= MAX_BUCKET_NCHANNELS; i++)
	{
	    sprintf(temp_buff, "%d", i);
	    add_PopupMenu(popup, temp_buff, i, NULL);
	}

	label = create_LabeldWindow(sound_conversion_base_table, GAP, GAP,
				    TextWidth(font, "ͥ: "),
				    8 + h + 2 * BD_WIDTH,
				    0, font, "ͥ: ",
				    WINOBJ_BORDER_NONE, Center);
	sound_conversion_channel =
	    create_SelectMenu(sound_conversion_base_table,
			      WinObjRight(label), GAP,
			      font, BD_WIDTH, popup, 1);
    }

    {
	WinObj* popup;
	WinObj* label;
	int i;
	popup = create_PopupMenu(disp, NULL, BD_WIDTH,
				 WinObj_fg, WinObj_bg, font);
	for(i = 1; i < USFFD_NENCODINGS; i++)
	    add_PopupMenu(popup, usffd_encoding_jname(i), i, NULL);
	label = create_LabeldWindow(sound_conversion_base_table,
				    GAP,
				    WinObjBottom(sound_conversion_channel)
				    + GAP,
				    TextWidth(font, "沽: "),
				    8 + h + 2 * BD_WIDTH,
				    0, font, "沽: ",
				    WINOBJ_BORDER_NONE, Center);
	sound_conversion_encoding =
	    create_SelectMenu(sound_conversion_base_table,
			      WinObjRight(label),
			      WinObjBottom(sound_conversion_channel) + GAP,
			      font, BD_WIDTH, popup, USFFD_G711_ULAW);
    }

    {
	WinObj* popup;
	WinObj* label;
	int i;
	long sample_rate_list[] = {
	    5512, 6615,	8000, 9600, 11025, 16000,
	    18900, 22050, 32000, 37800, 44100, 48000, 0};
	popup = create_PopupMenu(disp, NULL, BD_WIDTH,
				 WinObj_fg, WinObj_bg, font);
	for(i = 0; sample_rate_list[i]; i++)
	{
	    sprintf(temp_buff, "%ld", sample_rate_list[i]);
	    add_PopupMenu(popup, temp_buff, sample_rate_list[i], NULL);
	}
	label = create_LabeldWindow(sound_conversion_base_table,
				    GAP,
				    WinObjBottom(sound_conversion_encoding)
				    + GAP,
				    TextWidth(font, "ץ졼: "),
				    8 + h + 2 * BD_WIDTH,
				    0, font, "ץ졼: ",
				    WINOBJ_BORDER_NONE, Center);
	sound_conversion_sample_rate =
	    create_SelectMenu(sound_conversion_base_table,
			      WinObjRight(label),
			      WinObjBottom(sound_conversion_encoding) + GAP,
			      font, BD_WIDTH, popup, 8000);
    }

    ResizeWinObj(sound_conversion_base_table,
		 WinObjRight(sound_conversion_encoding) + GAP,
		 WinObjBottom(sound_conversion_sample_rate) + GAP);
    if(WinObjRight(sound_conversion_base_table) + GAP > width)
    {
	width = WinObjRight(sound_conversion_base_table) + GAP;
	ResizeWinObj(top, width,
		     3 * height + 5 * GAP + 6 * BD_WIDTH
		     );
	ResizeWinObj(secondary_load_text, top->any.width - 2 * GAP, height);
    }

    ok_button = create_Button(top,
			      top->any.width - cancel_button_width
			      - GAP - 2 * BD_WIDTH,
			      top->any.height - GAP - height - 2 * BD_WIDTH,
			      cancel_button_width, height, BD_WIDTH,
			      font, "ϣ");
    cancel_button = create_Button(top,
				  WinObjLeft(ok_button) - cancel_button_width
				  - GAP - 2 * BD_WIDTH,
				  top->any.height - GAP - height
				  - 2 * BD_WIDTH,
				  cancel_button_width, height, BD_WIDTH,
				  font, "󥻥");
    SetCallBackProc(cancel_button, (call_back_t)setup_file_load_cancel);
    SetCallBackProc(ok_button, (call_back_t)setup_file_load_ok);

    sound_conversion_y1 = WinObjTop(ok_button);
    sound_conversion_y2 = sound_conversion_y1 +
	sound_conversion_base_table->any.height;
    sound_conversion_cancel_button = cancel_button;
    sound_conversion_ok_button = ok_button;

    sound_conversion_h1 = top->any.height;
    sound_conversion_h2 = top->any.height +
	sound_conversion_base_table->any.height;

    WinObjRealize(top);
}

PUBLIC void file_popup_setup_file_load(PopupMenuBase* m, char* name, int id)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(m);
    static Bool init_flag = False;

    GetLineText((LineText *)frame->dir_text, temp_buff, BUFSIZ);
    SetLineText((LineText *)secondary_load_text, temp_buff, strlen(temp_buff));

    if(!init_flag)
    {
	XSizeHints sizehints;

	init_flag = True;
	sizehints.flags = USPosition;
	QueryPointer(disp, DefaultRootWindow(disp),
		     &sizehints.x, &sizehints.y);
	XSetNormalHints(disp, file_load_window->any.win, &sizehints);
	XMoveWindow(disp, file_load_window->any.win, sizehints.x, sizehints.y);
    }
    if(sound_conversion_method->choice.selected == 0)
	XUnmapWindow(disp,
		     SkeletonWinObj(sound_conversion_base_table)->any.win);
    XMapWindow(disp, file_load_window->any.win);

    open_file_target_frame = frame;
    set_audio_file_lock(True);
}

PUBLIC void setup_file_load_cancel(WinObj* obj, XEvent* e)
{
    XUnmapWindow(disp, file_load_window->any.win);
    set_audio_file_lock(False);
}

PUBLIC void file_load_button_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);
    char filename[BUFSIZ];

    if(audio_file_lock_flag)
	return;

    open_file_target_frame = frame;
    GetLineText((LineText *)frame->dir_text, filename, BUFSIZ);
    if(filename[0] == '\0')
    {
	write_error_message("File name is empty");
	return;
    }

    if(usffd_open(ExpandHomeDirectory(filename), &file_usffd) == False)
    {
	if(usffd_errno == USFFD_SYS_ERROR)
	    write_syserror_message(filename);
	else
	{
	    sprintf(temp_buff, "%s: %s",
		    filename, usffd_strerror(usffd_errno));
	    write_error_message(temp_buff);
	}
	return;
    }

    set_audio_file_lock(True);

    if(!usffd_set_header(&file_usffd))
    {
	if(usffd_errno != USFFD_NO_ERROR)
	    write_error_message(usffd_strerror(usffd_errno));
	usffd_encoding_recognize(&file_usffd);
    }

    if(start_load_file_proc(&file_usffd,
		file_name_no_direcotry(file_usffd.filename)) == False)
	load_file_done();
}

PUBLIC void update_file_indicator(long total, long n)
{
    double r;
    static int call_cnt = 0;

    if(indicator_map_flag == False)
	return;

    if(total == 0)
	r = 1.0;
    else
	r = (double)n / (double)total;

    SetHSimpleSliderPos(file_indicator_slider,
			(int)(r * file_indicator_slider->sslider.maxlen));

    if(call_cnt++ > 20)
    {
	XRaiseWindow(disp, file_indicator_window->any.win);
	call_cnt = 0;
    }
}

static void file_load_success(void)
{
    SSFrame* frame = open_file_target_frame;
    HScrollbar* hs = (HScrollbar *)frame->edit_slider;

    frame->zoom[SCOPE_EDIT] = file_loader.read_samples;
    SetHScrollbarThumbPos(hs, 0.0);
    SetHScrollbarShown(hs, 1.0);

    SSFrameMapFuntion(reset_hlimited_slider);

    /* ɤ߹߽λ */
    load_file_done();
    ss_garbage_collection();
}

PUBLIC void file_load_slice_proc(void)
{
    long nchannels = file_loader.nchannels;
    long block_align = file_loader.block_align;
    long nsample;
    long i, j, n;
    long start = file_loader.read_samples;
    SSFrame* frame_list;
    int pos;
    char* readbuff;
    short* readbuff2 = (short *)(temp_buff + temp_buff_size / 2);

#ifdef DEBUG
    static double load_start_time;
    if(start == 0)
	load_start_time = get_current_calender_time();
#endif /* DEBUG */

#ifdef DEBUGALL
    puts("file_load_slice_proc()");
#endif

    nsample = DATA_FRAGMENTATION_SIZE / block_align;
    n = block_align * nsample;

#ifdef USE_FILE_MAP
    if(file_loader.map_file != NULL)
    {
	if(n > file_loader.map_file_size - file_loader.map_file_offset)
	    n = file_loader.map_file_size - file_loader.map_file_offset;
	readbuff = (char *)(file_loader.map_file + file_loader.map_file_offset);

	if((file_loader.map_file_offset & 0x1) &&
	   block_align / nchannels > 1)
	{
	    memcpy(temp_buff, readbuff, n);
	    readbuff = temp_buff;
	}
	file_loader.map_file_offset += n;
    }
    else
#endif
    {
	readbuff = temp_buff;
	n = read(file_usffd.fd, readbuff, n);

	if(n < 0)
	{
	    write_syserror_message("read");
	    load_cancel_button_proc(NULL, NULL);
	    return;
	}
    }

    nsample = n / block_align;
    n = block_align * nsample;

    if(file_loader.rest_samples >= nsample)
    {
	file_loader.read_samples += nsample;
	file_loader.rest_samples -= nsample;
    }
    else
    {
	nsample = file_loader.rest_samples;
	file_loader.read_samples += nsample;
	file_loader.rest_samples = 0;
	n = block_align * nsample;
    }
       
    if(nchannels == 1)
	file_loader.converter(readbuff,
			      file_loader.data[0] + start,
			      n);
    else
    {
	file_loader.converter(readbuff, readbuff2, n);
	for(j = 0; j < nchannels; j++)
	    for(i = 0; i < nsample; i++)
		file_loader.data[j][i + start]
		    = readbuff2[i * nchannels + j];
    }

    update_file_indicator(file_loader.read_samples +
	file_loader.rest_samples, file_loader.read_samples);

    pos = sample_to_play_slider_pos(open_file_target_frame,
				    file_loader.read_samples);
    for(frame_list = all_frame_list; frame_list; frame_list = frame_list->next)
    {
	SSFrame* frame = frame_list;
	if(SSFRAME_AVAILABLE(frame) &&
	   frame->player->aubuff == open_file_target_frame->player->aubuff)
	    SetHLimitedSliderLimit(frame->play_slider, pos);
    }

    if(file_loader.rest_samples == 0 || n == 0)
    {
	if(file_loader.rest_samples > 0)
	{
	    n = file_loader.rest_samples + file_loader.read_samples;
	    start = file_loader.read_samples;
	    if(nchannels == 1)
		for(i = start; i < n; i++)
		    file_loader.data[0][i] = 0;
	    else
		for(j = 0; j < nchannels; j++)
		    for(i = start; i < n; i++)
			file_loader.data[j][i] = 0;
	}
#ifdef DEBUG
	printf("load file: %.1f Kbps\n",
	       (double)file_loader.read_samples /
	       (get_current_calender_time() - load_start_time) * 0.008);
#endif /* DEBUG */
	file_load_success();
    }
#ifdef DEBUGALL
    puts("file_load_slice_proc() - ok");
#endif
}


PUBLIC void soundscope_usffd_recognizer(USFFD* usffd)
{
    extern void sear(int fd, int start_seek, int size, int* enc, int* chan);
    int enc, chan;

    usffd->hdr_size = 0;
    usffd->data_size = usffd->file_size;
    sear(usffd->fd, 0, usffd->data_size, &enc, &chan);
    usffd->data_encoding = enc;
    usffd->nchannels     = chan;
    usffd->sample_rate   = open_file_target_frame->player->sample_rate;
}
