#include <stdio.h>
#include "soundscope.h"
#include "ssframe.h"

#define RECOG_SIZE 3
#define MINLEN 12

PUBLIC void draw_measure(Display* disp, Drawable win,
			 scope_wave_measure_t* measure)
{
    XDrawLine(disp, win, gc,
	      measure->x_left,
	      measure->y,
	      measure->x_right,
	      measure->y);

    XDrawLine(disp, win, gc,
	      measure->x_left,
	      measure->y,
	      measure->x_left + 2,
	      measure->y + 2);

    XDrawLine(disp, win, gc,
	      measure->x_left,
	      measure->y,
	      measure->x_left + 2,
	      measure->y - 2);

    XDrawLine(disp, win, gc,
	      measure->x_right,
	      measure->y,
	      measure->x_right - 2,
	      measure->y + 2);

    XDrawLine(disp, win, gc,
	      measure->x_right,
	      measure->y,
	      measure->x_right - 2,
	      measure->y - 2);
}

static Bool region_check(int x, int y, int px, int py)
{
    int dx, dy;

    dx = x - px;
    dy = y - py;
    if(dx < 0) dx = -dx;
    if(dy < 0) dy = -dy;

    return dx < RECOG_SIZE && dy < RECOG_SIZE;
}

PUBLIC Bool measure_event_proc(SSFrame* frame, XEvent* e)
{
    int x, y;
    int dx, dy;
    scope_wave_measure_t* measure = &frame->measure;

    switch(e->type)
    {
      case ButtonPress:
	measure->status = None;

	if(e->xbutton.button != Button1)
	    return False;

	x = measure->px = e->xbutton.x;
	y = measure->py = e->xbutton.y;

	if(region_check(measure->x_right, measure->y, x, y))
	{
	    measure->status = MEASURE_EXTEND_RIGHT;
	    return True;
	}

	if(region_check(measure->x_left, measure->y, x, y))
	{
	    measure->status = MEASURE_EXTEND_LEFT;
	    return True;
	}

	dy = measure->y - y;
	if(dy < 0) dy = -dy;

	if(dy < RECOG_SIZE &&
	   measure->x_left <= x &&
	   x <= measure->x_right)
	{
	    measure->status = MEASURE_MOVE;
	    return True;
	}
	break;
      case ButtonRelease:
	if(e->xbutton.button != Button1)
	    return False;
	if(measure->status != None)
	{
	    measure->status = None;
	    return True;
	}
	break;
      case MotionNotify:
	if(measure->status == None || !(e->xmotion.state & Button1Mask))
	    return False;

	CompressMotionEvent(e->xany.display, e->xany.window, e);
	x = e->xmotion.x;
	y = e->xmotion.y;
	dx = x - measure->px;
	dy = y - measure->py;
	measure->px = x;
	measure->py = y;

	switch(measure->status)
	{
	  case MEASURE_MOVE:
	    if(measure->x_left + dx < 0)
		dx = -measure->x_left;
	    else if(measure->x_right + dx >= frame->scope_width)
		dx = frame->scope_width - 1 - measure->x_right;

	    if(measure->y + dy < 0)
		dy = -measure->y;
	    else if(measure->y + dy >= frame->scope_height)
		dy = frame->scope_height - 1 - measure->y;

	    measure->x_left  += dx;
	    measure->x_right += dx;
	    measure->y       += dy;
	    break;

	  case MEASURE_EXTEND_LEFT:
	    if(measure->x_right - (measure->x_left + dx) + 1 < MINLEN)
		measure->x_left = measure->x_right - MINLEN + 1;
	    else
	    {
		measure->x_left += dx;
		if(measure->x_left < 0)
		    measure->x_left = 0;
	    }
	    break;
	  case MEASURE_EXTEND_RIGHT:
	    if((measure->x_right + dx) - measure->x_left + 1 < MINLEN)
		measure->x_right = measure->x_left + MINLEN - 1;
	    else
	    {
		measure->x_right  += dx;
		if(measure->x_right >= frame->scope_width)
		    measure->x_right = frame->scope_width - 1;
	    }
	    break;
	}
	return True;
    }
    return False;
}
