#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "soundscope.h"
#include "aubuff.h"
#include "ssframe.h"
#include "audriv.h"

char* temp_buff = NULL;
long  temp_buff_size = 0;
extern WinObj* message_winobj;

//extern char *sys_errlist[];
void write_error_message(const char* msg)
{
    if(message_winobj == NULL)
	fprintf(stderr, "%s\n", msg);
    else
    {
#ifdef DEBUG
	fprintf(stderr, "***\n");
	fprintf(stderr, "***>%s<***\n", msg);
	fprintf(stderr, "***\n");
#endif
	XBell(disp, 0);
	DisplayMessageDialog(message_winobj, msg);
    }
}

void write_syserror_message(const char* msg)
{
    if(message_winobj == NULL)
	fprintf(stderr, "%s:%s\n", msg, sys_errlist[errno]);
    else
    {
	sprintf(temp_buff, "%s:%s", msg, sys_errlist[errno]);
	write_error_message(temp_buff);
    }
}

void* xmalloc(unsigned long n)
{
    void* p;

    if((p = (void *)malloc(n)) == NULL)
    {
	if(errno == ENOMEM)
	{
	    write_error_message("꤬ԽʬǥǡݤǤޤ");
#ifdef DEBUG
	    fprintf(stderr, "## n = %ld\n", n);
#endif /* DEBUG */
	}
	else
	    write_syserror_message("malloc");

	return NULL;
    }

    return p;
}

void xfree(void* p)
{
    if(p != NULL)
	free(p);
}

char* xstrdup(const char* str)
{
    char* s;

    if((s = strdup(str)) == NULL)
    {
	if(errno == ENOMEM)
	    write_error_message("꤬ԽʬǥǡݤǤޤ");
	else
	    write_syserror_message("strdup");

	return NULL;
    }

    return s;
}

void initialize_misc_funcs(void)
{
    temp_buff_size = 1024*1024;	/* 1M */

    if((temp_buff = xmalloc(temp_buff_size)) == NULL)
	exit(1);
}

Bool winobj_call_error_handler(int code, const char* msg)
{
    write_error_message(msg);
    return False;
}

void set_error_handlers(void)
{
    audriv_error_handler = write_error_message;
    WinObjSetErrorHandler(winobj_call_error_handler);
}

void** xmalloc2(unsigned long n1, unsigned long n2)
{
    void** p;
    int i;

    p = (void **)xmalloc(n1 * sizeof(void *));
    if(p == NULL)
	return NULL;
    if(n2 == 0)
	for(i = 0; i < n1; i++)
	    p[i] = NULL;
    else
	for(i = 0; i < n1; i++)
	    if((p[i] = (void *)xmalloc(n2)) == NULL)
	    {
		while(--i >= 0)
		    xfree(p[i]);
		xfree(p);
		return NULL;
	    }
    return p;
}

void xfree2(void* p, int n1)
{
    int i;
    void** p2 = (void **)p;

    if(p == NULL)
	return;

    for(i = 0; i < n1; i++)
	xfree(p2[i]);
    xfree(p2);
}
