#include <stdio.h>
#include <stdlib.h>
#include "soundscope.h"
#include "player.h"
#include "ssframe.h"
#include "aubuff.h"
#include "audriv.h"
#include "bucket.h"
#include "timer.h"

long device_sample_rate = 8000;
long device_data_encoding = USFFD_G711_ULAW;
long device_nchannels = 1;

void audio_player_initialize(AudioPlayer* player, SSFrame* frame)
{
    int i;

    player->aubuff = get_audio_buffer(AUDIO_DEFAULT_BUFF_NAME);
    if(player->aubuff == NULL)
    {
	fprintf(stderr, AUDIO_DEFAULT_BUFF_NAME " is NULL\n");
	exit(1);
    }
    player->pos = 0;
    player->pos_offset = 0;
    player->region_start = 0;
    player->region_end = 0;
    player->looping = False;
    player->load_flag = False;
    player->stop_and_rewind_button = NULL;
    player->fast_rewind_button = NULL;
    player->fast_forward_button = NULL;
    player->play_button = NULL;
    player->stop_button = NULL;
    player->looping_button = NULL;
    player->sample_rate = 8000;
    player->custom_sample_rate = False;
    player->play_flag = False;
    player->ff_rw_type = FF_RW_UNUSED;
    for(i = 0; i < MAX_BUCKET_NCHANNELS; i++)
	player->bucket_reader[i] = NULL;

    /* frame  NULL ʤ顤Υץ쥤䡼ϥե졼⤿
     * դΤߤεǽ̤
     */
    player->frame_reference = frame;
}

void player_get_region(AudioPlayer* player, long* pos_min, long* pos_max)
{
    /* ղǽΰμ */
    if(AUDIO_PLAYER_REGIONP(player)) /* ΰ */
    {
	*pos_min = player->region_start;
	*pos_max = player->region_end;
    }
    else
    {
	*pos_min = 0;
	*pos_max = player->aubuff->nsamples;
    }
}

long calc_player_samples(AudioPlayer* player, long samples)
{
    long pos_min, pos_max, s, d;

    if(AUDIO_PLAYER_REGIONP(player))
    {
	pos_min = player->region_start;
	pos_max = player->region_end;
    }
    else
    {
	pos_min = 0;
	pos_max = player->aubuff->nsamples;
    }
    d = pos_max - pos_min;

    if(player->sample_rate != device_sample_rate ||
       player->ff_rw_type != FF_RW_UNUSED)
    {
	if(player->ff_rw_type == FF_RW_UNUSED)
	    samples = (long)((double)samples *
		(double)player->sample_rate / (double)device_sample_rate);
	else
	    samples = (long)((double)samples *
		(double)player->sample_rate / (double)device_sample_rate *
			     (double)player->ffre_step);
    }

    samples -= player->pos_offset;
    if(player->looping)
	if(d == 0)
	    s = pos_min;
	else
	    if(player->ff_rw_type == FAST_REWIND_PLAY)
	    {
		samples = pos_max - player->pos + samples;
		while(samples > d)
		{
		    samples -= d;
		    player->pos_offset += d;
		}
		s = pos_max - samples;
	    }
	    else
	    {
		samples = samples - pos_min + player->pos;
		while(samples > d)
		{
		    samples -= d;
		    player->pos_offset += d;
		}
		    
		s = pos_min + samples;
	    }
    else
	if(player->ff_rw_type == FAST_REWIND_PLAY)
	{
	    s = player->pos - samples;
	    if(s < pos_min)
		s = pos_min;
	}
	else
	{
	    s = player->pos + samples;
	    if(s > pos_max)
		s = pos_max;
	}

    return s;
}

void set_initial_player(AudioPlayer* player)
{
    if(player->play_flag)
	player_stop_and_rewind(player);

    if(player->frame_reference)
	move_play_slider_position(player->frame_reference, 0);

    SetPressdIconButton(player->play_button, False);
    SetPressdIconButton(player->stop_button, False);
    player->pos = 0;
    player->pos_offset = 0;
    player->region_start = 0;
    player->region_end = 0;
}

Bool audio_player_check_finish(AudioPlayer* player, long sample)
{
    if(!player->play_flag)
	return True;
    if(player->looping == True)
	return False;
    if(sample >= AUDIO_PLAYER_POS_MAX(player))
	return True;
    return False;
}

static void set_player_sample_rate(AudioPlayer* player, long sample_rate)
{
    int i, n;
    const long* sr;
    SSFrame* frame = player->frame_reference;

    FIXREGION(sample_rate, MIN_SAMPLE_RATE, MAX_SAMPLE_RATE);

    player->sample_rate = sample_rate;
    player->custom_sample_rate = True;
    sr = audriv_available_sample_rates(&n);
    if(frame != NULL)
	redraw_equlizer_window(frame);
    for(i = 0; i < n; i++)
    {
	if(sr[i] == sample_rate)
	{
	    player->custom_sample_rate = False;
	    if(frame != NULL)
		PopupMenuChangeMarker(frame->sample_rate_popup,
				      sample_rate);
	    return;
	}
    }
    if(frame != NULL)
	PopupMenuChangeMarker(frame->sample_rate_popup, 0);
}

Bool player_start_play(AudioPlayer* player)
{
    Bool status;

    if(player->aubuff->record_flag)
	return False;

    abc_stop_audio();
    status = abc_add_player(player);
    abc_restart_audio();

    return status;
}

void player_stop_play(AudioPlayer* player)
{
    abc_stop_audio();
    abc_remove_player(player);
    abc_restart_audio();
    if(player->frame_reference != NULL)
	repaint_frame_scope(player->frame_reference, player->pos, False);
}

void player_stop_and_rewind(AudioPlayer* player)
{
    if(is_audio_buffer_empty(player->aubuff))
	return;
    player_stop_play(player);
    player->pos = AUDIO_PLAYER_POS_MIN(player);
    if(player->frame_reference != NULL)
	repaint_frame_scope(player->frame_reference, player->pos, False);
    SetPressdIconButton(player->play_button, False);
    SetPressdIconButton(player->stop_button, False);
}

void player_change_sample_rate(AudioPlayer* player, long sample_rate)
{
#ifdef DEBUG
    printf("Play sample rate: %ld\n", sample_rate);
#endif

    if(!player->play_flag)
    {
	set_player_sample_rate(player, sample_rate);
	return;
    }

    abc_stop_audio();
    set_player_sample_rate(player, sample_rate);
    abc_restart_audio();
}


void player_change_output(long output)
{
    SSFrame* frame;
    for(frame = all_frame_list; frame; frame = frame->next)
	PopupMenuChangeMarker(frame->output_popup, output);
    audriv_set_play_output(output);
}


void player_change_encoding(long encoding)
{
    SSFrame* frame;
    const long* elist;
    int i, n;

    elist = audriv_available_encodings(&n);
    if(n == 1)
	encoding = elist[0];
    else
    {
	for(i = 0; i < n; i++)
	    if(elist[i] == encoding)
		break;
	if(i == n)
	{
	    if(usffd_encoding_sample_size[elist[n - 1]] == 1)
		encoding = elist[0];
	    else
		encoding = elist[n - 1];
	}
    }

    for(frame = all_frame_list; frame; frame = frame->next)
	PopupMenuChangeMarker(frame->encoding_popup, encoding);
    device_data_encoding = encoding;

    if(audio_active_flag)
    {
	abc_stop_audio();
	abc_restart_audio();	/* abc_restart_audio() calls
				   abc_update_audio_converter() */
    }
    else
	abc_update_audio_converter();
}

void player_change_channel(long ch)
{
    SSFrame* frame;
    const long* chlist;
    int i, n;

    chlist = audriv_available_channels(&n);

    for(i = 0; i < n; i++)
	if(chlist[i] == ch)
	    break;
    if(i == n)
	if(n == 1)
	    ch = chlist[0];
	else
	    ch = chlist[n - 1];

    for(frame = all_frame_list; frame; frame = frame->next)
	PopupMenuChangeMarker(frame->channel_popup, ch);
    device_nchannels = ch;
    if(audio_active_flag)
    {
	abc_stop_audio();
	abc_restart_audio();
    }
}

void player_start_fast_forward_play(AudioPlayer* player,
				    int step)
{
    if(audio_active_flag)
	abc_stop_audio();

    if(!player->play_flag)
    {
	if(!abc_add_player(player))
	    return;
	player->is_ff_rw_stop_flag = True;
    }
    else
	player->is_ff_rw_stop_flag = False;

    player->ffre_step = step;
    player->start_ffre_time = get_current_calender_time();
    player->ff_rw_type = FAST_FORWARD_PLAY;
    if(!abc_restart_audio())
	player->ff_rw_type = FF_RW_UNUSED;
}

void player_stop_fast_forward_play(AudioPlayer* player)
{
    if(!player->play_flag)
	return;
    abc_stop_audio();
    if(player->is_ff_rw_stop_flag == True)
    {
	if(player->frame_reference != NULL)
	    repaint_frame_scope(player->frame_reference, player->pos, False);
	abc_remove_player(player);
    }
    else
	player->ff_rw_type = FF_RW_UNUSED;
    abc_restart_audio();
}

void player_start_fast_rewind_play(AudioPlayer* player, int step)
{
    if(audio_active_flag)
	abc_stop_audio();

    if(!player->play_flag)
    {
	if(!abc_add_player(player))
	    return;
	player->is_ff_rw_stop_flag = True;
    }
    else
	player->is_ff_rw_stop_flag = False;

    player->ffre_step = step;
    player->start_ffre_time = get_current_calender_time();
    player->ff_rw_type = FAST_REWIND_PLAY;
    if(!abc_restart_audio())
	player->ff_rw_type = FF_RW_UNUSED;
}

void player_stop_fast_rewind_play(AudioPlayer* player)
{
    if(!player->play_flag)
	return;
    abc_stop_audio();
    if(player->is_ff_rw_stop_flag == True)
    {
	if(player->frame_reference != NULL)
	    repaint_frame_scope(player->frame_reference, player->pos, False);
	abc_remove_player(player);
    }
    else
	player->ff_rw_type = FF_RW_UNUSED;
    abc_restart_audio();
}
