#include <stdio.h>
#include <string.h>
#include "soundscope.h"
#include "ssframe.h"
#include "bucket.h"
#include "usffd.h"
#include "aubuff.h"
#include "audriv.h"
#include "soundex.h"
#include "player.h"

enum record_status
{
    RECORD_NONE,
    RECORD_INPUT,
    RECORD_CONVERT
} record_status = RECORD_NONE;

#define RECORD_WINDOW_WIDTH 500

AudioBuffer* record_backup_buffer;
static WinObj* record_window;
static WinObj* record_close_btn;
static AudioBucket* record_bucket_head[MAX_BUCKET_NCHANNELS];
static AudioBucket* record_bucket_tail[MAX_BUCKET_NCHANNELS];
static AudioBuffer* record_aubuff_temp = NULL;
static AudioBuffer* record_new_buffer = NULL;
static void (* record_converter)(void *, void *, int);

/* recording format */
static int record_nchannels = 1;
static int record_encoding = USFFD_G711_ULAW;
static int record_block_align = 1;
static int record_sample_rate = 8000;
static long record_total_samples;
static long record_conv_samples;

static void record_start_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = open_file_target_frame;
    int i;

#ifdef DEBUG
    puts("record_start_proc()");
#endif

    open_file_target_frame = frame;

    if(!audriv_set_record_encoding(record_encoding))
	return;
    if(!audriv_set_record_sample_rate(record_sample_rate))
	return;
    if(!audriv_set_record_channels(record_nchannels))
	return;
    record_block_align = record_nchannels
	* usffd_encoding_sample_size[record_encoding];
    if(audriv_record_open() == False)
	return;

    if(record_aubuff_temp == NULL)
    {
	record_aubuff_temp = make_new_audio_buffer(" *Recording Temporary*");
	if(record_aubuff_temp == NULL)
	    return;
	if(!audio_buffer_make_buffer(record_aubuff_temp, 0,
				     MAX_BUCKET_NCHANNELS))
	{
	    kill_audio_buffer(record_aubuff_temp);
	    record_aubuff_temp = NULL;
	    return;
	}
    }

    XLowerWindow(disp, SkeletonWinObj(obj)->any.win);
    XUnmapWindow(disp, SkeletonWinObj(record_close_btn)->any.win);
    XSync(disp, False);

    player_stop_and_rewind(frame->player);
    player_change_sample_rate(frame->player, record_sample_rate);
    record_backup_buffer = frame->player->aubuff;
    record_aubuff_temp->nchannels = record_nchannels;
    record_aubuff_temp->sample_rate = record_sample_rate;
    record_aubuff_temp->file_encoding = record_encoding;
    record_aubuff_temp->nsamples = 0;
    record_aubuff_temp->hdr_type = USFFD_HDR_AU;
    record_aubuff_temp->record_flag = True;
    frame->player->aubuff = record_aubuff_temp;
    record_converter = select_converter(record_encoding, internal_encoding,
					NULL, NULL);
    record_total_samples = 0;
    record_status = RECORD_INPUT;

    for(i = 0; i < record_nchannels; i++)
	BUCKET_LIST_INIT(record_bucket_head[i], record_bucket_tail[i]);

    audio_record_flag = True;
    update_ss_active_flag();
#ifdef DEBUG
    puts("Recording start!!!");
#endif
}

static void record_done(void)
{
    int i;

    for(i = 0; i < record_nchannels; i++)
    {
	reuse_audio_bucket_list(record_bucket_head[i]);
	record_bucket_head[i] = NULL;
    }
    close_file_indicator();
    player_stop_and_rewind(open_file_target_frame->player);
    audio_record_flag = False;
    update_ss_active_flag();
}

static void record_success(void)
{
    record_new_buffer->record_flag = False;
    record_done();
    ss_garbage_collection();
}

static void record_stop_proc(WinObj* obj, XEvent* e)
{
    long datasize, bucket_len;

    audriv_record_close();
    XLowerWindow(disp, SkeletonWinObj(obj)->any.win);
    XMapWindow(disp, SkeletonWinObj(record_close_btn)->any.win);
    XSync(disp, False);

    record_new_buffer = make_new_audio_buffer("*Record Buffer*");
    if(record_new_buffer == NULL)
    {
	open_file_target_frame->player->aubuff = record_backup_buffer;
	record_done();
    }
    if(!audio_buffer_make_buffer(record_new_buffer,
				 record_total_samples,
				 record_nchannels))
    {
	kill_audio_buffer(record_new_buffer);
	open_file_target_frame->player->aubuff = record_backup_buffer;
	record_done();
    }

    open_file_target_frame->player->aubuff = record_new_buffer;
    record_new_buffer->record_flag = True;

    record_status = RECORD_CONVERT;
    record_conv_samples = 0;
    datasize = record_total_samples * record_block_align;

    bucket_len = audio_bucket_length(record_bucket_head[0]);
    if(bucket_len > 5)
    {
	if(datasize >= 1000000)
	    sprintf(temp_buff, "Converting (%.1f Mbytes)",
		    (double)datasize / 1000000.0);
	else
	    sprintf(temp_buff, "Converting (%.1f kbytes)",
		    (double)datasize / 1000.0);
	open_file_indicator(temp_buff, NULL);
    }
}

static void close_record_window_proc(WinObj* obj, XEvent* e)
{
    if(audio_record_flag)
	return;
    XUnmapWindow(disp, record_window->any.win);
    XSync(disp, False);
    set_audio_file_lock(False);
    open_file_target_frame = NULL;
}

PUBLIC void initialize_record_window(void)
{
    WinObj* top;
    WinObj* record_start_btn;
    WinObj* record_stop_btn;
    int h, height, width;
    int start_stop_btn_width;

    h = TextHeight(font);
    height = h + GAP;
    start_stop_btn_width = MAX(TextWidth(font, "Ͽλ"),
			       TextWidth(font, "Ͽ")) + GAP;
    width = RECORD_WINDOW_WIDTH;
    top = record_window =
	toplevel_WinObj(disp, "Recorder", "Recorder",
			4 * GAP + 4 * BD_WIDTH
			+ TextWidth(font, "Ĥ") + start_stop_btn_width,
			height + 2 * GAP + 2 * BD_WIDTH,
			WinObj_fg, WinObj_bg);
    top->any.when_realize_mapping = False;

    record_close_btn = create_Button(top, GAP, GAP,
				     TextWidth(font, "Ĥ") + GAP, height,
				     BD_WIDTH,
				     font, "Ĥ");
    SetCallBackProc(record_close_btn, (call_back_t)close_record_window_proc);

    record_stop_btn = create_Button(top,
			       WinObjRight(record_close_btn) + GAP,
			       GAP,
			       start_stop_btn_width, height,
			       BD_WIDTH,
			       font, "Ͽλ");
    record_start_btn = create_Button(top,
			       WinObjRight(record_close_btn) + GAP,
			       GAP,
			       start_stop_btn_width, height,
			       BD_WIDTH,
			       font, "Ͽ");
    SetCallBackProc(record_start_btn, (call_back_t)record_start_proc);
    SetCallBackProc(record_stop_btn, (call_back_t)record_stop_proc);

    WinObjRealize(top);
}

PUBLIC void file_popup_record(PopupMenuBase* m, char* name, int id)
{
    static Bool init_flag = False;

    open_file_target_frame = WINOBJ_EVENT_FRAME(m);

    if(!init_flag)
    {
	XSizeHints sizehints;

	init_flag = True;
	sizehints.flags = USPosition;
	QueryPointer(disp, DefaultRootWindow(disp),
		     &sizehints.x, &sizehints.y);
	XSetNormalHints(disp, record_window->any.win, &sizehints);
	XMoveWindow(disp, record_window->any.win, sizehints.x, sizehints.y);
    }
    XMapWindow(disp, record_window->any.win);
    set_audio_file_lock(True);
}

static void record_proc_input(void)
{
    int i, j, n;
    AudioBucket* bucket;
    short* buff16 = (short *)(temp_buff + temp_buff_size / 2);
    long nsamples;

#ifdef DEBUG
    puts("record_proc_input()");
#endif /* DEBUG */

    n = AUDIO_BUCKET_LEN * sizeof(short);
    if(usffd_encoding_sample_size[record_encoding] == 1)
	n /= 2;
    n -= (n % record_block_align);
    nsamples = n / record_block_align;
    record_total_samples += nsamples;

    for(i = 0; i < record_nchannels; i++)
    {
	bucket = new_allocated_bucket();
	if(bucket == NULL)
	{
	    record_done();
	    return;
	}
	bucket->len = nsamples;
	record_aubuff_temp->buffer[i] = bucket->data;
	ADD_BUCKET(record_bucket_head[i], record_bucket_tail[i],
		   bucket);
    }

    record_aubuff_temp->nsamples = nsamples;
    audriv_read(temp_buff, n);
    if(record_nchannels == 1)
	record_converter(temp_buff,
			 record_aubuff_temp->buffer[0],
			 n);
    else
    {
	record_converter(temp_buff, buff16, n);
	for(j = 0; j < record_nchannels; j++)
	    for(i = 0; i < nsamples; i++)
		record_aubuff_temp->buffer[j][i]
		    = buff16[i * record_nchannels + j];
    }
    repaint_frame_scope(open_file_target_frame, nsamples / 2, False);

#ifdef DEBUG
    {
	long s = record_total_samples * record_block_align;
	if(s < 1000)
	    printf("Record %ld bytes\n", s);
	else if(s < 1000000)
	    printf("Record %ld Kbytes\n", s / 1000);
	else
	    printf("Record %.2f Mbytes\n", s / 1000000.0);
    }
#endif
}

static void record_proc_convert(void)
{
    long len, ch;
    AudioBucket* bucket;

    len = record_bucket_head[0]->len;
    for(ch = 0; ch < record_nchannels; ch++)
    {
	memcpy(record_new_buffer->buffer[ch] + record_conv_samples,
	       record_bucket_head[ch]->data,
	       len * sizeof(short));
	bucket = record_bucket_head[ch];
	record_bucket_head[ch] = record_bucket_head[ch]->next;
	reuse_audio_bucket(bucket);
    }

    record_conv_samples += len;
    update_file_indicator(record_total_samples, record_conv_samples);
    if(record_conv_samples == record_total_samples)
	record_success();
}

PUBLIC void record_slice_proc(void)
{
    switch(record_status)
    {
      case RECORD_NONE:
	break;
      case RECORD_INPUT:
	record_proc_input();
	break;
      case RECORD_CONVERT:
	record_proc_convert();
	break;
    }
}
