#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "soundscope.h"
#include "ssframe.h"
#include "audriv.h"
#include "aubuff.h"
#include "player.h"
#include "timer.h"
#include "fft.h"

#define FITSIZ 40
#define MAX_BACK_SEARCH 256
#define DIFF_STEP 2

XPoint wave_points[MAX_NPOINTS]; /* ȷ */
double fft_data[FFTSIZE];         /* ڥȥ */
int repaint_y_min, repaint_y_max;

void initialize_log_hz_table(double zoom, double* table)
{
    int i;
    double x1, x2;
    double tx1, tx2;

    x1 = 1.0;
    x2 = x1 + LOG_HZ_MAX_X1 - zoom + LOG_HZ_MIN_X1;
    tx1 = log(x1);
    tx2 = log(x2);

    for(i = 0; i < FFTSIZE/2; i++)
    {
	double tx;

	tx = (double)i / (FFTSIZE/2.0);
	/* 0 <= tx < 1 */

	tx = x1 + tx * (x2 - x1);
	/* x1 <= tx < x2 */

	tx = log(tx);
	/* tx1 <= tx < tx2 */

	tx = (tx - tx1) / (tx2 - tx1);
	/* 0 <= tx < 1 */

	table[i] = tx;
    }
}

void initialize_exp_hz_table(double zoom, double* table)
{
    int j;
    double x1, x2;
    double tx1, tx2, tx;

    x1 = 1.0;
    x2 = x1 + LOG_HZ_MAX_X1 - zoom + LOG_HZ_MIN_X1;
    tx1 = log(x1);
    tx2 = log(x2);


    /*
      tx = (log(x1 + i / (FFTSIZE/2.0) * (x2 - x1)) - tx1) / (tx2 - tx1)
      tx * (tx2 - tx1) = log(x1 + i / (FFTSIZE/2.0) * (x2 - x1)) - tx1
      tx1 + tx * (tx2 - tx1) = log(x1 + i / (FFTSIZE/2.0) * (x2 - x1))
      exp(tx1 + tx * (tx2 - tx1)) = x1 + i / (FFTSIZE/2.0) * (x2 - x1)
      exp(tx1 + tx * (tx2 - tx1)) - x1 = i / (FFTSIZE/2.0) * (x2 - x1)
      (exp(tx1 + tx * (tx2 - tx1)) - x1) / (x2 - x1) * (FFTSIZE/2.0) = i
      */

    for(j = 0; j < (FFTSIZE/2.0); j++)
    {
	tx = j / (double)(FFTSIZE/2.0);
	table[j] =
	    (exp(tx1 + tx * (tx2 - tx1)) - x1) / (x2 - x1) * (FFTSIZE/2.0);
    }
}


static unsigned long calc_wave_diff(const short* prev, const short* x,
				    unsigned long dmin)
{
    unsigned long diff;
    long t1, t2;
    const short *p3;

    p3 = prev + FITSIZ;
    diff = 0;
    while(diff < dmin && prev < p3)
    {
	t1 = (long)*prev - (long)*x;
	prev += DIFF_STEP;
	x += DIFF_STEP;

	t2 = (long)*prev - (long)*x;
	prev += DIFF_STEP;
	x += DIFF_STEP;

	diff += (unsigned long)(t1 * t1 + t2 * t2);
    }

    return diff;
}

static long synchronize_display_position(SSFrame* frame, int ch, long sample)
{
    unsigned long dmin;
    long prev_pos = MAX(0, frame->prev_pos[ch] - FITSIZ / 2);
    short* di;
    short* bottomp;
    short* prev;
    short* start;
    short* start_orig;

    if(sample <= prev_pos ||
       frame->player->aubuff->nsamples <= sample + FITSIZ)
	return sample;
    prev  = frame->player->aubuff->buffer[ch] + prev_pos;
    start = frame->player->aubuff->buffer[ch] + sample;

    if(sample - prev_pos < MAX_BACK_SEARCH)
	bottomp = prev;
    else
	bottomp = start - MAX_BACK_SEARCH;
    start_orig = start;

    di = start;
    dmin = 0xffffffff;
    for(; start >= bottomp; start--)
    {
	unsigned long diff;

	diff = calc_wave_diff(prev, start, dmin);
	if(dmin > diff)
	{
	    dmin = diff;
	    di = start;
	}
    }
    return sample - (long)(start_orig - di);
}

#define SCOPE_MSG_H 32
/* ꤵ줿ե졼Υפ sample ֤补
 * sync  True ʤ sample Ŭ˰ư롥
 */
PUBLIC void repaint_frame_scope(SSFrame* frame, long sample, Bool sync)
{
    int mode = SSFRAME_SCOPE_MAIN_MODE(frame);
    static long call_cnt = 0;
    Bool specgpaint =
	((mode == SCOPE_SPECTOROGRAM) &&
	 (frame->local_mode_selector[SCOPE_SPECGTYPE]->choice.selected == 0));
    static int msg_w = 1;
    int prev_repaint_y_min, prev_repaint_y_max;

    if(frame->player->aubuff->nsamples == 0)
	return;
    if(sample < 0)
	sample = 0;
    else if(sample > frame->player->aubuff->nsamples)
	sample = frame->player->aubuff->nsamples;

    prev_repaint_y_min = frame->repaint_y_min;
    prev_repaint_y_max = frame->repaint_y_max;

    if(!ss_active_flag)
    {
	repaint_y_min = 0;
	repaint_y_max = frame->scope_height - 1;
    }
    else
    {
	if(mode == SCOPE_WAVE)
	    repaint_y_min = repaint_y_max = frame->scope_height / 2;
	else if(mode == SCOPE_SPECTORAM)
	    repaint_y_min = repaint_y_max = frame->scope_height - 1;
	else
	{
	    repaint_y_min = 0;
	    repaint_y_max = frame->scope_height - 1;
	}
    }

    if(mode == SCOPE_EDIT)
	scope_edit_move_cur(frame, sample);
    else
    {
	int ch;

	if(!specgpaint)
	{
	    if(!ss_active_flag)
		XCopyArea(disp, frame->scope_background, frame->scope_pm, gc,
			  0, 0, frame->scope_width, frame->scope_height, 0, 0);
	    else
	    {
		XCopyArea(disp, frame->scope_background, frame->scope_pm, gc,
			  9, 0, msg_w, SCOPE_MSG_H, 9, 0);
		XCopyArea(disp, frame->scope_background, frame->scope_pm, gc,
			  0, prev_repaint_y_min,
			  frame->scope_width,
			  prev_repaint_y_max - prev_repaint_y_min + 1,
			  0, prev_repaint_y_min);
	    }
	}

	if(specgpaint)
	{
#ifdef FAST_SOUNDSPECTROGRAM
	    if(call_cnt % FAST_SOUNDSPECTROGRAM == 0)
	    {
		XCopyArea(disp, frame->scope_pm, frame->scope_pm, gc,
			  FAST_SOUNDSPECTROGRAM, 0,
			  frame->scope_width - FAST_SOUNDSPECTROGRAM,
			  frame->scope_height,
			  0, 0);
		XSetForeground(disp, gc, scope_bg);
		XFillRectangle(disp, frame->scope_pm, gc,
			       frame->scope_width - FAST_SOUNDSPECTROGRAM, 0,
			       FAST_SOUNDSPECTROGRAM, frame->scope_height);
	    }
#else
	    XCopyArea(disp, frame->scope_pm, frame->scope_pm, gc,
		      1, 0,
		      frame->scope_width - 1,
		      frame->scope_height,
		      0, 0);
#endif /* FAST_SOUNDSPECTROGRAM */
	}

	for(ch = 0; ch < frame->player->aubuff->nchannels; ch++)
	{
	    long scope_pos;

	    if(sync)
		scope_pos = FITSIZ / 2 +
		    synchronize_display_position(frame,
						 ch, sample - FITSIZ / 2);
	    else
		scope_pos = sample;

	    switch(mode)
	    {
	      case SCOPE_WAVE:
		draw_scope_wave(frame, ch, scope_pos);
		break;
	      case SCOPE_SPECTORAM:
		draw_scope_spectram(frame, ch, scope_pos);
		break;
	      case SCOPE_SPECTOROGRAM:
		if(specgpaint)
		    draw_scope_color_spectrogram(frame, ch, scope_pos, call_cnt);
		else
		    draw_scope_spectrogram(frame, ch, scope_pos);
		break;
	    }
		
	    frame->prev_pos[ch] = scope_pos;
	}

	if(mode == SCOPE_SPECTOROGRAM && !specgpaint)
	    XCopyArea(disp, frame->scope_pm, frame->scope_background, gc,
		      0, 0, frame->scope_width, frame->scope_height,
		      -5, 5);

#ifdef FAST_SOUNDSPECTROGRAM
	if(!specgpaint || call_cnt % FAST_SOUNDSPECTROGRAM == 0)
#endif /* FAST_SOUNDSPECTROGRAM */
	{
	    int ymin, ymax;

	    ymin = MIN(repaint_y_min, prev_repaint_y_min);
	    ymax = MAX(repaint_y_max, prev_repaint_y_max);
	    XCopyArea(disp, frame->scope_pm, frame->scope_win, gc,
		      0, ymin,
		      frame->scope_width, ymax - ymin + 1,
		      0, ymin);
	    frame->repaint_y_min = repaint_y_min;
	    frame->repaint_y_max = repaint_y_max;
	}
    }

    if(mode != SCOPE_EDIT
#ifdef FAST_SOUNDSPECTROGRAM
       && (!specgpaint || call_cnt % FAST_SOUNDSPECTROGRAM == 0)
#endif /* FAST_SOUNDSPECTROGRAM */
	)
    {
	double t, dt;
	static double calen_time = 0.0;
	static int ncall = 0;

	t = get_current_calender_time();
#ifdef FAST_SOUNDSPECTROGRAM
	if(specgpaint)
	    dt = (t - calen_time) / FAST_SOUNDSPECTROGRAM;
	else
#endif /* FAST_SOUNDSPECTROGRAM */	    
	    dt = t - calen_time;

	if(dt > 0)
	{
	    calen_time = t;

	    if(dt <= 10.0)
		sprintf(temp_buff, "%.2f frame/sec. (No. %d)", 1.0 / dt,
			++ncall);
	    else
		sprintf(temp_buff, "%.1f frame/sec. (No. %d)", 1.0 / dt,
			++ncall);

	    if(specgpaint)
	    {
		XSetForeground(disp, gc, 1);
		XDrawString(disp, frame->scope_win, gc,
			    11, 11, temp_buff, strlen(temp_buff));
		XSetForeground(disp, gc, 0);
		XDrawString(disp, frame->scope_win, gc,
			    10, 10, temp_buff, strlen(temp_buff));
	    }
	    else
	    {
		if(mode == SCOPE_WAVE || mode == SCOPE_SPECTORAM)
		{
		    int w = (strlen(temp_buff) + 2) * 6;
		    msg_w = MAX(w, msg_w);
		    XCopyArea(disp, frame->scope_pm, frame->scope_win, gc,
			      9, 0, msg_w, SCOPE_MSG_H, 9, 0);
		}
		XDrawString(disp, frame->scope_win, gc,
			    10, 10, temp_buff, strlen(temp_buff));
	    }
	}
    }

    move_play_slider_position(frame, sample);
    XSync(disp, False);
    call_cnt++;
    if(call_cnt < 0)
	call_cnt = 0;
}

/* ɽη׻(ԥñ) */
PUBLIC void get_scope_draw_region(SSFrame* frame, int ch,
				  int* x, int* y, int* width, int* height)
{
    int nch = frame->player->aubuff->nchannels;

    if(x != NULL)	*x = 0;
    if(width != NULL)	*width = frame->scope_width;

    if(nch == 1 || ch == -1)
    {
	if(y != NULL)		*y = 0;
	if(height != NULL)	*height = frame->scope_height - 1;
    }
    else
    {
	int h;
	int r;

	h = frame->scope_height / nch;
	r = frame->scope_height % nch;
	if(y != NULL)		*y = h * ch + MIN(r, ch);
	if(height != NULL)	*height = h + (r > ch) - 1;
    }
}

PUBLIC void scope_event_handler(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);
    int mode = SSFRAME_SCOPE_MAIN_MODE(frame);

#ifdef DEBUG
    printf("scope_event_handler(): %s\n", EventTypeString(e->type));
#endif

    if(mode == SCOPE_WAVE || mode == SCOPE_EDIT)
    {
	if(measure_event_proc(frame, e))
	{
	    init_scope_pixmap(frame);
	    if(ss_active_flag)
		soundscope_slice_proc();
	    return;
	}
    }

    if(e->type == Expose)
    {
	if(e->xexpose.count > 0)
	    return;
#ifdef DEBUG
	printf("copy frame->scope_pm to frame->scope_win\n");
#endif
	XCopyArea(disp, frame->scope_pm, frame->scope_win, gc, 0, 0,
		  frame->scope_width, frame->scope_height, 0, 0);
    }
    else if(mode == SCOPE_EDIT)
	scope_edit_event_proc(frame, e);
#ifdef DEBUG
    puts("End of scope_event_handler()");
#endif
}

PUBLIC void get_current_fft_data(SSFrame* frame, int ch, long sample)
{
    long i, nsamples;
    short* data;
    static double* w_table = NULL;

    if(frame == NULL)
    {
	double t;

	if((w_table = (double *)malloc(FFTSIZE * sizeof(double))) == NULL)
	{
	    perror("malloc");
	    exit(1);
	}
	t = -M_PI;
	for(i = 0; i < FFTSIZE; i++)
	{
	    w_table[i] = 0.50 + 0.50 * cos(t);
	    t += 2.0 * M_PI / FFTSIZE;
	}
	return;
    }

    nsamples = frame->player->aubuff->nsamples;
    sample -= FFTSIZE / 2;

    if(frame->equlizing_flag)
    {
	data = (short *)temp_buff;
	convolution_proc(frame->player->aubuff->buffer[ch],
			 nsamples, sample, data, FFTSIZE,
			 frame->convolution);
	for(i = 0; i < FFTSIZE; i++)
	    fft_data[i] = (double)data[i] * (1.0/32768.0);
    }
    else
    {
	data = frame->player->aubuff->buffer[ch] + sample;

	if(0 <= sample && sample + FFTSIZE < nsamples)
	    for(i = 0; i < FFTSIZE; i++)
		fft_data[i] = (double)data[i] * (1.0/32768.0);
	else
	{
	    for(i = 0; sample + i < 0; i++)
		fft_data[i] = 0.0;
	    for(; i < FFTSIZE && sample + i < nsamples; i++)
		fft_data[i] = (double)data[i] * (1.0/32768.0);
	    for(; i < FFTSIZE; i++)
		fft_data[i] = 0.0;
	}
    }

    for(i = 0; i < FFTSIZE; i++)
	fft_data[i] *= w_table[i];
    realfft(fft_data, FFTSIZE);
}
