#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include "soundscope.h"
#include <X11/Xresource.h>

#ifdef DEBUG
#include "timer.h"
#endif

static Bool filename_exist(const char* f)
{
    int fd;

    if((fd = open(f, O_RDONLY)) < 0)
	return False;
    close(fd);
    return True;
}

static Bool yorn(const char* s)
{
    static const char * const yes_rep[] =
    {"yes", "Yes", "true", "True", "ok", "OK", "on", "ON", "1"};
    int i;

    for(i = 0; i < sizeof(yes_rep) / sizeof(yes_rep[0]); i++)
        if(!strcmp(s, yes_rep[i]))
            return True;
    return False;
}

void get_resource(void)
{
    XrmDatabase db;
    char* home;
    char* str;
    XrmValue vl;
#ifdef DEBUG
    double t = get_current_calender_time();
#endif

    if((home = getenv("HOME")) == NULL)
    {
	fprintf(stderr, "Warning: Undefined environment `HOME'\n");
	return;
    }

    XrmInitialize();
    sprintf(temp_buff, "%s/.Xdefaults", home);
    if(filename_exist(temp_buff))
	db = XrmGetFileDatabase(temp_buff);
    else
    {
	sprintf(temp_buff, "%s/.Xresources", home);
	if(filename_exist(temp_buff))
	    db = XrmGetFileDatabase(temp_buff);
	else
	    return;
    }

    if(XrmGetResource(db, "soundscope*background", "soundscope*Background",
		      &str, &vl))
        WinObj_bg_name = xstrdup((char *)vl.addr);

    if(XrmGetResource(db, "soundscope*foreground", "soundscope*Foreground",
		      &str, &vl))
        WinObj_fg_name = xstrdup((char *)vl.addr);

    if(XrmGetResource(db, "soundscope*font", "soundscope*Font",
		      &str, &vl))
        WinObj_font8 = xstrdup((char *)vl.addr);

    if(XrmGetResource(db, "soundscope*font_jp", "soundscope*FontJP",
		      &str, &vl))
        WinObj_font16 = xstrdup((char *)vl.addr);

    if(XrmGetResource(db, "soundscope*defaultText", "soundscope*DefaultText",
		      &str, &vl))
        default_text_string = xstrdup((char *)vl.addr);

    if(XrmGetResource(db, "soundscope*reverseVideo", "soundscope*ReverseVideo",
		      &str, &vl))
	WinObj_reverse_color_flag = yorn((char *)vl.addr);

    if(XrmGetResource(db, "soundscope*monoColor", "soundscope*MonoColor",
		      &str, &vl))
    {
	if(!strcmp((char *)vl.addr, "default"))
	    display_visual_type = DISPLAY_VISUAL_DEFAULT;
	else if(yorn((char *)vl.addr))
	    display_visual_type = DISPLAY_VISUAL_MONO;
    }

    if(XrmGetResource(db, "soundscope*display", "soundscope*Display",
		      &str, &vl))
	disp_name = xstrdup((char *)vl.addr);

    if(XrmGetResource(db, "soundscope*shortCutList", "soundscope*ShortCutList",
		      &str, &vl))
	default_short_cut_list = xstrdup((char *)vl.addr);

    XrmDestroyDatabase(db);

#ifdef DEBUG
    printf("Load X-Resource: %g msec.\n",
	   (get_current_calender_time() - t) * 1000);
#endif
}
