#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include "soundscope.h"
#include "ssframe.h"
#include "aubuff.h"
#include "player.h"
#include "usffd.h"
#include "timer.h"
#include "soundex.h"
extern int errno;

struct _file_saver file_saver =
{
    False,			/* save_flag */
    NULL,			/* filename */
    -1,				/* fd */
    NULL,			/* data */
    1,				/* nchannels */
    USFFD_UNDEFINED,		/* file_encoding */
    1,				/* block_align */
    0,				/* write_samples */
    0,				/* rest_samples */
    NULL,			/* save_buffer */
    NULL,			/* converter */
};


#define FILE_SAVE_WINDOW_WIDTH 500
static WinObj* file_save_window;
static WinObj* save_text;
static WinObj* save_format;
static char filename[BUFSIZ];
static WinObj* save_yn;
static WinObj* save_encoding_select_menu;

static void save_file_done(void)
{
    close_file_indicator();
    if(number_of_ssframes() > 1)
	PopupMenuResetNoselect(open_file_target_frame->sound_win_popup,
			       WIN_POPUP_WINCLOSE);

    file_saver.save_flag = False;
    if(file_saver.fd >= 0)
    {
	close(file_saver.fd);
	file_saver.fd = -1;
    }
    file_saver.save_buffer = NULL;
    set_audio_file_lock(False);
    update_ss_active_flag();
}

static enum usffd_audio_header_t save_selected_format(void)
{
    return (enum usffd_audio_header_t)save_format->choice.selected;
}

static long save_selected_encoding(void)
{
    return save_encoding_select_menu->select_menu.selected_id;
}

static void save_cancel_button_proc(WinObj* obj, XEvent* e)
{
#ifdef DEBUG
    puts("save_cancel_button_proc()");
#endif

    save_file_done();
}

static Bool start_save_file_proc(void)
{
    SSFrame* frame = open_file_target_frame;
    AudioBuffer* b = frame->player->aubuff;
    USFFD usffd;
    int fd;
    long save_encoding;

#ifdef DEBUG
    puts("start_save_file_proc()");
#endif
    PopupMenuNoselect(open_file_target_frame->sound_win_popup,
		      WIN_POPUP_WINCLOSE);
    XUnmapWindow(disp, file_save_window->any.win);

    save_encoding = save_selected_encoding();

    usffd_make_header(&usffd,
		      save_encoding, b->nchannels, b->sample_rate);
    usffd.filename = filename;
    usffd.hdr_type = save_selected_format();
    usffd.data_size = b->nsamples * AENC_SAMPW(save_encoding) * b->nchannels;

    if((fd = open(ExpandHomeDirectory(filename),
		  O_WRONLY | O_CREAT | O_TRUNC, 0644)) < 0)
    {
	write_syserror_message(filename);
	return False;
    }

    file_saver.save_flag = True;
    file_saver.filename = filename;
    file_saver.fd = fd;
    file_saver.data = b->buffer;
    file_saver.nchannels = b->nchannels;
    file_saver.write_samples = 0;
    file_saver.rest_samples = b->nsamples;
    file_saver.save_buffer = b;

    usffd_write_header(fd, &usffd, &file_saver.file_encoding);
    file_saver.converter = select_converter(internal_encoding,
					    file_saver.file_encoding,
					    NULL, NULL);
    file_saver.block_align = b->nchannels *
	usffd_encoding_sample_size[file_saver.file_encoding];
    usffd.data_size = b->nsamples * file_saver.block_align;

    if(DATA_FRAGMENTATION_SIZE / file_saver.block_align * 5
       < file_saver.rest_samples)
    {
	if(usffd.data_size >= 1000000)
	    sprintf(temp_buff, "Saving %s (%.1f Mbytes)",
		    b->name, (double)usffd.data_size / 1000000.0);
	else
	    sprintf(temp_buff, "Saving %s (%.1f kbytes)",
		    b->name, (double)usffd.data_size / 1000.0);
	open_file_indicator(temp_buff, save_cancel_button_proc);
    }

    update_ss_active_flag();
    return True;
}

static void save_yes(QueryYesNo* qyn)
{
    start_save_file_proc();
}
static void save_no(QueryYesNo* qyn)
{
    /* Do nothing */
}

static void open_save_overrite_yn(void)
{
    sprintf(temp_buff, "%s is exist. Overrite it ?", filename);
    QueryYesNoProc(save_yn, temp_buff);
}

static void setup_file_save_ok(WinObj* obj, XEvent* e)
{
    struct stat sbuf;

    GetLineText((LineText *)save_text, filename, BUFSIZ);

    if(filename[0] == '\0')
    {
	write_error_message("File name is empty");
	return;
    }

    if(stat(ExpandHomeDirectory(filename), &sbuf) < 0)
    {
	if(errno == ENOENT)
	    start_save_file_proc();
	else
	    write_syserror_message(filename);
    }
    else if(S_ISDIR(sbuf.st_mode))
    {
	filename[BUFSIZ - 32] = '\0';
	sprintf(temp_buff, "%s: Is a directory", filename);
	write_error_message(temp_buff);
    }
    else
    {
	int fd;
	if((fd = open(ExpandHomeDirectory(filename), O_WRONLY)) >= 0)
	{
	    close(fd);
	    open_save_overrite_yn();
	}
	else
	    write_syserror_message(filename);
    }
}

PUBLIC void setup_file_save_cancel(WinObj* obj, XEvent* e)
{
    XUnmapWindow(disp, file_save_window->any.win);
    XUnmapWindow(disp, save_yn->any.win);
    set_audio_file_lock(False);
}

static void rename_file_extension(char* filename, const char* ext)
{
    char* p;
    if((p = strrchr(filename, '/')) == NULL)
	p = filename;
    if((p = strrchr(p, '.')) != NULL)
	memcpy(p, ext, strlen(ext) + 1);
    else
	strcat(filename, ext);
}

static void make_file_name_extension(void)
{
    AudioBuffer* b;
    USFFD usffd;

#ifdef DEBUG
    puts("make_file_name_extension()");
#endif /* DEBUG */

    b = open_file_target_frame->player->aubuff;
    GetLineText((LineText *)save_text, filename, BUFSIZ - 100);
    usffd_make_header(&usffd,
		      b->file_encoding, b->nchannels, b->sample_rate);
    usffd.hdr_type = save_selected_format();
    rename_file_extension(filename, usffd_hdr_extension(&usffd));
    SetLineText((LineText *)save_text, filename, strlen(filename));
}

static long select_save_encoding(enum usffd_audio_header_t type, long encoding)
{
    long i;
    WinObj* popup;

    encoding = usffd_select_encoding(type, encoding);
    SetSelectMenuItem(save_encoding_select_menu, encoding);
    popup = save_encoding_select_menu->select_menu.select_popup;
    for(i = 1; i < USFFD_NENCODINGS; i++)
    {
	if(usffd_encoding_available(type, i))
	    PopupMenuResetNoselect(popup, i);
	else
	    PopupMenuNoselect(popup, i);
    }
    return encoding;
}

static void save_format_proc(ChoiceButton* c, XEvent* e)
{
    make_file_name_extension();
    select_save_encoding((enum usffd_audio_header_t)c->selected,
			 save_selected_encoding());
}

void change_save_encoding(PopupMenuBase* m, char* name, int id)
{
    USFFD usffd;

    if(save_selected_format() != 0)
	return;

    usffd.hdr_type = USFFD_HDR_NONE;
    usffd.data_encoding = id;
    usffd.nchannels = open_file_target_frame->player->aubuff->nchannels;
    usffd_hdr_extension(&usffd);
    rename_file_extension(filename, usffd_hdr_extension(&usffd));
    SetLineText((LineText *)save_text, filename, strlen(filename));
}

PUBLIC void initialize_setup_file_save_window(void)
{
    WinObj* top;
#define SAVE_FORMAT_KINDS 3
    char* format_msg[SAVE_FORMAT_KINDS] = {"إå̵", "au", "wav"};
    WinObj* cancel_button;
    WinObj* ok_button;
    int cancel_button_width;
    int h, height, width;

    cancel_button_width = TextWidth(font, "󥻥") + GAP;
    h = TextHeight(font);
    height = h + GAP;

    width = FILE_SAVE_WINDOW_WIDTH;
    top = file_save_window =
	toplevel_WinObj(disp, "Sound File Save", "Sound File Save",
			width,
			3 * height + 5 * GAP + 8 * BD_WIDTH + h + 8,
			WinObj_fg, WinObj_bg);
    top->any.when_realize_mapping = False;

    save_text =
	create_DirLineText(top,
			   GAP, GAP,
			   top->any.width - 2 * GAP, height,
			   BD_WIDTH, font, " ¸ե̾:");
    SetCallBackProc(save_text, (call_back_t)setup_file_save_ok);
    save_text->dirtext.short_cut_list =
	(MCPopupMenu *)text_short_cut_list_popup;
    SetPopupSelectingHook(save_text, popup_waiting_hook);

    save_format =
	create_ChoiceButton(top,
			    GAP, WinObjBottom(save_text) + GAP,
			    BD_WIDTH, height, font,
			    "إåեޥå: ",
			    format_msg, SAVE_FORMAT_KINDS, 1);
    SetCallBackProc(save_format, (call_back_t)save_format_proc);

    {
	WinObj* popup;
	WinObj* label;
	int i;
	popup = create_PopupMenu(disp, NULL, BD_WIDTH,
				 WinObj_fg, WinObj_bg, font);
	for(i = 1; i < USFFD_NENCODINGS; i++)
	    add_PopupMenu(popup, usffd_encoding_jname(i), i,
			  change_save_encoding);
	label = create_LabeldWindow(top,  GAP,
				    WinObjBottom(save_format) + GAP,
				    TextWidth(font, "沽: "),
				    8 + h + 2 * BD_WIDTH,
				    0, font, "沽: ",
				    WINOBJ_BORDER_NONE, Center);
	save_encoding_select_menu =
	    create_SelectMenu(top, WinObjRight(label),
			      WinObjBottom(save_format) + GAP,
			      font, BD_WIDTH, popup, USFFD_G711_ULAW);
    }

    ok_button = create_Button(top,
			      top->any.width - cancel_button_width
			      - GAP - 2 * BD_WIDTH,
			      top->any.height - GAP - height - 2 * BD_WIDTH,
			      cancel_button_width, height, BD_WIDTH,
			      font, "ϣ");
    cancel_button = create_Button(top,
				  WinObjLeft(ok_button) - cancel_button_width
				  - GAP - 2 * BD_WIDTH,
				  top->any.height - GAP - height
				  - 2 * BD_WIDTH,
				  cancel_button_width, height, BD_WIDTH,
				  font, "󥻥");
    SetCallBackProc(cancel_button, (call_back_t)setup_file_save_cancel);
    SetCallBackProc(ok_button, (call_back_t)setup_file_save_ok);

    save_yn = create_QueryYesNo(disp, BD_WIDTH, WinObj_fg, WinObj_bg, font);
    save_yn->qyn.yes_selected = save_yes;
    save_yn->qyn.no_selected = save_no;

    WinObjRealize(top);
}

PUBLIC void file_popup_setup_file_save(PopupMenuBase* menu,
				       char* name, int id)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(menu);
    AudioBuffer* b;
    static Bool init_flag = False;

#ifdef DEBUG
    puts("file_popup_setup_file_save()");
#endif

    b = frame->player->aubuff;
    if(b == NULL || b->buffer == NULL)
    {
	write_error_message("ǡޤ");
	return;
    }

    if(!init_flag)
    {
	XSizeHints sizehints;

	init_flag = True;
	sizehints.flags = USPosition;
	QueryPointer(disp, DefaultRootWindow(disp),
		     &sizehints.x, &sizehints.y);
	XSetNormalHints(disp, file_save_window->any.win, &sizehints);
	XMoveWindow(disp, file_save_window->any.win, sizehints.x, sizehints.y);
    }
    XMapWindow(disp, file_save_window->any.win);

    open_file_target_frame = frame;

    if(b->filename == NULL)
	SetLineText((LineText *)save_text, "untitled.au", 11);
    else
	SetLineText((LineText *)save_text,
		    b->filename, strlen(b->filename));
    if(b->hdr_type == -1)
	b->hdr_type = USFFD_HDR_AU;
    change_ChoiceButton(save_format, b->hdr_type);
    select_save_encoding((enum usffd_audio_header_t)b->hdr_type,
			 b->file_encoding);
    make_file_name_extension();
    set_audio_file_lock(True);
}

PUBLIC void file_save_slice_proc(void)
{
    long nchannels = file_saver.nchannels;
    long block_align = file_saver.block_align;
    long nsample;
    long i, j, n;
    long start = file_saver.write_samples;
    char* writebuff;
    short** data = file_saver.save_buffer->buffer;
    short* buff16 = (short *)(temp_buff + temp_buff_size / 2);

#ifdef DEBUG
    static double save_start_time;
    if(start == 0)
	save_start_time = get_current_calender_time();
#endif /* DEBUG */

    nsample = DATA_FRAGMENTATION_SIZE / block_align;
    if(nsample > file_saver.rest_samples)
	nsample = file_saver.rest_samples;

    n = nsample * block_align;
    writebuff = temp_buff;

    if(nchannels == 1)
	file_saver.converter(data[0] + start,
			     writebuff,
			     (int)(nsample * sizeof(short)));
    else
    {
	for(j = 0; j < nchannels; j++)
	    for(i = 0; i < nsample; i++)
		buff16[i * nchannels + j] = data[j][i + start];
	file_saver.converter(buff16, writebuff,
			     (int)(nsample * nchannels * sizeof(short)));
    }

    if(write(file_saver.fd, writebuff, n) < 0)
    {
	write_syserror_message("write");
	sleep(3);
	return;
    }

    file_saver.write_samples += nsample;
    file_saver.rest_samples -= nsample;

    update_file_indicator(file_saver.write_samples +
	file_saver.rest_samples, file_saver.write_samples);

    if(file_saver.rest_samples == 0)
    {
#ifdef DEBUG
	printf("load file: %.1f Kbps\n",
	       (double)file_loader.read_samples /
	       (get_current_calender_time() - save_start_time) * 0.008);
#endif /* DEBUG */
	save_file_done();
    }
}
