#include <stdio.h>
#include <math.h>
#include "soundscope.h"
#include "ssframe.h"
#include "player.h"
#include "aubuff.h"
#include "audriv.h"

#define RVIEW_WIDTH(frame) ((frame)->scope_width - 4 * BD_WIDTH)
#define RVIEW_STARTX       (2 * BD_WIDTH)

static void draw_edit_rview_region(SSFrame* frame)
{
    long n, start, end, diff;
    int startp, w;
    XSetForeground(disp, gc, scope_bg);
    XFillRectangle(disp, frame->edit_region_viewer, gc,
		   0, 0, frame->scope_width, 8);
    if(is_audio_buffer_empty(frame->player->aubuff))
	return;
    start  = frame->player->region_start;
    end    = frame->player->region_end;
    n      = frame->player->aubuff->nsamples;
    diff   = end - start;
    if(diff == 0)
	return;
    startp = (int)((double)start / n * RVIEW_WIDTH(frame) + RVIEW_STARTX);
    w      = (int)((double)diff / n * RVIEW_WIDTH(frame));
    if(w == 0)
	w = 1;
    XSetForeground(disp, gc, color_pixel4);
    XFillRectangle(disp, frame->edit_region_viewer, gc,
		   startp, 0, w, 8);
}

static void update_scope_edit(SSFrame* frame)
{
    long n;
    int mode = SSFRAME_SCOPE_MAIN_MODE(frame);

    n = frame->player->aubuff->nsamples;

    if(frame->zoom[SCOPE_WAVE] > n) frame->zoom[SCOPE_WAVE] = n;
    if(frame->zoom[SCOPE_EDIT] > n) frame->zoom[SCOPE_EDIT] = n;

    if(mode == SCOPE_EDIT)
    {
	draw_edit_rview_region(frame);
	draw_edit_scope_wave(frame);
	if(n > 0)
	    SetHScrollbarShown((HScrollbar *)frame->edit_slider,
			       frame->zoom[SCOPE_EDIT] / n);
	else
	    SetHScrollbarShown((HScrollbar *)frame->edit_slider, 1.0);
	SetHScrollbarThumbPos((HScrollbar *)frame->edit_slider,
			      frame->ed_vratio);
    }
    if(mode == SCOPE_WAVE || mode == SCOPE_EDIT)
	SetHSimpleSliderPos(frame->zoom_slider,
			    calc_log_wave_zoom(frame->zoom[mode],
				frame->zoom_slider->sslider.maxlen, n));
}

static long calc_display_start(SSFrame* frame)
{
    long nsamples;

    nsamples = frame->player->aubuff->nsamples;
    if(frame->player->load_flag && file_loader.read_samples < nsamples)
	nsamples = file_loader.read_samples;
    return (long)((nsamples - frame->zoom[SCOPE_EDIT]) * frame->ed_vratio);
}

static Bool sample_to_window_x(SSFrame* frame, long sample, int* x)
{
    long ds;
    int x0, width;
    double scale;

    ds = calc_display_start(frame);
    scale = frame->zoom[SCOPE_EDIT];

    if(sample < ds || sample - ds > scale)
	return False;
    get_scope_draw_region(frame, -1, &x0, NULL, &width, NULL);
    *x = x0 + (int)((sample - ds) * width / scale);
    return True;
}

static long window_x_to_sample(SSFrame* frame, int x)
{
    long ds, sample;
    int x0, width;
    double scale;

    ds = calc_display_start(frame);
    scale = frame->zoom[SCOPE_EDIT];
    get_scope_draw_region(frame, -1, &x0, NULL, &width, NULL);

    sample = (long)(ds + (x - x0) * scale / width);
    if(sample < 0)
	sample = 0;
    else if(sample > frame->player->aubuff->nsamples)
	sample = frame->player->aubuff->nsamples;
    return sample;
}

static void clear_edit_cursor(SSFrame* frame)
{
    XCopyArea(disp, frame->ed_cur_pm, frame->scope_pm, gc,
	      0, 0, 1, frame->scope_height, frame->ed_cur, 0);
    XCopyArea(disp, frame->ed_cur_pm, frame->scope_win, gc,
	      0, 0, 1, frame->scope_height, frame->ed_cur, 0);
    frame->ed_cur = -1;
}

static void set_edit_cursor(SSFrame* frame, int x)
{
    XCopyArea(disp, frame->scope_pm, frame->ed_cur_pm, gc,
	      x, 0, 1, frame->scope_height, 0, 0);
    frame->ed_cur = x;
    XSetForeground(disp, gc, color_pixel1);
    XDrawLine(disp, frame->scope_pm, gc,
	      x, 0, x, frame->scope_height - 1);
    XDrawLine(disp, frame->scope_win, gc,
	      x, 0, x, frame->scope_height - 1);
}

PUBLIC void scope_edit_move_cur(SSFrame* frame, long sample)
{
    int x;

    if(frame->ed_cur == -1)
    {
	if(sample_to_window_x(frame, sample, &x) == False)
	    return;
	set_edit_cursor(frame, x);
    }
    else
    {
	if(sample_to_window_x(frame, sample, &x) == False)
	    clear_edit_cursor(frame);
	else if(frame->ed_cur != x)
	{
	    clear_edit_cursor(frame);
	    set_edit_cursor(frame, x);
	}
    }
}

static int edit_set_wave_points(SSFrame* frame, int ch, XPoint* xp)
{
    int x, y, width, height;
    int np;
    long i, ds;
    double wr, hr, scale;
    short* data;

    get_scope_draw_region(frame, ch, &x, &y, &width, &height);
    x++;
    width -= 2;
    scale = frame->zoom[SCOPE_EDIT];
    hr = (double)height / 65536.0 * frame->yzoom;
    np = 0;
    ds = calc_display_start(frame);

    data = frame->player->aubuff->buffer[ch] + ds;

    if(scale <= MAX_NPOINTS)
    {
	wr = (double)width / scale;

	for(i = 0; i < scale; i++)
	{
	    xp[i].x = (short)(x + i * wr);
	    xp[i].y = (short)(y + height/2 - data[i] * hr);
	}
	np = scale;
    }
    else
    {
	double r = scale / (double)MAX_NPOINTS; /* r >= 1.0 */
	wr = (double)width / (double)MAX_NPOINTS;
	if(r > 10.0)
	    r = floor(r);
	data -= ds % (long)r;
	for(i = 0; i < MAX_NPOINTS; i++)
	{
	    xp[i].x = (short)(x + i * wr);
	    xp[i].y = (short)(y + height/2
			      - data[((long)((double)i * r))] * hr);
	}
	np = MAX_NPOINTS;
    }
    return np;
}

static void draw_edit_scope_region_mark(SSFrame* frame, Drawable d,
					long start, long end)
{
    int xs, xe, x0, y0, width, height;
    long ds;
    double scale;

    get_scope_draw_region(frame, -1, &x0, &y0, &width, &height);

    ds = calc_display_start(frame);
    scale = frame->zoom[SCOPE_EDIT];

    if(start < ds)
	xs = 0;
    else if(start - ds > scale)
	xs = width;
    else
	xs = x0 + (int)((start - ds) * width / scale);
    if(end < ds)
	xe = 0;
    else if(end - ds > scale)
	xe = width;
    else
	xe = x0 + (int)((end - ds) * width / scale);
    if(xs >= xe)
	return;

    XSetForeground(disp, gc, color_pixel4);
    XFillRectangle(disp, d, gc, xs, y0, xe - xs, height);
}

PUBLIC void draw_edit_scope_wave(SSFrame* frame)
{
    Pixmap pm = frame->scope_pm;
    int np, ch;
    XPoint* xp;

    XSetForeground(disp, gc, scope_bg);
    XFillRectangle(disp, pm, gc, 0, 0,
		   frame->scope_width, frame->scope_height);

    if(is_audio_buffer_empty(frame->player->aubuff))
    {
	scope_wave_draw_measure(frame, pm);
	XCopyArea(disp, pm, frame->scope_win, gc,
		  0, 0, frame->scope_width, frame->scope_height, 0, 0);
	XSync(disp, False);
	return;
    }

    if(frame->player->region_start < frame->player->region_end)
	draw_edit_scope_region_mark(frame, pm,
				    frame->player->region_start,
				    frame->player->region_end);

    scope_wave_draw_measure(frame, pm);
    XSetForeground(disp, gc, scope_fg);
    xp = (XPoint *)temp_buff;
    for(ch = 0; ch < frame->player->aubuff->nchannels; ch++)
    {
	np = edit_set_wave_points(frame, ch, xp);
	XDrawLines(disp, pm, gc, xp, np, CoordModeOrigin);
    }

    XCopyArea(disp, pm, frame->scope_win, gc,
	      0, 0, frame->scope_width, frame->scope_height, 0, 0);
    frame->ed_cur = -1;

    if(!audio_active_flag)
	scope_edit_move_cur(frame, frame->player->pos);
    XSync(disp, False);
}

static void scope_edit_event_proc_set_region(SSFrame* frame,
					     long start_sample,
					     long end_sample)
{
    long cur;

    if(start_sample > end_sample)
    {
	int tmp;
	SWAP(start_sample, end_sample, tmp);
    }

    if(start_sample == end_sample)
    {
	frame->player->region_start = 
	    frame->player->region_end = start_sample;
	draw_edit_scope_wave(frame);
	soundscope_goto_sample(frame, start_sample);
    }
    else
    {
	if(frame->player->region_start == start_sample &&
	   frame->player->region_end == end_sample)
	    return;

	if(audio_active_flag)
	    cur = calc_player_samples(frame->player,audriv_play_samples());
	else
	    cur = frame->player->pos;
	frame->player->region_start = start_sample;
	frame->player->region_end = end_sample;
	draw_edit_scope_wave(frame);
	if(cur < start_sample || end_sample <= cur)
	    soundscope_goto_sample(frame, start_sample);
	else
	    scope_edit_move_cur(frame, cur);
    }
    draw_edit_rview_region(frame);
}

PUBLIC void scope_edit_event_proc(SSFrame* frame, XEvent* e)
{
    static long x_start, x_end;
    static Bool drag = False;

    switch(e->type)
    {
      case ButtonPress:
	if(frame->player->aubuff->nsamples == 0)
	    break;
	if(drag == True)
	    break;
	drag = True;

	if(e->xbutton.button == Button2 ||
	   (frame->player->region_start == frame->player->region_end))
	    x_start = x_end = window_x_to_sample(frame, e->xbutton.x);
	else if(e->xbutton.button == Button1)
	{
	    x_start = frame->player->region_end;
	    x_end   = window_x_to_sample(frame, e->xbutton.x);
	}
	else			/* Button3 */
	{
	    x_start = frame->player->region_start;
	    x_end   = window_x_to_sample(frame, e->xbutton.x);
	}
	scope_edit_event_proc_set_region(frame, x_start, x_end);
#ifdef DEBUG
	printf("ButtonPress at %ld\n", x_start);
#endif
	break;
      case ButtonRelease:
	if(drag == False)
	    break;
	drag = False;
	x_end = window_x_to_sample(frame, e->xbutton.x);
#ifdef DEBUG
	printf("ButtonRelease at %ld\n", x_end);
#endif

	if(x_start > x_end)
	{
	    int tmp;
	    SWAP(x_start, x_end, tmp);
	}

	scope_edit_event_proc_set_region(frame, x_start, x_end);

#ifdef DEBUG
	printf("region: %ld, %ld\n",
	       frame->player->region_start, frame->player->region_end);
#endif
	break;
      case MotionNotify:
	CompressMotionEvent(e->xany.display, e->xany.window, e);

	if(drag == False)
	    break;

#ifdef DEBUG
	printf("MotionNotify at %d\n", e->xmotion.x);
#endif

	x_end = window_x_to_sample(frame, e->xbutton.x);
	scope_edit_event_proc_set_region(frame, x_start, x_end);
	break;
    }
}

PUBLIC void sound_edit_copy_popup(PopupMenuBase* menu, char* name, int id)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(menu);
    AudioPlayer* player = frame->player;
    Bool plflag = audio_active_flag;

#ifdef DEBUG
    puts("sound_edit_copy_popup()");
#endif

    if(player->region_start == player->region_end)
	return;			/* do nothing */
    if(plflag)
	abc_stop_audio();
    audio_buffer_copy_region(player->aubuff,
			     player->region_start,
			     player->region_end);
    draw_edit_scope_wave(frame);
    if(plflag)
	abc_restart_audio();
}

PUBLIC void sound_edit_cut_popup(PopupMenuBase* menu, char* name, int id)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(menu);
    AudioPlayer* player = frame->player;
    Bool plflag = audio_active_flag;

#ifdef DEBUG
    puts("sound_edit_cut_popup()");
#endif

    if(player->region_start == player->region_end)
	return;			/* do nothing */
    if(plflag)
	abc_stop_audio();
    audio_buffer_cut_region(player->aubuff,
			    player->region_start,
			    player->region_end);
    player->region_end = player->region_start;
    update_scope_edit(frame);
    if(plflag)
	abc_restart_audio();
}

PUBLIC void sound_edit_paste_popup(PopupMenuBase* menu, char* name, int id)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(menu);
    AudioPlayer* player = frame->player;
    AudioBuffer* copy;
    Bool plflag = audio_active_flag;
    long n;

#ifdef DEBUG
    puts("sound_edit_paste_popup()");
#endif

    copy = get_audio_buffer(AUDIO_COPY_BUFF_NAME);
    if(is_audio_buffer_empty(copy))
	return;

    if(plflag)
	abc_stop_audio();

    if(player->region_start != player->region_end)
	audio_buffer_delete_region(player->aubuff,
				   player->region_start,
				   player->region_end);
    n = insert_audio_buffer(copy, player->aubuff, player->region_start);
#ifdef DEBUG
    printf("paste to %ld\n", player->region_start);
#endif

    if(n >= 0)
	player->region_end = player->region_start + n;
    update_scope_edit(frame);
    if(plflag)
	abc_restart_audio();
}

PUBLIC void sound_edit_merge_popup(PopupMenuBase* menu, char* name, int id)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(menu);
    AudioPlayer* player = frame->player;
    AudioBuffer* copy;
    Bool plflag = audio_active_flag;

#ifdef DEBUG
    puts("sound_edit_merge_popup()");
#endif

    copy = get_audio_buffer(AUDIO_COPY_BUFF_NAME);
    if(is_audio_buffer_empty(copy))
	return;

    if(plflag)
	abc_stop_audio();

    merge_audio_buffer(copy, player->aubuff, player->region_start);
#ifdef DEBUG
    printf("merge beggin with %ld\n", player->region_start);
#endif

    update_scope_edit(frame);
    if(plflag)
	abc_restart_audio();
}

PUBLIC void sound_edit_overwrite_popup(PopupMenuBase* menu, char* name, int id)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(menu);
    AudioPlayer* player = frame->player;
    AudioBuffer* copy;
    Bool plflag = audio_active_flag;

#ifdef DEBUG
    puts("sound_edit_overwrite_popup()");
#endif

    copy = get_audio_buffer(AUDIO_COPY_BUFF_NAME);
    if(is_audio_buffer_empty(copy))
	return;

    if(plflag)
	abc_stop_audio();

    overwrite_audio_buffer(copy, player->aubuff, player->region_start);
#ifdef DEBUG
    printf("overwrite beggin with %ld\n", player->region_start);
#endif

    update_scope_edit(frame);
    if(plflag)
	abc_restart_audio();
}

static void move_edit_rview_region(SSFrame* frame, long start)
{
    long n, diff;

    n = frame->player->aubuff->nsamples;
    diff = frame->player->region_end - frame->player->region_start;
    if(start < 0)
	start = 0;
    else if(start + diff > n)
	start = n - diff;
    frame->player->region_start = start;
    frame->player->region_end   = start + diff;
    if(!audio_active_flag)
	if(frame->player->pos < start)
	    frame->player->pos = start;
	else if(frame->player->pos > start + diff)
	    frame->player->pos = start + diff;
    draw_edit_scope_wave(frame);
    draw_edit_rview_region(frame);
}

PUBLIC void edit_rview_event_handler(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);
    static int button_press_x, start_init;
    int dx;
    long start;

    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count > 0)
	    break;
	draw_edit_rview_region(frame);
	break;
      case ButtonPress:
	if(is_audio_buffer_empty(frame->player->aubuff))
	    return;
	button_press_x = e->xbutton.x;
	start_init = frame->player->region_start;
	break;
      case MotionNotify:
	CompressMotionEvent(e->xany.display, e->xany.window, e);
	if(is_audio_buffer_empty(frame->player->aubuff))
	    return;
	dx = e->xbutton.x - button_press_x;
	start = (long)
	    ((double)dx / RVIEW_WIDTH(frame)
	     * frame->player->aubuff->nsamples + start_init);
	move_edit_rview_region(frame, start);
	break;
    }
}

PUBLIC void sound_edit_fin_popup(PopupMenuBase* menu, char* name, int id)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(menu);
    AudioPlayer* player = frame->player;
    Bool plflag = audio_active_flag;
    long start, end;
    double d;
    int ch, nch;

#ifdef DEBUG
    puts("sound_edit_fin_popup");
#endif

    start = player->region_start;
    end = player->region_end;
    if(start == end || is_audio_buffer_empty(player->aubuff))
	return;			/* do nothing */
    if(plflag)
	abc_stop_audio();

    nch = player->aubuff->nchannels;
    d = 1.0 / (end - start);
    for(ch = 0; ch < nch; ch++)
    {
	short* data;
	double r;
	long i;

	data = player->aubuff->buffer[ch];
	r = d;
	for(i = start; i < end; i++)
	{
	    data[i] = (short)(data[i] * r);
	    r += d;
	}
    }

    draw_edit_scope_wave(frame);
    if(plflag)
	abc_restart_audio();
}

PUBLIC void sound_edit_fout_popup(PopupMenuBase* menu, char* name, int id)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(menu);
    AudioPlayer* player = frame->player;
    Bool plflag = audio_active_flag;
    long start, end;
    double d;
    int ch, nch;

#ifdef DEBUG
    puts("sound_edit_fout_popup");
#endif

    start = player->region_start;
    end = player->region_end;
    if(start == end || is_audio_buffer_empty(player->aubuff))
	return;			/* do nothing */
    if(plflag)
	abc_stop_audio();

    nch = player->aubuff->nchannels;
    d = 1.0 / (end - start);
    for(ch = 0; ch < nch; ch++)
    {
	short* data;
	double r;
	long i;

	data = player->aubuff->buffer[ch];
	r = 1.0;
	for(i = start; i < end; i++)
	{
	    data[i] = (short)(data[i] * r);
	    r -= d;
	}
    }

    draw_edit_scope_wave(frame);
    if(plflag)
	abc_restart_audio();
}
