#include <stdio.h>
#include <math.h>
#include "soundscope.h"
#include "ssframe.h"
#include "player.h"
#include "aubuff.h"


#define FFT_SCALE 0.07

void draw_scope_spectrogram(SSFrame* frame, int ch, long sample)
{
    int x, y, width, height, hh, ty;
    int i, j;
    double wr, hr;
    double t, u, p;
    double yz;

    get_current_fft_data(frame, ch, sample);
    yz = frame->yzoom;
    get_scope_draw_region(frame, ch, &x, &y, &width, &height);

    width -= 256;
    x     += 256;
    hh = height/5;

    hr = (0.8 * FFT_SCALE) * height * yz;

    if(fft_data[0] >= 0)
	p = fft_data[0];
    else
	p = -fft_data[0];
    p = log(1.0 + (128.0 / FFTSIZE) * p);
    ty = (int)(hr * p);
    if(ty > hh)
	ty = hh;
    wave_points[0].y = y + hh - 1 - ty;
    for(i = 1, j = FFTSIZE - 1; i < FFTSIZE/2; i++, j--)
    {
	t = fft_data[i];
	u = fft_data[j];
	p = log(1.0 + (128.0 / FFTSIZE) * sqrt(t*t + u*u));

	ty = (int)(hr * p);
	if(ty > hh)
	    ty = hh;
	wave_points[i].y = y + hh - 1 - ty;
    }

    if(frame->local_mode_selector[SCOPE_HZ_LOG]->choice.selected == 1)
    {
	wr = (double)width / (double)(FFTSIZE/2 + 1);
	for(i = 0; i < FFTSIZE/2; i++)
	    wave_points[i].x = x + (int)(wr * i);
    }
    else
    {
	wr = (double)width;
	for(i = 0; i < FFTSIZE/2; i++)
	    wave_points[i].x = x + (int)(wr * frame->log_hz_table[i]);
    }
    XSetForeground(disp, gc, scope_fg);
    XDrawLines(disp, frame->scope_pm, gc, wave_points, FFTSIZE/2,
	       CoordModeOrigin);
}



/*
 * COLOR PICTURE MODE
 */
#include "colcnv.h"
#include "misc.h"
#define NCOLOR    40
static unsigned long color_ring[NCOLOR];

static void set_color_ring(void)
{
    int i;

    if(display_visual_type == DISPLAY_VISUAL_MONO)
    {
	for(i = 0; i < NCOLOR/4; i++)
	    color_ring[i] = scope_fg;
	for(; i < NCOLOR; i++)
	    color_ring[i] = scope_bg;
    }
    else
    {
	for(i = 0; i < NCOLOR; i++)
	{
	    rgb_t rgb;
	    double h = 240 - 360 * sqrt((1.0/NCOLOR) * i);
	    double s;

	    if(h >= 0)
		s = 0.9;
	    else
	    {
		s = 0.9 + h / 120.0 * 0.9;
		h = 0.0;
	    }

	    hsv_to_rgb(h, s, 1.0, &rgb);
	    if(WinObj_reverse_color_flag)
		color_ring[NCOLOR - i - 1]
		    = AllocRGBColor(disp, rgb.r, rgb.g, rgb.b);
	    else
		color_ring[i] = AllocRGBColor(disp, rgb.r, rgb.g, rgb.b);
	}
    }
}

void draw_scope_color_spectrogram(SSFrame* frame, int ch, long sample,
				  long call_cnt)
{
    static Bool init_flag = False;
    static unsigned char image_buff[FFTSIZE / 2 * 4];
    static XImage* spectram_image;
    int x, y, width, height;
    int i, j;
    double t, u, p, hr;
    double* p1 = (double *)temp_buff;
    double* p2 = (double *)(temp_buff + temp_buff_size / 2);
    double* pt;
    int nch = frame->player->aubuff->nchannels;
    int np;
    double yz;
#ifdef FAST_SOUNDSPECTROGRAM
    int offset = call_cnt % FAST_SOUNDSPECTROGRAM;
#endif /* FAST_SOUNDSPECTROGRAM */

    if(!init_flag)
    {
	int scr;
	set_color_ring();
	scr = DefaultScreen(disp);
	spectram_image =
	    XCreateImage(disp, DefaultVisual(disp, scr),
			 DefaultDepth(disp, scr),
			 ZPixmap, 0, 0,
			 1, FFTSIZE / 2, 8, 0);
	spectram_image->data = (char *)image_buff;
	init_flag = True;
    }

    yz = frame->yzoom;
    get_current_fft_data(frame, ch, sample);
    get_scope_draw_region(frame, ch, &x, &y, &width, &height);

    hr = yz * (0.27 * NCOLOR);
    if(fft_data[0] >= 0)
	p = fft_data[0];
    else
	p = -fft_data[0];
    p1[0] = log(1.0 + (128.0 / FFTSIZE) * p) * hr;

    for(i = 1, j = FFTSIZE - 1; i < FFTSIZE/2; i++, j--)
    {
	t = fft_data[i];
	u = fft_data[j];
	p1[i] = log(1.0 + (128.0 / FFTSIZE) * sqrt(t*t + u*u)) * hr;
    }

    if(frame->local_mode_selector[SCOPE_HZ_LOG]->choice.selected == 1)
	pt = p1;
    else
    {
	double px, tx, a, b, s;
	int x1, n;

	for(i = 0; i < FFTSIZE / 2; i++)
	    p2[i] = 0.0;

	p2[0] = p1[0];
	px = 0.0;

	for(i = 1; i < FFTSIZE / 2 - 1; i++)
	{
	    tx = frame->exp_hz_table[i];
	    x1 = (int)px;
	    n = 0;
	    s = 0.0;
	    do
	    {
		a = p1[x1];
		b = p1[x1 + 1];
		s += (a * (x1 + 1 - tx) + b * (tx - x1));
		n++;
		x1++;
	    } while(x1 < tx);
	    if(s <= -0.0)
		p2[i] = 0.0;
	    else
		p2[i] = s / n;
	    px = tx;
	}

	p2[FFTSIZE / 2 - 1] = p1[FFTSIZE / 2 - 1];
	pt = p2;
    }

    switch (spectram_image->bits_per_pixel) {
    case 8:
      if(nch == 1) {
	for(i = 0; i < FFTSIZE / 2; i++)
	  image_buff[i] = (unsigned char)
	    color_ring[((unsigned)pt[FFTSIZE / 2 - i - 1]) % NCOLOR];
	np = FFTSIZE / 2;
      } else {
	np = FFTSIZE / 2 / nch;
	for(i = 0; i < np; i++) {
	  double s;
	  s = 0;
	  for(j = 0; j < nch; j++)
	    s += pt[i * nch + j];
	  s /= nch;
	  if (s >= NCOLOR - 1)
	    s = NCOLOR - 1;
	  image_buff[np - i - 1] = (unsigned char)color_ring[(int)s];
	}
      }
      break;

    case 24:
      if(nch == 1) {
	for(i = 0; i < FFTSIZE / 2; i++) {
	  unsigned long c =
	    color_ring[((unsigned)pt[FFTSIZE / 2 - i - 1]) % NCOLOR];
	  j = i * 3;
	  image_buff[j] = c & 0xff;
	  image_buff[j+1] = (c>>8) & 0xff;
	  image_buff[i+2] = (c>>16) & 0xff;
	}
	np = FFTSIZE / 2;
      } else {
	np = FFTSIZE / 2 / nch;
	for(i = 0; i < np; i++) {
	  double s;
	  unsigned long c;

	  s = 0;
	  for(j = 0; j < nch; j++)
	    s += pt[i * nch + j];
	  s /= nch;
	  if (s >= NCOLOR - 1)
	    s = NCOLOR - 1;
	  c = color_ring[(int)s];
	  j = (np - i - 1) * 3;
	  image_buff[j] = c & 0xff;
	  image_buff[j+1] = (c>>8) & 0xff;
	  image_buff[j+2] = (c>>16) & 0xff;
	}
      }
      break;

    case 32:
      if(nch == 1) {
	for(i = 0; i < FFTSIZE / 2; i++) {
	  unsigned long c =
	    color_ring[((unsigned)pt[FFTSIZE / 2 - i - 1]) % NCOLOR];
	  j = i * 4;
	  image_buff[j] = c & 0xff;
	  image_buff[j+1] = (c>>8) & 0xff;
	  image_buff[i+2] = (c>>16) & 0xff;
	  image_buff[i+3] = 0;
	}
	np = FFTSIZE / 2;
      } else {
	np = FFTSIZE / 2 / nch;
	for(i = 0; i < np; i++) {
	  double s;
	  unsigned long c;

	  s = 0;
	  for(j = 0; j < nch; j++)
	    s += pt[i * nch + j];
	  s /= nch;
	  if (s >= NCOLOR - 1)
	    s = NCOLOR - 1;
	  c = color_ring[(int)s];
	  j = (np - i - 1) * 4;
	  image_buff[j] = c & 0xff;
	  image_buff[j+1] = (c>>8) & 0xff;
	  image_buff[j+2] = (c>>16) & 0xff;
	  image_buff[j+3] = 0;
	}
      }
      break;
    }
#ifdef FAST_SOUNDSPECTROGRAM
    XPutImage(disp, frame->scope_pm, gc, spectram_image, 0, 0,
	      x + width - FAST_SOUNDSPECTROGRAM + offset, np * ch,
	      1, np);
    XCopyArea(disp, frame->scope_pm, frame->scope_win, gc,
	      x + width - FAST_SOUNDSPECTROGRAM + offset, np * ch,
	      1, np, x + width - FAST_SOUNDSPECTROGRAM + offset, np * ch);
#else
    XPutImage(disp, frame->scope_pm, gc, spectram_image, 0, 0,
	      x + width - 1, np * ch, 1, np);
#endif /* FAST_SOUNDSPECTROGRAM */
}
