#include <stdio.h>
#include <math.h>
#include <string.h>
#include "soundscope.h"
#include "ssframe.h"
#include "player.h"
#include "aubuff.h"

static void draw_dash_vline(Drawable win, int x, int y, int len)
{
    XSegment* s = (XSegment *)temp_buff;
    int i, j, end;

    end = y + len - 2;
    for(j = 0, i = y; i < end; i += 4, j++)
    {
	s[j].x1 = s[j].x2 = x;
	s[j].y1 = i;
	s[j].y2 = i+1;
    }

    XDrawSegments(disp, win, gc, s, j);
}

static void draw_dash_hline(Drawable win, int x, int y, int len)
{
    XSegment* s = (XSegment *)temp_buff;
    int i, j, end;

    end = x + len - 2;
    for(j = 0, i = x; i < end; i += 4, j++)
    {
	s[j].y1 = s[j].y2 = y;
	s[j].x1 = i;
	s[j].x2 = i+1;
    }

    XDrawSegments(disp, win, gc, s, j);
}

static char* make_time_string(char* buff, double sec)
{
    if(sec < 0.01)
	sprintf(buff, "%.1f ms", (sec * 1000));
    else if(sec < 1.0)
	sprintf(buff, "%d ms", (int)(sec * 1000));
    else if(sec < 10.0)
	sprintf(temp_buff, "%.2f s", sec);
    else if(sec < 60.0)
	sprintf(buff, "%.1f s", sec);
    else
	sprintf(buff, "%d:%02d", ((int)sec) / 60, ((int)sec) % 60);
    return buff;
}

PUBLIC void scope_wave_draw_measure(SSFrame* frame, Drawable win)
{
    int i, ch, width, height;
    double dx, xm, t, scale;
    char* s;

    width  = frame->scope_width;
    XSetForeground(disp, gc, color_pixel2);
    for(ch = 0; ch < frame->player->aubuff->nchannels; ch++)
    {
	int y;

	get_scope_draw_region(frame, ch, NULL, &y, NULL, &height);
	for(i = -3; i <= 3; i++)
	{
	    int y0 = (int)(y + height / 2.0 + (i / 8.0) * height);
	    if(i != 0)
		draw_dash_hline(win, 0, y0, width);
	}
    }

    height = frame->scope_height;
    dx = (double)width / 32.0;
    for(i = 0, xm = width / 2.0; xm < width; xm += dx, i++)
    {
	int x0 = (int)(xm);
	if(i % 10 == 0)
	{
	    if(!ISMONODISP)
		XSetForeground(disp, gc,
			       WhitePixel(disp, DefaultScreen(disp)));
	    draw_dash_vline(win, x0, 0, height);
	    draw_dash_vline(win, width - x0, 0, height);
	    if(ISMONODISP)
	    {
		draw_dash_vline(win, x0 + 1, 0, height);
		draw_dash_vline(win, width - x0 + 1, 0, height);
	    }
	    XSetForeground(disp, gc, color_pixel2);
	}
	else
	{
	    draw_dash_vline(win, x0, 0, height);
	    draw_dash_vline(win, width - x0, 0, height);
	}
    }
    if(!ISMONODISP)
	XSetForeground(disp, gc,
		       WhitePixel(disp, DefaultScreen(disp)));
    draw_dash_hline(win, 0, height / 2, width);

    if(ISMONODISP)
	draw_dash_hline(win, 0, height / 2 + 1, width);

    scale = frame->zoom[SSFRAME_SCOPE_MAIN_MODE(frame)];
    draw_display_scope_time(frame, win, scale);

    XSetForeground(disp, gc, color_pixel1);
    draw_measure(disp, win, &frame->measure);
    t = floor(0.5 + scale) / frame->player->sample_rate;
    s = make_time_string(temp_buff,
			 t / frame->scope_width *
			 (frame->measure.x_right -
			  frame->measure.x_left + 1));
    XDrawString(disp, win, gc,
		frame->measure.x_left + 2,
		frame->measure.y + 14, s, strlen(s));
}
				    
PUBLIC void init_scope_wave(SSFrame* frame)
{
    Pixmap pm = frame->scope_background;

    scope_wave_draw_measure(frame, pm);

    XCopyArea(disp, pm, frame->scope_pm, gc, 0, 0,
	      frame->scope_width, frame->scope_height, 0, 0);
    XCopyArea(disp, pm, frame->scope_win, gc, 0, 0,
	      frame->scope_width, frame->scope_height, 0, 0);
}

#define SCALED(n) ((long)((double)(n) * r))

static int c_set_wave_points(SSFrame* frame, int ch, long sample, long scale)
{
    int x, y, width, height;
    int np;
    long i, nsamples;
    double wr, hr;

    get_scope_draw_region(frame, ch, &x, &y, &width, &height);

    nsamples = frame->player->aubuff->nsamples;
    if(frame->player->load_flag && file_loader.read_samples < nsamples)
	nsamples = file_loader.read_samples;

    hr = (double)height / 65536.0 * frame->yzoom;
    np = 0;

    if(scale <= MAX_NPOINTS)
    {
	short* p = (short *)temp_buff;
	wr = (double)width / (double)scale;
	sample -= scale / 2;
	convolution_proc(frame->player->aubuff->buffer[ch], nsamples, sample,
			 p, scale, frame->convolution);

	if(0 <= sample && sample + scale < nsamples)
	{
	    for(i = 0; i < scale; i++)
	    {
		wave_points[i].x = (short)(x + i * wr);
		wave_points[i].y = (short)(y + height/2 - p[i] * hr);
	    }
	    np = scale;
	}
	else
	{
	    i = MAX(-sample, 0);
	    np = 0;
	    for(; i < scale && sample + i < nsamples; i++, np++)
	    {
		wave_points[np].x = (short)(x + i * wr);
		wave_points[np].y = (short)(y + height/2 - p[i] * hr);
	    }
	}
    }
    else
    {
	double r = (double)scale / (double)MAX_NPOINTS; /* r >= 1.0 */
	short* data = frame->player->aubuff->buffer[ch];
	wr = (double)width / (double)MAX_NPOINTS;

	if(r > 10.0)
	    r = floor(r);

	sample -= SCALED(MAX_NPOINTS / 2.0);
	sample -= sample % SCALED(1);

	np = 0;
	for(i = 0; sample + SCALED(i) < 0; i++)
	    ;
	for(; i < MAX_NPOINTS && sample + SCALED(i) < nsamples; i++, np++)
	{
	    short si;
	    convolution_proc(data, nsamples, (long)(sample + SCALED(i)),
			     &si, 1, frame->convolution);
	    wave_points[np].x = (short)(x + i * wr);
	    wave_points[np].y = (short)(y + height/2 - si * hr);
	}
    }
    return np;
}

static int set_wave_points(SSFrame* frame, int ch, long sample, long scale)
{
    int x, y, width, height;
    int np;
    long i, nsamples;
    short* data;
    double wr, hr;

    if(frame->equlizing_flag)
	return c_set_wave_points(frame, ch, sample, scale);

    get_scope_draw_region(frame, ch, &x, &y, &width, &height);

    nsamples = frame->player->aubuff->nsamples;
    if(frame->player->load_flag && file_loader.read_samples < nsamples)
	nsamples = file_loader.read_samples;

    hr = (double)height / 65536.0 * frame->yzoom;
    np = 0;

    if(scale <= MAX_NPOINTS)
    {
	sample -= scale / 2;
	data = frame->player->aubuff->buffer[ch] + sample;
	wr = (double)width / (double)scale;

	if(0 <= sample && sample + scale < nsamples)
	{
	    for(i = 0; i < scale; i++)
	    {
		wave_points[i].x = (short)(x + i * wr);
		wave_points[i].y = (short)(y + height/2 - data[i] * hr);
	    }
	    np = scale;
	}
	else
	{
	    i = MAX(-sample, 0);
	    np = 0;
	    for(; i < scale && sample + i < nsamples; i++, np++)
	    {
		wave_points[np].x = (short)(x + i * wr);
		wave_points[np].y = (short)(y + height/2 - data[i] * hr);
	    }
	}
    }
    else
    {
	double r = (double)scale / (double)MAX_NPOINTS; /* r >= 1.0 */
	wr = (double)width / (double)MAX_NPOINTS;

	if(r > 10.0)
	    r = floor(r);

	sample -= SCALED(MAX_NPOINTS / 2.0);
	sample -= sample % SCALED(1);
	data = frame->player->aubuff->buffer[ch] + sample;

	if(0 <= sample && sample + SCALED(MAX_NPOINTS) < nsamples)
	{
	    for(i = 0; i < MAX_NPOINTS; i++)
	    {
		wave_points[i].x = (short)(x + i * wr);
		wave_points[i].y = (short)(y + height/2
					   - data[SCALED(i)] * hr);
	    }
	    np = MAX_NPOINTS;
	}
	else
	{
	    for(i = 0; sample + SCALED(i) < 0; i++)
		;
	    np = 0;
	    for(; i < MAX_NPOINTS && sample + SCALED(i) < nsamples; i++, np++)
	    {
		wave_points[np].x = (short)(x + i * wr);
		wave_points[np].y = (short)(y + height/2
					    - data[SCALED(i)] * hr);
	    }
	}
    }
    return np;
}

PUBLIC void draw_scope_wave(SSFrame* frame, int ch, long sample)
{
    int np, i;
    long scale = (long)(0.5 + frame->zoom[SCOPE_WAVE]);

    np = set_wave_points(frame, ch, sample, scale);
    XSetForeground(disp, gc, scope_fg);
    XDrawLines(disp, frame->scope_pm, gc, wave_points, np, CoordModeOrigin);
    for(i = 0; i < np; i++)
    {
	short y;
	y = wave_points[i].y;
	if(y < repaint_y_min)
	    repaint_y_min = y;
	if(y > repaint_y_max)
	    repaint_y_max = y;
    }
}

PUBLIC void change_display_scope_time(SSFrame* frame, Drawable win,
				      double dscale)
{
    XSetForeground(disp, gc, scope_bg);
    XFillRectangle(disp, win, gc,
		   10, 20, 50, 10);
    XSetForeground(disp, gc, color_pixel2);
    draw_dash_vline(win, 24, 0, frame->scope_height);
    draw_dash_vline(win, 48, 0, frame->scope_height);
    draw_display_scope_time(frame, win, dscale);
}

PUBLIC void draw_display_scope_time(SSFrame* frame, Drawable win,
				    double dscale)
{
    long scale;
    double t;
    char* s;

    scale = (long)(0.5 + dscale);
    t = (double)scale / frame->player->sample_rate;
    s = make_time_string(temp_buff, t);
    XSetForeground(disp, gc, scope_fg);
    XDrawString(disp, win, gc,
		10, 30, s, strlen(s));
}
