/* sear - sound encoding auto recognizer ưǧץ */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#if defined(sgi)
#include <alloca.h>
#endif
#include "aenc.h"
#include "audio_cnv.h"

#ifdef NoDRand48
#include "fdrand48.c"
#define drand48 fdrand48
#endif

#ifndef lint
char* sear_version = "1.1";
#endif

int sear_try_samples = 512;
int sear_retry_seek = 10;
void sear(int fd,		/* եǥץ */
	  long start_seek,	/* ǧκǽΥХȰ */
	  long size,		/* ǧǡΥХȿ */
	  int* encoding_return,	/* ǧ줿粽 */
	  int* nchannels_return); /* ǧ줿ͥ */

void sear_tty(int* encoding_return,   /* ǧ줿粽 */
	      int* nchannels_return); /* ǧ줿ͥ */

void sear_string(const char* buff, /* ǡ */
		 long buff_size,/* ǡ(Хȿ) */
		 int* enc,	/* ǧ줿粽 */
		 int* chan);	/* ǧ줿ͥ */

/* ơ粽ˤĤơʬ PCM ¤ȤäơǤ⾮Τ
 * ꤹ粽ǤȤޤơΥץͤ 0  1 ϰ
 * ΤƱ٤ˤʤ褦Ʒ׻ޤ
 * ͥ 1  2 ѤݡȤƤޤ
 */

#ifndef ABS
#define ABS(a) ((a) >= 0 ? (a) : -(a))
#endif
#define MAXSAMPLEBYTE 2
#define ENDIAN16(type, s) ((type)((((unsigned short)(s)) >> 8) | (((unsigned short)(s)) << 8)))
#define NEED_BUFFER (MAXSAMPLEBYTE * (sear_try_samples + 4))

#ifdef DEBUG
#ifdef sun
int fprintf(FILE *, const char *, ...);
int printf(const char *, ...);
void perror(const char *);
#endif
#endif

extern double drand48(void);

static double sear_sigbyte(const char* buff_data, int nchannels)
{
    const char* p = buff_data;
    double d = 0.0;
    int i;

    if(nchannels == 1)
    {
	for(i = 0; i < sear_try_samples; i++)
	{
	    int t;

	    t = (int)p[i + 1] - (int)p[i];
	    d += ABS(t);
	}
    }
    else
    {
	for(i = 0; i < sear_try_samples; i += 2)
	{
	    int t1, t2;

	    t1 = (int)p[i + 2] - (int)p[i];
	    t2 = (int)p[i + 3] - (int)p[i + 1];
	    d += ABS(t1);
	    d += ABS(t2);
	}
    }

    return d / 256.0;
}

static double sear_unsigbyte(const char* buff_data, int nchannels)
{
    const unsigned char* p = (const unsigned char *)buff_data;
    double d = 0.0;
    int i;

    if(nchannels == 1)
    {
	for(i = 0; i < sear_try_samples; i++)
	{
	    int t;

	    t = (int)p[i + 1] - (int)p[i];
	    d += ABS(t);
	}
    }
    else
    {
	for(i = 0; i < sear_try_samples; i += 2)
	{
	    int t1, t2;

	    t1 = (int)p[i + 2] - (int)p[i];
	    t2 = (int)p[i + 3] - (int)p[i + 1];
	    d += ABS(t1);
	    d += ABS(t2);
	}
    }

    return d / 256.0;
}

static double sear_ulaw(const char* buff_data, int nchannels)
{
    const char* p = buff_data;
    double d = 0.0;
    int i;

    if(nchannels == 1)
    {
	for(i = 0; i < sear_try_samples; i++)
	{
	    int t;

	    t = (int)AUDIO_U2C(p[i + 1]) - (int)AUDIO_U2C(p[i]);
	    d += ABS(t);
	}
    }
    else
    {
	for(i = 0; i < sear_try_samples; i += 2)
	{
	    int t1, t2;

	    t1 = (int)AUDIO_U2C(p[i + 2]) - (int)AUDIO_U2C(p[i]);
	    t2 = (int)AUDIO_U2C(p[i + 3]) - (int)AUDIO_U2C(p[i + 1]);
	    d += ABS(t1);
	    d += ABS(t2);
	}
    }

    return d / 256.0;
}

static double sear_alaw(const char* buff_data, int nchannels)
{
    const char* p = buff_data;
    double d = 0.0;
    int i;

    if(nchannels == 1)
    {
	for(i = 0; i < sear_try_samples; i++)
	{
	    int t;

	    t = (int)AUDIO_A2C(p[i + 1]) - (int)AUDIO_A2C(p[i]);
	    d += ABS(t);
	}
    }
    else
    {
	for(i = 0; i < sear_try_samples; i += 2)
	{
	    int t1, t2;

	    t1 = (int)AUDIO_A2C(p[i + 2]) - (int)AUDIO_A2C(p[i]);
	    t2 = (int)AUDIO_A2C(p[i + 3]) - (int)AUDIO_A2C(p[i + 1]);
	    d += ABS(t1);
	    d += ABS(t2);
	}
    }

    return d / 256.0;
}

static double sear_sigwardb(const char* buff_data, int nchannels)
{
    const short* p = (const short *)buff_data;
    double d = 0.0;
    int i;

    if(nchannels == 1)
    {
	for(i = 0; i < sear_try_samples; i++)
	{
	    int t;
#ifdef BIGENDIAN
	    t = (int)p[i + 1] - (int)p[i];
#else
	    t = (int)ENDIAN16(short, p[i + 1]) - (int)ENDIAN16(short, p[i]);
#endif
	    d += ABS(t);
	}
    }
    else
    {
	for(i = 0; i < sear_try_samples; i += 2)
	{
	    int t1, t2;
#ifdef BIGENDIAN
	    t1 = (int)p[i + 2] - (int)p[i];
	    t2 = (int)p[i + 3] - (int)p[i + 1];
#else
	    t1 = (int)ENDIAN16(short, p[i + 2]) - (int)ENDIAN16(short, p[i]);
	    t2 = (int)ENDIAN16(short, p[i + 3]) - (int)ENDIAN16(short, p[i + 1]);
#endif
	    d += ABS(t1);
	    d += ABS(t2);
	}
    }

    return d / 65536.0;
}

static double sear_unsigwardb(const char* buff_data, int nchannels)
{
    const unsigned short* p = (const unsigned short *)buff_data;
    double d = 0.0;
    int i;

    if(nchannels == 1)
    {
	for(i = 0; i < sear_try_samples; i++)
	{
	    int t;
#ifdef BIGENDIAN
	    t = (int)ENDIAN16(unsigned short, p[i + 1])
		- (int)ENDIAN16(unsigned short, p[i]);
#else
	    t = (int)p[i + 1] - (int)p[i];
#endif
	    d += ABS(t);
	}
    }
    else
    {
	for(i = 0; i < sear_try_samples; i += 2)
	{
	    int t1, t2;
#ifdef BIGENDIAN
	    t1 = (int)p[i + 2] - (int)p[i];
	    t2 = (int)p[i + 3] - (int)p[i + 1];
#else
	    t1 = (int)ENDIAN16(unsigned short, p[i + 2])
		- (int)ENDIAN16(unsigned short, p[i]);
	    t2 = (int)ENDIAN16(unsigned short, p[i + 3])
		- (int)ENDIAN16(unsigned short, p[i + 1]);
#endif
	    d += ABS(t1);
	    d += ABS(t2);
	}
    }

    return d / 65536.0;
}

static double sear_sigwardl(const char* buff_data, int nchannels)
{
    const short* p = (const short *)buff_data;
    double d = 0.0;
    int i;

    if(nchannels == 1)
    {
	for(i = 0; i < sear_try_samples; i++)
	{
	    int t;
#ifdef BIGENDIAN
	    t = (int)ENDIAN16(short, p[i + 1]) - (int)ENDIAN16(short, p[i]);
#else
	    t = (int)p[i + 1] - (int)p[i];
#endif
	    d += ABS(t);
	}
    }
    else
    {
	for(i = 0; i < sear_try_samples; i += 2)
	{
	    int t1, t2;
#ifdef BIGENDIAN
	    t1 = (int)ENDIAN16(short, p[i + 2]) - (int)ENDIAN16(short, p[i]);
	    t2 = (int)ENDIAN16(short, p[i + 3]) - (int)ENDIAN16(short, p[i + 1]);
#else
	    t1 = (int)p[i + 2] - (int)p[i];
	    t2 = (int)p[i + 3] - (int)p[i + 1];
#endif
	    d += ABS(t1);
	    d += ABS(t2);
	}
    }

    return d / 65536.0;
}

static double sear_unsigwardl(const char* buff_data, int nchannels)
{
    const unsigned short* p = (const unsigned short *)buff_data;
    double d = 0.0;
    int i;

    if(nchannels == 1)
    {
	for(i = 0; i < sear_try_samples; i++)
	{
	    int t;
#ifdef BIGENDIAN
	    t = (int)ENDIAN16(unsigned short, p[i + 1])
		- (int)ENDIAN16(unsigned short, p[i]);
#else
	    t = (int)p[i + 1] - (int)p[i];
#endif
	    d += ABS(t);
	}
    }
    else
    {
	for(i = 0; i < sear_try_samples; i += 2)
	{
	    int t1, t2;
#ifdef BIGENDIAN
	    t1 = (int)ENDIAN16(unsigned short, p[i + 2])
		- (int)ENDIAN16(unsigned short, p[i]);
	    t2 = (int)ENDIAN16(unsigned short, p[i + 3])
		- (int)ENDIAN16(unsigned short, p[i + 1]);
#else
	    t1 = (int)p[i + 2] - (int)p[i];
	    t2 = (int)p[i + 3] - (int)p[i + 1];
#endif
	    d += ABS(t1);
	    d += ABS(t2);
	}
    }

    return d / 65536.0;
}

static struct
{
    double (* tester)(const char *, int);
    int encoding;
    double d1, d2;
} sear_testers[] =
{
    {sear_sigbyte,	AENC_SIGBYTE,    0.0, 0.0},
    {sear_unsigbyte,	AENC_UNSIGBYTE,  0.0, 0.0},
    {sear_ulaw,		AENC_G711_ULAW,  0.0, 0.0},
    {sear_alaw,		AENC_G711_ALAW,  0.0, 0.0},
    {sear_sigwardb,	AENC_SIGWORDB,   0.0, 0.0},
    {sear_unsigwardb,	AENC_UNSIGWORDB, 0.0, 0.0},
    {sear_sigwardl,	AENC_SIGWORDL,   0.0, 0.0},
    {sear_unsigwardl,	AENC_UNSIGWORDL, 0.0, 0.0},
    {NULL,		AENC_UNDEFINED,  0.0, 0.0},
};

static void sear_start_init(void)
{
    int i;
    for(i = 0; sear_testers[i].tester != NULL; i++)
	sear_testers[i].d1 = sear_testers[i].d2 = 0.0;
}

static void sear_guess(int* idx, int* chan)
{
    int i;
    double d, dmin;

    dmin = sear_testers[0].d1;
    *chan = 1;
    *idx = 0;
#ifdef DEBUG
	printf("%f\t%s/1\n",
	       dmin, AENC_NAME(sear_testers[0].encoding));
#endif
    for(i = 1; sear_testers[i].tester != NULL; i++)
    {
	d = sear_testers[i].d1;

#ifdef DEBUG
	printf("%f\t%s/1\n",
	       d, AENC_NAME(sear_testers[i].encoding));
#endif
	if(d < dmin)
	{
	    dmin = d;
	    *idx = i;
	}
    }
    for(i = 0; sear_testers[i].tester != NULL; i++)
    {
	d = sear_testers[i].d2;

#ifdef DEBUG
	printf("%f\t%s/2\n",
	       d, AENC_NAME(sear_testers[i].encoding));
#endif
	if(d < dmin)
	{
	    dmin = d;
	    *chan = 2;
	    *idx = i;
	}
    }
}

void sear(int fd, long start_seek, long size, int* enc, int* chan)
{
    long seek_pos;
    char* buff;
    int sear_idx, i;

    if(size < NEED_BUFFER)
    {
	*enc = AENC_G711_ULAW;
	*chan = 1;
	return;
    }

    buff = (char *)alloca(NEED_BUFFER);
    sear_start_init();
    for(i = 0; i < sear_retry_seek; i++)
    {
	int j;

	seek_pos = start_seek + (long)(drand48() * (size - NEED_BUFFER));
	seek_pos -= (seek_pos & 1);
	lseek(fd, seek_pos, SEEK_SET);
	read(fd, buff, NEED_BUFFER);

	for(j = 0; sear_testers[j].tester != NULL; j++)
	{
	    sear_testers[j].d1 += sear_testers[j].tester(buff, 1);
	    sear_testers[j].d2 += sear_testers[j].tester(buff, 2);
	}
    }
    sear_guess(&sear_idx, chan);
    *enc = sear_testers[sear_idx].encoding;
}

void sear_tty(int* enc, int* chan)
{
    int i;
    char* buff;
    int sear_idx;

    *enc = AENC_G711_ULAW;
    *chan = 1;

    buff = (char *)alloca(NEED_BUFFER);
    sear_start_init();
    for(i = 0; i < sear_retry_seek; i++)
    {
	int j;

	if(read(0, buff, NEED_BUFFER) != NEED_BUFFER)
	    break;

	for(j = 0; sear_testers[j].tester != NULL; j++)
	{
	    sear_testers[j].d1 += sear_testers[j].tester(buff, 1);
	    sear_testers[j].d2 += sear_testers[j].tester(buff, 2);
	}
    }
    sear_guess(&sear_idx, chan);
    *enc = sear_testers[sear_idx].encoding;
}


void sear_string(const char* buff, long buff_size, int* enc, int* chan)
{
    int i, sear_idx;
    long pos;

    if(buff_size < NEED_BUFFER)
    {
	*enc = AENC_G711_ULAW;
	*chan = 1;
	return;
    }

    sear_start_init();
    for(i = 0; i < sear_retry_seek; i++)
    {
	int j;
	pos = (long)(drand48() * (buff_size - NEED_BUFFER));
	pos -= (pos & 1);

	for(j = 0; sear_testers[j].tester != NULL; j++)
	{
	    sear_testers[j].d1 += sear_testers[j].tester(buff + pos, 1);
	    sear_testers[j].d2 += sear_testers[j].tester(buff + pos, 2);
	}
    }
    sear_guess(&sear_idx, chan);
    *enc = sear_testers[sear_idx].encoding;
}

#ifdef SEAR_MAIN
#include <fcntl.h>
#include <sys/stat.h>

void main(int argc, char** argv)
{
    int i;
    int enc, chan;

    if(argc <= 1)
    {
	sear_tty(&enc, &chan);
	printf("%s/%d\n", AENC_NAME(enc), chan);
	exit(0);
    }

    for(i = 1; i < argc; i++)
    {
	int fd;
	struct stat buf;

	if((fd = open(argv[i], O_RDONLY)) < 0)
	{
	    perror(argv[i]);
	    continue;
	}

	if(fstat(fd, &buf) < 0)
	{
	    perror(argv[i]);
	    close(fd);
	    continue;
	}

	if(!S_ISREG(buf.st_mode))
	{
	    fprintf(stderr, "%s: Not a regular file.\n", argv[i]);
	    close(fd);
	    continue;
	}

	sear(fd, 0, buf.st_size, &enc, &chan);
	printf("%s: %s/%d\n",
	       argv[i], AENC_NAME(enc), chan);
	close(fd);
    }
}
#endif
