#include <stdio.h>
#include <string.h>

#define CHAR_PTR		 char*
#define UNSIGNED_CHAR_PTR	 unsigned char*
#define SHORT_PTR		 short*
#define UNSIGNED_SHORT_PTR	 unsigned short*
#define CONST_CHAR_PTR		 const char*
#define CONST_UNSIGNED_CHAR_PTR	 const unsigned char*
#define CONST_SHORT_PTR		 const short*
#define CONST_UNSIGNED_SHORT_PTR const unsigned short*

#include "usffd.h"
#include "soundex.h"
#include "audio_cnv.h"

#ifdef sun
#ifndef memmove
#define memmove(a, b, n) bcopy(b, a, n)
#endif /* memmove */
#endif /* sun */


#define ENDIAN16(type, s) ((type)((((unsigned short)(s)) >> 8) | (((unsigned short)(s)) << 8)))
#define MEAN(x1, x2) (((long)(x1) + (long)(x2)) / 2)

static void swap_endian(const void* source, void* target, int nbytes)
{
    unsigned short* s;
    unsigned short* t;
    unsigned short* last;
    int n;

    s = (unsigned short *)source;
    t = (unsigned short *)target;
    n = nbytes / 2;
    last = s + n - 3;
    while(s < last)
    {
	*t++ = ENDIAN16(unsigned short, s[0]);
	*t++ = ENDIAN16(unsigned short, s[1]);
	*t++ = ENDIAN16(unsigned short, s[2]);
	*t++ = ENDIAN16(unsigned short, s[3]);
	s += 4;
    }
    while(s < last + 3)
    {
	*t++ = ENDIAN16(unsigned short, *s);
	s++;
    }
}

/************************************************************
 * FIXME 衢ǥХåɬס
 ***********************************************************/
/* դ1Х		  => դ1Х */
void sig1_to_sig1(CONST_CHAR_PTR source,
		  CHAR_PTR	 target,
		  int		 nbytes)
{
    memmove(target, source, nbytes);
}

/* ʤ1Х		  => դ1Х */
void unsig1_to_sig1(CONST_UNSIGNED_CHAR_PTR source,
		    CHAR_PTR		    target,
		    int			    nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
	target[i] = (source[i] ^ 0x80);
}

/* ˡ§			  => դ1Х */
void ulaw_to_sig1(CONST_CHAR_PTR source,
		  CHAR_PTR	 target,
		  int		 nbytes)
{
    convert_u2c(source, target, nbytes);
}

/* ˡ§			  => դ1Х */
void alaw_to_sig1(CONST_CHAR_PTR source,
		  CHAR_PTR	 target,
		  int		 nbytes)
{
    convert_a2c(source, target, nbytes);
}

/* դ2Х(big-endian)	  => դ1Х */
void sig2b_to_sig1(CONST_SHORT_PTR source,
		   CHAR_PTR	   target,
		   int		   nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = source[i] / 256;
#else
	target[i] = ENDIAN16(short, source[i]) / 256;
#endif
}

/* ʤ2Х(big-endian)	  => դ1Х */
void unsig2b_to_sig1(CONST_UNSIGNED_SHORT_PTR source,
		     CHAR_PTR		      target,
		     int		      nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
    {
#ifdef BIGENDIAN
	target[i] = (source[i] ^ 0x8000) / 256;
#else
	short s = source[i] ^ 0x8000;
	target[i] = ENDIAN16(short, s) / 256;
#endif
    }
}

/* դ2Х(little-endian) => դ1Х */
void sig2l_to_sig1(CONST_SHORT_PTR source,
		   CHAR_PTR	   target,
		   int		   nbytes)
{
    int i, n = nbytes / 2;

    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = ENDIAN16(short, source[i]) / 256;
#else
	target[i] = source[i] / 256;
#endif
}

/* ʤ2Х(little-endian) => դ1Х */
void unsig2l_to_sig1(CONST_UNSIGNED_SHORT_PTR source,
		     CHAR_PTR		      target,
		     int		      nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
    {
#ifdef BIGENDIAN
	short s = ENDIAN16(short, source[i]) ^ 0x8000;
#else
	short s = source[i] ^ 0x8000;
#endif
	target[i] = s / 256;
    }
}

/* դ1Х		  => ʤ1Х */
void sig1_to_unsig1(CONST_CHAR_PTR    source,
		    UNSIGNED_CHAR_PTR target,
		    int		      nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
	target[i] = (source[i] ^ 0x80);
}

/* ʤ1Х		  => ʤ1Х */
void unsig1_to_unsig1(CONST_UNSIGNED_CHAR_PTR source,
		      UNSIGNED_CHAR_PTR	      target,
		      int		      nbytes)
{
    memmove(target, source, nbytes);
}

/* ˡ§			  => ʤ1Х */
void ulaw_to_unsig1(CONST_CHAR_PTR    source,
		    UNSIGNED_CHAR_PTR target,
		    int		      nbytes)
{
    int i;

    for(i = 0; i < nbytes; i++)
	target[i] = (AUDIO_U2C(source[i]) ^ 0x80);
}

/* ˡ§			  => ʤ1Х */
void alaw_to_unsig1(CONST_CHAR_PTR    source,
		    UNSIGNED_CHAR_PTR target,
		    int		      nbytes)
{
    int i;

    for(i = 0; i < nbytes; i++)
	target[i] = (AUDIO_A2C(source[i]) ^ 0x80);
}

/* դ2Х(big-endian)	  => ʤ1Х */
void sig2b_to_unsig1(CONST_SHORT_PTR   source,
		     UNSIGNED_CHAR_PTR target,
		     int	       nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = ((source[i] / 256) ^ 0x80);
#else
	target[i] = ((ENDIAN16(unsigned short, source[i]) / 256) ^ 0x80);
#endif
}

/* ʤ2Х(big-endian)	  => ʤ1Х */
void unsig2b_to_unsig1(CONST_UNSIGNED_SHORT_PTR source,
		       UNSIGNED_CHAR_PTR	target,
		       int			nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = source[i] / 256;
#else
	target[i] = ENDIAN16(unsigned short, source[i]) / 256;
#endif
}

/* դ2Х(little-endian) => ʤ1Х */
void sig2l_to_unsig1(CONST_SHORT_PTR   source,
		     UNSIGNED_CHAR_PTR target,
		     int	       nbytes)
{
    int i, n = nbytes / 2;

    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = ((ENDIAN16(unsigned short, source[i]) / 256) ^ 0x80);
#else
	target[i] = ((source[i] / 256) ^ 0x80);
#endif
}

/* ʤ2Х(little-endian) => ʤ1Х */
void unsig2l_to_unsig1(CONST_UNSIGNED_SHORT_PTR source,
		       UNSIGNED_CHAR_PTR	target,
		       int			nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = ENDIAN16(unsigned short, source[i]) / 256;
#else
	target[i] = source[i] / 256;
#endif
}

/* դ1Х		  => ʤ2Х(big-endian) */
void sig1_to_unsig2b(CONST_CHAR_PTR   source,
		   UNSIGNED_SHORT_PTR target,
		   int		      nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	target[i] = (source[i] ^ 0x80) * 256;
#else
	unsigned short s = (source[i] ^ 0x80) * 256;
	target[i] = ENDIAN16(unsigned short, s);
#endif
    }
}

/* ʤ1Х		  => ʤ2Х(big-endian) */
void unsig1_to_unsig2b(CONST_UNSIGNED_CHAR_PTR source,
		     UNSIGNED_SHORT_PTR	       target,
		     int		       nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	target[i] = source[i] * 256;
#else
	unsigned short s = source[i] * 256;
	target[i] = ENDIAN16(unsigned short, s);
#endif
    }
}

/* ˡ§			  => ʤ2Х(big-endian) */
void ulaw_to_unsig2b(CONST_CHAR_PTR   source,
		   UNSIGNED_SHORT_PTR target,
		   int		      nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	target[i] = (AUDIO_U2S(source[i]) ^ 0x8000);
#else
	unsigned short s = (AUDIO_U2S(source[i]) ^ 0x8000);
	target[i] = ENDIAN16(unsigned short, s);
#endif
    }
}

/* ˡ§			  => ʤ2Х(big-endian) */
void alaw_to_unsig2b(CONST_CHAR_PTR   source,
		   UNSIGNED_SHORT_PTR target,
		   int		      nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	target[i] = (AUDIO_A2S(source[i]) ^ 0x8000);
#else
	unsigned short s = (AUDIO_A2S(source[i]) ^ 0x8000);
	target[i] = ENDIAN16(unsigned short, s);
#endif
    }
}

/* դ2Х(big-endian)	  => ʤ2Х(big-endian) */
void sig2b_to_unsig2b(CONST_SHORT_PTR  source,
		    UNSIGNED_SHORT_PTR target,
		    int		       nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = (source[i] ^ 0x8000);
#else
	target[i] = (source[i] ^ 0x80);
#endif
}

/* ʤ2Х(big-endian)	  => ʤ2Х(big-endian) */
void unsig2b_to_unsig2b(CONST_UNSIGNED_SHORT_PTR source,
		      UNSIGNED_SHORT_PTR	 target,
		      int			 nbytes)
{
    memmove(target, source, nbytes);
}

/* դ2Х(little-endian) => ʤ2Х(big-endian) */
void sig2l_to_unsig2b(CONST_SHORT_PTR  source,
		    UNSIGNED_SHORT_PTR target,
		    int		       nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = ((ENDIAN16(unsigned short, source[i]) ^ 0x8000));
#else
	target[i] = ((ENDIAN16(unsigned short, source[i]) ^ 0x80));
#endif
}

/* ʤ2Х(little-endian) => ʤ2Х(big-endian) */
void unsig2l_to_unsig2b(CONST_UNSIGNED_SHORT_PTR source,
		      UNSIGNED_SHORT_PTR	 target,
		      int			 nbytes)
{
    swap_endian(source, target, nbytes);
}

/* դ1Х		  => ʤ2Х(little-endian) */
void sig1_to_unsig2l(CONST_CHAR_PTR   source,
		   UNSIGNED_SHORT_PTR target,
		   int		      nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	unsigned short s = (source[i] ^ 0x80) * 256;
	target[i] = ENDIAN16(unsigned short, s);
#else
	target[i] = (source[i] ^ 0x80) * 256;
#endif
    }
}

/* ʤ1Х		  => ʤ2Х(little-endian) */
void unsig1_to_unsig2l(CONST_UNSIGNED_CHAR_PTR source,
		     UNSIGNED_SHORT_PTR	       target,
		     int		       nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	unsigned short s = source[i] * 256;
	target[i] = ENDIAN16(unsigned short, s);
#else
	target[i] = source[i] * 256;
#endif
    }
}

/* ˡ§			  => ʤ2Х(little-endian) */
void ulaw_to_unsig2l(CONST_CHAR_PTR   source,
		   UNSIGNED_SHORT_PTR target,
		   int		      nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	unsigned short s = (AUDIO_U2S(source[i]) ^ 0x8000);
	target[i] = ENDIAN16(unsigned short, s);
#else
	target[i] = (AUDIO_U2S(source[i]) ^ 0x8000);
#endif
    }
}

/* ˡ§			  => ʤ2Х(little-endian) */
void alaw_to_unsig2l(CONST_CHAR_PTR   source,
		   UNSIGNED_SHORT_PTR target,
		   int		      nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	unsigned short s = (AUDIO_A2S(source[i]) ^ 0x8000);
	target[i] = ENDIAN16(unsigned short, s);
#else
	target[i] = (AUDIO_A2S(source[i]) ^ 0x8000);
#endif
    }
}

/* դ2Х(big-endian)	  => ʤ2Х(little-endian) */
void sig2b_to_unsig2l(CONST_SHORT_PTR  source,
		    UNSIGNED_SHORT_PTR target,
		    int		       nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = (source[i] ^ 0x80);
#else
	target[i] = (source[i] ^ 0x8000);
#endif
}

/* ʤ2Х(big-endian)	  => ʤ2Х(little-endian) */
void unsig2b_to_unsig2l(CONST_UNSIGNED_SHORT_PTR source,
		      UNSIGNED_SHORT_PTR	 target,
		      int			 nbytes)
{
    swap_endian(source, target, nbytes);
}

/* դ2Х(little-endian) => ʤ2Х(little-endian) */
void sig2l_to_unsig2l(CONST_SHORT_PTR  source,
		    UNSIGNED_SHORT_PTR target,
		    int		       nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = ((ENDIAN16(unsigned short, source[i]) ^ 0x80));
#else
	target[i] = ((ENDIAN16(unsigned short, source[i]) ^ 0x8000));
#endif
}

/* ʤ2Х(little-endian) => ʤ2Х(little-endian) */
void unsig2l_to_unsig2l(CONST_UNSIGNED_SHORT_PTR source,
		      UNSIGNED_SHORT_PTR	 target,
		      int			 nbytes)
{
    memmove(target, source, nbytes);
}

/************************************************************
 * FIXME ޤǡǥХåɬס
 ***********************************************************/

/* դ1Х		  => ˡ§ */
void sig1_to_ulaw(CONST_CHAR_PTR source,
		  CHAR_PTR	 target,
		  int		 nbytes)
{
    convert_c2u(source, target, nbytes);
}

/* ʤ1Х		  => ˡ§ */
void unsig1_to_ulaw(CONST_UNSIGNED_CHAR_PTR source,
		    CHAR_PTR		    target,
		    int			    nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
	target[i] = AUDIO_C2U(source[i] ^ 0x80);
}

/* ˡ§			  => ˡ§ */
void ulaw_to_ulaw(CONST_CHAR_PTR source,
		  CHAR_PTR	 target,
		  int		 nbytes)
{
    memmove(target, source, nbytes);
}

/* ˡ§			  => ˡ§ */
void alaw_to_ulaw(CONST_CHAR_PTR source,
		  CHAR_PTR	 target,
		  int		 nbytes)
{
    convert_a2u(source, target, nbytes);
}

/* դ2Х(big-endian)	  => ˡ§ */
void sig2b_to_ulaw(CONST_SHORT_PTR source,
		   CHAR_PTR	   target,
		   int		   nbytes)
{
#ifdef BIGENDIAN
    convert_s2u(source, target, nbytes / 2);
#else
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
	target[i] = AUDIO_S2U(ENDIAN16(short, source[i]));
#endif
}

/* ʤ2Х(big-endian)	  => ˡ§ */
void unsig2b_to_ulaw(CONST_UNSIGNED_SHORT_PTR source,
		     CHAR_PTR		      target,
		     int		      nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
    {
#ifdef BIGENDIAN
	target[i] = AUDIO_S2U(source[i] ^ 0x8000);
#else
	short s = source[i] ^ 0x8000;
	target[i] = AUDIO_S2U(ENDIAN16(short, s));
#endif
    }
}

/* դ2Х(little-endian) => ˡ§ */
void sig2l_to_ulaw(CONST_SHORT_PTR source,
		   CHAR_PTR	   target,
		   int		   nbytes)
{
#ifdef BIGENDIAN
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
	target[i] = AUDIO_S2U(ENDIAN16(short, source[i]));
#else
    convert_s2u(source, target, nbytes / 2);
#endif
}

/* ʤ2Х(little-endian) => ˡ§ */
void unsig2l_to_ulaw(CONST_UNSIGNED_SHORT_PTR source,
		     CHAR_PTR		      target,
		     int		      nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
    {
#ifdef BIGENDIAN
	short s = ENDIAN16(short, source[i]) ^ 0x8000;
#else
	short s = source[i] ^ 0x8000;
#endif
	target[i] = AUDIO_S2U(s);
    }
}

/* դ1Х		  => ˡ§ */
void sig1_to_alaw(CONST_CHAR_PTR source,
		  CHAR_PTR	 target,
		  int		 nbytes)
{
    convert_c2a(source, target, nbytes);
}

/* ʤ1Х		  => ˡ§ */
void unsig1_to_alaw(CONST_UNSIGNED_CHAR_PTR source,
		       CHAR_PTR		       target,
		       int		       nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
	target[i] = AUDIO_C2A(source[i] ^ 0x80);
}

/* ˡ§			  => ˡ§ */
void ulaw_to_alaw(CONST_CHAR_PTR source,
		  CHAR_PTR	 target,
		  int		 nbytes)
{
    convert_u2a(source, target, nbytes);
}

/* ˡ§			  => ˡ§ */
void alaw_to_alaw(CONST_CHAR_PTR source,
		  CHAR_PTR	 target,
		  int		 nbytes)
{
    memmove(target, source, nbytes);
}

/* դ2Х(big-endian)	  => ˡ§ */
void sig2b_to_alaw(CONST_SHORT_PTR source,
		   CHAR_PTR	   target,
		   int		   nbytes)
{
#ifdef BIGENDIAN
	convert_s2a(source, target, nbytes / 2);
#else
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
	target[i] = AUDIO_S2A(ENDIAN16(short, source[i]));
#endif
}

/* ʤ2Х(big-endian)	  => ˡ§ */
void unsig2b_to_alaw(CONST_UNSIGNED_SHORT_PTR source,
		     CHAR_PTR		      target,
		     int		      nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
    {
#ifdef BIGENDIAN
	target[i] = AUDIO_S2A(source[i] ^ 0x8000);
#else
	short s = source[i] ^ 0x8000;
	target[i] = AUDIO_S2A(ENDIAN16(short, s));
#endif
    }
}

/* դ2Х(little-endian) => ˡ§ */
void sig2l_to_alaw(CONST_SHORT_PTR source,
		   CHAR_PTR	   target,
		   int		   nbytes)
{
#ifdef BIGENDIAN
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
	target[i] = AUDIO_S2A(ENDIAN16(short, source[i]));
#else
    convert_s2a(source, target, nbytes / 2);
#endif
}

/* ʤ2Х(little-endian) => ˡ§ */
void unsig2l_to_alaw(CONST_UNSIGNED_SHORT_PTR source,
		     CHAR_PTR		      target,
		     int		      nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
    {
#ifdef BIGENDIAN
	short s = ENDIAN16(short, source[i]) ^ 0x8000;
#else
	short s = source[i] ^ 0x8000;
#endif
	target[i] = AUDIO_S2A(s);
    }
}

/* դ1Х		  => դ2Х(big-endian) */
void sig1_to_sig2b(CONST_CHAR_PTR source,
		   SHORT_PTR	  target,
		   int		  nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	target[i] = source[i] * 256;
#else
	short s = source[i] * 256;
	target[i] = ENDIAN16(short, s);
#endif
    }
}

/* ʤ1Х		  => դ2Х(big-endian) */
void unsig1_to_sig2b(CONST_UNSIGNED_CHAR_PTR source,
		     SHORT_PTR		     target,
		     int		     nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	target[i] = (source[i] ^ 0x80) * 256;
#else
	short s = (source[i] ^ 0x80) * 256;
	target[i] = ENDIAN16(short, s);
#endif
    }
}

/* ˡ§			  => դ2Х(big-endian) */
void ulaw_to_sig2b(CONST_CHAR_PTR source,
		   SHORT_PTR	  target,
		   int		  nbytes)
{
#ifdef BIGENDIAN
    convert_u2s(source, target, nbytes);
#else
    int i;
    for(i = 0; i < nbytes; i++)
    {
	short s = AUDIO_U2S(source[i]);
	target[i] = ENDIAN16(short, s);
    }
#endif /* BIGENDIAN */
}

/* ˡ§			  => դ2Х(big-endian) */
void alaw_to_sig2b(CONST_CHAR_PTR source,
		   SHORT_PTR	  target,
		   int		  nbytes)
{
#ifdef BIGENDIAN
    convert_a2s(source, target, nbytes);
#else
    int i;
    for(i = 0; i < nbytes; i++)
    {
	short s = AUDIO_A2S(source[i]);
	target[i] = ENDIAN16(short, s);
    }
#endif
}

/* դ2Х(big-endian)	  => դ2Х(big-endian) */
void sig2b_to_sig2b(CONST_SHORT_PTR source,
		    SHORT_PTR	    target,
		    int		    nbytes)
{
  memmove(target, source, nbytes);
}

/* ʤ2Х(big-endian)	  => դ2Х(big-endian) */
void unsig2b_to_sig2b(CONST_UNSIGNED_SHORT_PTR source,
		      SHORT_PTR		       target,
		      int		       nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = (source[i] ^ 0x8000);
#else
	target[i] = (source[i] ^ 0x80);
#endif
}

/* դ2Х(little-endian) => դ2Х(big-endian) */
void sig2l_to_sig2b(CONST_SHORT_PTR source,
		    SHORT_PTR	    target,
		    int		    nbytes)
{
    swap_endian(source, target, nbytes);
}

/* ʤ2Х(little-endian) => դ2Х(big-endian) */
void unsig2l_to_sig2b(CONST_UNSIGNED_SHORT_PTR source,
		      SHORT_PTR		       target,
		      int		       nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
    {
#ifdef BIGENDIAN
	short s = (source[i] ^ 0x80);
#else
	short s = (source[i] ^ 0x8000);
#endif
	target[i] = ENDIAN16(short, s);
    }
}

/* դ1Х		  => դ2Х(little-endian) */
void sig1_to_sig2l(CONST_CHAR_PTR source,
		   SHORT_PTR	  target,
		   int		  nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	short s = source[i] * 256;
	target[i] = ENDIAN16(short, s);
#else
	target[i] = source[i] * 256;
#endif
    }
}

/* ʤ1Х		  => դ2Х(little-endian) */
void unsig1_to_sig2l(CONST_UNSIGNED_CHAR_PTR source,
		     SHORT_PTR		     target,
		     int		     nbytes)
{
    int i;
    for(i = 0; i < nbytes; i++)
    {
#ifdef BIGENDIAN
	short s = (source[i] ^ 0x80) * 256;
	target[i] = ENDIAN16(short, s);
#else
	target[i] = (source[i] ^ 0x80) * 256;
#endif
    }
}

/* ˡ§			  => դ2Х(little-endian) */
void ulaw_to_sig2l(CONST_CHAR_PTR source,
		   SHORT_PTR	  target,
		   int		  nbytes)
{
#ifdef BIGENDIAN
    int i;
    for(i = 0; i < nbytes; i++)
    {
	short s = AUDIO_U2S(source[i]);
	target[i] = ENDIAN16(short, s);
    }
#else
    convert_u2s(source, target, nbytes);
#endif

}

/* ˡ§			  => դ2Х(little-endian) */
void alaw_to_sig2l(CONST_CHAR_PTR source,
		   SHORT_PTR	  target,
		   int		  nbytes)
{
#ifdef BIGENDIAN
    int i;
    for(i = 0; i < nbytes; i++)
    {
	short s = AUDIO_A2S(source[i]);
	target[i] = ENDIAN16(short, s);
    }
#else
    convert_a2s(source, target, nbytes);
#endif
}

/* դ2Х(big-endian)	  => դ2Х(little-endian) */
void sig2b_to_sig2l(CONST_SHORT_PTR source,
		    SHORT_PTR	    target,
		    int		    nbytes)
{
    swap_endian(source, target, nbytes);
}

/* ʤ2Х(big-endian)	  => դ2Х(little-endian) */
void unsig2b_to_sig2l(CONST_UNSIGNED_SHORT_PTR source,
		      SHORT_PTR		       target,
		      int		       nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
    {
#ifdef BIGENDIAN
	short s = (source[i] ^ 0x8000);
#else
	short s = (source[i] ^ 0x80);
#endif
	target[i] = ENDIAN16(short, s);
    }
}

/* դ2Х(little-endian) => դ2Х(little-endian) */
void sig2l_to_sig2l(CONST_SHORT_PTR source,
		    SHORT_PTR	    target,
		    int		    nbytes)
{
    memmove(target, source, nbytes);
}

/* ʤ2Х(little-endian) => դ2Х(little-endian) */
void unsig2l_to_sig2l(CONST_UNSIGNED_SHORT_PTR source,
		      SHORT_PTR		       target,
		      int		       nbytes)
{
    int i, n = nbytes / 2;
    for(i = 0; i < n; i++)
#ifdef BIGENDIAN
	target[i] = (source[i] ^ 0x80);
#else
	target[i] = (source[i] ^ 0x8000);
#endif
}

void (* select_converter(long encoding_from, long encoding_to, 
			 int* from_bpu, int* to_bpu))
			(void *, void *, int)
{
    if(from_bpu != NULL)
	*from_bpu = usffd_encoding_sample_size[encoding_from];
    if(to_bpu != NULL)
	*to_bpu	  = usffd_encoding_sample_size[encoding_to];

    switch(encoding_to)
    {
      case USFFD_SIGBYTE:
	switch(encoding_from)
	{
	  case USFFD_SIGBYTE:
	    return (void (*)(void *, void *, int))sig1_to_sig1;
	  case USFFD_UNSIGBYTE:
	    return (void (*)(void *, void *, int))unsig1_to_sig1;
	  case USFFD_G711_ULAW:
	    return (void (*)(void *, void *, int))ulaw_to_sig1;
	  case USFFD_G711_ALAW:
	    return (void (*)(void *, void *, int))alaw_to_sig1;
	  case USFFD_SIGWORDB:
	    return (void (*)(void *, void *, int))sig2b_to_sig1;
	  case USFFD_UNSIGWORDB:
	    return (void (*)(void *, void *, int))unsig2b_to_sig1;
	  case USFFD_SIGWORDL:
	    return (void (*)(void *, void *, int))sig2l_to_sig1;
	  case USFFD_UNSIGWORDL:
	    return (void (*)(void *, void *, int))unsig2l_to_sig1;
	}
	break;
      case USFFD_UNSIGBYTE:
	switch(encoding_from)
	{
	  case USFFD_SIGBYTE:
	    return (void (*)(void *, void *, int))sig1_to_unsig1;
	  case USFFD_UNSIGBYTE:
	    return (void (*)(void *, void *, int))unsig1_to_unsig1;
	  case USFFD_G711_ULAW:
	    return (void (*)(void *, void *, int))ulaw_to_unsig1;
	  case USFFD_G711_ALAW:
	    return (void (*)(void *, void *, int))alaw_to_unsig1;
	  case USFFD_SIGWORDB:
	    return (void (*)(void *, void *, int))sig2b_to_unsig1;
	  case USFFD_UNSIGWORDB:
	    return (void (*)(void *, void *, int))unsig2b_to_unsig1;
	  case USFFD_SIGWORDL:
	    return (void (*)(void *, void *, int))sig2l_to_unsig1;
	  case USFFD_UNSIGWORDL:
	    return (void (*)(void *, void *, int))unsig2l_to_unsig1;
	}
	break;
      case USFFD_G711_ULAW:
	switch(encoding_from)
	{
	  case USFFD_SIGBYTE:
	    return (void (*)(void *, void *, int))sig1_to_ulaw;
	  case USFFD_UNSIGBYTE:
	    return (void (*)(void *, void *, int))unsig1_to_ulaw;
	  case USFFD_G711_ULAW:
	    return (void (*)(void *, void *, int))ulaw_to_ulaw;
	  case USFFD_G711_ALAW:
	    return (void (*)(void *, void *, int))alaw_to_ulaw;
	  case USFFD_SIGWORDB:
	    return (void (*)(void *, void *, int))sig2b_to_ulaw;
	  case USFFD_UNSIGWORDB:
	    return (void (*)(void *, void *, int))unsig2b_to_ulaw;
	  case USFFD_SIGWORDL:
	    return (void (*)(void *, void *, int))sig2l_to_ulaw;
	  case USFFD_UNSIGWORDL:
	    return (void (*)(void *, void *, int))unsig2l_to_ulaw;
	}
	break;
      case USFFD_G711_ALAW:
	switch(encoding_from)
	{
	  case USFFD_SIGBYTE:
	    return (void (*)(void *, void *, int))sig1_to_alaw;
	  case USFFD_UNSIGBYTE:
	    return (void (*)(void *, void *, int))unsig1_to_alaw;
	  case USFFD_G711_ULAW:
	    return (void (*)(void *, void *, int))ulaw_to_alaw;
	  case USFFD_G711_ALAW:
	    return (void (*)(void *, void *, int))alaw_to_alaw;
	  case USFFD_SIGWORDB:
	    return (void (*)(void *, void *, int))sig2b_to_alaw;
	  case USFFD_UNSIGWORDB:
	    return (void (*)(void *, void *, int))unsig2b_to_alaw;
	  case USFFD_SIGWORDL:
	    return (void (*)(void *, void *, int))sig2l_to_alaw;
	  case USFFD_UNSIGWORDL:
	    return (void (*)(void *, void *, int))unsig2l_to_alaw;
	}
	break;
      case USFFD_SIGWORDB:
	switch(encoding_from)
	{
	  case USFFD_SIGBYTE:
	    return (void (*)(void *, void *, int))sig1_to_sig2b;
	  case USFFD_UNSIGBYTE:
	    return (void (*)(void *, void *, int))unsig1_to_sig2b;
	  case USFFD_G711_ULAW:
	    return (void (*)(void *, void *, int))ulaw_to_sig2b;
	  case USFFD_G711_ALAW:
	    return (void (*)(void *, void *, int))alaw_to_sig2b;
	  case USFFD_SIGWORDB:
	    return (void (*)(void *, void *, int))sig2b_to_sig2b;
	  case USFFD_UNSIGWORDB:
	    return (void (*)(void *, void *, int))unsig2b_to_sig2b;
	  case USFFD_SIGWORDL:
	    return (void (*)(void *, void *, int))sig2l_to_sig2b;
	  case USFFD_UNSIGWORDL:
	    return (void (*)(void *, void *, int))unsig2l_to_sig2b;
	}
	break;
      case USFFD_UNSIGWORDB:
	switch(encoding_from)
	{
	  case USFFD_SIGBYTE:
	    return (void (*)(void *, void *, int))sig1_to_unsig2b;
	  case USFFD_UNSIGBYTE:
	    return (void (*)(void *, void *, int))unsig1_to_unsig2b;
	  case USFFD_G711_ULAW:
	    return (void (*)(void *, void *, int))ulaw_to_unsig2b;
	  case USFFD_G711_ALAW:
	    return (void (*)(void *, void *, int))alaw_to_unsig2b;
	  case USFFD_SIGWORDB:
	    return (void (*)(void *, void *, int))sig2b_to_unsig2b;
	  case USFFD_UNSIGWORDB:
	    return (void (*)(void *, void *, int))unsig2b_to_unsig2b;
	  case USFFD_SIGWORDL:
	    return (void (*)(void *, void *, int))sig2l_to_unsig2b;
	  case USFFD_UNSIGWORDL:
	    return (void (*)(void *, void *, int))unsig2l_to_unsig2b;
	}
	break;
      case USFFD_SIGWORDL:
	switch(encoding_from)
	{
	  case USFFD_SIGBYTE:
	    return (void (*)(void *, void *, int))sig1_to_sig2l;
	  case USFFD_UNSIGBYTE:
	    return (void (*)(void *, void *, int))unsig1_to_sig2l;
	  case USFFD_G711_ULAW:
	    return (void (*)(void *, void *, int))ulaw_to_sig2l;
	  case USFFD_G711_ALAW:
	    return (void (*)(void *, void *, int))alaw_to_sig2l;
	  case USFFD_SIGWORDB:
	    return (void (*)(void *, void *, int))sig2b_to_sig2l;
	  case USFFD_UNSIGWORDB:
	    return (void (*)(void *, void *, int))unsig2b_to_sig2l;
	  case USFFD_SIGWORDL:
	    return (void (*)(void *, void *, int))sig2l_to_sig2l;
	  case USFFD_UNSIGWORDL:
	    return (void (*)(void *, void *, int))unsig2l_to_sig2l;
	}
	break;
      case USFFD_UNSIGWORDL:
	switch(encoding_from)
	{
	  case USFFD_SIGBYTE:
	    return (void (*)(void *, void *, int))sig1_to_unsig2l;
	  case USFFD_UNSIGBYTE:
	    return (void (*)(void *, void *, int))unsig1_to_unsig2l;
	  case USFFD_G711_ULAW:
	    return (void (*)(void *, void *, int))ulaw_to_unsig2l;
	  case USFFD_G711_ALAW:
	    return (void (*)(void *, void *, int))alaw_to_unsig2l;
	  case USFFD_SIGWORDB:
	    return (void (*)(void *, void *, int))sig2b_to_unsig2l;
	  case USFFD_UNSIGWORDB:
	    return (void (*)(void *, void *, int))unsig2b_to_unsig2l;
	  case USFFD_SIGWORDL:
	    return (void (*)(void *, void *, int))sig2l_to_unsig2l;
	  case USFFD_UNSIGWORDL:
	    return (void (*)(void *, void *, int))unsig2l_to_unsig2l;
	}
	break;
    }

    return NULL;
}

/* ƥ쥪 =>  */
void stereo_to_mono(long encoding,
		    void* from,
		    void* to,
		    int nbytes)
{
    int i, n = nbytes / usffd_encoding_sample_size[encoding];

    switch(encoding)
    {
      case USFFD_SIGBYTE:
	{
	    char* p1 = (char *)from;
	    char* p2 = (char *)to;
	    for(i = 0; i < n; i += 2)
		p2[i] = (char)MEAN(p1[i], p1[i + 1]);
	}
	break;
      case USFFD_UNSIGBYTE:
	{
	    unsigned char* p1 = (unsigned char *)from;
	    unsigned char* p2 = (unsigned char *)to;
	    for(i = 0; i < n; i += 2)
		p2[i] = (unsigned char)MEAN(p1[i], p1[i + 1]);
	}
	break;
      case USFFD_G711_ULAW:
	{
	    char* p1 = (char *)from;
	    char* p2 = (char *)to;
	    for(i = 0; i < n; i += 2)
	    {
		/* ٤ˤä꤬뤫 */
		char c = (char)MEAN(AUDIO_U2C(p1[i]), AUDIO_U2C(p1[i + 1]));
		p2[i] = AUDIO_C2U(c);
	    }
	}
	break;
      case USFFD_G711_ALAW:
	{
	    char* p1 = (char *)from;
	    char* p2 = (char *)to;
	    for(i = 0; i < n; i += 2)
	    {
		/* ٤ˤä꤬뤫 */
		char c = (char)MEAN(AUDIO_A2C(p1[i]), AUDIO_A2C(p1[i + 1]));
		p2[i] = AUDIO_C2A(c);
	    }
	}
	break;
      case USFFD_SIGWORDB:
	{
	    short* p1 = (short *)from;
	    short* p2 = (short *)to;
	    for(i = 0; i < n; i += 2)
	    {
#ifdef BIGENDIAN
		p2[i] = (short)MEAN(p1[i], p2[i]);
#else
		short s = (short)MEAN(ENDIAN16(short, p1[i]),
				      ENDIAN16(short, p2[i]));
		p2[i] = ENDIAN16(short, s);
#endif
	    }
	}
	break;
      case USFFD_UNSIGWORDB:
	{
	    unsigned short* p1 = (unsigned short *)from;
	    unsigned short* p2 = (unsigned short *)to;
	    for(i = 0; i < n; i += 2)
	    {
#ifdef BIGENDIAN
		p2[i] = (unsigned short)MEAN(p1[i], p2[i]);
#else
		short s = (short)MEAN(ENDIAN16(unsigned short, p1[i]),
				      ENDIAN16(unsigned short, p2[i]));
		p2[i] = ENDIAN16(unsigned short, s);
#endif
	    }
	}
	break;
      case USFFD_SIGWORDL:
	{
	    short* p1 = (short *)from;
	    short* p2 = (short *)to;
	    for(i = 0; i < n; i += 2)
	    {
#ifdef BIGENDIAN
		short s = (short)MEAN(ENDIAN16(short, p1[i]),
				      ENDIAN16(short, p2[i]));
		p2[i] = ENDIAN16(short, s);
#else
		p2[i] = (short)MEAN(p1[i], p2[i]);
#endif
	    }
	}
	break;
      case USFFD_UNSIGWORDL:
	{
	    unsigned short* p1 = (unsigned short *)from;
	    unsigned short* p2 = (unsigned short *)to;
	    for(i = 0; i < n; i += 2)
	    {
#ifdef BIGENDIAN
		short s = (short)MEAN(ENDIAN16(unsigned short, p1[i]),
				      ENDIAN16(unsigned short, p2[i]));
		p2[i] = ENDIAN16(unsigned short, s);
#else
		p2[i] = (unsigned short)MEAN(p1[i], p2[i]);
#endif
	    }
	}
	break;
    }
}

/*  => ƥ쥪 */
void mono_to_stereo(long encoding,
		    void* from,
		    void* to,
		    int nbytes)
{
    int i;

    /* duplicate samples */
    if(usffd_encoding_sample_size[encoding] == 1)
    {
	unsigned char* p1 = (unsigned char *)from;
	unsigned short* p2 = (unsigned short *)to;

	for(i = 0; i < nbytes; i++)
	{
	    unsigned short x = p1[i];
	    p2[i] = (x | (x << 8));
	}
    }
    else
    {
	int n = nbytes / 2;
	unsigned short* p1 = (unsigned short *)from;
	unsigned long* p2 = (unsigned long *)to;

	for(i = 0; i < n; i++)
	{
	    unsigned long x = p1[i];
	    p2[i] = (x | (x << 16));
	}
    }
}

/* ͥΤߤФ */
void stereo_left_channel(long encoding,
			 void* from,
			 void* to,
			 int nbytes)
{
    int i, j;

    if(usffd_encoding_sample_size[encoding] == 1)
    {
	char* p1 = (char *)from;
	char* p2 = (char *)to;

	for(i = j = 0; i < nbytes; i += 2, j++)
	    p2[j] = p1[i];
    }
    else
    {
	int n = nbytes / 2;
	short* p1 = (short *)from;
	short* p2 = (short *)to;

	for(i = j = 0; i < n; i += 2, j++)
	    p2[j] = p1[i];
    }
}

/* ͥΤߤФ */
void stereo_right_channel(long encoding,
			  void* from,
			  void* to,
			  int nbytes)
{
    int i, j;

    if(usffd_encoding_sample_size[encoding] == 1)
    {
	char* p1 = ((char *)from + 1);
	char* p2 = (char *)to;

	for(i = j = 0; i < nbytes; i += 2, j++)
	    p2[j] = p1[i];
    }
    else
    {
	int n = nbytes / 2;
	short* p1 = ((short *)from + 1);
	short* p2 = (short *)to;

	for(i = j = 0; i < n; i += 2, j++)
	    p2[j] = p1[i];
    }
}

#ifdef SOUNDEX_DEBUG_MAIN

#define NBYTES 1024
#define ENCODING_FROM USFFD_UNSIGWORDL
#define ENCODING_TO   USFFD_SIGWORDL

void main(void)
{
    void (* converter)(void *, void *, int);
    char buff1[2 * NBYTES], buff2[2 * NBYTES];
    int bpu1, bpu2;

    converter = select_converter(ENCODING_FROM, ENCODING_TO,
				 &bpu1, &bpu2);
    if(converter == NULL)
    {
	fprintf(stderr, "Can't convert %s to %s\n",
		usffd_encoding_name(ENCODING_FROM),
		usffd_encoding_name(ENCODING_TO));
	exit(1);
    }

    if(bpu1 != 1 && bpu1 != 2 &&
       bpu2 != 1 && bpu2 != 2)
    {
	fprintf(stderr, "bpu1 = %d, bpu2 = %d\n", bpu1, bpu2);
	exit(1);
    }


    fread(buff1, 1, NBYTES * bpu1, stdin);
    converter(buff1, buff2, NBYTES * bpu1);
    fwrite(buff2, 1, NBYTES * bpu2, stdout);

    exit(0);
}

#endif
