#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include "soundscope.h"
#include "audriv.h"
#include "usffd.h"
#include "ssframe.h"
#include "player.h"
#include "fft.h"
#include "aubuff.h"
#include "bucket.h"
#include "timer.h"

char* soundscope_version = VERSION;

#ifdef BIGENDIAN
long internal_encoding = USFFD_SIGWORDB;
#else
long internal_encoding = USFFD_SIGWORDL;
#endif

Bool audio_active_flag = False;	/* audio ƥ֤ʤ True */
Bool audio_record_flag = False;	/* audio Ͽʤ True */
Bool ss_active_flag    = False;	/* soundscope νƥ֤ʤ True */
Bool audio_file_lock_flag = False; /* ԽΥåե饰 */
char* default_text_string = NULL;
char* default_short_cut_list = NULL;

PUBLIC void exit_soundscope()
{
#ifdef DEBUG
    printf("Terminate soundscope (pid=%d)\n", getpid());
#endif
    audriv_play_stop();
    audriv_free_audio();
    exit(0);
}

static void init_signals(void)
{
    signal(SIGINT, exit_soundscope);
    signal(SIGTERM, exit_soundscope);
}

static void update_volatile_values(void){}

PUBLIC void soundscope_event_proc(Bool nonblocking)
{
    XEvent e;

#ifdef DEBUGALL
	printf("soundscope_event_proc(%d)\n", nonblocking);
#endif

    if(nonblocking)
    {
	while(XSync(disp, False), QLength(disp) > 0)
	{
	    XNextEvent(disp, &e);
#ifdef DEBUGALL
	    printf("DispatchEvent: %s\n", EventTypeString(e.type));
#endif
	    DispatchEvent(&e);
	}
    }
    else
    {
	do
	{
	    XNextEvent(disp, &e);
#ifdef DEBUGALL
	    printf("DispatchEvent: %s\n", EventTypeString(e.type));
#endif
	    DispatchEvent(&e);
	} while(QLength(disp) > 0);
    }
}

static void active_frame_processing(void)
{
    SSFrame* frame;
    long sample;
    double cur_time;

#ifdef DEBUGALL
    printf("active_frame_processing()\n");
#endif

    sample = audriv_play_samples();
    cur_time = get_current_calender_time();
    for(frame = all_frame_list; frame; frame = frame->next)
	if(SSFRAME_AVAILABLE(frame))
	{
	    AudioPlayer* player = frame->player;
	    if(player->play_flag && player->aubuff->nsamples > 0)
	    {
		long frame_sample = calc_player_samples(frame->player,
							sample);
		if(!frame->play_slider_grab_flag)
		{
#ifdef DEBUGALL
		    printf("repaint_frame_scope()\n");
#endif
		    if(frame->mapflag)
			repaint_frame_scope(frame, frame_sample, True);
		    else
			audriv_wait_play();
		}
		
		if(audio_player_check_finish(player, frame_sample))
		{
		    player_stop_and_rewind(player);
		    sample = 0;	/* FIXME  sample = 0? */
		}
		else if(player->ff_rw_type != FF_RW_UNUSED &&
			player->ffre_step < FAST_MAX_STEP &&
			player->start_ffre_time + FFRW_SWITCH_SEC < cur_time)
		{
		    int step = player->ffre_step * 4;
		    if(player->ff_rw_type == FAST_FORWARD_PLAY)
		    {
			player_stop_fast_forward_play(player);
			player_start_fast_forward_play(player, step);
		    }
		    else
		    {
			player_stop_fast_rewind_play(player);
			player_start_fast_rewind_play(player, step);
		    }
		}

	    }
	}
}

PUBLIC void soundscope_slice_proc(void)
{
#ifdef DEBUGALL
    printf("soundscope_slice_proc()\n");
#endif
    if(audio_active_flag)
    {
#ifdef DEBUGALL
	printf("play_buffer()\n");
#endif

	if(!file_loader.load_flag)
	    play_buffer();
	else
	{
	    if(play_buffer() > 0)
		play_buffer();	/* play once more */
	}

	active_frame_processing();
	update_volatile_values();
    }
#ifdef DEBUGALL
    printf("flags: %d %d %d %d %d\n",
	   file_loader.load_flag, audio_record_flag, file_saver.save_flag,
	   forward_backward_button_mode, forward_backward_button_mode);
#endif
    if(file_loader.load_flag)
	file_load_slice_proc();
    if(audio_record_flag)
	record_slice_proc();
    if(file_saver.save_flag)
	file_save_slice_proc();
    if(forward_backward_button_mode == FAST_FORWARD_PLAY)
	move_view_position(1);
    else if(forward_backward_button_mode == FAST_REWIND_PLAY)
	move_view_position(-1);
}

static void frame_set_audio_file_lock(SSFrame* frame)
{
    if(audio_file_lock_flag)
    {
	PopupMenuNoselect(frame->file_popup, FILE_POPUP_OPEN);
	PopupMenuNoselect(frame->file_popup, FILE_POPUP_SAVE);
	PopupMenuNoselect(frame->file_popup, FILE_POPUP_RECORD);
	PopupMenuNoselect(frame->sound_edit_popup, SOUND_EDIT_COPY);
	PopupMenuNoselect(frame->sound_edit_popup, SOUND_EDIT_CUT);
	PopupMenuNoselect(frame->sound_edit_popup, SOUND_EDIT_PASTE);
	PopupMenuNoselect(frame->sound_edit_popup, SOUND_EDIT_MERGE);
	PopupMenuNoselect(frame->sound_edit_popup, SOUND_EDIT_OVERWRITE);
	PopupMenuNoselect(frame->sound_edit_popup, SOUND_EDIT_FIN);
	PopupMenuNoselect(frame->sound_edit_popup, SOUND_EDIT_FOUT);
    }
    else
    {
	PopupMenuResetNoselect(frame->file_popup, FILE_POPUP_OPEN);
	PopupMenuResetNoselect(frame->file_popup, FILE_POPUP_SAVE);
	PopupMenuResetNoselect(frame->file_popup, FILE_POPUP_RECORD);
	PopupMenuResetNoselect(frame->sound_edit_popup, SOUND_EDIT_COPY);
	PopupMenuResetNoselect(frame->sound_edit_popup, SOUND_EDIT_CUT);
	PopupMenuResetNoselect(frame->sound_edit_popup, SOUND_EDIT_PASTE);
	PopupMenuResetNoselect(frame->sound_edit_popup, SOUND_EDIT_MERGE);
	PopupMenuResetNoselect(frame->sound_edit_popup, SOUND_EDIT_OVERWRITE);
	PopupMenuResetNoselect(frame->sound_edit_popup, SOUND_EDIT_FIN);
	PopupMenuResetNoselect(frame->sound_edit_popup, SOUND_EDIT_FOUT);
    }
}

void set_audio_file_lock(Bool flag)
{
    audio_file_lock_flag = flag;
    SSFrameMapFuntion(frame_set_audio_file_lock);
}

static void main_loop(void)
{
#ifdef DEBUG
    puts("Enter main_loop()");
#endif

    while(1)
    {
	soundscope_event_proc(ss_active_flag);
	if(ss_active_flag)
	    soundscope_slice_proc();
    }
}

PUBLIC void update_ss_active_flag(void)
{
    if(audio_wave_adder->u.wave_adder.nchannels > 0)
    {
	audio_active_flag = True;
	ss_active_flag = True;
    }
    else if(file_loader.load_flag == True ||
	    file_saver.save_flag  == True ||
	    audio_record_flag     == True ||
	    forward_backward_button_mode != FF_RW_UNUSED)
    {
	audio_active_flag = False;
	ss_active_flag = True;
    }
    else
    {
	audio_active_flag = False;
	ss_active_flag = False;
    }
}

PUBLIC int sample_to_play_slider_pos(SSFrame* frame, long sample)
{
    if(frame->player->aubuff->nsamples == 0)
	return 0;
    return (long)(0.5 + (((double)sample / (double)frame->player->aubuff->nsamples) * frame->play_slider->sslider.maxlen));
}

PUBLIC long play_slider_pos_to_sample(SSFrame* frame, int pos)
{
    return (long)(0.5 + ((double)pos / (double)frame->play_slider->sslider.maxlen) * frame->player->aubuff->nsamples);
}

PUBLIC void move_play_slider_position(SSFrame* frame, long sample)
{
    if(frame->play_slider_grab_flag == False)
    {
	int pos;
	pos = sample_to_play_slider_pos(frame, sample);
	if(frame->play_slider->limited_slider.pos != pos)
	    SetHLimitedSliderPos(frame->play_slider, pos);
    }
}

PUBLIC void soundscope_goto_sample(SSFrame* frame, long sample)
{
    if(sample < 0)
	sample = 0;
    else if(sample > frame->player->aubuff->nsamples)
	sample = frame->player->aubuff->nsamples;

    move_play_slider_position(frame, sample);
    if(frame->player->play_flag)
    {
	int i;

	abc_stop_audio();
	frame->player->pos = sample;
	for(i = 0; i < frame->player->aubuff->nchannels; i++)
	    frame->player->bucket_reader[i]->u.reader.seek_pointer = sample;
	abc_restart_audio();
    }
    else
    {
	frame->player->pos = sample;
	repaint_frame_scope(frame, sample, False);
    }
}

PUBLIC void ss_garbage_collection(void)
{
    SSFrame* cur;

    /* ǥХåեΥ١쥯 */
    /* ƤΥޡ򥯥ꥢ */
    clear_audio_buffer_mark();

    /* ȲǽʥХåեƤ˥ޡĤ */
    for(cur = all_frame_list; cur; cur = cur->next)
	if(SSFRAME_AVAILABLE(cur))
	    cur->player->aubuff->mark = True;
    get_audio_buffer(AUDIO_DEFAULT_BUFF_NAME)->mark = True;
    if(file_saver.save_buffer != NULL)
	file_saver.save_buffer->mark = True;
    if(record_backup_buffer != NULL)
	record_backup_buffer->mark = True;
    if(open_file_target_frame != NULL &&
       open_file_target_frame->player != NULL &&
       open_file_target_frame->player->aubuff != NULL)
	open_file_target_frame->player->aubuff->mark = True;

    /* ޡդƤʤХåեȤΥǡ free
     */
    audio_buffer_garbage_collect();
}

static void usage(void)
{
    fputs("Usage: soundscope [-v] [audio-filename] [short-cut-list]\n", stderr);
}

int main(int argc, char** argv)
{
    char* filename = NULL;
    char** short_cut_list = NULL;
    int i, nshort_cut = 0;

#ifdef DEBUG
    printf("%s (version %s, pid=%d) start.\n", argv[0], VERSION, getpid());
    printf("sizeof(SSFrame) = %d\n", sizeof(SSFrame));
#endif

#ifdef DEBUG
    puts("initialize_misc_funcs()");
#endif
    initialize_misc_funcs();

    if(argc >= 2 && !strcmp(argv[1], "-q"))
	i = 2;
    else
    {
	i = 1;
	get_resource();
    }

    for(; i < argc; i++)
    {
	if(!strcmp(argv[i], "-v"))
	{
	    printf("soundscope %s\n", soundscope_version);
	}
	else if(argv[i][0] == '-')
	{
	    usage();
	    exit(1);
	}
	else
	    break;
    }

    if(i < argc)
	filename = argv[i];
    i++;
    if(i < argc)
    {
	nshort_cut = argc - i;
	short_cut_list = argv + i;
    }

#ifdef DEBUG
    puts("initialize_audio_buffer()");
#endif
    initialize_audio_buffer();

#ifdef DEBUG
    puts("initialize_audio_bucket()");
#endif
    initialize_audio_bucket();

#ifdef DEBUG
    puts("initialize FFT processing");
#endif
    get_current_fft_data(NULL, 0, 0); /* initialize */

    convolution_proc(NULL, 0, 0, NULL, 0, NULL);

    set_error_handlers();
    usffd_encoding_auto_recognizer = soundscope_usffd_recognizer;

    /* initialize fft procedure call */
    realfft(NULL, FFTSIZE);
    fft(NULL, NULL, CONVOLUTION_SIZE);

    init_signals();

    if(audriv_setup_audio() == False)
    {
	fprintf(stderr, "soundscope: %s\n", audriv_errmsg);
	exit(1);
    }

    player_change_encoding(USFFD_G711_ULAW);

    if(audriv_set_noblock_write(True) == False)
    {
	fprintf(stderr, "soundscope: %s\n", audriv_errmsg);
	exit(1);
    }

    if(audriv_set_noblock_read(False) == False)
    {
	fprintf(stderr, "soundscope: %s\n", audriv_errmsg);
	exit(1);
    }

#ifdef DEBUG
    puts("initialize_abc()");
#endif
    initialize_abc();

#ifdef DEBUG
    puts("open_new_ssframe()");
#endif

    open_new_ssframe(NULL);

#ifdef DEBUG
    puts("open_new_ssframe() - Done");
#endif

    prohibit_close_window_proc();

    if(filename != NULL)
    {
	SetLineText((LineText *)all_frame_list->dir_text,
		    filename, strlen(filename));
	file_load_button_proc(all_frame_list->dir_text, NULL);
    }
    else if(default_text_string != NULL)
	SetLineText((LineText *)all_frame_list->dir_text,
		    default_text_string, strlen(default_text_string));

    for(i = 0; i < nshort_cut; i++)
	AddDirLineTextShortCut(all_frame_list->dir_text, short_cut_list[i]);
    if(default_short_cut_list != NULL)
    {
	char *p, *q;

	p = default_short_cut_list;

	while(1)
	{
	    while(*p == ' ' || *p == '\t' || *p == ',')
		p++;
	    if(*p == '\0')
		break;
	    q = p;
	    while(*q && *q != ' ' && *q != '\t' && *q != ',')
		q++;
	    if(*q == '\0')
	    {
		AddDirLineTextShortCut(all_frame_list->dir_text, p);
		break;
	    }
	    *q = '\0';
	    AddDirLineTextShortCut(all_frame_list->dir_text, p);
	    p = q + 1;
	}
    }

    main_loop();
    return 0;
}
