#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "soundscope.h"
#include "ssframe.h"
#include "player.h"
#include "misc.h"

#define SRCUSTOM_SLIDER_LEN 300

static void set_srtext_value(LineText* text, long sample_rate)
{
    sprintf(temp_buff, "%ld", sample_rate);
    SetLineText(text, temp_buff, strlen(temp_buff));
    while(LineTextForwardChar(text))
	;
}

static long get_srtext_value(LineText* text)
{
    GetLineText(text, temp_buff, temp_buff_size);
    return atoi(temp_buff);
}

static void set_srslider_value(HSimpleSlider* slider, long sample_rate)
{
#if 0
    double r;
    r = (double)sample_rate / MAX_SAMPLE_RATE;

sample_rate == srmin + (exp(3.0 * slider->pos / slider->maxlen) - 1) / (exp(3.0) - 1.0) * (srmax - srmin)

(sample_rate - srmin) / (srmax - srmin) * (exp(3.0) - 1.0) + 1
    == exp(3.0 * slider->pos / slider->maxlen)

3.0 * slider->pos / slider->maxlen
    == log((sample_rate - srmin) / (srmax - srmin) * (exp(3.0) - 1.0) + 1)

slider->pos
    == log((sample_rate - srmin) / (srmax - srmin) * (exp(3.0) - 1.0) + 1)
	* slider->maxlen / 3.0
#endif

    SetHSimpleSliderPos((WinObj *)slider,
			(int)(0.5 + log((double)(sample_rate - MIN_SAMPLE_RATE) / (MAX_SAMPLE_RATE - MIN_SAMPLE_RATE) * (exp(3.0) - 1.0) + 1) * slider->maxlen / 3.0));
}

PUBLIC void set_srvalue(SSFrame* frame)
{
    set_srtext_value((LineText *)frame->srcustom_text,
		     frame->player->sample_rate);
    set_srslider_value((HSimpleSlider *)frame->srcustom_slider,
		       frame->player->sample_rate);
}


PUBLIC void open_custom_window_popup(PopupMenuBase* menu, char* name, int id)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(menu);
    set_srvalue(frame);
    XMapWindow(disp, frame->srcustom->any.win);
}

static void close_srcustom_proc(ButtonWindow* btn, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(btn);
    XUnmapWindow(disp, frame->srcustom->any.win);
}

static long sr_calc_log_zoom(HSimpleSlider* slider, long srmin, long srmax)
{
    double x;

    x = exp(3.0 * (double)slider->pos / (double)slider->maxlen) - 1;
    return (long)(0.5 + srmin + x / (exp(3.0) - 1.0) * (srmax - srmin));
}

static void srslider_callback_handler(HSimpleSlider* slider, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(slider);
    long sample_rate;

    sample_rate = sr_calc_log_zoom(slider, MIN_SAMPLE_RATE, MAX_SAMPLE_RATE);
    switch(e->type)
    {
      case ButtonRelease:
	set_srtext_value((LineText *)frame->srcustom_text, sample_rate);
	player_change_sample_rate(frame->player, sample_rate);
	break;
      case MotionNotify:
	set_srtext_value((LineText *)frame->srcustom_text, sample_rate);
	break;
    }
}

static void srtext_callback_handler(LineText* text, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(text);
    long sample_rate, x;

    x = sample_rate = get_srtext_value(text);
    FIXREGION(sample_rate, MIN_SAMPLE_RATE, MAX_SAMPLE_RATE);
    if(x != sample_rate)
	set_srtext_value(text, sample_rate);
    set_srslider_value((HSimpleSlider *)frame->srcustom_slider, sample_rate);
    player_change_sample_rate(frame->player, sample_rate);
}

PUBLIC void create_srcustom(struct _SSFrame* frame)
{
    WinObj* srcustom;
    WinObj* close_btn;
    WinObj* text;
    WinObj* slider;
    int h;

    h = TextHeight(font) + GAP;

    srcustom = toplevel_WinObj(disp, "sample rate", "Sample rate", 1, 1,
			       WinObj_fg, WinObj_bg);
    if(ISMONODISP)
    {
	srcustom->any.mono_color = True;
	srcustom->any.mark_color = WinObj_fg;
    }

    srcustom->any.when_realize_mapping = False;
    close_btn = create_Button(srcustom, GAP, GAP,
			      TextWidth(font, "Ĥ") + GAP, h,
			      BD_WIDTH, font, "Ĥ");
    SetCallBackProc(close_btn, (call_back_t)close_srcustom_proc);

    text = create_LineText(srcustom,
			   WinObjRight(close_btn) + GAP, GAP,
			   TextWidth(font, "Sample Rate: 000000"), h,
			   BD_WIDTH, font, "Sample Rate: ");
    text->text.maxlen = text->text.textlen = 5;
    SetCallBackProc(text, (call_back_t)srtext_callback_handler);
    SetLineText((LineText *)text, "8000", 4);

    slider = create_HSimpleSlider(srcustom,
				  WinObjRight(text) + GAP, GAP,
				  SRCUSTOM_SLIDER_LEN, h,
				  BD_WIDTH, 0);
    SetCallBackProc(slider, (call_back_t)srslider_callback_handler);
    set_srslider_value((HSimpleSlider *)slider, 8000);

    frame->srcustom = srcustom;
    frame->srcustom_slider = slider;
    frame->srcustom_text = text;
    ResizeWinObj(srcustom,
		 WinObjRight(slider) + GAP,
		 WinObjBottom(slider) + GAP);

    WinObjRealize(srcustom);
}
