#ifndef ___USFFD_H_
#define ___USFFD_H_

/* universal sound file format discripter */

#include "datatype.h"
#include "aenc.h"

enum usffd_data_encoding
{
    USFFD_UNDEFINED = 0,

    USFFD_SIGBYTE,		/* signed byte */
    USFFD_UNSIGBYTE,		/* unsigned byte */
    USFFD_G711_ULAW,		/* G.711 U-Law 8-bit */
    USFFD_G711_ALAW,		/* G.711 A-Law 8-bit */

    USFFD_SIGWORDB,		/* signed word big-endian */
    USFFD_UNSIGWORDB,		/* unsigned word big-endian */
    USFFD_SIGWORDL,		/* signed word little-endian */
    USFFD_UNSIGWORDL,		/* unsigned word little-endian */

    USFFD_NENCODINGS		/* number of encodints */
};

enum usffd_error_code
{
    USFFD_NO_ERROR = 0,

    USFFD_SYS_ERROR,
    USFFD_NOT_REGULAR_FILE,
    USFFD_INVALID_FILE_HEADER,
    USFFD_UNRECOGNIZED_FILE_HEADER,
    USFFD_UNKNOWN_FILE_ENCODING
};

enum usffd_audio_header_t
{
    USFFD_HDR_NONE = 0,

    USFFD_HDR_AU,		/* Sun audio file */
    USFFD_HDR_WAV,		/* Windows WAV file */
    USFFD_HDR_AIFF		/* AIFF */
};

typedef struct _usffd
{
    /* setting in usffd_open() */
    int fd;
    char* filename;
    long file_size;

    /* setting in usffd_set_header() */
    long hdr_size;
    long data_size;		/* nsamples * block_align */
    long data_encoding;
    long sample_rate;
    long nchannels;
    enum usffd_audio_header_t hdr_type;
} USFFD;

Bool usffd_open(const char* filename, USFFD* usffd_return);
/* ɥե򥪡ץ󤷤ޤ True 򡤼Ԥ False 
֤ޤ*/

void usffd_close(USFFD* usffd);
/* ɥե򥯥ޤ*/

Bool usffd_set_header(USFFD* usffd);
/* եĥҤ˴Ťơɥեηꤷޤ
Ǥʤäϡ顼  usffd_errno ꤵ졤False ֤ޤ*/

void usffd_make_header(USFFD* usffd,
		       long encoding, long nchannels, long sample_rate);
/* usffd Υإåޤ*/
		       

void usffd_encoding_recognize(USFFD* usffd);
/* եȤ˴Ťơɥեηꤷޤ
ɤΤ褦ʾǤäƤ⡤ĤΥɷ˷ꤷޤ*/

void usffd_print(const USFFD* usffd);
/* usffd ¤Τξɽޤ*/

const char* usffd_strerror(int error_code);
/* usffd_errno б륨顼å֤ޤ*/

extern int usffd_errno;
/* 顼ֹǤ*/

const char* usffd_encoding_name(long encoding);
const char* usffd_encoding_jname(long encoding);
/* encoding 򤢤魯ʸ֤ޤ*/

const char* usffd_hdr_extension(const USFFD* usffd);
/* եĥҤ֤ޤ*/

void usffd_write_header(int fd, const USFFD* usffd, long* file_encoding);
/* fd ˡإå񤭹ߤޤ
   usffd ΰʲΥФꤵƤʤФʤޤ

   data_size
   data_encoding
   sample_rate
   nchannels
   hdr_type
   filename (hdr_type  USFFD_HDR_AU ΤȤξ) */

long usffd_select_encoding(enum usffd_audio_header_t type, long encoding);
/* type  encoding ɽΤ˺ǤŬ沽֤ޤ*/

Bool usffd_encoding_available(enum usffd_audio_header_t type, long encoding);
/* type  encoding Ѳǽʤ True, Ǥʤʤ False ֤ޤ*/

extern long usffd_default_sample_rate;
/* ǥեȤΥץ졼ȤǤ.sw ʤɡإåΤʤɷ
Фơץ졼ȤꤹΤѤޤ*/

extern void (* usffd_encoding_auto_recognizer)(USFFD* usffd);
/* ưǧؿꤷޤǥեȤǤ̵˦ˡ§Ѵޤ*/

extern const int usffd_encoding_sample_size[USFFD_NENCODINGS];
/* 粽ˤ 1 ץΥХȿɽǤ */

#endif /* ___USFFD_H_ */
