#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define UNIT_SIZE 1048576	/* 1M */
#define READ_UNIT (1024*64)	/* 64k */

struct list
{
    char* buffer;
    int   n;
    struct list* next;
};

static struct list* first = NULL;
static struct list* last = NULL;

static struct list* newlist(void)
{
    struct list* node;

    if((node = (struct list *)malloc(sizeof(struct list))) == NULL ||
       (node->buffer = (char *)malloc(UNIT_SIZE)) == NULL)
    {
	fprintf(stderr, "Can't allocate memory\n");
	exit(1);
    }

    node->n = 0;
    node->next = NULL;

    return node;
}

static void next_string(char* s, int n)
{
    struct list* p;

    if(first == NULL)
	last = first = newlist();
    p = last;
    if(p->n + n < UNIT_SIZE)
    {
	memcpy(p->buffer + p->n, s, n);
	p->n += n;
    }
    else
    {
	int n0;

	n0 = UNIT_SIZE - p->n;
	memcpy(p->buffer + p->n, s, n0);
	p->n = UNIT_SIZE;
	last = p->next = newlist();
	p = last;
	memcpy(p->buffer, s + n0, n - n0);
	p->n = n - n0;
    }
}

static char* concat_string(char* s, long* nread)
{
    struct list* p;
    struct list* cur;
    char* start = s;

    cur = first;
    *nread = 0;
    while(cur != NULL)
    {
	p = cur;
	cur = cur->next;
	memcpy(s, p->buffer, p->n);
	s += p->n;
	*nread += p->n;
	free(p->buffer);
	free(p);
    }

    last = first = NULL;
    start[*nread] = '\0';
    return start;
}

char* readtty(long* nread)
{
    char buff[READ_UNIT];
    int fd = 0;
    long n;
    char* s;
    long t;

    t = 0;
    while((n = read(fd, buff, READ_UNIT)) > 0)
    {
	t += n;
	next_string(buff, n);
    }

    if(n < 0)
    {
	perror("read");
	exit(1);
    }

    if((s = (char *)malloc(t + 1)) == NULL)
    {
	fprintf(stderr, "Can't allocate memory\n");
	exit(1);
    }

    return concat_string(s, nread);
}
