#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>

void usage(void)
{
    fprintf(stderr, "Usage: swapendian16 infile outfile\n");
    fprintf(stderr, "       swapendian16 infile > outfile\n");
    fprintf(stderr, "       swapendian16 < infile > outfile\n");
    fprintf(stderr, "       swapendian16 -w inoutfile\n");
}

int open_file(char* filename, int mode)
{
    int fd;

    if((fd = open(filename, mode)) < 0)
    {
	perror(filename);
	exit(1);
    }

    return fd;
}

void main(int argc, char** argv)
{
    int ofd, ifd;
    char buff[BUFSIZ], c;
    int i, n;

    if(argc == 1)
    {
	ifd = 0;
	ofd = 1;
    }
    else if(argc == 2)
    {
	ifd = open_file(argv[1], O_RDONLY);
	ofd = 1;
    }
    else if(argc == 3 && !strcmp(argv[1], "-w"))
    {
	ifd = ofd = open_file(argv[2], O_RDWR);
    }
    else if(argc == 3)
    {
	ifd = open_file(argv[1], O_RDONLY);
	ofd = open_file(argv[2], O_WRONLY | O_CREAT | O_TRUNC);
    }
    else
    {
	usage();
	exit(1);
    }

    if(isatty(ofd))
    {
	usage();
	exit(1);
    }

    if(ifd != ofd)
	while((n = read(ifd, buff, BUFSIZ)) > 0)
	{
	    for(i = 0; i < n - 1; i += 2)
	    {
		c = buff[i];
		buff[i] = buff[i + 1];
		buff[i + 1] = c;
	    }
	    if(write(ofd, buff, n) < 0)
	    {
		perror("write");
		close(ifd);
		close(ofd);
		exit(1);
	    }
	}
    else
	while((n = read(ifd, buff, BUFSIZ)) > 0)
	{
	    for(i = 0; i < n; i += 2)
	    {
		c = buff[i];
		buff[i] = buff[i + 1];
		buff[i + 1] = c;
	    }
	    lseek(ifd, -n, SEEK_CUR);
	    if(write(ifd, buff, n) < 0)
	    {
		perror("write");
		close(ifd);
		close(ofd);
		exit(1);
	    }
	}

    if(n < 0)
    {
	perror("read");
	close(ifd);
	close(ofd);
	exit(1);
    }
    close(ifd);
    close(ofd);
    exit(0);
}
