#include <stdio.h>
#include <math.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "WinObj.h"
#include "ssdefs.h"
#include "audriv.h"
#include "aubuff.h"
#include "ssframe.h"
#include "soundscope.h"
#include "player.h"

#ifdef sun5
#include <values.h>
#endif

/* #define BIGFONT */

enum slider_unit_t
{
    SLIDER_UNIT_TIME,
    SLIDER_UNIT_SAMPLE
};

#ifdef BIGFONT 
#define FONT8 "r24"
#define FONT16 "kanji24"
#else
#define FONT8 "a14"
#define FONT16 "k14"
#endif

char* WinObj_font8 = FONT8;
char* WinObj_font16 = FONT16;
Bool WinObj_reverse_color_flag = False;
char* disp_name = NULL;

#define SLIDER_HEIGHT 25
#define TOPLEVEL_DEFAULT_WINDOW_WIDTH (DEFAULT_SCOPE_WIDTH+2*GAP+2*BD_WIDTH)
#define GAPMIN 4
#define VOLUME_SLIDER_LEN 256

#define SCOPE_MENU_KINDS 4

#ifdef DEBUG
/* ... */
#endif

Display*  disp = NULL;
GC        gc   = NULL;		/* ǻȤ󤷤 gc. ե̤ */
GC        dashgc = NULL;	/*  */
CharFont* font = NULL;		/* ƥѥե () */
int display_width, display_height;
/*
int scope_base_pixmap_width  = DEFAULT_SCOPE_WIDTH;
int scope_base_pixmap_height = DEFAULT_SCOPE_HEIGHT;
*/
Pixel color_pixel1;
Pixel color_pixel2;
Pixel color_pixel3;
Pixel color_pixel4;
Pixel scope_fg;
Pixel scope_bg;
Pixel WinObj_fg;
Pixel WinObj_bg;
Pixel WinObj_mark_color;

/* start options */
char* WinObj_fg_name = "black";
char* WinObj_bg_name = "gray";

/* WinObj */
static WinObj* soundscope_help_popup;
WinObj* text_short_cut_list_popup;
WinObj* message_winobj = NULL;


/* icon pixmaps */
#define PIXMAP_WIDTH 16
#define PIXMAP_HEIGHT 16
extern Pixmap rectangle_icon1_pixmap;
extern Pixmap rectangle_icon2_pixmap;
extern Pixmap ltlt_icon1_pixmap;
extern Pixmap ltlt_icon2_pixmap;
extern Pixmap triangle_icon1_pixmap;
extern Pixmap triangle_icon2_pixmap;
extern Pixmap gtgt_icon1_pixmap;
extern Pixmap gtgt_icon2_pixmap;
extern Pixmap recrec_icon1_pixmap;
extern Pixmap recrec_icon2_pixmap;
extern Pixmap looping_icon1_pixmap;
extern Pixmap looping_icon2_pixmap;

static void allow_close_window_proc1(SSFrame* frame)
{
    PopupMenuResetNoselect(frame->sound_win_popup, WIN_POPUP_WINCLOSE);
}

static void prohibit_close_window_proc1(SSFrame* frame)
{
    PopupMenuNoselect(frame->sound_win_popup, WIN_POPUP_WINCLOSE);
}

PUBLIC void allow_close_window_proc(void)
{
    SSFrameMapFuntion(allow_close_window_proc1);
}

PUBLIC void prohibit_close_window_proc(void)
{
    SSFrameMapFuntion(prohibit_close_window_proc1);
}

static void initialize_help_popup(void)
{
    soundscope_help_popup = create_DrawerPopupMenu(disp, NULL, 2,
						   WinObj_fg, WinObj_bg, font);
    add_PopupMenu(soundscope_help_popup, "soundscope ", 0, NULL);
}

PUBLIC void map_local_mode(SSFrame* frame)
{
    static Bool map_flags[SCOPE_NMODE][SCOPE_NLOCAL_MODE] =
    {
	{False, False, False},	/* SCOPE_WAVE */
	{True,  True,  False},	/* SCOPE_SPECTORAM */
	{True,  False, True},	/* SCOPE_SPECTOROGRAM */
	{False, False, False},	/* SCOPE_EDIT */
    };
    int i, mode;

    mode = SSFRAME_SCOPE_MAIN_MODE(frame);
    for(i = 0; i < SCOPE_NLOCAL_MODE; i++)
	if(map_flags[mode][i])
	    XMapWindow(disp,
		SkeletonWinObj(frame->local_mode_selector[i])->any.win);
	else
	    XUnmapWindow(disp,
		SkeletonWinObj(frame->local_mode_selector[i])->any.win);

    if(mode == SCOPE_EDIT)
    {
	XMapWindow(disp, SkeletonWinObj(frame->edit_slider)->any.win);
	XMapWindow(disp, frame->edit_region_viewer);
    }
    else
    {
	XUnmapWindow(disp, SkeletonWinObj(frame->edit_slider)->any.win);
	XUnmapWindow(disp, frame->edit_region_viewer);
    }
}

static void x_setup_once(void)
{
    static int setup_flag = False;
    int visual_class;
    XGCValues gv;
    int scr;

    if(setup_flag)
	return;
    setup_flag = True;

    disp = OpenDisplay(disp_name);
    visual_class = DefaultVisual(disp, DefaultScreen(disp))->class;
    if(display_visual_type == DISPLAY_VISUAL_DEFAULT &&
       (visual_class == StaticGray || visual_class == GrayScale))
      {
	display_visual_type = DISPLAY_VISUAL_MONO;
      }

    scr = DefaultScreen(disp);
    display_width  = DisplayWidth(disp, scr);
    display_height = DisplayHeight(disp, scr);

    if(ISMONODISP)
    {
	if(WinObj_reverse_color_flag)
	{
	    scope_fg = WinObj_fg = WhitePixel(disp, scr);
	    scope_bg = WinObj_bg = BlackPixel(disp, scr);
	}
	else
	{
	    scope_fg = WinObj_fg = BlackPixel(disp, scr);
	    scope_bg = WinObj_bg = WhitePixel(disp, scr);
	}
	color_pixel1 =
	    color_pixel2 =
		color_pixel3 =
		    color_pixel4 = scope_fg;
    }
    else
    {
	if(WinObj_reverse_color_flag)
	{
	    WinObj_bg = AllocNamedColor(disp, WinObj_fg_name);
	    WinObj_fg = AllocNamedColor(disp, WinObj_bg_name);
	}
	else
	{
	    WinObj_fg = AllocNamedColor(disp, WinObj_fg_name);
	    WinObj_bg = AllocNamedColor(disp, WinObj_bg_name);
	}
	scope_bg = AllocRGBColor(disp, 32.0/255, 32.0/255, 0.0);
	scope_fg = AllocRGBColor(disp, 43.0/255, 1.0, 1.0);
	color_pixel1 = AllocNamedColor(disp, "gold");
	color_pixel2 = AllocNamedColor(disp, "gray43");
	color_pixel3 = AllocNamedColor(disp, "white");
	color_pixel4 = AllocNamedColor(disp, "IndianRed1");
    }
    font = CreateCharFont(disp, WinObj_font8, WinObj_font16);
    gv.graphics_exposures = False;
    gc = XCreateGC(disp, DefaultRootWindow(disp), GCGraphicsExposures, &gv);
    dashgc = XCreateGC(disp, DefaultRootWindow(disp),
		       GCGraphicsExposures, &gv);

#if defined(sgi)
    {
	Font f;

	f = XLoadFont(disp, "6x12");
	if(!f)
	    f = XLoadFont(disp, "6x10");
	if(f)
	    XSetFont(disp, gc, f);
    }
#endif

#if 0				/* FIXME */
    XGetGCValues(disp, dashgc, GCLineWidth | GCCapStyle
		 | GCJoinStyle, &gv);
#if defined(__alpha) || defined(__linux__)
    XSetLineAttributes(disp, dashgc, 1, LineOnOffDash,
		       gv.cap_style, gv.join_style);
#else
    XSetLineAttributes(disp, dashgc, 0, LineOnOffDash,
		       gv.cap_style, gv.join_style);
#endif
#endif

    {
	static char dash_list[2] = {2, 2};
	XSetDashes(disp, dashgc, 0, dash_list, 2);
    }


    text_short_cut_list_popup =
	create_MCPopupMenu(disp, "Short Cut", BD_WIDTH, WinObj_fg, WinObj_bg,
			   font);
    initialize_help_popup();
    initialize_setup_file_load_window();
    initialize_setup_file_save_window();
    initialize_record_window();
    initialize_file_indicator_window();
    message_winobj =
	create_MessageDialog(disp, BD_WIDTH, WinObj_fg, WinObj_bg, font);
    WinObj_mark_color = message_winobj->any.mark_color;
    initialize_button_icon_pixmaps(disp, gc, WinObj_fg, WinObj_bg,
				   WinObj_mark_color,
				   ISMONODISP);
}

static void x_secondary_setup_once(void){}

static void init_sample_rate_popup_menu(WinObj* srp)
{
    int i, n;
    const long* samples;

    samples = audriv_available_sample_rates(&n);
    for(i = 0; i < n; i++)
    {
	sprintf(temp_buff, "%d", (int)samples[i]);
	add_PopupMenu(srp, temp_buff, (int)samples[i], sample_rate_popup_proc);
    }
    add_PopupMenu(srp, "", 0, open_custom_window_popup);
}

static void init_output_popup_menu(WinObj* op)
{
    add_PopupMenu(op, "ԡ", AUDRIV_OUTPUT_SPEAKER,
		  output_popup_proc);
    add_PopupMenu(op, "إåɥۥ", AUDRIV_OUTPUT_HEADPHONE,
		  output_popup_proc);
    add_PopupMenu(op, "饤", AUDRIV_OUTPUT_LINE_OUT,
		  output_popup_proc);
}

static void init_encoding_rate_popup_menu(WinObj* ep)
{
    const long* enc;
    const char* name;
    int i, n;

    enc = audriv_available_encodings(&n);
    for(i = 0; i < n; i++)
    {
	name = usffd_encoding_name(enc[i]);
#ifdef DEBUG
	printf("%s is available.\n", name);
#endif
	add_PopupMenu(ep, name, enc[i], encoding_popup_proc);
    }
}

static void init_channel_popup_menu(WinObj* cp)
{
    const long* ch;
    int i, n;

    ch = audriv_available_channels(&n);
    for(i = 0; i < n; i++)
    {
#ifdef DEBUG
	printf("%ld channel is available.\n", ch[i]);
#endif /* DEBUG */
	sprintf(temp_buff, "%ld ͥ", ch[i]);
	add_PopupMenu(cp, temp_buff, ch[i], channel_popup_proc);
    }
}

static SSFrame* private_ssframe_reference;
static void set_ssframe_reference_pointer1(WinObj* obj)
{
    obj->any.ext_data = (void *)private_ssframe_reference;
}
static void set_ssframe_reference_pointer(SSFrame* frame, WinObj* root)
{
    private_ssframe_reference = frame;
    VisitEachWinObj(root, set_ssframe_reference_pointer1);
}
static void close_ssframe_from_popup(struct _PopupMenuBase* m, char* name, int id)
{
    SSFrame* frame = (SSFrame *)m->ext_data;

    set_initial_player(frame->player);
    if(open_file_target_frame == frame)
    {
	setup_file_load_cancel(NULL, NULL);
	setup_file_save_cancel(NULL, NULL);
    }
    delete_ssframe(frame);
    if(number_of_ssframes() == 1)
	prohibit_close_window_proc();
}

static void open_new_ssframe_popup_proc(struct _PopupMenuBase* m, char* name, int id)
{
    open_new_ssframe((SSFrame *)m->ext_data);
}

static WinObj* create_scope_bar(SSFrame* frame, WinObj* parent)
{
    MenuBarItem baritem[SCOPE_MENU_KINDS];
    WinObj* file_popup;		/* ե */
    WinObj* sound_attri_popup;	/* ° */
    WinObj* sound_win_popup;	/* ɥ */
    WinObj* sound_edit_popup;	/* Խ */
    WinObj* sample_rate_popup;	/* ץ졼 */
    WinObj* output_popup;	/*  */
    WinObj* encoding_popup;	/* 沽 */
    WinObj* channel_popup;	/* ͥ */

    file_popup = create_DrawerPopupMenu(disp, NULL, BD_WIDTH,
					parent->any.fg, parent->any.bg, font);
    sound_attri_popup = create_DrawerPopupMenu(disp, NULL, BD_WIDTH,
					 parent->any.fg, parent->any.bg, font);
    sound_edit_popup = create_DrawerPopupMenu(disp, NULL, BD_WIDTH,
					      parent->any.fg, parent->any.bg,
					      font);
    sound_win_popup = create_DrawerPopupMenu(disp, NULL, BD_WIDTH,
					 parent->any.fg, parent->any.bg, font);
    sample_rate_popup =
	create_DrawerPopupMenu(disp, "ץ졼", BD_WIDTH,
			       parent->any.fg, parent->any.bg, font);
    output_popup = create_DrawerPopupMenu(disp, "ϥݡ", BD_WIDTH,
					  parent->any.fg, parent->any.bg,
					  font);
    encoding_popup = create_DrawerPopupMenu(disp, "沽", BD_WIDTH,
					    parent->any.fg, parent->any.bg,
					    font);
    channel_popup =  create_DrawerPopupMenu(disp, "ͥ", BD_WIDTH,
					    parent->any.fg, parent->any.bg,
					    font);

/**** file_popup ****/
    add_PopupMenu(file_popup, "", FILE_POPUP_OPEN,
		  file_popup_setup_file_load);
    add_PopupMenu(file_popup, "¸", FILE_POPUP_SAVE,
		  file_popup_setup_file_save);
    add_PopupMenu(file_popup, "Ͽ", FILE_POPUP_RECORD,
		  file_popup_record);
    add_PopupMenu(file_popup, NULL,   -1, NULL);
    add_PopupMenu(file_popup, "λ", FILE_POPUP_QUIT, exit_soundscope);

/**** sound_attri_popup ****/
    add_PopupMenu(sound_attri_popup, "ץ졼", SOUND_ATTRI_SAMPLERATE,
		  PopupMenuHierarchicalHook);
#ifndef MMEAUDIO
    add_PopupMenu(sound_attri_popup, "ϥݡ", SOUND_ATTRI_OUTPUT,
		  PopupMenuHierarchicalHook);
#endif /* MMEAUDIO */

    add_PopupMenu(sound_attri_popup, "沽", SOUND_ATTRI_ENCODING,
		  PopupMenuHierarchicalHook);
    add_PopupMenu(sound_attri_popup, "ͥ", SOUND_ATTRI_CHANNEL,
		  PopupMenuHierarchicalHook);

/**** sound_edit_popup ****/
    add_PopupMenu(sound_edit_popup, "ԡ",   SOUND_EDIT_COPY,
		  sound_edit_copy_popup);
    add_PopupMenu(sound_edit_popup, "å",   SOUND_EDIT_CUT,
		  sound_edit_cut_popup);
    add_PopupMenu(sound_edit_popup, "ڡ", SOUND_EDIT_PASTE,
		  sound_edit_paste_popup);
    add_PopupMenu(sound_edit_popup, "ʻ", SOUND_EDIT_MERGE,
		  sound_edit_merge_popup);
    add_PopupMenu(sound_edit_popup, "", SOUND_EDIT_OVERWRITE,
		  sound_edit_overwrite_popup);
    add_PopupMenu(sound_edit_popup, "եɥ", SOUND_EDIT_FIN,
		  sound_edit_fin_popup);
    add_PopupMenu(sound_edit_popup, "եɥ", SOUND_EDIT_FOUT,
		  sound_edit_fout_popup);

/**** sound_win_popup ****/
    add_PopupMenu(sound_win_popup, "ɥ򳫤",
		  WIN_POPUP_NEWWINDOW, open_new_ssframe_popup_proc);
    add_PopupMenu(sound_win_popup, "ɥĤ",
		  WIN_POPUP_WINCLOSE, close_ssframe_from_popup);
    add_PopupMenu(sound_win_popup, "饤",
		  WIN_POPUP_FILTER, open_equlizer_win_popup_proc);
#if 0	/* system status */
    add_PopupMenu(sound_win_popup, "ƥॹơ",
		  WIN_POPUP_SYSSTATUS, NULL);
#endif

    init_sample_rate_popup_menu(sample_rate_popup);
    init_output_popup_menu(output_popup);
    init_encoding_rate_popup_menu(encoding_popup);
    init_channel_popup_menu(channel_popup);

    PopupMenuSetNextMenu(sound_attri_popup, SOUND_ATTRI_SAMPLERATE,
			 sample_rate_popup);

#ifndef MMEAUDIO
    PopupMenuSetNextMenu(sound_attri_popup, SOUND_ATTRI_OUTPUT,
			 output_popup);
#endif /* MMEAUDIO */

    PopupMenuSetNextMenu(sound_attri_popup, SOUND_ATTRI_ENCODING,
			 encoding_popup);
    PopupMenuSetNextMenu(sound_attri_popup, SOUND_ATTRI_CHANNEL,
			 channel_popup);

    baritem[0].name = "ե";
    baritem[0].popup = file_popup;
    baritem[2].name = "°";
    baritem[2].popup = sound_attri_popup;
    baritem[1].name = "Խ";
    baritem[1].popup = sound_edit_popup;
    baritem[3].name = "ɥ";
    baritem[3].popup = sound_win_popup;

    set_ssframe_reference_pointer(frame, file_popup);
    set_ssframe_reference_pointer(frame, sound_attri_popup);
    set_ssframe_reference_pointer(frame, sound_edit_popup);
    set_ssframe_reference_pointer(frame, sound_win_popup);
    set_ssframe_reference_pointer(frame, sample_rate_popup);
    set_ssframe_reference_pointer(frame, output_popup);
    set_ssframe_reference_pointer(frame, encoding_popup);
    set_ssframe_reference_pointer(frame, channel_popup);

    frame->file_popup = file_popup;
    frame->sound_attri_popup = sound_attri_popup;
    frame->sound_edit_popup = sound_edit_popup;
    frame->sound_win_popup = sound_win_popup;
    frame->sample_rate_popup = sample_rate_popup;
    frame->output_popup = output_popup;
    frame->encoding_popup = encoding_popup;
    frame->channel_popup = channel_popup;

    frame->file_popup->popup_menu_base.selecting_hook = popup_waiting_hook;
    frame->sound_attri_popup->popup_menu_base.selecting_hook
	= popup_waiting_hook;
    frame->sound_edit_popup->popup_menu_base.selecting_hook
	= popup_waiting_hook;
    frame->sound_win_popup->popup_menu_base.selecting_hook
	= popup_waiting_hook;
    frame->sample_rate_popup->popup_menu_base.selecting_hook
	= popup_waiting_hook;
    frame->output_popup->popup_menu_base.selecting_hook = popup_waiting_hook;
    frame->encoding_popup->popup_menu_base.selecting_hook = popup_waiting_hook;
    frame->channel_popup->popup_menu_base.selecting_hook = popup_waiting_hook;

#if 0 /* help menu */
    return create_MenuBar(parent, 0, 0, 0, font, 2, baritem, SCOPE_MENU_KINDS,
			  "إ", soundscope_help_popup);
#endif
    return create_MenuBar(parent, 0, 0, 0, font, 2, baritem, SCOPE_MENU_KINDS,
			  NULL, NULL);
}

static void create_scope_window(SSFrame* frame, WinObj* parent)
{
    WinObj* scope_win_obj;
    WinObj* rview;
    int h;
    int edit_slider_height = 16;
    int rview_height = 8;

    h = frame->scope_height;
    scope_win_obj = create_AnyWindow(parent, 0, 0,
				     frame->scope_width, h);
    SetEventHandler(scope_win_obj, scope_event_handler);
    SetEventMask(scope_win_obj, ExposureMask | ButtonPressMask |
		 ButtonReleaseMask | ButtonMotionMask);

    frame->edit_slider =
	create_HScrollbar(scope_win_obj, 0,
			  h - edit_slider_height - 2*BD_WIDTH - rview_height,
			  frame->scope_width - 2*BD_WIDTH, edit_slider_height,
			  BD_WIDTH, 1.0);
    SetCallBackProc(frame->edit_slider, (call_back_t)edit_slider_proc);

    rview =
	create_AnyWindow(scope_win_obj,
			 0, h - rview_height,
			 frame->scope_width, rview_height);
    frame->edit_region_viewer = rview->any.win;
    SetEventMask(rview, ExposureMask | ButtonPressMask |
		 ButtonReleaseMask | ButtonMotionMask);
    SetEventHandler(rview, edit_rview_event_handler);

    frame->scope_win = scope_win_obj->any.win;
    frame->scope_winobj = scope_win_obj;
}

static WinObj* create_sstop_window(SSFrame* parent_frame)
{
    WinObj* top;

    int width, height;

    if(parent_frame == NULL)
    {
	width = TOPLEVEL_DEFAULT_WINDOW_WIDTH;
	height = 540;
    }
    else
    {
	width = parent_frame->toplevel->any.width;
	height = parent_frame->toplevel->any.height;
    }

    top = toplevel_WinObj(disp, SOUNDSCOPE_WINDOW_TITLE, SOUNDSCOPE_ICON_TITLE,
			  width, height, WinObj_fg, WinObj_bg);
    SetEventMask(top, ExposureMask | StructureNotifyMask);
    SetEventHandler(top, sstoplevel_event_handler);

    return top;
}

static void create_play_buttons(SSFrame* frame, WinObj* top, int x, int y)
{
    int button_width = TextWidth(font, "") + 2 * GAP;
    int button_height = TextHeight(font) + GAP;

    frame->player->stop_and_rewind_button =
	create_IconButton(top, x, y, button_width, button_height, BD_WIDTH,
			  rectangle_icon1_pixmap, rectangle_icon2_pixmap,
			  PIXMAP_WIDTH, PIXMAP_HEIGHT);
    x = WinObjRight(frame->player->stop_and_rewind_button) + GAP;
    frame->player->fast_rewind_button =
	create_IconButton(top, x, y, button_width, button_height, BD_WIDTH,
			  ltlt_icon1_pixmap, ltlt_icon2_pixmap,
			  PIXMAP_WIDTH, PIXMAP_HEIGHT);
    x = WinObjRight(frame->player->fast_rewind_button) + GAP;
    frame->player->play_button =
	create_IconButton(top, x, y, button_width, button_height, BD_WIDTH,
			  triangle_icon1_pixmap, triangle_icon2_pixmap,
			  PIXMAP_WIDTH, PIXMAP_HEIGHT);
    frame->player->play_button->icon.reset_when_released = False;

    x = WinObjRight(frame->player->play_button) + GAP;
    frame->player->fast_forward_button =
	create_IconButton(top, x, y, button_width, button_height, BD_WIDTH,
			  gtgt_icon1_pixmap, gtgt_icon2_pixmap,
			  PIXMAP_WIDTH, PIXMAP_HEIGHT);
    x = WinObjRight(frame->player->fast_forward_button) + GAP;
    frame->player->stop_button =
	create_IconButton(top, x, y, button_width, button_height, BD_WIDTH,
			  recrec_icon1_pixmap, recrec_icon2_pixmap,
			  PIXMAP_WIDTH, PIXMAP_HEIGHT);
    frame->player->stop_button->icon.reset_when_released = False;
    x = WinObjRight(frame->player->stop_button) + GAP;
    frame->player->looping_button =
	create_IconButton(top, x, y, button_width, button_height, BD_WIDTH,
			  looping_icon1_pixmap, looping_icon2_pixmap,
			  PIXMAP_WIDTH, PIXMAP_HEIGHT);
    frame->player->looping_button->icon.reset_when_released = False;

    /* seting of call back handlers */
    SetCallBackProc(frame->player->play_button, (call_back_t)play_button_proc);
    SetCallBackProc(frame->player->stop_button, (call_back_t)stop_button_proc);
    SetCallBackProc(frame->player->stop_and_rewind_button,
		    (call_back_t)stop_and_rewind_button_proc);
    SetCallBackProc(frame->player->fast_forward_button,
		    (call_back_t)fast_forward_button_proc);
    SetCallBackProc(frame->player->fast_rewind_button,
		    (call_back_t)fast_rewind_button_proc);
    SetCallBackProc(frame->player->looping_button,
		    (call_back_t)looping_button_proc);

}

static void copy_parent_frame_members(SSFrame* pf, SSFrame* frame)
{
    int i;

    frame->player->aubuff = pf->player->aubuff;
    GetLineText((LineText *)pf->dir_text, temp_buff, BUFSIZ);
    SetLineText((LineText *)frame->dir_text, temp_buff, strlen(temp_buff));
    if(pf->player->aubuff->name != NULL)
	SetLabeldWindowLabel(frame->buffer_label,
			     pf->player->aubuff->name,
			     WINOBJ_BORDER_CONCAVE);

    /* ݥåץåץ˥塼Υԡ */
    player_change_sample_rate(frame->player,
			      PopupMenuMarkerID(pf->sample_rate_popup));
    PopupMenuChangeMarker(frame->output_popup,
			  PopupMenuMarkerID(pf->output_popup));
    PopupMenuChangeMarker(frame->encoding_popup,
			  PopupMenuMarkerID(pf->encoding_popup));
    PopupMenuChangeMarker(frame->channel_popup,
			  PopupMenuMarkerID(pf->channel_popup));

    /* ॹ饤 */
    SetHSimpleSliderPos(frame->zoom_slider,
			pf->zoom_slider->sslider.pos);
    memcpy(frame->zoom, pf->zoom, sizeof(frame->zoom));

    for(i = 0; i < SCOPE_NLOCAL_MODE; i++)
	change_ChoiceButton(frame->local_mode_selector[i],
			    pf->local_mode_selector[i]->choice.selected);
}

static void create_play_slider(SSFrame* frame, WinObj* top, int x, int y)
{
    int w, h;
    WinObj* backward_button;
    WinObj* forward_button;

    h = TextHeight(font) + GAP;
    x += TextWidth(font, "հ ") + h;

    w = top->any.width - x - 2 * BD_WIDTH - GAP - h;
    frame->play_slider =
	create_HLimitedSlider(top, x, y, w, h, BD_WIDTH, 0,  w);
    SetCallBackProc(frame->play_slider, (call_back_t)play_slider_proc);

    backward_button =
	create_TrigButton(top, WinObjLeft(frame->play_slider) - h, y, h, h,
			  BD_WIDTH, M_PI);
    forward_button =
	create_TrigButton(top, WinObjRight(frame->play_slider), y, h, h,
			  BD_WIDTH, 0.0);
    SetCallBackProc(backward_button, (call_back_t)backward_button_proc);
    SetCallBackProc(forward_button, (call_back_t)forward_button_proc);
}

static void repaint_scope_winobj_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);
    repaint_frame_scope(frame, frame->player->pos, False);
}

static void create_scope_mode_selector(SSFrame* frame, WinObj* top,
				       int x, int y, int mode)
{
    static char* names[SCOPE_NMODE] = {
	"Wave",			/* SCOPE_WAVE */
	"Spectram",		/* SCOPE_SPECTORAM */
	"Spectrogram",		/* SCOPE_SPECTOROGRAM */
	"Edit" };		/* SCOPE_EDIT */
    static char* local_modes[SCOPE_NLOCAL_MODE][2] =
    {
	{"Log", "Linear"},
	{"No wave", "Wave"},
	{"Paint", "3D"},
    };
    int i;

    frame->scope_main_mode =
	create_ChoiceButton(top, x, y, BD_WIDTH,
			    TextHeight(font) + 2, font,
			    NULL, names, SCOPE_NMODE, mode);
    SetCallBackProc(frame->scope_main_mode, (call_back_t)change_scope_mode_proc);

    x = WinObjRight(frame->scope_main_mode) + GAP;
    for(i = 0; i < SCOPE_NLOCAL_MODE; i++)
    {
	frame->local_mode_selector[i] =
	    create_ChoiceButton(top, x, y, BD_WIDTH,
			    TextHeight(font) + 2, font,
			    NULL, local_modes[i], 2, 0);
	SetCallBackProc(frame->local_mode_selector[i],
			(call_back_t)repaint_scope_winobj_proc);
	if(i + 2 < SCOPE_NLOCAL_MODE)
	    x = WinObjRight(frame->local_mode_selector[i]) + GAP/2;
    }
    MoveWinObj(frame->local_mode_selector[SCOPE_WAVEFORM],
	       WinObjRight(frame->local_mode_selector[SCOPE_HZ_LOG]) + GAP/2,
	       y);
}

PUBLIC void open_new_ssframe(SSFrame* parent_frame)
{
    SSFrame* frame;
    WinObj* top;
    WinObj* bar;
    WinObj* scope_simple_base;
    WinObj* volume_name_label;
    WinObj* scope_zoom_gear;

    x_setup_once();

    frame = alloc_ssframe();
    if(frame == NULL)
	return;

    frame->toplevel = top = create_sstop_window(parent_frame);

    if(parent_frame == NULL)
    {
	frame->scope_width  = DEFAULT_SCOPE_WIDTH;
	frame->scope_height = DEFAULT_SCOPE_HEIGHT;
    }
    else
    {
	frame->scope_width  = parent_frame->scope_width;
	frame->scope_height = parent_frame->scope_height;
    }

#ifdef DEBUG
    puts("create_scope_bar()");
#endif /* DEBUG */

    bar = create_scope_bar(frame, top);

#ifdef DEBUG
    puts("create_scope_bar() - done");
#endif /* DEBUG */

    frame->load_button =
	create_Button(top, GAP, WinObjBottom(bar) + GAP,
		      TextWidth(font, "  ") + GAP,
		      TextHeight(font) + GAP,
		      BD_WIDTH, font, "  ");
    SetCallBackProc(frame->load_button, (call_back_t)file_load_button_proc);

    frame->dir_text =
	create_DirLineText(top,
			   WinObjRight(frame->load_button) - BD_WIDTH,
			   WinObjBottom(bar) + GAP,
			   top->any.width - WinObjRight(frame->load_button) -
			   GAP - BD_WIDTH,
			   TextHeight(font) + GAP,
			   BD_WIDTH, font, " File:");
    SetCallBackProc(frame->dir_text, (call_back_t)file_load_button_proc);
    frame->dir_text->dirtext.short_cut_list =
	(MCPopupMenu *)text_short_cut_list_popup;
    SetPopupSelectingHook(frame->dir_text, popup_waiting_hook);

    create_play_buttons(frame, top, GAP,
			WinObjBottom(frame->load_button) + GAP);

#ifdef DEBUG
    puts("create_play_buttons() - done");
#endif /* DEBUG */

    create_play_slider(frame, top, GAP,
		       WinObjBottom(frame->player->play_button) + GAP);

    create_scope_mode_selector(frame, top, GAP,
			       WinObjBottom(frame->play_slider) + GAP,
			       parent_frame != NULL
			           ? SSFRAME_SCOPE_MAIN_MODE(parent_frame)
			           : DEFAULT_SCOPE_MAIN_MODE);

    /* volume */
    volume_name_label =
	create_LabeldWindow(top,
			    WinObjRight(frame->player->looping_button) + GAP,
			    WinObjTop(frame->player->looping_button),
			    TextWidth(font, " "), TextHeight(font) + GAP,
			    BD_WIDTH, font, " ", WINOBJ_BORDER_NONE,
			    Center);
    frame->volume_slider =
	create_HSimpleSlider(top,
			     WinObjRight(volume_name_label),
			     WinObjTop(frame->player->looping_button),
			     VOLUME_SLIDER_LEN + 2 * BD_WIDTH,
			     TextHeight(font) + GAP,
			     BD_WIDTH, audriv_get_play_volume());
    SetCallBackProc(frame->volume_slider, (call_back_t)volume_slider_proc);

    scope_simple_base =
	create_SimpleWindow(top, GAP, WinObjBottom(frame->scope_main_mode),
			    frame->scope_width, frame->scope_height,
			    BD_WIDTH, WINOBJ_BORDER_CONCAVE);
    create_scope_window(frame, scope_simple_base);
    scope_zoom_gear =
	create_VAdjustGear(top, WinObjRight(scope_simple_base),
			   WinObjTop(scope_simple_base), GAP - 2 * BD_WIDTH,
			   SkeletonWinObj(scope_simple_base)->any.height
			       -2 * BD_WIDTH,
			   BD_WIDTH, 10);
    SetCallBackProc(scope_zoom_gear, (call_back_t)scope_zoom_gear_proc);

    frame->zoom_slider =
	create_HSimpleSlider(top, GAP + TextWidth(font, " "),
			     GAP + WinObjBottom(scope_simple_base),
			     frame->scope_width - TextWidth(font, " "),
			     SLIDER_HEIGHT, BD_WIDTH, 0);
    SetCallBackProc(frame->zoom_slider, (call_back_t)zoom_slider_proc);
    
    /* Resize Toplevel Window */
    ResizeWinObj(top,
		 top->any.width, WinObjBottom(frame->zoom_slider) + GAP);

    frame->buffer_label =
	create_LabeldWindow(top,
			    WinObjLeft(frame->zoom_slider),
			    WinObjBottom(frame->zoom_slider) + GAP,
			    SkeletonWinObj(frame->zoom_slider)->any.width - 2,
			    TextHeight(font) + GAP,
			    1, font, SOUNDSCOPE_DEFAULT_BUFFER_NAME,
			    WINOBJ_BORDER_CONCAVE,
			    Center);

    create_equlizer(frame);

#ifdef DEBUG
    puts("create_equlizer() - done");
#endif /* DEBUG */

    create_srcustom(frame);

#ifdef DEBUG
    puts("create_srcustom() - done");
#endif /* DEBUG */

    set_ssframe_reference_pointer(frame, top);
    set_ssframe_reference_pointer(frame, frame->equlizer);
    set_ssframe_reference_pointer(frame, frame->srcustom);

    if(number_of_ssframes() == 2)
	allow_close_window_proc();

    frame->mapflag = True;
    WinObjRealize(top);

    map_local_mode(frame);

#ifdef DEBUG
    puts("map_local_mode() - done");
#endif /* DEBUG */

    x_secondary_setup_once();

#ifdef DEBUG
    puts("x_secondary_setup_once() - done");
#endif /* DEBUG */

    if(parent_frame != NULL)
    {
	copy_parent_frame_members(parent_frame, frame);
    }
    else
    {
	PopupMenuChangeMarker(frame->sample_rate_popup, device_sample_rate);
	PopupMenuChangeMarker(frame->output_popup, audriv_get_play_output());
	PopupMenuChangeMarker(frame->encoding_popup, device_data_encoding);
	PopupMenuChangeMarker(frame->channel_popup, device_nchannels);
    }

    if(SSFRAME_SCOPE_MAIN_MODE(frame) == SCOPE_SPECTORAM ||
       SSFRAME_SCOPE_MAIN_MODE(frame) == SCOPE_SPECTOROGRAM)
    {
	initialize_log_hz_table(frame->zoom[SSFRAME_SCOPE_MAIN_MODE(frame)],
				frame->log_hz_table);
	initialize_exp_hz_table(frame->zoom[SSFRAME_SCOPE_MAIN_MODE(frame)],
				frame->exp_hz_table);
    }

    frame->measure.x_left = 720;
    frame->measure.x_right = 743;
    frame->measure.y = 7;
    frame->measure.status = None;

    init_scope_pixmap(frame);

#if 0
    reset_scope_query_Hz();
#endif


    if(audio_file_lock_flag)
	set_audio_file_lock(True);
#ifdef DEBUG
    printf("WinObj: %g%% event hash table is used.\n",
	   get_dispatch_table_use_ratio() * 100);
#endif
}

PUBLIC void init_scope_pixmap(SSFrame* frame)
{
#ifdef DEBUG
    printf("init_scope_pixmap();\n");
#endif

    if(frame->scope_pm != None &&
       (frame->scope_pixmap_width < frame->scope_width ||
	frame->scope_pixmap_height < frame->scope_height))
    {
	XFreePixmap(disp, frame->scope_pm);
	XFreePixmap(disp, frame->scope_background);
	frame->scope_pm = None;
    }

    if(frame->ed_cur_pm != None &&
       frame->scope_pixmap_height < frame->scope_height)
    {
	XFreePixmap(disp, frame->ed_cur_pm);
	frame->ed_cur_pm = None;
    }

    if(frame->scope_pm == None)
    {
	frame->scope_pm =
	    XCreatePixmap(disp, frame->scope_win,
			  frame->scope_width, frame->scope_height,
			  DefaultDepth(disp, DefaultScreen(disp)));
	frame->scope_background = XCreatePixmap(disp, frame->scope_win,
				    frame->scope_width, frame->scope_width,
				    DefaultDepth(disp, DefaultScreen(disp)));
	frame->scope_pixmap_width = frame->scope_width;
	frame->scope_pixmap_height = frame->scope_height;
    }

    if(frame->ed_cur_pm == None)
    {
	frame->ed_cur_pm =
	    XCreatePixmap(disp, frame->scope_win,
			  1, frame->scope_height,
			  DefaultDepth(disp, DefaultScreen(disp)));
    }

    XSetForeground(disp, gc, scope_bg);
    XFillRectangle(disp, frame->scope_background, gc, 0, 0,
		   frame->scope_width, frame->scope_width);

    switch(SSFRAME_SCOPE_MAIN_MODE(frame))
    {
      case SCOPE_WAVE:

	init_scope_wave(frame);
	break;
      case SCOPE_SPECTORAM:
	init_scope_spectram(frame);
	break;
      case SCOPE_SPECTOROGRAM:
	break;
      case SCOPE_EDIT:
	draw_edit_scope_wave(frame);
	break;
    }

    if(SSFRAME_SCOPE_MAIN_MODE(frame) != SCOPE_EDIT)
    {
	XCopyArea(disp, frame->scope_background, frame->scope_pm, gc,
		  0, 0, frame->scope_width, frame->scope_height, 0, 0);
	XCopyArea(disp, frame->scope_background, frame->scope_win, gc,
		  0, 0, frame->scope_width, frame->scope_height, 0, 0);
    }

    if(!frame->player->play_flag)
	repaint_frame_scope(frame, frame->player->pos, False);
}
