#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include "soundscope.h"
#include "audriv.h"
#include "aubuff.h"
#include "ssframe.h"
#include "player.h"
#include "bucket.h"

enum ff_rw_type_t forward_backward_button_mode = FF_RW_UNUSED;
static SSFrame* forward_backward_frame;
static int forward_backward_cnt;

void fixme_here(WinObj* obj, XEvent* e)
{
    puts("## FIXME !");
}

void forward_button_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);

    switch(e->type)
    {
      case ButtonRelease:
	forward_backward_button_mode = FF_RW_UNUSED;
	break;
      case ButtonPress:
	if(!audio_active_flag || SSFRAME_SCOPE_MAIN_MODE(frame) == SCOPE_EDIT)
	{
	    forward_backward_button_mode = FAST_FORWARD_PLAY;
	    forward_backward_frame = frame;
	    forward_backward_cnt = 0;
	}
	break;
    }
    update_ss_active_flag();
}

void backward_button_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);

    switch(e->type)
    {
      case ButtonRelease:
	forward_backward_button_mode = FF_RW_UNUSED;
	break;
      case ButtonPress:
	if(!audio_active_flag || SSFRAME_SCOPE_MAIN_MODE(frame) == SCOPE_EDIT)
	{
	    forward_backward_button_mode = FAST_REWIND_PLAY;
	    forward_backward_frame = frame;
	    forward_backward_cnt = 0;
	}
	break;
    }
    update_ss_active_flag();
}

PUBLIC void move_view_position(int diff)
{
    long cur;
    double z;
    SSFrame* frame = forward_backward_frame;
    long nsamples;

    cur = frame->player->pos;
    nsamples = frame->player->aubuff->nsamples;
    if(frame->player->load_flag && file_loader.read_samples < nsamples)
	nsamples = file_loader.read_samples;

    if(SSFRAME_SCOPE_MAIN_MODE(frame) != SCOPE_EDIT)
    {
	z = frame->zoom[SCOPE_WAVE];
	diff = z * diff / 32;
	cur += diff;
	if(cur < 0)
	    cur = 0;
	else if(cur > nsamples)
	    cur = nsamples;

	frame->player->pos = cur;
	repaint_frame_scope(frame, cur, False);

	if(forward_backward_cnt == 0)
	    usleep(200000);
    }
    else
    {
	double vr, scale, d;

	vr = frame->ed_vratio;
	scale = frame->zoom[SCOPE_EDIT];

	d = nsamples - scale;
	if(fabs(d) < 1.0)
	    if(diff < 0)
		vr = 0.0;
	    else
		vr = 1.0;
	else
	{
	    vr += diff * 24 * scale / frame->scope_width / d;
	    if(vr < 0)
		vr = 0;
	    else if(vr > 1.0)
		vr = 1.0;
	}

	SetHScrollbarThumbPos((HScrollbar *)frame->edit_slider, vr);

	frame->ed_vratio = vr;
	draw_edit_scope_wave(frame);
	repaint_frame_scope(frame, cur, False);
	if(!audio_active_flag && forward_backward_cnt == 0)
	    usleep(200000);
    }

    forward_backward_cnt++;
}

static void sstoplevel_redraw_proc(WinObj* obj)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);

    DrawString(disp, obj->any.win, font, WinObj_fg,
	       GAP, WinObjTop(frame->play_slider),
	       0, WinObjBottom(frame->play_slider) -
	       WinObjTop(frame->play_slider),
	       "հ", West);

    DrawString(disp, obj->any.win, font, WinObj_fg,
	       GAP, WinObjTop(frame->zoom_slider),
	       0, WinObjBottom(frame->zoom_slider) -
	       WinObjTop(frame->zoom_slider),
	       "", West);
}

void sstoplevel_event_handler(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);

    if(e == NULL)
    {
	sstoplevel_redraw_proc(obj);
	return;
    }
    switch(e->type)
    {
      case Expose:
	if(e->xexpose.count == 0)
	{
	    sstoplevel_redraw_proc(obj);
	    if(frame->mapflag == False)
	    {
		SetEventMask(obj, StructureNotifyMask | ExposureMask);
		frame->mapflag = True;
	    }
	}
	break;
      case UnmapNotify:
        SetEventMask(obj, ExposureMask);
	frame->mapflag = False;
        break;
    }
}

static void redraw_toplevel_window(SSFrame* frame)
{
    XClearWindow(disp, frame->toplevel->any.win);
    sstoplevel_event_handler(frame->toplevel, NULL);
}

void play_slider_proc(HLimitedSlider* slider, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(slider);
    long sample;
    int i;

#ifdef DEBUG
    puts("play_slider_proc()");
#endif

    if(frame->player->aubuff->nsamples == 0)
	return;

    switch(e->type)
    {
      case ButtonPress:
	frame->play_slider_grab_flag = True;
	repaint_frame_scope(frame,
			    play_slider_pos_to_sample(frame, slider->pos),
			    False);
	break;
      case MotionNotify:
	repaint_frame_scope(frame,
			    play_slider_pos_to_sample(frame, slider->pos),
			    False);
	break;
      case ButtonRelease:
	frame->play_slider_grab_flag = False;
	sample = play_slider_pos_to_sample(frame, slider->pos);
	if(frame->player->play_flag)
	{
	    abc_stop_audio();
	    frame->player->pos = sample;
	    for(i = 0; i < frame->player->aubuff->nchannels; i++)
		frame->player->bucket_reader[i]->u.reader.seek_pointer =
		    sample;
	    abc_restart_audio();
	}
	else
	    frame->player->pos = sample;
	break;
    }
}

void edit_slider_proc(HSimpleSlider* slider, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(slider);
    HScrollbar* hs = (HScrollbar *)frame->edit_slider;

#ifdef DEBUG
    puts("edit_slider_proc()");
#endif

    if(frame->player->aubuff->nsamples == 0)
	return;
    frame->ed_vratio = GetHScrollbarThumbPos(hs);
    draw_edit_scope_wave(frame);
}


/************************************************************************
 *                                                                      *
 *                     play buttons                                     *
 *                                                                      *
 ************************************************************************/

void play_button_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);
#ifdef DEBUG
    puts("play_button_proc()");
#endif

    if(frame->player->aubuff->nsamples == 0)
    {
	if(e->type == ButtonRelease)
	    SetPressdIconButton(obj, False);
	return;			/* do nothing because of non-data */
    }

    if(e->type != ButtonPress || frame->player->play_flag)
	return;			/* do nothing because of button pressd */

    if(player_start_play(frame->player) == True)
	SetPressdIconButton(frame->player->stop_button, False);
    else
	SetPressdIconButton(frame->player->play_button, False);
}

void stop_button_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);

    if(frame->player->aubuff->nsamples == 0)
    {
	if(e->type == ButtonRelease)
	    SetPressdIconButton(obj, False);
	return;			/* do nothing because of non-data */
    }

    if(e->type != ButtonPress)
	return;

    if(!frame->player->play_button->icon.pressd)
	return;

    player_stop_play(frame->player);
    SetPressdIconButton(frame->player->play_button, False);
}

void stop_and_rewind_button_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);

    if(e->type != ButtonPress)
	return;

    player_stop_and_rewind(frame->player);
}

void fast_forward_button_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);
    AudioPlayer* player = frame->player;

    if(frame->player->aubuff->nsamples == 0)
        return;

    if(e->type == ButtonPress)
        player_start_fast_forward_play(player, FAST_FORWARD_STEP1);
    else
        player_stop_fast_forward_play(player);
}

void looping_button_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);
    AudioPlayer* player = frame->player;


    if(e->type != ButtonRelease)
	return;

    if(player->looping)
    {
	/* reset looping */
	SetPressdIconButton(obj, False);
	player->looping = False;
    }
    else
    {
	/* set looping */
	player->looping = True;
    }
}

void fast_rewind_button_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);
    AudioPlayer* player = frame->player;

    if(frame->player->aubuff->nsamples == 0)
	return;

    if(e->type == ButtonPress)
	player_start_fast_rewind_play(player, FAST_REWIND_STEP1);
    else
	player_stop_fast_rewind_play(player);
}


Bool popup_waiting_hook(PopupMenuBase* m)
{
    if(ss_active_flag)
	soundscope_slice_proc();
    return !ss_active_flag;
}

void sample_rate_popup_proc(PopupMenuBase* menu, char* name, int id)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(menu);

    if(id == PopupMenuMarkerID((WinObj *)menu))
	return;
    player_change_sample_rate(frame->player, (long)id);
    set_srvalue(frame);
}

void output_popup_proc(PopupMenuBase* menu, char* name, int id)
{
    if(id == PopupMenuMarkerID((WinObj *)menu))
	return;
    player_change_output((long)id);
}

void encoding_popup_proc(PopupMenuBase* menu, char* name, int id)
{
    if(id == PopupMenuMarkerID((WinObj *)menu))
	return;
    player_change_encoding((long)id);
}

void channel_popup_proc(PopupMenuBase* menu, char* name, int id)
{
    if(id == PopupMenuMarkerID((WinObj *)menu))
	return;
    player_change_channel((long)id);
}


#define WAVE_ZOOM_C 9.0
static double zoom_c = 0.0;	/* 1.0 / (exp(WAVE_ZOOM_C) - 1.0); */

PUBLIC int calc_log_wave_zoom(double zoom, int maxpos, long n)
{
    int pos;
    if(zoom_c == 0.0)
	zoom_c = 1.0 / (exp(WAVE_ZOOM_C) - 1.0);

    if(zoom <= (double)MIN_WAVE_ZOOM + 1e-6 || n <= MIN_WAVE_ZOOM)
	return 0;

    pos = (int)(0.5 + (log((zoom - MIN_WAVE_ZOOM) / zoom_c /
	(n - MIN_WAVE_ZOOM)) + 1) / WAVE_ZOOM_C * (double)maxpos);
    if(pos > maxpos)
	return maxpos;
    return pos;
}

static int zoom_to_slider_pos(SSFrame* frame, double zoom)
{
    HSimpleSlider* slider = (HSimpleSlider *)frame->zoom_slider;
    switch(SSFRAME_SCOPE_MAIN_MODE(frame))
    {
      case SCOPE_WAVE:
      case SCOPE_EDIT:
	{
	    return calc_log_wave_zoom(zoom, slider->maxlen,
				      frame->player->aubuff->nsamples);
#if 0
	    int n;

	    if(zoom_c == 0.0)
		zoom_c = 1.0 / (exp(WAVE_ZOOM_C) - 1.0);
	    /* WAVE_ZOOM_C > 0.0, so zoom_c > 0.0 */

	    n = frame->player->aubuff->nsamples;
	    if(zoom <= (double)MIN_WAVE_ZOOM + 1e-6 || n == 0)
		return 0;
	    if(zoom + 1e-6 >= n)
		return slider->maxlen;
/*
zoom = MIN_WAVE_ZOOM + exp(WAVE_ZOOM_C * ((double)slider->pos / (double)slider->maxlen) - 1) * zoom_c * (n - MIN_WAVE_ZOOM)

exp(WAVE_ZOOM_C * ((double)slider->pos / (double)slider->maxlen) - 1) =
	(zoom - MIN_WAVE_ZOOM) / zoom_c / (n - MIN_WAVE_ZOOM)

WAVE_ZOOM_C * ((double)slider->pos / (double)slider->maxlen) =
	log((zoom - MIN_WAVE_ZOOM) / zoom_c / (n - MIN_WAVE_ZOOM)) + 1

slider->pos = (log((zoom - MIN_WAVE_ZOOM) / zoom_c / (n - MIN_WAVE_ZOOM)) + 1) / WAVE_ZOOM_C * (double)slider->maxlen
*/
	    return (int)(0.5 + (log((zoom - MIN_WAVE_ZOOM) / zoom_c / (n - MIN_WAVE_ZOOM)) + 1) / WAVE_ZOOM_C * (double)slider->maxlen);
#endif
	}

      case SCOPE_SPECTORAM:
      case SCOPE_SPECTOROGRAM:
/*
zoom = LOG_HZ_MIN_X1 + ((double)slider->pos / (double)slider->maxlen) * (LOG_HZ_MAX_X1 - LOG_HZ_MIN_X1)

(slider->pos / slider->maxlen) * (LOG_HZ_MAX_X1 - LOG_HZ_MIN_X1)
	= zoom - LOG_HZ_MIN_X1

slider->pos / slider->maxlen = (zoom - LOG_HZ_MIN_X1) / (LOG_HZ_MAX_X1 - LOG_HZ_MIN_X1)

slider->pos = (zoom - LOG_HZ_MIN_X1) / (LOG_HZ_MAX_X1 - LOG_HZ_MIN_X1) * slider->maxlen
*/
	return (int)(0.5 + (zoom - LOG_HZ_MIN_X1) / (LOG_HZ_MAX_X1 - LOG_HZ_MIN_X1) * slider->maxlen);
    }
    return 0;
}

void zoom_slider_proc(WinObj* obj, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(obj);
    HSimpleSlider* slider = (HSimpleSlider *)obj;
    int mode = SSFRAME_SCOPE_MAIN_MODE(frame);

#ifdef DEBUG
    printf("zoom_slider_proc();\n");
#endif

    switch(mode)
    {
      case SCOPE_WAVE:
      case SCOPE_EDIT:
        {
	    double x;
	    long n;

	    if(zoom_c == 0.0)
		zoom_c = 1.0 / (exp(WAVE_ZOOM_C) - 1.0);
	    			/* WAVE_ZOOM_C > 0.0, so zoom_c > 0.0 */

	    n = frame->player->aubuff->nsamples;
	    if(n < MIN_WAVE_ZOOM)
	    {
		frame->zoom[mode] = MIN_WAVE_ZOOM;
		break;
	    }
	    x = exp(WAVE_ZOOM_C *
		    (double)slider->pos / (double)slider->maxlen) - 1;
	    /* 0 <= x <= exp(WAVE_ZOOM_C) - 1
	     * then
	     * 0 <= x * zoom_c <= 1
	     */
	    frame->zoom[mode] = MIN_WAVE_ZOOM +
		x * zoom_c * (n - MIN_WAVE_ZOOM);
	    if(mode == SCOPE_EDIT)
	    {
		draw_edit_scope_wave(frame);
		if(n > 0)
		    SetHScrollbarShown((HScrollbar *)frame->edit_slider,
				       frame->zoom[SCOPE_EDIT] / n);
	    }
	    else
		change_display_scope_time(frame, frame->scope_background,
					  frame->zoom[SCOPE_WAVE]);
	    if(mode == SCOPE_WAVE && e != NULL && e->type == ButtonRelease)
		init_scope_pixmap(frame);
	}
	break;
      case SCOPE_SPECTORAM:
      case SCOPE_SPECTOROGRAM:
	frame->zoom[mode] = LOG_HZ_MIN_X1 + ((double)slider->pos
	     / (double)slider->maxlen) * (LOG_HZ_MAX_X1 - LOG_HZ_MIN_X1);
	initialize_log_hz_table(frame->zoom[mode], frame->log_hz_table);
	initialize_exp_hz_table(frame->zoom[mode], frame->exp_hz_table);
	break;
    }

#ifdef DEBUG
    printf("zoom = %f\n", frame->zoom[mode]);
#endif

    if(!frame->player->play_flag)
	repaint_frame_scope(frame, frame->player->pos, False);
}

static void animate_box(SSFrame* frame)
{
    int w, h, win, dw, dh, i;

    w = frame->scope_width;
    h = frame->scope_height;
    win = frame->scope_win;

    dw = w / 16;
    dh = h / 16;

    for(i = 0; i < 8; i++)
    {
	XSetForeground(disp, gc, scope_fg);
	XDrawRectangle(disp, win, gc,
		       w/2 - dw * i - 1, h/2 - dh * i - 2,
		       dw * i * 2, dh * i * 2);
	XSync(disp, False);
	usleep(10000);
	XSetForeground(disp, gc, scope_bg);
	XDrawRectangle(disp, win, gc,
		       w/2 - dw * i - 1, h/2 - dh * i - 2,
		       dw * i * 2, dh * i * 2);
    }
}

void change_scope_mode_proc(ChoiceButton* c, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(c);
    int mode = SSFRAME_SCOPE_MAIN_MODE(frame);

    animate_box(frame);

#ifdef DEBUG
    printf("change_scope_mode_proc()\n");
#endif

    SetHSimpleSliderPos(frame->zoom_slider,
			zoom_to_slider_pos(frame, frame->zoom[c->selected]));

    if(mode == SCOPE_SPECTORAM || mode == SCOPE_SPECTOROGRAM)
    {
	initialize_log_hz_table(frame->zoom[mode], frame->log_hz_table);
	initialize_exp_hz_table(frame->zoom[mode], frame->exp_hz_table);
    }

    map_local_mode(frame);
    init_scope_pixmap(frame);
    redraw_toplevel_window(frame);
}

void volume_slider_proc(HSimpleSlider* slider, XEvent* e)
{
    audriv_set_play_volume(slider->pos);
    SetHSimpleSliderPos((WinObj *)slider, audriv_get_play_volume());
}

void scope_zoom_gear_proc(VAdjustGear* zoom, XEvent* e)
{
    SSFrame* frame = WINOBJ_EVENT_FRAME(zoom);

    if(zoom->diff > 0)
    {
	frame->yzoom -= (double)zoom->diff * 0.005;
	if(frame->yzoom < 0)
	    frame->yzoom = 0;
    }
    else if(zoom->diff < 0)
	frame->yzoom -= (double)zoom->diff * 0.005;

    if(!audio_active_flag)
    {
	XEvent ev;
	repaint_frame_scope(frame, frame->player->pos, False);
	ev = *e;
	ev.type = Expose;
	ev.xexpose.count = 0;
	scope_event_handler((WinObj *)zoom, &ev);

	sprintf(temp_buff, "Zoom: %g", frame->yzoom);
	DrawString(disp, frame->scope_win, font, scope_bg,
		   0, 1, frame->scope_width - GAP + 1, frame->scope_height,
		   temp_buff, East);
	DrawString(disp, frame->scope_win, font, scope_bg,
		   0, -1, frame->scope_width - GAP - 1, frame->scope_height,
		   temp_buff, East);
	DrawString(disp, frame->scope_win, font, scope_fg,
		   0, 0, frame->scope_width - GAP, frame->scope_height,
		   temp_buff, East);
    }

#ifdef DEBUG
    printf("Y zoom: %f\n", frame->yzoom);
#endif /* DEBUG */
}
